/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1997, 2005 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

buffer int (*when_net_activity)();	// called by epsilon during net job
char url_services[50] = "ftp|http|telnet|scp|ssh";
	// Names of supported services.

user char anon_ftp_password[FNAMELEN] = "-EpsilonUser@unknown.host";
user char ftp_ascii_transfers;	// Set nonzero for VMS and similar systems.
user char ftp_compatible_dirs = 2;	// Bits make directory ops more compatible.
user char ftp_passive_transfers = 1;
zeroed buffer ftp_go_to_line;	// Go to this line after reading file.

user char http_proxy_server[FNAMELEN];	// Send all requests here.
user int http_proxy_port;		// Zero means no proxy.
user char http_proxy_exceptions[FNAMELEN] = "|localhost|127.0.0.1|";
user char http_user_agent[FNAMELEN];	// Identify us to web servers.
user char http_log_request;	// Copy request to HTTP Headers buffer too.
char *http_force_headers;	// Send just these header lines.

char _telnet_mode_name[] = "Telnet";
buffer int telnet_id;
user char telnet_interpret_output = 7;
keytable telnet_tab, telnet_ctrlc_tab;
char _ssh_mode_name[] = "Ssh";
keytable ssh_tab;
buffer int (*buffer_ftp_activity)();
char *get_url_file_part(char *url, int separator);
#define CONNECT_LIST_BUF	"-connections"

// Parsing the URL puts the data in one of these.

struct url_parts {
	char service[FNAMELEN];		// "http", "ftp", etc.
	char host[FNAMELEN];		// The host name.
	char usr[FNAMELEN];			// User name (can be "anonymous").
	char pwd[FNAMELEN];			// Password (can be empty).
	char port[FNAMELEN];		// Port number if any, as text.
	char fname[FNAMELEN * 3];		// The file name.
	int have_password, have_usr; // These items can be specififed but empty.
};

// We must send a password to the DLL on every FTP operation.
// Maintain the passwords here so we don't have to keep asking
// the user.  The structure is marked "zeroed" to ensure that it's
// never saved in an Epsilon state file.

zeroed struct password_data {
	struct password_data *next;
	char host[FNAMELEN];
	char usr[FNAMELEN];
	char pwd[FNAMELEN];
	int invalid;		// If nonzero, prompt for a new pwd next time.
} *password_data;

struct ftp_job {
	int id;				// A code from the DLL.
	int operation;		// FTP_SEND, FTP_RECV, etc.
	int buf;			// Data goes in this buffer.
	int log_buf;		// The log file with status messages.
	int file_size;		// We need to send/retrieve this many bytes
	int so_far;			// and we've done this many bytes so far.
	int time_started;	// Value of time_ms() when we started.
};
buffer struct ftp_job *ftp_job;

// Data for scp jobs.
struct scp_job {
	int busy;			// Using this buffer for an operation?
	int buf;			// The buffer where we'll get status messages.
	char *host;			// The host we're connecting to.
	char *uname;		// The user name to use.
	int op;				// Operation we're doing (send, receive, dired list, misc).
	int destbuf;		// Which buffer gets sent text, status msgs, etc.
	char file[FNAMELEN];	// Temp file, or extra path text to insert.
	int flags;			// Optional behaviors.
	char *cmd;			// Next cmd to execute, delimited by \n.
	char *cmdlist;		// List of cmds to execute.
	struct scp_job *next;
	char *homedir;		// Initial directory.
	char *lastdir;		// Directory of last pwd response.
};
zeroed struct scp_job *scp_job;	// List of scp jobs.

#define SCP_CLIENT_MODE		1	// Use protocol of custom client.
#define SCP_OLD_SFTP		2	// Using old sftp client?
#define SCP_SCP_XFER		4	// Use scp for xfers, not sftp?
#define SCP_TMP_FILE		16	// Delete tmp file after xfer finishes?
#define SCP_CHECK_IS_DIR	32	// Checking if a URL refers to a directory?
#define SCP_CMD_ERROR		64	// Got error from command, ignore results.
#define SCP_IS_XFER			128	// Currently doing xfer with sftp?

user char scp_write_file_template[FNAMELEN] = "scp -q %a %l %u@%h:%f";
user char scp_read_file_template[FNAMELEN] = "scp -q %a %u@%h:%f %l";
user char scp_write_file_no_user_template[FNAMELEN] = "scp -q %a %l %h:%f";
user char scp_read_file_no_user_template[FNAMELEN] = "scp -q %a %h:%f %l";

user char scp_run_helper_template[FNAMELEN] = "sftp %a %u@%h";
user char scp_run_helper_no_user_template[FNAMELEN] = "sftp %a %h";
user char scp_client_style = 0;
// Use "ssh -q %a -l %u %h epsilon-xfer-helper" or "ssh -q %a %h
// epsilon-xfer-helper" for custom client, and set scp_client_style to
// 1.

user char scp_unix_options[FNAMELEN] = "";
user char scp_windows_options[FNAMELEN] = "-oBATCHMODE=yes";
// Add -oBATCHMODE=yes to the ssh command line, for Windows versions of
// ssh that prompt for a password in a way we can't intercept and
// respond to.

user char ssh_template[FNAMELEN] = "ssh -l %u %h";
user char ssh_no_user_template[FNAMELEN] = "ssh %h";
