/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1985, 2003 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// This file defines the set-color command, which changes Epsilon's
// screen colors.

#include "eel.h"
#include "proc.h"

#define FG(c)		((c) & 15)	/* extract foreground & background */
#define BG(c)		(((c) >> 4) & 7)
#define COLOR_COL	56
#define COLOR_LINE	2
#define SCRPLACE(c)	(COLOR_COL + FG(c)), (COLOR_LINE + BG(c))

// Is the name table entry m a color scheme?
int is_color_scheme(int m)
{
	return name_type(m) == NT_COLSCHEME
			&& strcmp(name_name(m), "color-defaults");

}
// Fill buffer b with a list of all color schemes.
make_color_scheme_list(int b)
{
	int i;

	buf_zap(b);
	for (i = 1; i <= final_index(); i++)
		if (is_color_scheme(i))
			buf_printf(b, "%s\n", name_name(i));
}

int count_color_schemes()
{
	int i, j = 0;

	for (i = 1; i <= final_index(); i++)
		if (is_color_scheme(i))
			j++;
	return j;
}

record_color_settings(int **orig_fg, int **orig_bg, int **orig_schemes)
{
	int i, j, k, classes = number_of_color_classes();
	int schemes = count_color_schemes();

	*orig_schemes = (int *) malloc(schemes * sizeof(int));
	*orig_fg = (int *) malloc(classes * schemes * sizeof(int));
	*orig_bg = (int *) malloc(classes * schemes * sizeof(int));
	for (i = 1, k = 0; i <= final_index(); i++)
		if (is_color_scheme(i)) {
			selected_color_scheme = i;
			for (j = 0; j < classes; j++) {
				(*orig_fg)[k * classes + j] = get_foreground_color(j, 1);
				(*orig_bg)[k * classes + j] = get_background_color(j, 1);
			}
			(*orig_schemes)[k++] = i;
		}
}

restore_color_settings(int *orig_fg, int *orig_bg, int *orig_schemes)
{
	int i, j, k, classes = number_of_color_classes();
	int schemes = count_color_schemes();

	for (i = 0; i < schemes; i++) {
		selected_color_scheme = orig_schemes[i];
		for (j = 0; j < classes; j++) {
			k = i * classes + j;
			set_color_pair(j, orig_fg[k], orig_bg[k]);
		}
	}
}

preserve_color_settings(int *orig_fg, int *orig_bg, int *orig_schemes,
						int orig_scheme)
{
	int i, j, k, n, classes = number_of_color_classes();
	int schemes = count_color_schemes(), fg_color, bg_color;

	*get_color_scheme_variable() = selected_color_scheme;
	if (want_record_customizations()) {
		k = prep_customizations();
		if (selected_color_scheme != orig_scheme)
			list_one(k, color_scheme_variable_index(), 1);
		for (i = 0; i < schemes; i++) {
			selected_color_scheme = orig_schemes[i];
			for (j = 0; j < classes; j++) {
				fg_color = get_foreground_color(j, 1);
				bg_color = get_background_color(j, 1);
				n = i * classes+ j;
				if (fg_color != orig_fg[n] || bg_color != orig_bg[n])
					list_one_color_setting(k, j, 1);
			}
		}
		selected_color_scheme = *get_color_scheme_variable();
		say("");
	} else
		show_binding("To remember these colors for "
					 "future sessions, use %s.", "list-customizations");
}

// Respond to output from dialog helper: Set a color's values, switch to
// a color scheme.

set_color_handler(int activity, int buf, int from, int to)
{
	char msg[FNAMELEN], scheme[FNAMELEN];
	int i, nums[4];

	save_var bufnum = buf;
	if (activity != NET_RECV)
		return;
	save_var narrow_start, narrow_end, point = to;
	to_begin_line();
	narrow_end = size() - point;
	point = from;
	to_begin_line();
	do {
		if (parse_string(1, "=")) {
			point++;
			parse_string(1, "[^\n ]+", scheme);
			i = find_index(scheme);
			point++;
			parse_string(1, ".*", msg);
			if (break_into_numbers(msg, nums) == 3 && is_color_scheme(i)) {
				save_var selected_color_scheme = i;
				set_color_pair(nums[0], nums[1], nums[2]);
			}
		} else if (parse_string(1, "@.*", scheme)) {
			i = find_index(scheme + 1);
			if (is_color_scheme(i))
				selected_color_scheme = i;
		} else if (parse_string(1, ".*", msg))
			say("%s", msg);	// Display the error message.
	} while (nl_forward());
	refresh();
}

set_color_xwin()
{
	int i, j, classes = number_of_color_classes();
	int *orig_fg, *orig_bg, *orig_schemes;
	int orig_scheme = selected_color_scheme, this_scheme = orig_scheme;
	char *s, cmd[FNAMELEN];

	record_color_settings(&orig_fg, &orig_bg, &orig_schemes);
	on_exit free(orig_schemes), free(orig_fg), free(orig_bg);
	save_var bufnum;
	get_executable_directory(cmd);
	sprintf(cmd + strlen(cmd), "epsilon-dialogs --parent %x --color",
		   screen_to_window_id(0));

	for (;;) {
		bufnum = zap("-epsilon-dialog-data");
		bprintf("%d\n%s\n", font_styles, font_fixed);
		for (i = 1; i <= final_index(); i++)
			if (is_color_scheme(i))
				bprintf("%s\n", name_name(i));
		bprintf("\n");
		for (i = 0; i < classes; i++)
			bprintf("%s\n", name_color_class(i));
		bprintf("\n");
		bprintf("%s\n", (s = lookpath("colclass.txt")) ? s : "");
		bprintf("%s\n", name_name(this_scheme));

		selected_color_scheme = find_index("color-defaults");
		for (j = 0; j < classes; j++)
			bprintf("default %x %x\n", get_foreground_color(j, 1),
					get_background_color(j, 1));
		for (i = 1; i <= final_index(); i++)
			if (is_color_scheme(i)) {
				selected_color_scheme = i;
				for (j = 0; j < classes; j++)
					bprintf("%d %d %x %x\n", i, j, get_foreground_color(j, 1),
							get_background_color(j, 1));
			}
		selected_color_scheme = this_scheme;
		if (!pipe_text("-epsilon-dialog-data", "-epsilon-dialog-out", cmd,
					   NULL, PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH,
					   set_color_handler))
			file_error(errno, "Couldn't run dialog helper", "Error");
		this_scheme = selected_color_scheme;
		bufname = "-epsilon-dialog-out";
		i = process_exit_status;
		if (i < 0 || i > 2)
			i = !ask_yn("Keep changes", "The dialog process exited "
						"unexpectedly.  Keep these settings? ",
						"Yes", "No", 0);
		if (i == 0) {
			preserve_color_settings(orig_fg, orig_bg, orig_schemes,
									orig_scheme);
			return;
		} else if (i == 1) {
			restore_color_settings(orig_fg, orig_bg, orig_schemes);
			selected_color_scheme = orig_scheme;
			say("");
			return;
		} else {
			show_binding("Continue changing colors "
						 "by typing %s", "exit-level");
			recursive_edit();
		}
	}
}

command set_color()		/* query user to set colors */
{
	int class, color, i, j, *orig, classes, docbuf, wantwin;
	int swin, ccwin, cwin, infowin, newcolor, newscheme, orig_scheme;
	int *orig_fg, *orig_bg, *orig_schemes;
	int fg_color, bg_color;
	char buf[100];
	int sbuf, ccbuf, cbuf, infobuf, orig_win = window_handle, win, page;
	char *new, *old;
	jmp_buf this_level;

	if (is_unix == IS_UNIX_XWIN) {
		set_color_xwin();
		return;
	}
	save_var top_level = &this_level;
	classes = number_of_color_classes();

	sbuf = tmp_buf();	/* list of color schemes */
	bufnum = sbuf;	/* now highlight current line in buffers */
	add_region(point_spot, point_spot, color_class menu_highlight,
			   REGLINE);
	make_color_scheme_list(sbuf);

	orig_scheme = selected_color_scheme;
	record_color_settings(&orig_fg, &orig_bg, &orig_schemes);
	selected_color_scheme = orig_scheme;
	orig = (int *) malloc(classes * sizeof(int));
	for (i = 0; i < classes; i++)
		orig[i] = alter_color(i, -1);

	class = iter = page = 0;

	bufnum = docbuf = tmp_buf();	// read help text for classes into buf
	on_exit buf_delete(docbuf);
	if (new = lookpath("colclass.txt"))
		file_read(new, FILETYPE_AUTO);
	ccbuf = tmp_buf();	/* list of color classes */
	cbuf = tmp_buf();	/* window with colors */ 
	infobuf = tmp_buf();	/* help buffer with options */ 

	bufnum = ccbuf;
	add_region(point_spot, point_spot, color_class menu_highlight,
			   REGLINE);
	for (i = 0; i < classes; i++)
		bprintf("%s\n", name_color_class(i));
	delete(point - 1, point);

winagain:
	swin = add_popup(4, 1, 20, 10, ALL_BORD(BSINGLE), sbuf);
	window_title(swin, BTOP, TITLECENTER, " Color Schemes ");
	set_menu_look(swin);

	ccwin = add_popup(29, 1, 21, 10, ALL_BORD(BSINGLE), ccbuf);
	display_column = 0;		/* prevent continuation lines */
	window_title(ccwin, BTOP, TITLECENTER, " Color Classes ");
	set_menu_look(ccwin);

	if (!(has_feature & FEAT_ANYCOLOR)) {		// Matrix of color choices.
		cwin = add_popup(55, 1, 18, 10, ALL_BORD(BSINGLE), cbuf);
		window_title(cwin, BTOP, TITLECENTER, " Colors ");
		invisible_window = 1;
		insert('\n');
		if (selectable_colors > 128)
			window_title(cwin, BBOTTOM, TITLERIGHT(1), " More ");
	}

	infowin = add_popup(2, 14, 75, 8, _view_border, infobuf);
	display_column = 0;		/* prevent continuation lines */
	mouse_cursor = &left_pointer;
	text_color = is_gui ? color_class text : color_class menu;
	display_scroll_bar = 0;
	newscheme = 2;
	if (has_feature & FEAT_COLOR_DIALOG) {
		display_dialog_box("SetColorBox", "Set colors",
						   swin, ccwin, infowin, "OK", "Cancel", "Help");
		save_var permit_window_keys = PERMIT_SCROLL_KEY;
		save_var return_raw_buttons = 1;
		wantwin = ccwin;
	}

	if (setjmp(top_level))	/* come here if subcommand's aborted */
		;
	for (;;) {
		color = alter_color(class, -1);
		fg_color = get_foreground_color(class, 0);
		bg_color = get_background_color(class, 0);

		if ((color >= 128) != page && !newscheme)
			newscheme = 1;
		page = (color >= 128);
		if (newscheme) {
			if (!(has_feature & FEAT_ANYCOLOR)) {	// Sample window has 2 pages.
				window_handle = cwin;
				point = page;
			}
			window_handle = swin;
			if (newscheme == 2) {
				make_color_scheme_list(bufnum);
				delete(size() - 1, size());	/* no final \n */
			}
			point = 0;
			col_search(name_name(selected_color_scheme), 0);
			if (!(has_feature & FEAT_ANYCOLOR)) {
				j = page ? 128 : 0;
				for (i = 0; i < 128; i++, j++)
					term_write_attr(SCRPLACE(i), '+', j);
			}
		}
		window_handle = ccwin;
		point = 0;
		sprintf(buf, "^%s$", name_color_class(class));
		re_search(1, buf);
		to_begin_line();
		buf_zap(infobuf);
		window_handle = infowin;
		if (!(has_feature & FEAT_COLOR_DIALOG)) {
			bprintf("Type F or B to change the color scheme, D to "
					"define a new color scheme.\n");
			bprintf("Or use the arrow keys to change the "
					"%s color class.\n", name_color_class(class));
		} else
			bprintf("Sample text using the %s color class:\n",
					name_color_class(class));
		i = point;
		bufnum = docbuf;
		point = 0;
		sprintf(buf, "^%s\t+", name_color_class(class));
		if (re_search(1, buf))
			buf_xfer(infobuf, point, give_end_line());
		else
			buf_printf(infobuf, "This is sample text.");
		bufnum = infobuf;
		j = point + 1;
		if (!(has_feature & FEAT_COLOR_DIALOG)) {
			bprintf("\nType N or P to set other color classes, "
					"T to try out\n");
			bprintf("these colors temporarily, S to set "
					"these colors permanently,\n");
			bprintf("or C to cancel.  Color was initially "
					"0x%02x, this color is 0x%02x.",
					orig[class], color);
		}
		set_character_color(i, j, class);
		point = 0;

		dialog_checkboxes = ((fg_color & EFONT_BOLD) ? 1 : 0)
							| ((fg_color & EFONT_ITALIC) ? 2 : 0)
							| ((fg_color & EFONT_UNDERLINED) ? 4 : 0);

		if (has_feature & FEAT_ANYCOLOR) {
			window_handle = wantwin;
			maybe_refresh();
			getkey();
			wantwin = window_handle;
		} else {
			window_handle = cwin;
			do {
				maybe_refresh();
				term_write_attr(SCRPLACE(color), 'C', color);
				full_getkey("Your choice: ", 0);
				term_write_attr(SCRPLACE(color), '+', color);
			} while (!IS_TRUE_KEY(key) && !IS_MOUSE_DOWN(key));
		}
		key = toupper(key);
		newcolor = -100;	// Assume it's an invalid key.
		newscheme = 0;
		if (binding_of(key, reg_tab) == (short) next_page
			  || binding_of(key, reg_tab) == (short) previous_page) {
			do_command(binding_of(key, reg_tab));
			newcolor = -99;		// Mark as a valid key.
		}
		if (has_feature & FEAT_COLOR_DIALOG) {
			window_handle = ccwin;	// Scrolling might have changed these.
			class = lines_between(0, point);

			window_handle = swin;
			grab_full_line(bufnum, buf);
			i = find_index(buf);
			if (i > 0 && name_type(i) == NT_COLSCHEME)
				selected_color_scheme = i;
		}
		if (key_is_button == 4) {
			i = comm_dlg_color(fg_color, "Foreground color");
			if (i != -1)
				set_color_pair(class, i | (fg_color & EFONT_STYLE), bg_color);
		} else if (key_is_button == 5) {
			i = comm_dlg_color(bg_color, "Background color");
			if (i != -1)
				set_color_pair(class, fg_color, i);
		} else if (key_is_button == 8) {
			set_color_pair(class, fg_color ^= EFONT_BOLD, bg_color);
		} else if (key_is_button == 9) {
			set_color_pair(class, fg_color ^= EFONT_ITALIC, bg_color);
		} else if (key_is_button == 10) {
			set_color_pair(class, fg_color ^= EFONT_UNDERLINED, bg_color);
		} else if (key == 'C' || key == 'Q' || key == ESC || user_abort
				   || key_is_button == 2) {
			restore_color_settings(orig_fg, orig_bg, orig_schemes);
			selected_color_scheme = orig_scheme;
			say("");
			break;
		} else if (key == 'S' || key == ' ' || IS_ENTER_KEY(key)
				   || key_is_button == 1) {
			preserve_color_settings(orig_fg, orig_bg, orig_schemes,
									orig_scheme);
			break;
		} else if (IS_MOUSE_DOWN(key)) {
			win = window_at_coords(mouse_y, mouse_x, mouse_screen);
			if (!is_window(win))
				continue;
			window_handle = win;
			get_text_area(win);
			if (win == cwin) {
				if (bar_on_border()) {
					i = scroll_bar_code();
					if ((i ? (i > 0) : !page) != point)
						newcolor = color ^ 128;
				} else {
					newcolor = (mouse_y - COLOR_LINE) * 16;
					newcolor += (mouse_x - COLOR_COL);
					if (page)
						newcolor += 128;
				}
			} else if (in_text_area(mouse_x, mouse_y))
				point_to_pos(mouse_x, mouse_y, mouse_screen);
			else if (display_scroll_bar && bar_on_border())
				mouse_scroll_bar();
			else
				continue;
			if (win == ccwin)
				class = lines_between(0, point);
			else if (win == swin)
				newscheme = 1;
			else if (win == infowin && isupper(curchar())
					 && character(point - 1) == ' '
					 && character(point + 1) == ' ')
				ungot_key = curchar();
			if (win != infowin)
				wantwin = win;
		} else if ((has_feature & FEAT_COLOR_DIALOG)
				   && full_key == NUMSHIFT(GREYTAB)) {
			prev_dialog_item();
			wantwin = window_handle;
			continue;
		} else if ((has_feature & FEAT_COLOR_DIALOG) && generic_key(key) == '\t') {
			next_dialog_item();
			wantwin = window_handle;
			continue;
		} else if ((has_feature & FEAT_COLOR_DIALOG) && key == GETFOCUS) {
			win = window_at_coords(mouse_y, mouse_x, mouse_screen);
			if (!is_window(win) || win == infowin)
				continue;
			wantwin = win;
		}
		if (!(has_feature & FEAT_ANYCOLOR))
			switch (get_direction()) {
				case BTOP:		newcolor = color + 256 - 16; break;
				case BBOTTOM:	newcolor = color + 16; break;
				case BLEFT:		newcolor = color + 256 - 1; break;
				case BRIGHT:	newcolor = color + 1; break;
			}
		else
			switch (get_direction()) {
				case BTOP: case BLEFT:
					key = (win == ccwin) ? 'P' : 'B'; break;
				case BBOTTOM: case BRIGHT:
					key = (win == ccwin) ? 'N' : 'F'; break;
			}
		if ((has_feature & FEAT_COLOR_DIALOG) && key == 'N' && win == swin)
			key = 'F';
		if ((has_feature & FEAT_COLOR_DIALOG) && key == 'P' && win == swin)
			key = 'B';
		switch (key) {
			case 'N': class++; break;
			case 'P': class--; break;
			case 'T':
				show_binding("Continue changing colors "
					"by typing %s", "exit-level");
				remove_window(swin);
				remove_window(ccwin);
				remove_window(cwin);
				remove_window(infowin);
				recursive_edit();
				goto winagain;
				break;
			case 'F':
				wantwin = window_handle = swin;
				if (!nl_forward())
					point = 0;
				newscheme = 1;
				break;
			case 'B':
				wantwin = window_handle = swin;
				if (!nl_reverse())
					point = size();
				newscheme = 1;
				break;
			case 'D':
				wantwin = swin;
				maybe_refresh();
				get_strpopup(buf, "New scheme name: ", "", "set-color");
				if (*buf) {
					new_variable(buf, NT_COLSCHEME, 0, 0);
					new = varptr(find_index(buf));
					old = varptr(selected_color_scheme);
					for (i = 0; i < ptrlen(old, 1); i++)
						*new++ = *old++;
					newscheme = 2;
				}
				break;
			default:
				if (binding_of(key, root_keys) == (short) help
					|| key == 'h' || key == '?'
					|| key_is_button == 3) {
					if (!(has_feature & FEAT_COLOR_DIALOG))
						say("Press <Space> to continue.");
					help_on_command("set-color");
					if (!(has_feature & FEAT_COLOR_DIALOG))
						newscheme = 1;
				} else if (IS_TRUE_KEY(key) && newcolor == -100) {
					maybe_ding(bell_on_bad_key);
					newscheme = 1;
				}
				break;
		}
		if (newcolor > -99)	/* did it change? */
			alter_color(class, newcolor % selectable_colors);
		else if (newscheme) {
			window_handle = swin;
			grab_full_line(bufnum, buf);
			i = find_index(buf);
			if (i > 0 && name_type(i) == NT_COLSCHEME)
				selected_color_scheme = i;
		} else
			class = (class + classes) % classes;
	}
	remove_window(swin);
	remove_window(ccwin);
	if (!(has_feature & FEAT_ANYCOLOR))
		remove_window(cwin);
	remove_window(infowin);
	buf_delete(ccbuf);
	buf_delete(cbuf);
	buf_delete(infobuf);
	if (is_window(orig_win))
		window_handle = orig_win;
	buf_delete(sbuf);
	full_redraw = 1;
	*get_color_scheme_variable() = selected_color_scheme;
	free(orig);
	free(orig_fg);
	free(orig_bg);
}
