/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1988, 2005 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// This file has the commands for search and replacing in multiple
// files.

#include "eel.h"
#include "grep.h"
#include "colcode.h"
#include "dired.h"

when_loading()	/* disable normal keys, copy some dired definitions */
{

	fix_key_table(reg_tab, (short) normal_character, grep_tab, -1);
	set_list_keys(grep_tab);
	fix_key_table(dired_tab, (short) dired_examine_in_window,
		grep_tab, (short) dired_examine_in_window);
}

command grep() on reg_tab[FALT(7)]
{
	grep_prompt(REGEX | UNKNOWN, has_arg);
}

grep_prompt(flags, inbufs)
{
	char filepat[FNAMELEN], old_cd[FNAMELEN], buf_dir[FNAMELEN];
	jmp_buf *old_level = top_level, this_level;

	_last_was_grep = 1;
	tiled_only();
	iter = 0;
	error_if_input(1);	/* error if we're in the middle of input */
	flags = get_search_string("List lines with: ",
				  flags | (inbufs ? BUFGREP : FILEGREP));
	flags &= ~(INCR | FAILING | BAD | REVERSE);
	if (!inbufs)
		save_var full_path_on_mode_line = 1;
	all_must_build_mode = 1;
	if (!get_grep_file(inbufs, filepat))
		return;
	restore_vars();
	all_must_build_mode = 1;
	top_level = &this_level;
	getcd(old_cd);
	if (grep_default_directory == 3 && get_buffer_directory(buf_dir))
		chdir(buf_dir);
	if (!setjmp(top_level))
		do_grep("grep", default_search_string(flags), flags, filepat, inbufs);
	top_level = old_level;
	if (grep_default_directory == 3)
		chdir(old_cd);
}

char *(*give_grep_finder)()(inbufs, filepat, ps)
char *filepat, **ps;	// put 1st match in ps & return finder for next matches
{
	char *(*finder)();

	finder = inbufs ? b_match : do_file_match;
	*ps = (*finder)(filepat, STARTMATCH | EXACTONLY
					| (inbufs ? ((preserve_filename_case & 4) ? FM_FOLD : 0)
							  : FM_NO_DIRS));
	return finder;
}

get_grep_file(inbufs, filepat)
char *filepat;
{
	char pr[FNAMELEN], dir[FNAMELEN];
	int show_rel = 1;		// Show the relative path of the default.

	if (!inbufs) {
		if (is_dired_buf() && !inp_get_other_name)
			save_var inp_get_other_name = get_dired_buffer_pattern();
		if (grep_default_directory == 1 && !grep_prompt_with_buffer_directory
			  || grep_default_directory == 3)
			show_rel = 0;
		build_prompt(pr, "In file(s)", _grep_files, 0, show_rel);
		save_var prompt_with_buffer_directory = grep_prompt_with_buffer_directory;
		get_file_dir(filepat, pr);
		if (!*filepat && grep_default_directory) {
			save_var _help_on_cmd = last_index;
			if (grep_default_directory == 3)
				strcpy(filepat, _grep_files);
			else
				get_last_recall(f_match, filepat);
			if (grep_default_directory >= 2 && *filepat
				  && get_buffer_directory(dir)) {
				if (grep_default_directory != 3)
					absolute(filepat, dir);
				return 1;
			}
		}
		if (*filepat) {
			if (is_directory(filepat))
				sprintf(filepat + strlen(filepat), "%c*", path_sep);
			strcpy(_grep_files, filepat);
		}
		if (grep_default_directory == 3)
			strcpy(filepat, _grep_files);
		else
			relative(_grep_files, filepat);
		if (grep_default_directory == 1 && !grep_prompt_with_buffer_directory)
			strcpy(_grep_files, filepat);
		if (!*filepat)
			return 0;
	} else
		comp_read(filepat, "In buffers [pattern]: ", b_match,
				  ((preserve_filename_case & 4) ? FM_FOLD : 0), bufname);
	return 1;
}

// Is this file name acceptable in grep (not excluded by
// grep-ignore-file-extentions)?

ok_grep_file_match(char *s, char *abs)
{
	char pat[FNAMELEN];
	struct file_info info;

	if (!use_grep_ignore_file_variables)
		return 1;
	save_var case_fold = (FOLDFILE != 0);
	sprintf(pat, "|%s|", get_extension(s));
	if (strstr(grep_ignore_file_extensions, pat))
		return 0;
	if (*grep_ignore_file_pattern
		  && string_matches_pattern(abs, grep_ignore_file_pattern))
		return 0;
	if (*grep_ignore_file_basename
		  && string_matches_pattern(get_tail(abs, 1), grep_ignore_file_basename))
		return 0;
	if (grep_ignore_file_types && check_file(abs, &info)) {
		if ((grep_ignore_file_types & 1) && info.check_type == CHECK_DEVICE)
			return 0;
		if ((grep_ignore_file_types & 2) && info.check_type == CHECK_PIPE)
			return 0;
		if ((grep_ignore_file_types & 4) && info.check_type == CHECK_OTHER)
			return 0;
	}
	return 1;
}

grep_date_warning(char *buf, char *file)
{
	int i;

	save_var bufname = buf;
	i = point;
	bprintf("Searching version of %s in memory, "
			  "but version on disk has changed.\n", file);
	set_character_color(i, point, color_class grep_header);
}

prepare_grep_buffer(char *buf, int flags, char *pat, char *filepat)
{
	char search_type[FNAMELEN], dir[FNAMELEN];
	int before_head;

	create(buf);
	if (grep_empties_buffer)
		zap(buf);
	locate_window(buf, "");
	point = size();
	if (grep_empties_buffer)
		modified = 0;
	save_var modified;
	if (size())
		bprintf("\n");	/* separate from previous query */
	make_search_prompt(search_type, flags & SMODES);
	before_head = point;
	getcd(dir);
	bprintf("Found 0 lines with %s\"%s\" in %s%s (cur dir \"%s\")\n",
			search_type, pat, (flags & BUFGREP) ? "(buffers) " : "",
			filepat, dir);
	set_character_color(before_head, point, color_class grep_header);
	if (display_column.default != -1)
		display_column = 0;	/* horizontally scroll this window */
	grep_mode();
	return before_head;
}

update_grep_title(int before_head, int cnt)
{
	point = before_head;
	delete(point, point + parse_string(1, "Found [0-9]+ "));
	bprintf("Found %d ", cnt);
	set_character_color(before_head, point, color_class grep_header);
	point = size();
}

do_grep(buf, pat, flags, filepat, inbufs)
char *buf, *pat, *filepat;
{
	int res = 0, aborted = 0, cnt = 0, tempbuf = 0, i, skipped = 0;
	int orig, before_head, this_err;
	char *s, rel[FNAMELEN], abs[FNAMELEN], *(*finder)();
	char *months = "Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec ";
	char extra[FNAMELEN];
	spot start;

	flags |= (inbufs ? BUFGREP : FILEGREP);
	save_var synchronous_file_io = 1;
	before_head = prepare_grep_buffer(buf, flags, pat, filepat);
	start = alloc_spot(0);
	save_var modified;
	save_var abort_file_matching = ABORT_ERROR;
	finder = give_grep_finder(inbufs, filepat, &s);	// get 1st match
	save_var readonly_warning, no_popup_errors = 1;
	if (grep_on_changed_file & 1)
		readonly_warning |= ROWARN_GREP;
	readonly_warning &= ~(ROWARN_MSG | ROWARN_BELL);
	orig = bufnum;
	for (; s; s = (*finder)(filepat, 0)) {
		if (user_abort) {
			aborted = 1;
			break;
		}
		grep_make_names(rel, abs, s, inbufs);
		if (inbufs)
			bufname = s;
		else if (!ok_grep_file_match(s, abs)) {
			skipped++;
			continue;
		}
		noteput("%d matches, searching %f...", cnt, rel);
		if (inbufs)	// skip over grep results buffer
			this_err = !strcmp(s, buf);
		else if (!look_file(abs)) {
			this_err = find_it(abs, default_translation_type);
			if (!grep_keeps_files)
				tempbuf = bufnum;
		} else {
			this_err = 0;
			if (grep_on_changed_file & 2)
				save_var auto_read_changed_file = 1;
			if (check_dates(0) == 2)
				grep_date_warning(buf, filename);
		}
		if (this_err == EREADABORT || user_abort) {
			aborted = 1;
			break;
		} else if (this_err) {
			window_bufnum = orig;
			if (tempbuf)
				buf_delete(tempbuf);
			tempbuf = 0;
			continue;
		}
		*extra = 0;
		if (grep_include_timestamp && !inbufs && file_info.check_type == CHECK_FILE)
			if (grep_include_timestamp == 2)
				sprintf(extra, "[%04d-%02d-%02d %02d:%02d:%02d] ",
						file_info.year, file_info.month, file_info.day,
						file_info.hour, file_info.minute, file_info.second);
			else
				sprintf(extra, "[%02d-%.3s-%04d %02d:%02d:%02d] ",
						file_info.day,
						(file_info.month >= 1 && file_info.month <= 12)
						? months + (file_info.month - 1) * 4 : "???",
						file_info.year,
						file_info.hour, file_info.minute, file_info.second);
		if (grep_show_absolute_path == 2)
			convert_to_8_3_filename(abs);
		res = copy_matches(name_to_bufnum(buf), flags, pat,
						   grep_show_absolute_path ? abs : rel, extra);
		window_bufnum = orig;
		if (res < 0)	/* error occurred during search */
			break;
		else if (res > 0) {
			cnt += res;
			build_first = 1;
			update_grep_title(before_head, cnt);
			refresh();
		}
		if (tempbuf)
			buf_delete(tempbuf);
		tempbuf = 0;
	}
	window_bufnum = orig;
	point = i = *start;
	free_spot(start);
	build_first = 1;
	if (tempbuf)
		buf_delete(tempbuf);
	say("");
	if (aborted)
		quick_abort();
	if (res == -DSBAD)
		error("Illegal search pattern: %s", pat);
	else if (res != -DSABORT) {
		locate_window(buf, "");
		point = i;
		if (skipped) {
			say("Skipped %d file%s due to grep-ignore-file- variables",
				  skipped, (skipped == 1) ? "" : "s");
			bprintf("%pSkipped %d file%s due to grep-ignore-file- variables\n",
					color_class grep_header,
					skipped, (skipped == 1) ? "" : "s");
		}
	}
}

grep_make_names(rel, abs, s, inbufs)	// put name to use in grep results
char *rel, *abs, *s;				// for any matches in rel
{
	if (inbufs) {
		save_var bufname = s;
		if (*filename)
			relative(filename, rel);
		else
			sprintf(rel, BUFPREFIX "%s", s);
	} else
		relative(s, rel);
	strcpy(abs, rel);
	if (!inbufs)
		absolute(abs);
}

/* copy matches in current buf to other buffer */
/* returns 0=no matches, >0=# of matches, <0= -(error code) */

copy_matches(int tobuf, int flags, char *pat, char *rel, char *extra)
{
	int sbuf = bufnum, cnt = 0, i, more, srcherr, lines;
	int before, file, do_color;

	do_color = want_code_coloring && recolor_range;
	save_var point = 0;
	while ((srcherr = do_color_searching(flags, pat)) == 1) {
		i = give_position_at_column(matchstart, 0);
		more = nl_forward();
		lines = lines_between(i, point, 0);
		before = get_buf_point(tobuf);
		if (lines > 1)
			buf_printf(tobuf, "%s%s [%d]: ", extra, rel, lines);
		else
			buf_printf(tobuf, "%s%s: ", extra, rel);
		file = get_buf_point(tobuf);
		if (do_color) {			// Color in place, then copy.
			if (do_color == 1 && idle_coloring_size > 1000
				  && (recolor_from_here != recolor_by_lines)) {
				recolor_buffer_range(0, size());
				do_color = 2;		// Color all at once if coloring pieces is
			}						// slow.
			if (do_color == 1)
				recolor_buffer_range(i, point - more);
		}
		buf_xfer_colors(tobuf, i, point - more);
		bufnum = tobuf;
		set_character_color(before, file, color_class grep_filename);
		if (!do_color)
			set_character_color(file, point, color_class grep_text);
		bprintf("\n");
		set_character_color(point - 1, point, color_class grep_filename);
		point = file;	// Now color matches within line.
		while (do_searching(flags, pat) == 1 && point > file) {
			set_character_color(matchstart, matchend, color_class grep_match);
			file = point;
		}
		point = size();
		bufnum = sbuf;
		cnt++;
		if (!more)
			return cnt;
	}
	return srcherr ? -srcherr : cnt;
}

grep_mode()
{
	do_mode_default_settings(1);
	mode_keys = grep_tab;
	major_mode = _grep_mode_name;
	mouse_dbl_selects = 1;
	make_mode();
}

grep_help() on grep_tab['h'], grep_tab['?']
{
	help_on_command("grep-mode");
}

grep_examine() on grep_tab[' '], grep_tab['\n'],
		   grep_tab['e'], grep_tab['\r']
{
	do_grep_examine(NULL);
}

grep_examine_in_other_window() on grep_tab['o']
{
	do_grep_examine("next-window");
}

do_grep_examine(char *cmd)
{
	int start, end, colon, old, instance, i, toolong = 0, extra;
	int lines = 1, precolon;
	char file[FNAMELEN], string[FNAMELEN], line[FNAMELEN], curdir[FNAMELEN], *p;
	char prefix[FNAMELEN];

	save_var synchronous_file_io = 1;
	_last_was_grep = 1;
	to_begin_line();
	save_var case_fold = 0;		/* turn off for grep buffer */
	if (!search(1, ": "))
		error("Can't read file name");
	colon = point;
	precolon = colon - 2;
	if (parse_string(-1, " %[[0-9]+%]: ")) {
		precolon = matchend;
		grab(precolon + 2, point, line);
		lines = numtoi(line);
	}
	to_begin_line();
	extra = parse_string(1, "%[[0-9][0-9]+-[^]]*[0-9]%] ");
	start = point;
	for (i = lines; i > 1; i--)
		nl_forward();
	to_end_line();
	end = point;
	point = start;
	grab(start, precolon, prefix);
	grab(start + extra, precolon, file);
	if (start + FNAMELEN - 3 < end)
		end = toolong = start + FNAMELEN - 3;
	grab(colon, end, string);
	line[0] = '\n';		// Put \n at start and end of line.
	grab(start, end + 1, line + 1);
	instance = count_instances(line) + 1;
	grep_get_curdir(curdir);
	if (cmd && *cmd)
		try_calling(cmd);
	if (strstr(file, MANPREFIX) == file) {
		grep_examine_man_page(file + sizeof(MANPREFIX) / sizeof(char) - 1);
		return;
	} else if (strstr(file, BUFPREFIX) != file) {
		absolute(file, curdir);
		if (find_it(file, ask_line_translate()))
			return;
	} else if (exist(p = file + sizeof(BUFPREFIX) / sizeof(char) - 1))
		to_buffer(p);
	else
		error("Can't find buffer %s.", p);
	if (strlen(line) != end - start + 2)
		error("Can't find matching lines with null characters.");
	old = point;
	save_var case_fold = 0;		/* turn off for user buffer */
	point = 0;
	for (i = 0; i < instance; i++)
		if (toolong ? !search(1, string) : !line_search(1, string)) {
			point = old;
			if (i == 0 || instance == 1)
				error("Can't find \"%s\"", string);
			else
				error("Can't find this copy of repeated line");
		} else {
			old = point;
			if (i + 1 < instance)
				nl_forward();
		}
	to_begin_line();
	if (toolong)
		say("Line too long; this may not be the correct match.");
}

line_search(dir, s)	/* search for line with only s on it */
char *s;		/* go to its end (if reverse: start) */
{
	int start, end;

	if (!*s)
		return re_search(dir, "^$");
	if (dir < 0)
		to_end_line();
	while (search(dir, s)) {
		start = (dir > 0) ? matchstart : point;
		end = (dir > 0) ? point : matchstart;
		if ((end >= size() - narrow_end || character(end) == '\n')
				&& (start <= narrow_start
					|| character(start - 1) == '\n'))
			return 1;
	}
	return 0;
}

grep_get_curdir(char *dir)
{
	save_var point;
	search(-1, "\nFound ");
	point += parse_string(1, "\nFound .* lines with \"");
	search(1, "\"");
	if (parse_string(1, ".*%(cur dir \"(.*)\"%)\n"))
		grab(find_group(1, 1), find_group(1, 0), dir);
	else
		getcd(dir);
}

count_instances(s)	/* how many times does s occur before pt in this */
char *s;		/* search? */
{
	int count = 0;

	save_var point, narrow_end = size() - point;
	search(-1, "\nFound ");
	save_var narrow_start = point;
	for (; search(1, s); point--)
		count++;
	return count;
}

command next_match()
{
	by_matches(iter);
	iter = 0;
}

command previous_match()
{
	by_matches(-iter);
	iter = 0;
}

by_matches(val)
{
	int owin, win, pt, dir = RE_FORWARD, buf = name_to_bufnum("grep");

	tiled_only();
	_last_was_grep = 1;
	if (!buf)
		error("No grep buffer.");
	win = buf_in_window(buf);
	if (win >= 0) {
		owin = window_handle;
		window_handle = win;
	} else
		bufnum = buf;
	if (val < 0)
		val = -val, dir = RE_REVERSE;
	else
		to_end_line();
	while (val-- > 0)
		if (!re_search(dir, "^(<(>buffer<)> )?[^\n\t]+: "))
			error("No more matches.");
	if (win >= 0) {
		pt = point;
		window_handle = owin;
		bufnum = buf;
		point = pt;
	}
	has_arg = iter = 0;
	grep_examine();
}

command next_position() on cx_tab[CTRL('N')]
{
	if (_last_was_grep)
		next_match();
	else
		next_error();
}

command previous_position() on cx_tab[CTRL('P')]
{
	if (_last_was_grep)
		previous_match();
	else
		previous_error();
}

grep_examine_man_page(char *pageref)
{
	char buf[FNAMELEN], *s = strrchr(pageref, '.');

	if (s) {
		*s = 0;
		sprintf(buf, "%s %s", s + 1, pageref);
	} else
		strcpy(buf, pageref);
	make_man_page(buf, 0);
}

copy_man_page_matches(int flags, char *pat)
{
	int before = buffer_size("grep");

	xfer("grep", give_begin_line(), give_end_line());
	save_var bufname = "grep";
	set_character_color(before, size(), color_class grep_text);
	bprintf("%p\n", color_class grep_filename);
	save_var point = before;
	while (do_color_searching(flags, pat) == 1)
		set_character_color(matchstart, matchend, color_class grep_match);
}

int search_one_man_page(char *file, int flags, char *pat, char *match)
{
	int cnt = 0;

	save_var bufnum = zap("-search-man-page");
	if (err_file_read(file, default_translation_type))
		return 0;
	save_var uncompress_files = 1;
	if (!strcmp(get_extension(file), ".gz"))
		call_mode(file);
	while (do_searching(flags, pat) == 1) {
		buffer_printf("grep", "%p%s%s: ",
					  color_class grep_filename, MANPREFIX, match);
		copy_man_page_matches(flags, pat);
		cnt++;
		if (!search_man_pages_shows_all)
			return cnt;
	}
	return cnt;
}

command search_man_pages()
{
	char pagepat[FNAMELEN], section[FNAMELEN], pat[FNAMELEN], *searchpat;
	int flags, before_head, cnt = 0, this_cnt;
	char man_path[FNAMELEN * 2], match[FNAMELEN], *path, *t, *q, *f;

	flags = get_search_string("Search man pages for text: ", UNKNOWN | FILEGREP);
	flags &= ~(INCR | FAILING | BAD | REVERSE);
	searchpat = default_search_string(flags);

	save_var _help_on_cmd = last_index;
	get_last_recall((char *(*)()) 0, pat);
	if (!*pat)
		strcpy(pat, "*");
	get_strnone(section, "Only search these man page sections, "
			   "such as 1 2 3, or * for all: ", pat);
	for (t = section, q = match; *t; t++)
		if (isdigit(*t))
			*q++ = *t;
	*q = 0;
	if (*match)
		sprintf(section, "[%s]", match);
	else
		*section = 0;

	save_var recall_id = "search-man-pages pat";
	get_last_recall((char *(*)()) 0, pat);
	if (!*pat)
		strcpy(pat, "*");
	get_strnone(pagepat, "Restrict to man pages matching pattern: ", pat);
	if (!*pagepat)
		strcpy(pagepat, "*");

	save_var cygwin_filenames |= 32;
	save_var show_status = 0;
	build_man_path(path = man_path);
	before_head = prepare_grep_buffer("grep", flags, searchpat, "man pages");

	while (!user_abort) {
		t = strchr(path, ':');
		if (t)
			strncpy(pat, path, t - path);
		else
			strcpy(pat, path);
		if (*pat && pat[strlen(pat) - 1] != '/')
			strcat(pat, "/");
		convert_cygwin_filename(pat, 0);
		sprintf(pat + strlen(pat), "man%s*/%s", section, pagepat);
		f = do_file_match(pat, STARTMATCH);
		for (; f && !user_abort; f = do_file_match(pat, 0)) {
			q = get_extension(f);
			if (!q || !*q)
				continue;
			strcpy(match, get_tail(f, 0));
			noteput("Searching %s...", match);
			q = get_extension(match);
			if (q && !strcmp(q, ".gz"))
				*q = 0;
			q = get_extension(match);
			if (!q || !*q)
				continue;
			this_cnt = search_one_man_page(f, flags, searchpat, match);
			if (this_cnt) {
				update_grep_title(before_head, cnt += this_cnt);
				build_first = 1;
				refresh();
			}
		}
		if (!t)
			break;
		path = t + 1;
	}
	point = before_head;
	nl_forward();
	if (!user_abort)
		modified = 0;
	say("");
}

command keep_matching_lines()
{
	drop_matching_lines(0, "Delete lines not containing: ");
}

command delete_matching_lines()
{
	drop_matching_lines(1, "Delete lines containing: ");
}

drop_matching_lines(drop, pr)
char *pr;
{
	int flags;

	tiled_only();
	iter = 0;
	error_if_input(1);	/* error if we're in the middle of input */
	save_var point;
	flags = get_search_string(pr, UNKNOWN | REGEX | BUFGREP);
	restore_vars();
	flags &= ~(INCR | FAILING | BAD | REVERSE);
	do_drop_matching_lines(flags, default_search_string(flags), drop);
}

// Keep or drop lines after pt matching pattern.
// This version handles patterns that may match \n, but it's much slower.
slow_drop_matching_lines(flags, pat, drop)
char *pat;
{
	int res, cnt = 0;

	save_spot point, narrow_start, narrow_end;
	to_begin_line();
	do {
		narrow_start = point;
		narrow_end = size() - give_end_line();
		res = do_color_searching(flags, pat);
		narrow_end = 0;
		nl_forward();
		if (res != 0 && res != 1)
			break;
		if (drop && res || !drop && !res)
			delete(narrow_start, point);
		if (res)
			  cnt++;
	} while (point < size() - narrow_end);
	return cnt;
}

do_drop_matching_lines(flags, pat, drop)
char *pat;		// keep or drop lines after pt matching pattern
{
	int res, here, there, cnt = 0;

	flags &= ~REVERSE;
	save_spot point;
	to_begin_line();
	for (;; cnt++) {
		here = point;
		res = do_color_searching(flags, pat);
		if (res && lines_between(matchstart, matchend, 0)) {
			point = here;	// Must use slow version instead.
			cnt += slow_drop_matching_lines(flags, pat, drop);
			res = 1;
			break;
		}
		if (res != 1 || point == size() - narrow_end)
			break;
		there = give_begin_line();
		nl_forward();
		delete(there, drop ? point : here);
	}
	if (!res && !drop)
		delete(here, size());
	if (!res || res == 1) {
		if (show_status)
			say("Deleted %s%d line%s.", drop ? "" : "all but ", cnt,
			   cnt == 1 ? "" : "s");
	} else if (res == DSBAD)
		error("Bad search pattern: %s", pat);
	return cnt;
}

command file_query_replace() on reg_tab[FSHIFT(7)]
{
	prompt_file_replace(QUERY | UNKNOWN, has_arg);
}

prompt_file_replace(flags, inbufs)
{		// prompt for replacing in files/buffers, then do it
	char *str, with[FNAMELEN], filepat[FNAMELEN];
	char old_cd[FNAMELEN], buf_dir[FNAMELEN];
	jmp_buf *old_level = top_level, this_level;

	flags = get_replace_text(flags | (inbufs ? BUFGREP : FILEGREP),
				 with);
	if (flags == -1 || !get_grep_file(inbufs, filepat))
		return;
	str = default_search_string(flags);
	top_level = &this_level;
	getcd(old_cd);
	if (grep_default_directory == 3 && get_buffer_directory(buf_dir))
		chdir(buf_dir);
	getcd(buf_dir);
	if (!setjmp(top_level))
		do_file_replace(str, with, flags, filepat, inbufs);
	top_level = old_level;
	if (grep_default_directory == 3)
		chdir(old_cd);
}

do_file_replace(str, with, flags, filepat, inbufs)
char *str, *with, *filepat;	// (query-) replace across files/buffers
{
	int res, oldbuf = 1, prevbuf = 0, total = 0, cnt = 0;
	char rel[FNAMELEN], abs[FNAMELEN], *s, *(*finder)(), msg[FNAMELEN];
	int total_found = 0, total_changed = 0, total_readonly = 0, old_readonly;

	save_var synchronous_file_io = 1;
	old_readonly = readonly_warning;
	save_var readonly_warning;			// Don't show msg or ring bell
	readonly_warning &= ~(ROWARN_MSG | ROWARN_BELL); // unless matches.
	save_var abort_file_matching = ABORT_ERROR;
	save_var bufnum, window_bufnum;
	finder = give_grep_finder(inbufs, filepat, &s);	// Get 1st match
	for (; s; s = (*finder)(filepat, 0)) {
		grep_make_names(rel, abs, s, inbufs);
		if (inbufs)
			bufname = s;
		else if (!ok_grep_file_match(s, abs))
			continue;
		noteput("Searching %s...", rel);
		if (!inbufs) {
			oldbuf = look_file(abs);
			if (find_it(abs, default_translation_type) == EREADABORT)
				user_abort = 1;
		} else
			to_buffer(s);
		undo_mainloop();
		if (!user_abort) {
			res = one_file_replace(str, with, flags, oldbuf, old_readonly);
			total_found += replace_num_found;
			total_changed += replace_num_changed;
			total++;
			if (res == -2)
				total_readonly++;
			else if (res >= 0)
				cnt++;
		} else {
			res = DSABORT;
			user_abort = 0;
		}
		if (prevbuf && buf_in_window(prevbuf) == -1) {
			buf_delete(prevbuf);	// delete untouched new bufs
			prev_forget(_prev_num - 1); // fix default
		}
		prevbuf = (oldbuf || modified) ? 0 : bufnum;
		if (res == 0 && (flags & QUERY) && key == '!'
			&& ask_continue_replace("Replace without asking"))
			flags &= ~QUERY;	// no more querying
		else if (res == DSABORT || res == 1 && ungot_key != -1) {
			if (!ask_continue_replace("Continue replacing"))
				break;
		} else if (res > 0) {	// User said to stop, or bad pattern.
			if (res == DSBAD)
				show_replace(str, with, flags | BAD);
			user_abort = 0;
			return;
		}
		ungot_key = -1;
		if (user_abort)
			break;
	}
	ungot_key = -1;
	user_abort = 0;
	sprintf(msg, "Replaced %d of %d matches in %d of %d %s",
	    total_changed, total_found, cnt, total,
	    inbufs ? "buffers" : "files");
	if (total_readonly)
		sprintf(msg + strlen(msg), ", %d read-only, see %s buffer",
				total_readonly, MSG_BUFFER);
	say("%s.", msg);
	restore_vars();
	if (prevbuf && buf_in_window(prevbuf) == -1) {
		buf_delete(prevbuf);	// delete untouched new bufs
		prev_forget(_prev_num - 1); // fix default
	}
}

ask_continue_replace(pr)	// ask question, return 1 if default of yes,
char *pr;			// or 0 if user pressed N
{
	char msg[FNAMELEN];

	drop_pending_says();
	sprintf(msg, "%s in remaining files? ", pr);
	return ask_yn(pr, msg, "Yes", "No", 1);
}

one_file_replace(str, with, flags, oldbuf, old_readonly)
char *str, *with;	// return 0=ok, 1=user said to stop,
{			// DSABORT=aborted, DSBAD=bad pat, -1=no matches, -2=readonly
	int res;

	replace_num_found = replace_num_changed = 0;
	save_var point = (flags & REVERSE) ? size() : 0;
	res = do_color_searching(flags, str);
	if (res != 1)
		return res ? res : -1;
	if (!oldbuf && (flags & FILEGREP))	// had to read file to search
		replace_in_readonly_hook(old_readonly);
	else
		replace_in_existing_hook(old_readonly);
	if (_buf_readonly) {
		if ((flags & FILEGREP) && file_info.check_type == CHECK_FILE
			  && (file_info.attr & ATTR_READONLY))
			show_text(0, 1, "File %r is read-only.", filename);
		else
			show_text(0, 1, "Buffer %s is read-only.", bufname);
		return -2;
	}
	point = matchstart + ((flags & REVERSE) ? 1 : -1);
	show_replace(str, with, flags);
	return string_replace(str, with, flags);
}

replace_in_readonly_hook(old_readonly)	// called when file-query-replace
{			// needs to modify a file that might be read-only
	if (file_info.check_type == CHECK_FILE
				&& (file_info.attr & ATTR_READONLY)) {
		save_var readonly_warning = old_readonly & ~ROWARN_MSG;
		do_readonly_warning();
		make_mode();
	}
}

// Epsilon wants to replace in an existing buffer.  (Might be readonly.)

replace_in_existing_hook(old_readonly)
{
	old_readonly = old_readonly;		// prevent compiler warning
}
