/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1996, 2005 Lugaru Software Ltd.  All rights reserved.  *
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

#include "eel.h"
#include "proc.h"
#include "colcode.h"
#include "misclang.h"

// Move past a string in PostScript.

fwd_postscript_string()
{
	int level = 0;

	while (re_search(1, "[\\()]")) {
		switch (character(point - 1)) {
			case '\\':		// Quote next char.
				point++;
				break;
			case '(':		// Ignore balanced parens within.
				level++;
				break;
			case ')':
				if (--level <= 0)
					return 1;
				break;
		}
	}
	return 0;
}

color_postscript_range(from, to)
{
	int t = -1, talk, r, talk_at = 0;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend, case_fold = 1;
	set_character_color(from, to, -1);
	point = from;
	to_begin_line();	// Move back over continued lines.
	point -= parse_string(-1, "(.*\\\n)+");
	while (point > narrow_start
		   && get_tagged_region("color-as-unit", point - 1, &r) != -1)
		point = r;		// Move before binary blocks.
	talk = (to - point > 1000000);	// Show status during long delays.
	while (point < to) {
		if (!re_search(1, "[%(]|/[-_a-z0-9]+|</word>(doclutimage|doNimage)</word>")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(t)) {		// check first char
			case '%':
				nl_forward();
				set_character_color(t, point, color_class postscript_comment);
				if (parse_string(-1, "^%%%%BeginData:?\n")) {
					r = point;
					if (search(1, "\n%%EndData"))
						point = matchstart;
					set_character_color(r, point, color_class postscript_text);
					set_tagged_region("color-as-unit", t, point, 1);
				}
				break;
			case '(':		// A text string.
				point--;
				fwd_postscript_string();
				set_character_color(t, point, color_class postscript_text);
				break;
			case '/':		// Found /literalname.
				set_character_color(t, point, color_class postscript_literal);
				break;
			default:		// Found start of Windows PS driver binary code.
				re_search(1, "~>[ \n]*(Z|endimage)");	// Find end.
				set_character_color(t, point, color_class postscript_text);
				set_tagged_region("color-as-unit", t, point, 1);
				break;
		}
		if (talk && point > talk_at + 20000) {
			note("Coloring PostScript: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	if (talk_at)
		note("");
	return point;
}

command postscript_mode()
{
	mode_default_settings();
	major_mode = _postscript_mode_name;
	mode_keys = pscript_tab;		/* Use these keys. */
	if (auto_show_postscript_delimiters)
		auto_show_matching_characters = postscript_auto_show_delim_chars;
	strcpy(comment_start, "%%[ \t]*");
	strcpy(comment_pattern, "%%.*$");
	strcpy(comment_begin, "% ");
	strcpy(comment_end, "");
	comment_column = 0;
	recolor_range = color_postscript_range;	// set up coloring rules
	recolor_from_here = recolor_from_top;
	coloring_flags |= COLOR_INVALIDATE_FORWARD | COLOR_INVALIDATE_RESETS;
	idle_coloring_size = 10000;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	buffer_maybe_break_line = generic_maybe_break_line;
	fill_mode = misc_language_fill_mode;
	try_calling("postscript-mode-hook");
	make_mode();
}

when_loading()
{
	pscript_tab[ALT('q')] = (short) fill_comment;
}

suffix_ps()
{
	postscript_mode();
}

suffix_eps()
{
	postscript_mode();
}

color_conf_range(from, to)
{
	int t = -1;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	point = to;				// Color entire lines.
	nl_forward();
	to = point;
	save_var narrow_end = size() - to;
	set_character_color(from, to, -1);
	point = from;

	while (point < to) {
		if (!re_search(1, "([^ \t\n#:;,]+[ \t]*)="
					   "([ \t]*([^ \t\n#:;,]+|'[^\n']*'|\"[^\n']*\"))|^;|[#]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(point - 1)) {		// check last char
			case '#':
			case ';':
				nl_forward();
				set_character_color(t, point, color_class conf_comment);
				break;
			default:
				set_character_color(t, find_group(1, 0),
									color_class conf_attribute);
				set_character_color(find_group(2, 1), point,
									color_class conf_value);
				break;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	return point;
}

command conf_mode()
{
	mode_default_settings();
	major_mode = _conf_mode_name;
	mode_keys = conf_tab;		/* Use these keys. */
	strcpy(comment_start, "#[ \t]*");
	strcpy(comment_pattern, "#.*$");
	strcpy(comment_begin, "# ");
	strcpy(comment_end, "");
	comment_column = 0;
	recolor_range = color_conf_range;	// set up coloring rules
	recolor_from_here = recolor_by_lines;
	idle_coloring_size = 10000;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	buffer_maybe_break_line = generic_maybe_break_line;
	fill_mode = misc_language_fill_mode;
	try_calling("conf-mode-hook");
	make_mode();
}

when_loading()
{
	conf_tab[ALT('q')] = (short) fill_comment;
}

suffix_conf()
{
	conf_mode();
}

// Return code for word from here to point (something with alpha or
// digits).  Return for 1 for command, 2 for comment, 0 if unknown.
is_batch_keyword(int from)
{
	char buf[500], *s;

	save_var case_fold = 1;
	if (point - from > sizeof(buf) / sizeof(char) - 10)
		save_var point = from + sizeof(buf) / sizeof(char) - 10;
	buf[0] = '|';		// get identifier, between | chars
	grab(from, point, buf + 1);
	if ((s = strchr(buf + 1, '.')) && matches_at(from, -1, "\n[ \t]*"))
		*s = 0;
	strcat(buf, "|");
	if (strstr("|rem|", buf) && !matches_at(from, -1, "[^ \n\t@][@ \t]*"))
		return 2;
	if (strstr("|assoc|at|attrib|break|cacls|call|cd|chcp|chdir"
			   "|chkdsk|chkntfs|cls|cmd|cmdextversion|color|comp"
			   "|compact|convert|copy|date|defined|del|dir|diskcomp"
			   "|diskcopy|do|doskey|echo|else|endlocal|equ|erase"
			   "|errorlevel|exist|exit|fc|find|findstr|for|format"
			   "|ftype|geq|goto|graftabl|gtr|help|if|in|label|leq"
			   "|lss|md|mkdir|mode|more|move|neq|not|path|pause"
			   "|popd|print|prompt|pushd|rd|recover|ren|rename"
			   "|replace|rmdir|set|setlocal|shift|sort|start|subst"
			   "|time|title|tree|type|ver|verify|vol|xcopy|cmd"
			   "|command|", buf))
			return 1;
	return 0;
}

color_batch_range(from, to)
{
	int t = -1;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend, case_fold = 1;
	point = to;				// Color entire lines.
	nl_forward();
	to = point;
	save_var narrow_end = size() - to;
	set_character_color(from, to, -1);
	point = from;

	while (point < to) {
		if (!re_search(1, "[-a-z0-9_.%/\\]+|[<>|]|^::")) {
			t = size();
			break;
		}
		t = matchstart;
		if (strchr("<>|", character(point - 1))) {
			set_character_color(t, point, color_class batch_punctuation);
			continue;
		} else if (character(point - 1) == ':') {
			nl_forward();
			set_character_color(t, point, color_class batch_comment);
			continue;
		}
		switch (is_batch_keyword(t)) {
			case 1:
				set_character_color(t, point, color_class batch_keyword);
				break;
			case 2:
				nl_forward();
				set_character_color(t, point, color_class batch_comment);
				break;
			default:
				set_character_color(t, point, -1);
				break;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	return point;
}

command batch_mode()
{
	mode_default_settings();
	major_mode = _batch_mode_name;
	mode_keys = batch_tab;		/* Use these keys. */
	strcpy(comment_start, "rem[ \t]*");
	strcpy(comment_pattern, "rem.*$");
	strcpy(comment_begin, "rem ");
	strcpy(comment_end, "");
	comment_column = 0;
	recolor_range = color_batch_range;	// set up coloring rules
	recolor_from_here = recolor_by_lines;
	idle_coloring_size = 10000;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	buffer_maybe_break_line = generic_maybe_break_line;
	fill_mode = misc_language_fill_mode;
	try_calling("batch-mode-hook");
	make_mode();
}

when_loading()
{
	batch_tab[ALT('q')] = (short) fill_comment;
}

suffix_bat()
{
	batch_mode();
}

suffix_cmd()
{
	batch_mode();
}

suffix_btm()
{
	batch_mode();
}
