/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2004 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 12.00 to 12.01.

#include "eel.h"
#include "proc.h"
#include "html.h"
#include "c.h"
#include "colcode.h"
#include "vbasic.h"
#include "perl.h"
#include "filter.h"
#include "tags.h"
#include "hex.h"
#include "epsnet.h"
#include "proc.h"
#include "bufed.h"
#include "kill.h"
#include "session.h"
#include "dired.h"
#include "grep.h"
#include "vhdl.h"
#include "misclang.h"
#include "tex.h"
#include "browse.h"
#include "menu.h"
#include "info.h"

command goto_line() on cx_tab['g']
{
	char res[FNAMELEN], *p;
	int num = iter;

	got_bad_number = iter = 0;
	if (has_arg) {
		go_line(num);
		return;
	}
	get_string(res, "Go to line: ");
	if (tolower(*res) == 'p') {
		num = numtoi(res + 1);
		if (got_bad_number)
			error("Bad position %s.", res + 1);
		point = num;
		return;
	}
	num = numtoi(res);
	if (!got_bad_number) {
		if (strchr(res, '%')) {
			point = muldiv(size(), num, 100);	// handles intermediates > 32 bits
			if (num)
				nl_forward();
		} else
			go_line(num);
	}
	p = strchr(res, ':');		// a number after : or . is a column
	if (!p)
		p = strchr(res, '.');
	if (p) {
		num = numtoi(p + 1);
		if (!got_bad_number)
			to_column_and_center(num);
	}
}

ask_yn(char *title, char *question, char *yes_button,
	   char *no_button, int def_button)
{
	if (is_gui && want_gui_prompts) {
		show_text(-2, 1, "%s", question);
		return button_dialog(title, question, yes_button, no_button, "",
							 def_button) == 1;
	} else {
		if (windows_state() & WINSTATE_MINIMIZED)
			windows_restore();
		for (;;) {
			noteput("%s%s", question,
					def_button == 1 ? "[Y] " : def_button == 2 ? "[N] " : "");
			getkey();
			if (user_abort)
				return 0;
			switch (toupper(key)) {
				case 'Y': say(""); return 1;
				case 'N': say(""); return 0;
			}
			if (IS_ENTER_KEY(key) && def_button) {
				say("");
				return def_button == 1;
			}
			refresh();
		}
	}
}

// Ask a 3-way question.  The first letter of each answer is its
// shortcut key; put & before another letter to use it instead.  Title
// and question are for GUI, prompt is for others.  Return 1, 2, 3, or 0
// if aborted.
ask_3way(char *title, char *question, char *prompt,
		 char *a1, char *a2, char *a3, int def)
{
	char msg[FNAMELEN], keys[10];
	int res;

	keys[0] = strchr(a1, '&') ? strchr(a1, '&')[1] : a1[0];
	keys[1] = strchr(a2, '&') ? strchr(a2, '&')[1] : a2[0];
	keys[2] = strchr(a3, '&') ? strchr(a3, '&')[1] : a3[0];
	keys[3] = 0;
	if (is_gui && want_gui_prompts) {
		show_text(-2, 1, "%s", prompt);
		res = button_dialog(title, question, a1, a2, a3, def);
		if (user_abort)
			res = 0;
		user_abort = 0;
	} else {
		*msg = 0;
		if (def >= 1 && def <= 3)
			sprintf(msg, "[%c] ", keys[def - 1]);
		for (;;) {
			noteput("%s%s", prompt, msg);
			getkey();
			if (user_abort) {
				user_abort = 0;
				return 0;
			}
			if (strchr(keys, toupper(key))) {
				say("");
				return strchr(keys, toupper(key)) + 1 - keys;
			}
			if (IS_ENTER_KEY(key) && def) {
				say("");
				return def;
			}
			refresh();
		}
	}
	return res;
}

command quoted_insert();

short *ask_key(char *pr, char *keyname, int ask_if_ambiguous)
{
	int index, k, fk;
	short *tab, *alttab;
	char msg[FNAMELEN], *p;

	tab = mode_keys;
	alttab = root_keys;
	*keyname = 0;			// put readable form of key here
	key_binding[0] = 1;		// also save sequence in this global
	for (;;) {
		sprintf(msg, "%s%s", pr, keyname);
		full_getkey(msg, 0);
		if (root_keys[key] == (short) quoted_insert) {
			mention("%s<Quote> ", msg);
			getkey();
			user_abort = 0;
		}
		check_abort();
		p = keyname + strlen(keyname);
		k = full_key;
		index = (tab[k] == 0) ? alttab[k] : tab[k];
		if (index <= 0) {
			k = key;
			index = (tab[k] == 0) ? alttab[k] : tab[k];
		}
		show_char(keyname, ask_if_ambiguous ? k : full_key, 1);
		key_binding[key_binding[0]] = k; // add to constructed key
		key_binding[0]++;
		if (index <= 0 || name_type(index) != NT_TABLE)
			break;
		alttab = index_table((tab[k] && !alttab[k]) ?
							 tab[k] : alttab[k]);
		tab = index_table(index);
		if (!alttab)
			alttab = tab;
		strcat(keyname, " ");
	}
	say("");
	if (ask_if_ambiguous && full_key != key) {
		k = key;
		fk = full_key;
		sprintf(msg, "Bind only specific key %k (not general key %k)? ",
				fk, k);
		if (ask_yn("Specific key or general?", msg, "Yes", "No", 2)) {
			key = k = fk;
			*p = 0;
			show_char(keyname, k, 1);
		}
	} else if (full_key != key && k == full_key)
		key = k;
	return ((tab[k] == 0) ? alttab : tab) + k;
}

int strtoi(char *s, int base)
{	/* convert a string of digits to a number */
	char *p, *digits = "0123456789ABCDEF";
	int number, neg = 0;

	got_bad_number = 0;
	while (*s == '-')
		s++, neg = !neg;
	number = 0;
	if (!*s || !(p = index(digits, toupper(*s))) || p - digits >= base)
		got_bad_number = 1;
	while (*s && (p = index(digits, toupper(*s++))) && p - digits < base)
		number = number * base + (p - digits);
	return neg ? -number : number;
}

command brief_keyboard()
{
	load_keyboard_file("brief.kbd", "Brief");
	virtual_space.default = 2;	// Brief always uses virtual space
	display_column.default = 0;	// and horiz scrolling
	virtual_space = 2;		// Also set for current buffer & window
	display_column = 0;
	mark_rectangle_expands = 1;
	resize_rectangle_on_tab = 0;
	mark_unhighlights = 1;
	yank_rectangle_to_corner = 3;
	yank_line_retains_position = 1;
	typing_deletes_highlight = 2;	// Only at prompts.
	typing_hides_highlight = 0;
	search_positions_at_start = 1;
	search_in_region = 1;
	replace_in_region = 2;
	recall_prior_response_options = 3;	// At prompts, display old responses in-place.
	recall_response_selected = 1;	// Highlight such responses.
	search_defaults_from = 2;
	abort_key = ESC;
	strcpy(menu_file, "brief.mnu");
	rebuild_menu();
	say("Brief keyboard loaded.");
}

command cua_keyboard()
{
	load_keyboard_file("cua.kbd", "CUA");
	abort_key = ESC;
	alt_invokes_menu = 1;
	fwd_search_key = CTRL('F');
	yank_rectangle_to_corner = 1;
	yank_line_retains_position = 0;
	typing_deletes_highlight = 1;
	typing_hides_highlight = 1;
	search_positions_at_start = 0;
	search_in_region = 0;
	replace_in_region = 0;
	recall_prior_response_options = 0;
	recall_response_selected = 0;
	search_defaults_from = 1;
	strcpy(gui_menu_file, "cua.mnu");
	rebuild_menu();
	say("CUA keyboard loaded.");
}

command epsilon_keyboard()
{
	load_keyboard_file("epsilon.kbd", "Epsilon");
	virtual_space.default = 0;	// set back to factory defaults
	virtual_space = 0;		// Also set for current buffer & window
	mark_unhighlights = 0;
	mark_rectangle_expands = 0;
	resize_rectangle_on_tab = 0;
	typing_deletes_highlight = 1;
	abort_key = CTRL('G');
	yank_rectangle_to_corner = 1;
	yank_line_retains_position = 0;
	typing_deletes_highlight = 1;
	typing_hides_highlight = 1;
	search_positions_at_start = 0;
	search_in_region = 0;
	replace_in_region = 0;
	recall_prior_response_options = 0;
	recall_response_selected = 0;
	search_defaults_from = 1;
	strcpy(menu_file, "epsilon.mnu");
	rebuild_menu();
	say("Epsilon keyboard loaded.");
}

bufed_getname(s)
char *s;
{
	int start, oline;

	*s = 0;
	save_var abort_searching = 0;
	move_to_column(BNAME_COL - 1);
	start = point;
	if (character(point) == '"') {
		oline = lines_between(0, point, 0);
		start = ++point;
		if (!search(1, "\"") || oline != lines_between(0, point, 0))
			return 0;
		point--;
	} else {
		if (character(start) == ' ')
			start++;
		move_to_column(BNAME_COL + bufed_column_width);
		if (character(point - 1) != ' ')
			if (!search(1, " "))
				return 0;
		if (!re_search(-1, " *"))
			return 0;
	}
	grab(start, point, s);
	to_bufed_column();
	return 1;
}

list_line(inbuffer)	/* insert a line into the buffer list */
char *inbuffer;
{
	char file[FNAMELEN], msg[FNAMELEN * 2], bname[FNAMELEN * 2];
	int width;

	strcpy(bname, bufname);
	quote_bufname(bname);
	if (force_save_as == FSA_NEWFILE)
		*file = 0;
	else if (bufed_show_absolute_path) {
		strcpy(file, filename);
		if (bufed_show_absolute_path == 2)
			convert_to_8_3_filename(file);
	} else
		relative(filename, file);
	width = sprintf(msg, "%10d%9d %c%c%-*s %s\n",
					buf_accessed, size(), modified ? '*' : ' ', 
					_buf_readonly ? 'R' : ' ', bufed_column_width,
					bname, file);
	bufname = inbuffer;
	stuff(msg);
	width = width - ACCESS_WIDTH + 2;
	if (width > bufed_width)
		bufed_width = width;
}

sort_buffer_list(opos, buf)
char *buf;		/* sort from opos to the end, & return new end */
{
	int col, want_sort = 1, rev = 0;

	switch (tolower(_bufed_sort)) {
		case 'f':	col = BNAME_COL + ACCESS_WIDTH + bufed_column_width; break;
		case 'b':	col = BNAME_COL + ACCESS_WIDTH; break;
		case 'i':	col = ACCESS_WIDTH; break;
		case 'a':	col = 0; rev = 1; break;
		default:	want_sort = 0; break;
	}
	if (isupper(_bufed_sort))
		rev = !rev;
	save_var bufnum;
	bufname = buf;
	if (want_sort)
		do_sort_region(opos, size(), col, rev);
	save_var narrow_start = opos;
	strip_columns(0, ACCESS_WIDTH);		// remove access timestamps
	return size();
}

int bufed_when_displaying();

do_bufed(flags)
{
	char find[FNAMELEN * 2], msg[FNAMELEN], title[FNAMELEN];
	int count;

	strcpy(title, " Size     Buffer");
	bufed_column_width = MAX(MIN(bufed_column_width, 200), 10);
	for (count = 6; count < bufed_column_width; count++)
		strcat(title, " ");
	strcat(title, "File ");
	if (in_bufed()) {
		bufnum = window_bufnum;
		set_read_only(0);
		if (bufed_getname(find)) {
			quote_bufname(find);
			strcat(find, " ");
		} else
			*find = 0;
		bufed_restore_orig();
		make_buffer_list(flags, BUFED_BUF);
		window_handle = _bufed_window;
		window_bufnum = name_to_bufnum(BUFED_BUF);
		point = 0;
	} else {
		_bufed_orig_window = window_handle;
		_bufed_orig_buffer = window_bufnum;
		_bufed_orig_highlight =_highlight_control;
		strcpy(find, bufname);
		quote_bufname(find);
		strcat(find, " ");
		make_buffer_list(flags, BUFED_BUF);
		build_first = 1;
		_bufed_window = add_popup(2, 2, bufed_width, screen_lines - 6,
			ALL_BORD(BSINGLE), name_to_bufnum(BUFED_BUF));
		if (is_gui && want_gui_prompts) {
			count = lines_between(0, size(), 0) + 1;
			mouse_cursor = &left_pointer;
			_bufed_caption_window = add_popup(0, 0, bufed_width, 1, 0,
									  create(BUFED_CAPTION_BUF));
			buf_zap(bufnum);
			bprintf("   %s", title);
			point = 0;
			system_window = 1;
			undo_size = 0;
			mouse_cursor = &left_pointer;
			window_handle = _bufed_window;

			if (flags & BUFED_EXIT) {
				sprintf(msg, "%d buffer%s not saved.  Exit?", count,
						(count == 1) ? "" : "s");
				display_dialog_box("AskExitBox", msg,
								   _bufed_caption_window, _bufed_window, 0,
								   "E&xit", "&Cancel", "");
			} else if (flags & BUFED_SAVE) {
				sprintf(msg, "Save %d unsaved buffers first? "
						"(! to never ask)", count);
				display_dialog_box("AskSaveBox", msg,
								   _bufed_caption_window, _bufed_window, 0,
								   "&Save", "&No Save", "&Cancel");
			} else if (flags & BUFED_SWITCH)
				display_dialog_box("SwitchBufBox", "Select a Buffer",
								   _bufed_caption_window, _bufed_window, 0,
								   "OK", "Cancel", "Help");
			else
				display_dialog_box("CaptionBox", "Select a Buffer",
								   _bufed_caption_window, _bufed_window, 0,
								   "OK", "Cancel", "Help");
		}
		window_handle = _bufed_window;
	}
	set_menu_look(_bufed_window);
	if (flags & (BUFED_EXIT | BUFED_SAVE))
		remove_line_highlight();
	else
		make_line_highlight();
	window_title(_bufed_window, BTOP, TITLELEFT(4), title);
	undo_size = 0;
	display_column = 0;	/* horizontally scroll this window */
	popup_near_window(_bufed_window, _bufed_orig_window);
	bufed_mode();
	if (*find)
		col_search(find, BNAME_COL - 1);
	to_bufed_column();
	modified = 0;
	goal_column = BNAME_COL;
	when_displaying = bufed_when_displaying;
}

bufed_ask_exit()
{
	char save_it[FNAMELEN];

	if (windows_state() & WINSTATE_MINIMIZED)
		windows_restore();
	if (is_gui && want_gui_prompts) {
		tiled_only();
		save_var _doing_input = DI_VIEW;
		save_var bufnum, window_handle;
		save_var return_raw_buttons = 1;
		for (;;) {
			do_bufed(BUFED_EXIT | BUFED_ONLY_UNSAVED);
			go_to_link = 0;
			view_loop(_bufed_window);
			bufed_getname(save_it);
			remove_bufed();
			if (user_abort) {
				user_abort = 0;
				return 0;
			}
			switch (key_is_button) {
				case 1:	key = 'Y'; break;
				case 2: key = 'N'; break;
				case 5: key = 'S'; break;
				}
			switch (toupper(key)) {
				case 'X':
				case 'Y':	return 1;
				case 'E':
				case '\n':
				case '\r':	if (exist(save_it))
								to_buffer(save_it);
							return 0;
				case 'C':
				case 'N':	return 0;
				case 'S':	if (*save_it && exist(save_it))
								bufed_save_one(save_it);
							if (!unsaved_buffers())
								return 1;
							break;
			}
		}
	} else {
		do_bufed(BUFED_ONLY_UNSAVED);
		maybe_refresh();
		sayput("Buffers not saved.  Exit? [N] ");
		do {
			maybe_refresh();
			getkey();
		} while (!IS_TRUE_KEY(key));
		say("");
		switch (toupper(key)) {
			case 'X':
			case 'Y':	break;
			case 'N':
			case 'C':
			case '\n':
			case '\r':	bufed_msg();
					return 0;
			default:	check_abort();
					ungot_key = key;
					return 0;
		}
		remove_bufed();
		maybe_refresh();
	}
	return 1;
}

command show_point() on cx_tab['=']
{
	char msg[300], cname[300];
	int sz = MAX(point - mark, mark - point), c;

	if (point < size()) {
		sprintf(msg, "Column %d, char %d of %d is ", virtual_column(),
				point, size());
		if (c = curchar())
			sprintf(msg + strlen(msg),
				"'%c'=%d decimal=%02x hex", c, c, c);
		else
			sprintf(msg + strlen(msg), "'\\0'");
		if (is_highlight_on())
			sprintf(msg + strlen(msg), ", %d char%s in region",
					sz, (sz != 1) ? "s" : "");
		else if (c > 127) {
			*cname = 0;
			show_char(cname, c, 1);	// Show full name first, if it all won't fit.
			if (strlen(msg) + strlen(cname) + 2 >= screen_cols)
				show_text(0, MAX(see_delay, 200), "'%c'=%s", c, cname);
			sprintf(msg + strlen(msg), ", %s", cname);
		}
	} else
		sprintf(msg, "Column %d, at end of %d chars",
				virtual_column(), size());
	say("%s", msg);
	expire_message = 1;
	iter = 0;
}

char *preserve_buffer_type(char *buf)
{
	struct scp_job *p;
	save_var bufname = buf;

	for (p = scp_job; p; p = p->next)
		if (p->buf == bufnum)
			return p->busy ? "Scp" : 0;
	if (is_process_buffer(bufnum) || can_send_invisible())
		return strcmp(major_mode, "Fundamental") ? major_mode : "Process";
	return 0;
}

minimize_buf_accessed()
{
	int i, tmp = tmp_buf();
	int nums[2];

	save_var bufnum, abort_searching = 0;
	i = buf_list(0, 2);
	do {
		bufnum = i;
		if (buf_accessed)
			buf_printf(tmp, "%09d %d\n", buf_accessed, i);
	} while (i = buf_list(-1, 1));
	if (buf_accessed_clock <= buf_accessed)
		buf_accessed_clock = buf_accessed + 1;
	bufnum = tmp;
	do_sort_region(0, size(), 0, 0);
	point = 0;
	buf_accessed_clock = 1;
	while (grab_numbers(tmp, nums) == 2) {
		bufnum = nums[1];
		buf_accessed = buf_accessed_clock++;
	}
	restore_vars();
	buf_delete(tmp);
}

int c_smart_move_level(int dir, int comm, int stop_on_key)
{
	char delims[100], *pat = (dir > 0) ? "])}[({*/" : "[({])}*/";
	int c, i, depth = 0, cnt = 0;
	int in_block_comment = (comm == IN_OLD_COMMENT) ? comm : 0;

	while (re_search(dir, "[])}[({]|/%*|%*/")) {
		if (stop_on_key && (cnt++ % 20 == 0)
			  && (char_avail() || is_key_repeating()))
			return -2;		// Give up search now?
		c = character(point - (dir > 0));
		if (c == '/' && !in_c_comment(c_look_back))
			in_block_comment = 0;	// Leaving a block comment.
		else if (c == '*' && in_c_comment(c_look_back) == IN_OLD_COMMENT)
			in_block_comment = IN_OLD_COMMENT;	// Entering block comment.
		i = strchr(pat, c) - pat;
		if (i >= 6)	// Block comment chars, not delimiters.
			continue;
		if ((in_block_comment ? in_block_comment : in_c_comment(0)) != comm)
			if (comm)			// If we started in comment/string,
				return -1;		// stop if we ever search outside it.
			else				// We didn't start in comment/string, so
				continue;		// skip over delims in comments/strings.
		if (i < 3) {		// Closing delim.
			if (depth < 1 || c != delims[--depth])
				return -1;	// This didn't match what we expected.
			if (!depth)		// We found a match, success.
				return point;
		} else if (depth < sizeof(delims) / sizeof(*delims))
			delims[depth++] = pat[i - 3];	// Opening delim, record it.
		else		// Nesting too deep, we must have gotten lost, give up.
			return -1;
	}
	return -1;
}

int find_c_class_info(int start, char *type, char *class, char *func)
{
	char this[FNAMELEN], last[FNAMELEN];
	int class_start = point, i;

	point--;	// See if "class foo" is before this.
	strcpy(last, "");
	for (i = 20; i-- > 0; ) {	// Look through some words before the {.
		if (cind_try_space() || give_begin_line() > start)
			return 0;
		if (strchr("<>:,", character(point - 1)))
			point--;
		else {
			if (!parse_string(-1, (c_extra_keywords
								   & (CSHARP_KEYWORDS | JAVA_KEYWORDS))
							  ? "[a-zA-Z0-9_.]+" : "[a-zA-Z0-9_:]+", this))
				return 0;
			if (!strcmp(this, "class") || !strcmp(this, "struct")
				  || !strcmp(this, "union") || !strcmp(this, "namespace")
				  || !strcmp(this, "enum")
				  || (c_extra_keywords == CSHARP_KEYWORDS) && !strcmp(this, "interface")) {
				if (class) {
					if (*class && ((display_c_options & 1)
								   || strcmp(type, "namespace"))) {
						strcat(class, (c_extra_keywords == CSHARP_KEYWORDS) ? "." : "::");
						class += strlen(class);
					}
					strcpy(class, last);	// Is place where we started within func?
				}
				if (type)
					strcpy(type, this);		// Word to the right of type is class.
				if (find_inner_func_name(start, class_start, func))
					return CF_INFO_TYPE | CF_INFO_CLASS | CF_INFO_FUNC;
				if (func)
					*func = 0;
				return CF_INFO_TYPE | CF_INFO_CLASS;
			}
			strcpy(last, this);		// Go on to the next word.
		}
	}
	return 0;		// Couldn't recognize anything here.
}

c_compute_indent()
{
	int ind = 0;		/* indentation to use */
	int orig = point;
	int first_char = curchar(); /* cur line's first char */
	int pstart;		/* start position of prev line */
	int pafter;		/* after label on prev line */
	int pend;		/* end position of prev line */
	int prev_end = ';';	/* last char on previous line */
	int norm_narrow;	/* narrowing setting for most parsing */

	save_var point;
	if (setjmp(&c_indent_giveup))	/* if we hit end & gave up, */
		return ind;		/* use best guess so far */
	if (in_perl_buffer)
		return perl_compute_indent();
	norm_narrow = fix_narrowing(point - c_look_back);
	save_var narrow_start = norm_narrow;
	if (!reindent_c_comments && parse_string(1, "[ \t\f]*/<*>"))
		return get_indentation(point);
	if (!reindent_one_line_c_comments && parse_string(1, "[ \t\f]*//"))
		return get_indentation(point);
	if (!reindent_c_preprocessor_lines && parse_string(1, "[ \t\f]*#"))
		return get_indentation(point);
	if (first_char == '#' || first_char == '\f')
		return 0;
	else if (first_char == '}')
		return cind_close();

	to_begin_line();
	if (character(point - 2) == '\\')
		if (!indent_preprocessor_contin)
			return indent_after_backslash();
		else if (!prior_contin_code(orig))
			return indent_preprocessor_contin;
		else {	// Note we're indenting preprocessor continuation line now.
			save_var indent_preprocessor_contin = 255;
			save_var narrow_start = point - parse_string(-1, "(\n.*\\)*\n");
		}
	if (parse_string(1, "[ \t]*\n") && after_many_comments()) {
		re_search(-1, "[^ \t\n]");	/* find previous non-blank line */
		return get_indentation(point);
	}
	cind_space();			/* skip whitespace and comments */
	if (character(point - 1) == ':')
		move_before_labels();
	prev_end = character(point - 1); /* get last relevant char */
	pend = point;

	to_statement_start();		/* if contin line, back up */
	pstart = point;			/* now at start of biggest statement */
	past_labels();			/* containing original point */
	pafter = point;			/* just after label */
	ind = current_column();
	narrow_start = norm_narrow;	/* ok to look far back again */

	if (is_case(orig)) {
		ind = indent_case(orig);
	} else if (parse_string(1, "ON_")) {
		// Found MFC dispatch table, leave alone.
	} else if (have_label(orig)) {
		char type[FNAMELEN];		// Could it be a bitfield, not a label?
		find_c_func_info(type, NULL, NULL, 0);
		if (!strcmp(type, "struct") || !strcmp(type, "union"))	// Yes, a bitfield.
			ind += get_c_indent();
		else					// It's a label.
			ind = c_label_indent;
	} else if (prev_end == ':' && is_label(pstart)) {
		ind = get_column(pstart);
		if (!(Closeback && just_open(orig)))
			ind += get_c_indent();
		if (just_open(orig))
			ind += c_brace_offset;
	} else if (is_else(orig)) {
		point = orig;
		ind = indent_else(ind);
	} else if (prev_end == ';') {
		if (just_open(orig) && func_start_block(pstart))
			ind = c_top_braces;
		else
			ind = indent_after_statement(orig);
	} else if (prev_end == '}') {
		point = pend;
		ind = indent_after_block(orig);
	} else if (prev_end == '{') {
		if (!c_indent_after_extern_c
			  && parse_string(1, "[ \t]*extern[ \t]*\"C\"[ \t\n]*{"))
			return ind;
		if (!c_indent_after_namespace
			  && parse_string(1, "[ \t]*namespace</word>.*{"))
			return ind;
		if (ind <= c_top_braces	// special case if no previous indent
		    || get_column(pend - 1) <= c_top_braces) {
			ind += top_indent(prev_end) ? get_c_indent() : c_top_braces;
			if (just_open(orig))
				ind += c_brace_offset;
			if (ind < c_top_braces)
				ind = c_top_braces;
		} else if (have_struct_init(pend - 1)) {
			ind = indent_struct_init(pend - 1) + get_c_indent();
		} else if (Closeback || curchar() != '{')
			ind += get_c_indent();
		if (just_open(pafter))
			ind -= c_brace_offset;
	} else if (just_open(orig) || matches_at(orig, 1, "[ \t\f]*{[ \t\f]*}")) {
		if (have_struct_init(orig))
			return indent_struct_init(orig);
		else if (have_other_init(orig))
			return ind;
		else
			return indent_just_open(ind);
	} else if (in_c_initializer(orig)) {
		return ind;
	} else if (is_contin_statement(pend, 1, orig)) {
		return indent_c_contin(ind, orig, pend);
	} else if (!index(":);{}", c_line_before()) && !c_statement_start()
				&& !(Closeback && just_open(point)))
		return ind;
	else
		ind += get_c_indent();
	return ind;
}

indent_case(orig)	// indent case-like statement at orig
{
	char pat[FNAMELEN];

	sprintf(pat, "</word>case</word>|(</word>(default|public|"
			"private|protected%s)[ \t\n]*:)|[{}]",
			(c_extra_keywords & CSHARP_KEYWORDS) ? "|internal" : "");

	point = orig;
	for (;;) {		// look for case-like statement, or {}'s
		cind_search(pat);
		if (in_c_comment(c_look_back))
			continue;
		else if (curchar() != '}')
			break;
		point++;
		if (!c_rev_skip_level('{', '}')) // move before matching {
			longjmp(&c_indent_giveup, 1);
	}
	if (curchar() != '{')	// copy indentation from prior case
		return get_indentation(point);

	cind_space();	// skip over switch-like construct before {
	if (character(point - 1) == ')') { // back past () of switch
		if (!c_rev_skip_level('(', ')'))
			longjmp(&c_indent_giveup, 1);
	} else
		to_statement_start();
	to_indentation();
	past_labels();	// might be "case 3: switch ()", get to switch
	point += parse_string(1, "[ \t\f]*");
	sprintf(pat, "(public|private|protected%s)</word>",
			(c_extra_keywords & CSHARP_KEYWORDS) ? "|internal" : "");
	return current_column() + get_c_indent() +
			(matches_at(orig, 1, pat)
			 ? c_access_spec_offset : c_case_offset);
}

indent_c_contin(ind, orig, lastl)	/* indent continuation line */
{
	int c, plevel = 0, qlevel = 0, findc, alevel = 0;

	if (!c_align_contin_lines)	// never line up
		return c_default_indent(ind, orig, 1);
 	if (c_align_inherit >= 0 && matches_at(orig, -1, ":\n[ \t]*")
 		  && !c_matches_in(give_begin_line(), lastl, "%?"))
 		return get_indentation(point) + c_align_inherit;
	to_begin_line();
	save_var narrow_start = point;
	point = orig;
	findc = character(orig);
	if (c_align_open_paren && parse_string(1, "%(([ \t]|//|/%*)*$"))
		return ind;
	while (re_search(-1, "[()?:\"'\n<>]|[=!<>]?=")) {
		c = curchar();
		if (in_perl_buffer && (in_perl_comment(point) || strchr("\"'\n", c)))
			continue;	// Skip over comments differently.
		switch (c) {
			case '(':	/* indent after unmatched lparen */
				if (plevel-- > 0)
					break;
				if (c_align_open_paren && findc != ')')
					return ind + get_c_indent();
indhere:		if (findc != ')')
					point++;
				if (!parse_string(1, "[ \t\f]*/[*/]"))
					point += parse_string(1, "[ \t\f]*![^ \t\f\n]");
				return limit_c_contin_indent(ind);
			case ')':
				plevel++;
				break;
			case '?':	/* indent : to matching ? */
				if (qlevel-- > 0 || plevel > 0 || findc != ':')
					break;
				point--;
				goto indhere;
			case ':':
				qlevel++;
				break;
			case '=': case '!':
				if (character(point + 1) == '=' || plevel > 0
				    || character(lastl - 1) == ',' || findc == ')')
					break;
				goto indhere;
			case '\n':		 // skip over comments
				cind_try_space(); // at ends of lines
				break;
			case '"': case '\'':
				point++;
				c_skip_item(c, RE_REVERSE);
				break;
			case '<':
				if (character(point + 1) == '=')
					break;
				if (alevel-- > 0 || !(c_extra_keywords & (CPP_KEYWORDS
					| JAVA_KEYWORDS | CSHARP_KEYWORDS)))
					break;
				// Someday, test if there's an identifier before the <
				// that's the name of a template, not just if it's a
				// template keyword.
				if (parse_string(-1, "</word>template[ \t\n]*")) {
					point++;
					point += parse_string(1, "[ \t]*!<word>");
					return limit_c_contin_indent(ind);
				}
				break;
			case '>':
				if (character(point + 1) != '=')
					alevel++;
				break;
		}
	}
	if (c_align_inherit >= 0 && c_matches_in(point, orig, ":[ \t]*($|//|/%*)")
		  && !c_matches_in(point, orig, "%?"))
		return get_indentation(point) + c_align_inherit;

	if (character(lastl - 1) == ',') {
		if (character(lastl - 2) != ')') {
			point = lastl;
			to_begin_line();
		}
		point += parse_string(1, "(.*[ \t\f]on)?[ \t\f]+");
		if (in_c_comment(c_look_back))
			to_indentation();
		return limit_c_contin_indent(ind);
	}
	return c_default_indent(ind, orig, 2);
}

c_find_linked_file(char *file)
{
	char rel[FNAMELEN], curdir[FNAMELEN], tmp[FNAMELEN], *p, *classpath, *include;
	int flags = PATH_ADD_CUR_DIR, is_abs = 0;

	to_begin_line();
	if (parse_string(1, "[ \t]*#[ \t]*include[ \t]*[<\"](.*)([>\"])"
					 "[ \t]*(/[*/].*)?$")) {
		grab(find_group(1, 1), find_group(1, 0), rel);
		is_abs = character(find_group(2, 1)) == '>'
				 && !find_linked_file_ignores_angles;
		// Using #include "" or <> syntax?
	} else if (is_java_buffer() &&  parse_string(1, "[ \t]*(package|import)"
		"[ \t]*(<word>+(%.<word>+)*(%.%*)?)[ \t]*;")) {
		grab(find_group(2, 1), find_group(2, 0), rel);
		while (p = strchr(rel, '.'))	// Transform java.lang to
			*p = path_sep;				// java/lang filename.
		p = rel + strlen(rel) - 1;
		if (*p == '*') {
			strcpy(p, ".");
			flags |= PATH_PERMIT_DIRS;
		} else
			strcat(p, ".java");
	} else
		error("Can't find any link here.");
	strcpy(file, rel);
	get_buffer_directory(curdir);
	if (is_abs)				// Don't look for relative names at all.
		flags = 0;
	else {
		absolute(file, curdir);
		if (check_file(file))
			return;
	}
	flags |= PATH_PERMIT_WILDCARDS;
	include = getenv("INCLUDE");
	classpath = getenv("CLASSPATH");
	p = 0;
	if (!p && !strcmp(get_extension(filename), ".e")) {
		sprintf(tmp, "include%c%s", path_sep, rel);
		p = lookpath(tmp, 0);
	}
	if (!p && is_java_buffer() && classpath)
		p = look_on_path(rel, flags, classpath);
	if (!p && include)
		p = look_on_path(rel, flags, include);
	if (!p && *include_directories)
		p = look_on_path(rel, flags, include_directories);
	if (!p)
		p = look_on_path(rel, flags, (opsys == OS_UNIX)
						 ? INCL_PATH_UNIX : INCL_PATH_WIN);
	if (!p)
		error("%s not found in include-directories path", rel);
	strcpy(file, p);
}

restart_file_paths()
{
	char pat[FNAMELEN];
	int res, res2;

	if (insert_default_response != 3)
		return;
	save_var point = 0, abort_searching = 0, case_fold = 1;
	sprintf(pat, "(%s):[\\/][\\/][^\\/]+", url_services);
	point += parse_string(1, pat);
		// Don't consider ftp://host.com part, which may have \\ after.
	if (opsys == OS_UNIX) {
		res = parse_string(1, "/.*/!~");
		if (res)	// Delete /dir/name/ before a ~.
			delete(0, point + res);
	}
	sprintf(pat, "((%s%s):)?"	// optional a: or ftp:
			"[\\/].*[\\/]!"			// slash, stuff, slash
			"([\\/]|(%s|[a-z]):)",		// slash or a: or ftp:
			url_services, (opsys == OS_UNIX) ? "" : "|[a-z]", url_services);
	res = parse_string(1, pat);
	if (res) {	// Delete abc:\dirname\dir2\ or \dir\ or k:\dir\, etc.
		point += res;	// But exclude a trailing URL part.
		sprintf(pat, "(%s):[\\/]+", url_services);
		res2 = parse_string(-1, pat);
		if (res2 && res2 < point)
			point -= res2;
		if (character(point) == '\\' && parse_string(-1, "\\\\"))
			point--;	// Preserve a UNC name following a path.
		delete(0, point);
	}
	if (opsys != OS_UNIX) {
		sprintf(pat, "[a-z]:[\\/]!([\\/]|(%s|[a-z]):)", url_services);
		res = parse_string(1, pat);
		if (res)	// Delete c:\ or similar before an absolute path.
			delete(0, point + res);
	}
}

// Check for program prog (if nonnull), then run the command line to
// uncompress stdin to stdout.
do_decompression(char *prog, char *cmdline)
{
	char fname[FNAMELEN], *s;

	if (translation_type != FILETYPE_BINARY || is_remote_file(filename))
		return;
	if (prog && *prog) {
		sprintf(fname, "%s%s", prog, is_unix ? "" : ".exe");
		if (!(s = getenv("PATH")) || !look_on_path(fname, 0, s))
			return;
	}
	save_var _buf_readonly = 0;
	force_save_as = FSA_READONLY;
	if (!pipe_text(bufname, bufname, cmdline, NULL,
				   PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0))
		error("Could not uncompress %s", filename);
	point = modified = 0;
	strcpy(fname, filename);
	if (s = get_extension(fname)) {
		*s = 0;
		call_mode(fname);
	}
}

check_running_processes()
{
	char *buf, *open = 0, *s, msg[FNAMELEN];

	if (!process_warn_on_exit)
		return;
	buf = buffer_list(1);
	save_var bufnum;
	do {
		if (preserve_buffer_type(buf))
			if (process_warn_on_exit == 2 || !exit_this_process(buf)) {
				open = buf;
				break;
			}
	} while (buf = buffer_list(0));
	restore_vars();
	if (open) {
		tiled_only();
		locate_window(open, "");
		refresh();
		s = preserve_buffer_type(open);
		if (windows_state() & WINSTATE_MINIMIZED)
			windows_restore();
		if (process_warn_on_exit != 2) {
			if (s && !strcmp(s, "Process"))
				error("Couldn't exit from concurrent process.");
			else
				error("Couldn't exit %s job.", s);
		} else {
			if (s && !strcmp(s, "Process"))
				sprintf(msg, "A process is running, confirm exit? ");
			else
				sprintf(msg, "%s job is running, confirm exit? ", s);
			if (!ask_yn("Confirm exit", msg, "Yes", "No", 2))
				error("Exiting canceled.");
		}
	}
}

command fundamental_mode()
{
	mode_default_settings();
	try_calling("fundamental-mode-hook");
}

dired_mark_for_copy() on dired_tab['c']
{
	if (no_remote_copy(get_dired_buffer_pattern()))
		error("Copy operation not implemented for remote files.");
	if (dired_isdir() == 1)
		error("Can't copy directories.");
	dired_repl('C');
}

// Delete the directory dir and its contents.
dired_delete_hierarchy(char *dir)
{
	char pat[FNAMELEN], tail[FNAMELEN], *s;
	int cnt = -1, last;

	strcpy(pat, dir);
	add_final_slash(pat);
	save_var file_pattern_rules |= FPAT_SKIP_DIR_SYMLINKS
								   | FPAT_SKIP_RECUR_SYMLINKS;
	strcat(pat, "**/*");
	do {	// Delete everything that matches our pattern.
		s = do_file_match(pat, STARTMATCH | EXACTONLY);
		last = cnt;
		for (cnt = 0; s; s = do_file_match(pat, EXACTONLY), cnt++) {
			check_abort();
			strcpy(tail, get_tail(s, 0));
			drop_final_slash(tail);
			if (!strcmp(tail, ".") || !strcmp(tail, ".."))
				continue;
			if (get_file_read_only(s))
				set_file_read_only(s, 0);
			if (cnt % 10 == 0)
				noteput("Deleting %f ...", s);
			if (is_directory(s))
				rmdir(s);
			else
				delete_file(s);
		}	// Repeat until the number of matches stops decreasing.
	} while (cnt < last || last == -1);
	return rmdir(dir);
}

/* Return 1 if error, 2 if we didn't do it but already showed error. */
dired_do_copy(name, to, in_dired)
char *name, *to;
{
	char dest[FNAMELEN];

	make_copy_name(name, to, dest);
	if (no_remote_copy(name) || no_remote_copy(to)) {
		if (!in_dired)
			noteput("Copy operation not implemented for remote files.");
		return 2;
	}
	if (in_dired)
		noteput("Copying %f to %s.", name, dest);

	if (!strcmp(name, dest))	// disallow copying file to itself
		return 2;
	if (!is_remote_file(dest) && warn_existing_file(dest) > 0)
		return 2;
	if (is_remote_file(name) && is_remote_file(to)) {
		char *s = get_url_file_part(name, 0), *t = get_url_file_part(dest, 0);
		if (s - name != t - dest || strncmp(name, dest, s - name)) {
			noteput("Remote=>remote copying with different hosts not supported.");
			return 2;
		}
		buffer_printf(REMOTE_FTP_OP_BUFFER, "CPFR %s\nCPTO %s\n", s, t);
	} else if (is_remote_file(name)) {
		net_receive_file(name, dest);
	} else if (is_remote_file(dest)) {
		net_send_file(name, dest);
	} else {
		if (copyfile(name, dest))
			return 1;
	}
	if (in_dired) {
		save_var modified;	// leave modify bit alone
		dired_unmark();
		nl_reverse();
	}
	return 0;
}

/* Return 1 if error, 2 if we didn't do it but already showed error. */
dired_do_move(name, to, in_dired)
char *name, *to;
{
	char dest[FNAMELEN];

	make_copy_name(name, to, dest);
	if (is_remote_file(name)) {
		buffer_printf(REMOTE_FTP_OP_BUFFER, "RNFR %s\nRNTO %s\n",
				get_url_file_part(name, 0), get_url_file_part(dest, 0));
	} else {
		if (in_dired)
			noteput("Moving %f to %s.", name, dest);
		if (rename_file(name, dest)) {
			if (!strcmp(name, dest) || warn_existing_file(dest) > 0)
				return 2;
			if (copyfile(name, dest))
				return 1;
			if (delete_file(name))
				return 1;
		}
	}
	if (in_dired) {
		save_var modified;	// leave modify bit alone
		to_begin_line();	// but delete line
		delete(point - 1, give_end_line());
	}
	return 0;
}

dired_do_param(op, tag, past, func)
char *op, *tag, *past;
int (*func)();
{
	char to[FNAMELEN], recall_tag[FNAMELEN];
	int count;

	sprintf(recall_tag, "dired-%s", op);
	save_var recall_id = recall_tag;
	count = dired_ask(op, tag, " to: ", to);
	if (count > 1 && !is_remote_file(to))	// can only copy/move mult
		switch (check_file(to, NULL)) {		// to a dir
		case 0:		if (mkdir(to)) {
					file_error(errno, to, "make dir");
					quick_abort();
				}
		case CHECK_DIR:	break;
		default:	error("%s is not a directory.", to);
		}
	if (count > 0)
		dired_do(past, tag, count, func, to);
	return count;
}

command quick_dired_command() on reg_tab[ALT('o')]
{
	char file[FNAMELEN], rel[FNAMELEN], msg[FNAMELEN], dest[FNAMELEN];
	char dir[FNAMELEN];

	if (is_dired_buf()) {
		dired_getname(file, 3);
		dired_try_getdir(dir, 1);
	} else {
		strcpy(file, filename);
		strcpy(dir, filename);
		*get_tail(dir, 0) = 0;
	}
	if (!*file)
		error("This buffer has no associated file.");
	strcpy(rel, get_tail(file, 0));
again:
	do {
		sayput("%s: D=Delete, Move, Copy, View, T=Props, .=Dired, "
			   "G=Go dir, ?=Show more: ", rel);
		getkey();
		check_abort();
	} while (!IS_TRUE_KEY(key));
	say("");
	if (is_remote_file(file) && exist(REMOTE_FTP_OP_BUFFER))
		zap(REMOTE_FTP_OP_BUFFER);
	switch (toupper(key)) {
		case '.':
			dired_one(file);
			break;
		case 'D':
			sprintf(msg, "Delete %s? ", rel);
			if (!ask_yn("Delete file", msg, "Delete", "Cancel", 1))
				break;
			save_var dired_confirmation;
			if (dired_do_delete(file, NULL, 0)) {
				check_abort();
				file_error(errno, file, "Could not delete");
			} else if (!is_remote_file(file))
				say("%f deleted.", file);
			break;
		case 'C':
			if (no_remote_copy(file))
				error("Copy operation not implemented for remote files.");
			if (is_directory(file) == 1)
				error("Can't copy directories.");
			sprintf(msg, "Copy %s to: ", rel);
			get_file_dir(dest, msg);
			if (*dest) {
				switch (dired_do_copy(file, dest, 0)) {
					case 1:
						file_error(errno, dest, "Could not copy");
						break;
					case 0:
						say("%f copied.", file);
						break;
				}
			}
			break;
		case 'M':
			sprintf(msg, "Move %s to: ", rel);
			get_file_dir(dest, msg);
			if (*dest) {
				if (dired_do_move(file, dest, 0))
					file_error(errno, dest, "Could not move");
				else if (!is_remote_file(file))
					say("%f moved.", file);
			}
			break;
		case 'V':
			do_shell_execute(file, "Open", "run viewer");
			break;
		case 'T':
			do_shell_execute(file,  "Properties", "show properties");
			break;
		case 'F':
			do_shell_execute(dir, "Open", "run viewer");
			break;
		case 'G':
			if (chdir(dir))
				file_error(errno, dir, "unknown error");
			else {
				getcd(dir);
				say("Current directory: %f", dir);
			}
			break;
		case '+':
			do_create_directory(0);
			break;
		case '!':
			do_shell_command(file, dir);
			break;
		case 'A':
			dired_do_set_attributes(file, dir);
			check_dates(0);
			break;
		default:
			if (key == '?' || key == HELPKEY || key == FKEY(1))
				help_on_command("quick-dired-command");
			else
				say("Invalid key.");
			goto again;
	}
	if (is_remote_file(file))
		send_remote_ftp_op(file);
}

dired_group_sorter(start, end)
{
	int on_col = goal_column, stripwid = 0, i;

	if (!dired_sorts_files || user_abort)
		return;
	save_var narrow_start = start;
	save_var narrow_end = size() - end;
	save_var point;
	save_var abort_searching = ABORT_ERROR;
	switch (toupper(dired_sorts_files)) {
		case 'D': stripwid = dired_sort_by_date(); break;
		case 'E': stripwid = dired_sort_by_exten(); break;
		case 'S': i = get_dired_item("dired-sort-by-size-", 1);
				  if (i)		// Call style-specific function.
					  stripwid = (*(int (*)())i)();
				  else
					  stripwid = dired_sort_by_size_plain();
				  break;
	}
	do_sort_region(0, size(), on_col, isupper(dired_sorts_files));
	if (stripwid)
		strip_columns(goal_column, goal_column + stripwid);
}

// Insert size in each line, return length.
dired_sort_by_size_unix()
{
	char line[FNAMELEN * 2];

	point = 0;
	while (re_search(1, UNIX_DATE)) {
		point = matchstart;
		if (!parse_string(-1, "[0-9]+ *", line)) {
			note("Unrecognized format, can't sort.");
			strcpy(line, "0");
		}
		move_to_column(goal_column);
		bprintf("%15d", numtoi(line));
		to_end_line();
	}
	return 15;
}

// Some sizes will use commas, but larger sizes won't.  For sorting,
// remove commas from all sizes.
dired_sort_by_size_plain()
{
	char line[FNAMELEN * 2];

	point = 0;
	do {
		if (point == give_end_line())
			break;
		move_to_column(SIZE_COL);
		re_search(1, " *");		/* skip initial spaces */
		*line = 0;
		while (parse_string(1, "[0-9]+", line + strlen(line)))
			while (curchar() == ',')
				point++;
		move_to_column(goal_column);
		bprintf("%15d", numtoi(line));
	} while (nl_forward());
	return 15;
}

	/* insert numeric sortable date in each line, return length */
dired_sort_by_date_unix()
{
	char line[FNAMELEN * 2], *p, *months, *time;
	int month, year;
	struct time_info now;

	time_and_day(&now);
	months = "Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec ";
	point = 0;
	while (re_search(1, UNIX_DATE)) {
		grab(matchstart + 1, point, line);
		if (line[3] != ' ' || line[6] != ' ') {
			say("Unrecognized date format, can't sort.");
			strcpy(line, " Jan  1  1970 ");	// Fake default date.
		}
		line[3] = line[6] = 0;	// break into pieces
		p = strstr(months, line);
		if (!p) {
			say("Unrecognized month %s, can't sort.", line);
			p = months;
		}
		month = 1 + (p - months) / 4;	// Compute month 1..12.

		if (line[9] == ':') {	// Either a time (in the last 6 months)...
			time = line + 7;	// Year is this year or late last year.
			year = now.year - (month > now.month + 1);
		} else {			// Or a prior year, so make up a fake time.
			time = "00:00 ";
			year = numtoi(line + 7);
		}
		move_to_column(goal_column);
		bprintf("%5d%02d%2s%5s", year, month, line + 4, time);
		to_end_line();
	}
	return 15;
}

/* insert numeric sortable date in each line, return length */
dired_sort_by_date_oldnt()
{
	char line[FNAMELEN * 2];
	int hour;

	point = 0;
	while (re_search(1, OLDNT_DATE)) {
		grab(matchstart, point, line);
		hour = numtoi(line + 10);
		if (hour == 12)
			hour = 0;
		if (toupper(line[15]) == 'P')
			hour += 12;
		move_to_column(goal_column);
		bprintf("%.2s%.5s%02d%.2s", line + 6,
				line, hour, line + 13);
		to_end_line();
	}
	return 11;
}

dired_sort_by_date()	/* insert stuff in each line, return length */
{
	char line[FNAMELEN * 3], *p, *months;
	int end, hour, i, month;

	months = "Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec ";
	point = 0;
	i = get_dired_item("dired-sort-by-date-", 1);
	if (i)		// Call style-specific function.
		return (*(int (*)())i)();
	do {
		end = give_end_line();
		if (point == end)
			break;
		grab(point, end, line);
		hour = numtoi(line + 15);
		if (hour == 12 && strchr("ap", line[23]))
			hour = 0;
		if (line[23] == 'p')
			hour += 12;
		line[8] = 0;
		p = strstr(months, line + 5);
		if (!p) {
			say("Unrecognized month %s, can't sort.", line + 5);
			p = months;
		}
		month = 1 + (p - months) / 4;	// Compute month 1..12.

		move_to_column(goal_column);
		bprintf("%.4s%2d%.2s%2d%.5s",
				line + 9, month, line + 2, hour, line + 18);
		point = end;
	} while (nl_forward());
	return 15;
}

oem_ansi_translate(int to_oem)
{
	char cp[20];

	save_var narrow_start = 0, narrow_end = 0;
	save_var point = 0, mark;
	if (is_win32)
		sprintf(cp, "%d", call_dll("kernel32.dll", "GetOEMCP", "p", ""));
	else
		strcpy(cp, "437");
	perform_unicode_conversion(bufnum, 0, size(),
							   to_oem ? 0 : CONV_TO_16, cp);
}

int find_remote_file(char *file, int transl)
{
	int op = FTP_RECV;

	inet_only();
	buf_zap(bufnum);
	type_point = 0;
	_read_aborted = 0;
	modified = 0;
	filename = file;
	unique_file_identifier[0] = 0;
	translation_type = transl;
	if (ftp_ascii_transfers && GET_LINE_TRANSLATE(transl) != FILETYPE_BINARY)
		op |= FTP_ASCII;
	url_operation(file, op);
	if (synchronous_file_io)
		wait_for_remote_file(bufnum);
	return 0;
}

int save_remote_file(char *fname, int trans, struct file_info *finfo)
{
	int op = FTP_SEND;

	if (ftp_job)
		error("This buffer's FTP job has not yet completed.");
	inet_only();
	if (finfo)
		finfo->check_type = 0;		// Don't check file dates.
	translation_type = trans;
	modified = 0;
	if (ftp_ascii_transfers && GET_LINE_TRANSLATE(trans) != FILETYPE_BINARY)
		op |= FTP_ASCII;
	url_operation(fname, op);
	if (synchronous_file_io)
		wait_for_remote_file(bufnum);
	return 0;
}

// Feed next cmd to process.
scp_next_cmd(struct scp_job *p)
{
	char *s;

	if (p->flags & SCP_CMD_ERROR)
		p->cmd = 0;
	if (p->cmd && *p->cmd) {
		p->busy = 1;
		if (p->flags & SCP_CLIENT_MODE)
			buf_printf(p->buf, "## ");
		if (s = strchr(p->cmd, '\n')) {
			buf_stuff(p->buf, p->cmd, s++ - p->cmd + 1);
		} else {
			buf_stuff(p->buf, p->cmd, strlen(p->cmd));
			buf_stuff(p->buf, "\n", 1);
		}
		p->cmd = s;
	}
}

struct scp_job *prep_scp_buffer(struct url_parts *parts, int op);
char *scp_variable(struct url_parts *parts, char *varname);

struct scp_job *do_scp(struct url_parts *parts, int op)
{
	struct scp_job *p;

	p = prep_scp_buffer(parts, op);
	if (op == FTP_RECV) {
		do_scp_xfer(parts, p, scp_variable(parts, "read-file"), 0, 1);
	} else if (op == FTP_SEND) {
		do_scp_xfer(parts, p, scp_variable(parts, "write-file"), -1, 1);
	} else if (op == FTP_LIST || op == FTP_MISC) {
		scp_run_helper(parts, p);
		*p->file = 0;
		if (op == FTP_LIST)
			scp_send_some_command(parts, p, (p->flags & SCP_OLD_SFTP)
								  ? "ls %s" : "ls -l %s");
	}
	return p;
}

// Send a command based on fmt to helper and wait for a response in
// buffer b.  To work with both sftp and our xfer client, we recognize
// and replace particular values of fmt.  What info to get:
// 0=completions, 1=absolute form of relative dir name, 2=whether
// symbolic link is to a directory or file (only for non-client mode),
// 3=wildcarding (with no * automatically added as for completions).

scp_get_response(int code, char *url, struct url_parts *parts, int b)
{
	char cmd[FNAMELEN], *s;
	struct scp_job *p;

	save_var bufnum = b;
	buf_zap(bufnum);
	p = do_scp(parts, FTP_MISC);
	strcpy(p->file, url);
	*get_url_file_part(p->file, 0) = 0;
	if (!(p->flags & SCP_CLIENT_MODE)) {
		s = (p->flags & SCP_OLD_SFTP) ? "ls \"%s*\"" : "ls -l \"%s*\"";
		switch (code) {
			case 1:	s = "";
					break;
			case 2:	p->flags |= SCP_CHECK_IS_DIR;
					s = "cd \"%s\"";
					break;
			case 3:	s = (p->flags & SCP_OLD_SFTP) ? "ls \"%s\"" : "ls -l \"%s\"";
					break;
		}
		scp_send_some_command(parts, p, s);
	} else {
		sprintf(cmd, (code == 1 || code == 2) ? "prdir %s\n" : "comp %s\n",
				parts->fname);
		p->cmd = p->cmdlist = strsave(cmd);
		if (p->busy != 2)
			scp_next_cmd(p);
	}
	delayed_say(0, 100, 30, "Working...");
	while (p->busy && !user_abort)
		delay(1, COND_KEY);
	set_buf_mode_message(bufnum, NULL);
	user_abort = 0;
	delayed_say(0, 0, 0, "");
	say("");
	point = 0;
	if (!(scp_flags(parts) & SCP_CLIENT_MODE)) {
		switch (code) {
			case 0:		// Discard the extra info.
			case 3:
				delete_final_newlines();
				simple_re_replace(1, "^.*" UNIX_DATE " *", "");
				strcpy(cmd, url);
				*get_tail(get_url_file_part(cmd, 0), 1) = 0;
				if (size())
					simple_re_replace(1, "^", cmd);
				break;
			case 1:
				buf_printf(b, "%s%s", p->file, p->lastdir);
				set_buf_point(b, 0);
				break;
		}
	}
}

char *remote_file_match(char *s, int flags)
{
	struct url_parts parts;
	int b;

	if (parse_url(s, &parts) && !strcmp(parts.service, "scp")) {
		b = create(_MATCH_BUF);
		if (flags & STARTMATCH)
			scp_get_response((flags & EXACTONLY) ? 3 : 0, s, &parts, b);
		if (!grab_line(b, _this_match))
			return 0;
		return _this_match;
	}
	return NULL;
}

// Is buffer b currently used for scp?
is_busy_doing_scp(int b)
{
	struct scp_job *p;

	save_var bufnum = b;
	for (p = scp_job; p; p = p->next)
		if (p->busy && (p->destbuf == b || p->buf == b))
			return 1;
	return 0;
}

// If buffer b is currently actively retrieving a remote file, don't
// return until it finishes.

wait_for_remote_file(int b)
{
	save_var bufnum = b;
	for (;;) {
		if (!ftp_job && !is_busy_doing_scp(b))
			return;
		delay(10, 0);
		check_abort();
	}
}

// Download remote file and save locally.
net_receive_file(char *from, char *to)
{
	struct url_parts parts;
	struct scp_job *p;

	wait_for_remote_file(bufnum);
	prepare_url_operation(from, FTP_RECV, &parts);
	p = prep_scp_buffer(&parts, FTP_RECV);
	strcpy(p->file, to);
	do_scp_xfer(&parts, p, scp_variable(&parts, "read-file"), 0, 0);
}

// Upload local file.
net_send_file(char *from, char *to)
{
	struct url_parts parts;
	struct scp_job *p;

	wait_for_remote_file(bufnum);
	prepare_url_operation(to, FTP_SEND, &parts);
	p = prep_scp_buffer(&parts, FTP_SEND);
	strcpy(p->file, from);
	do_scp_xfer(&parts, p, scp_variable(&parts, "write-file"), 0, 0);
}

int url_operation(char *file, int op)
{
	struct url_parts parts;
	int port;

	prepare_url_operation(file, op, &parts);
	port = numtoi(parts.port);
	if (*parts.port && got_bad_number)
		say("Ignoring invalid port number %s", parts.port);
	if (!strfcmp(parts.service, "http")) {
		if (got_bad_number)
			port = 80;
		if ((op & FTP_OP_MASK) != FTP_RECV)
			error("Can't do that with HTTP files.");
		do_http(&parts, port);
	} else if (!strfcmp(parts.service, "telnet")) {
		if (got_bad_number)
			port = 23;
		if ((op & FTP_OP_MASK) != FTP_RECV)
			error("Can't do that with telnet buffers.");
		do_telnet(parts.host, port, bufname);
	} else if (!strfcmp(parts.service, "ssh")) {
		if ((op & FTP_OP_MASK) != FTP_RECV)
			error("Can't do that with ssh buffers.");
		do_ssh(parts.host, bufname);
	} else if (!strfcmp(parts.service, "scp")) {
		do_scp(&parts, op);
	} else
		do_ftp_op(bufname, parts.host, parts.port, parts.usr, parts.pwd,
				  parts.fname, op);
	return 0;		// No error possible yet (asynchronous).
}

telnet_activity(int activity, int buf, int from, int to)
{
	if (!buf_exist(buf))
		return;
	save_var iter, has_arg;
	save_var bufnum = buf;
	switch (activity) {
		case NET_DONE:
			if (!telnet_id)		// A ssh buffer.
				break;
			stuff("Connection closed.\n");
			refresh();
			telnet_id = 0;
			break;
		case NET_RECV:
			set_character_color(from, to, color_class process_output);
			if (telnet_interpret_output & 8) {
				point = from - 2;
				while (re_search(1, ".\b."))
					if (character(point - 3) == character(point - 1)) {
						delete(point - 3, point - 1);
						set_character_color(point - 1, point, color_class man_bold);
					}
			}
			if (telnet_interpret_output & 1) {
				save_var point = from - 2;
				while (re_search(1, "_\b.|o\b%+")) {
					delete(point - 3, point - 1);
					set_character_color(point - 1, point, color_class process_prompt);
				}
			}
			point = to;
			save_var case_fold = 1;
			if (ungot_key == -1 && (recognize_password_prompt & 1)
				  && parse_string(-1, recognize_password_pattern)) {
				ungot_key = WIN_HELP_REQUEST;	// Make Epsilon run this cmd:
				menu_command = "password-prompt";
			}
			if (telnet_interpret_output & 2) {
				point = from;
				to_begin_line();
				while (re_search(1, "^[^\r\n]*\r+![^\n]"))
					delete(matchstart, matchend);
			}
			if (telnet_interpret_output & 4)
				color_ansi_output(from);
	}
}

telnet_mode()
{
	mode_default_settings();
	mode_keys = telnet_tab;
	telnet_tab[CTRL('C')] = find_index("telnet-ctrlc-tab");
	telnet_tab[ALT('p')] = find_index("process-previous-cmd");
	telnet_tab[ALT('n')] = find_index("process-next-cmd");
	major_mode = _telnet_mode_name;
	tab_size = 8;
	try_calling("telnet-mode-hook");
	make_mode();
}

// For sending chars to ssh job, act like a process, but for getting
// chars, act like telnet.
ssh_activity(int activity, int buf, int from, int to)
{
	save_var process_coloring_rules = ssh_coloring_rules;
	if (activity == NET_SEND) {
		save_var bufnum = buf;
		color_process_range(from, to);
		do_when_process_input(buf, from, to);
	} else
		telnet_activity(activity, buf, from, to);
}

// A variation on process mode for ssh buffers.
ssh_mode()
{
	process_mode();
	mode_keys = ssh_tab;
	major_mode = _ssh_mode_name;
	try_calling("ssh-mode-hook");
	make_mode();
}

 /* Compare this file's time info with version on disk. */
/* Return 0 to indicate ok, 1 to indicate user said to cancel saving
 * the file, or 2 to tell grep that it skipped asking to read a modified
 * file from disk. */
check_dates(save)
{
	struct file_info now, *then = &file_info;
	int ret = 1, temp, orig = bufnum, pos;

	if (num_of_popups() > 0)	// Don't pop up over bufed or similar.
		return 0;
	if (file_date_skip_drives && check_dates_skip_drive())
		return 0;
	ignoring_file_change = 0;	// Resume checking on getting focus.
	if (!want_warn || !filename || then->check_type != CHECK_FILE
			|| check_file(filename, &now) != CHECK_FILE)
		return 0;
	update_readonly_warning(&now);
	if ((readonly_warning & ROWARN_REFRESH) && !save
		  && (file_info.attr ^ now.attr) & ATTR_READONLY)
		return 2;

	if (!compare_dates(&now, then)) {
		maybe_add_readonly_once(&now);
		return 0;
	}
	switch_to_buffer(bufnum);
	maybe_add_readonly_once(&now);

	if ((auto_read_changed_file || _force_auto_read_changed_file)
		&& !save && maybe_auto_read(&now))
		return 1;
	if ((readonly_warning & (ROWARN_GREP | ROWARN_REFRESH)) && !save)
		return 2;
	tiled_only();
	switch_to_buffer(bufnum);
	maybe_ding(bell_on_date_warning);
	temp = tmp_buf();
	save_var _view_right = 0, _view_bottom = screen_lines - 30;
	save_var return_raw_buttons = 1;
	while (!user_abort) {
		show_date_warning(temp, orig, then, &now, save);
		bufnum = orig;
		pos = point;
		if (is_gui && want_gui_prompts) {
			int win, owin = window_handle;

			remove_final_view();
			save_var _doing_input = DI_VIEW;
			win = add_popup(0, 0, 75, 15, _view_border, temp);
			go_to_link = 0;
			display_dialog_box("FileDateBox2", "File has changed on disk",
							   win, 0, 0,
							   "Compare", "Read", save ? "Save" : "Ignore");
			text_color = (is_gui && want_gui_prompts) ? color_class text
						 : color_class viewed_text;
			point = 0;
			view_loop(win);
			switch (key_is_button) {
				case 1:	key = 'C'; break;
				case 2:	key = 'R'; break;
				case 3:	key = save ? 'S' : 'I'; break;
			}
			window_handle = owin;
			remove_window(win);
			if (user_abort)
				break;
		} else {
			char junk[FNAMELEN];
			view_buf(temp, 0);
			if (key == abort_key)
				user_abort = 1;
			if (user_abort)
				break;
			getkey();
			if (key == WIN_DRAG_DROP)	// Ignore dropped files.
				while(drag_drop_result(junk))
					;
		}
		key = toupper(key);
		if (key == 'C') {
			ignoring_file_change = 2;	// Don't check for a while.
			refresh();
			split_window_vertically();
			window_number--;	/* to 1st window */
			find_in_other_buf(filename, default_translation_type);
			point = pos;
			to_begin_line();
			set_buffer_filename("");
			error("This buffer has disk version.");
			break;		/* abort out of calling func */
		} else if (key == 'R') {
			refresh();
			if (!read_file(filename, reading_translation_type()))
				say("%s: Read from disk.", filename);
			point = pos;
			to_begin_line();	// Don't check immediately after dialog ends.
			ignoring_file_change = is_gui ? 2 : 0;
			break;
		} else if (key == (save ? 'S' : 'I')) {
			if (key == 'I')
				ignoring_file_change = 100;	// Don't check on getting focus.
			else
				ignoring_file_change = 2;	// Don't check immediately after dialog ends.
			refresh();
			ret = 0;
			break;
		} else if (key == '!')
			want_warn = !want_warn;
		else if (key == 'A')
			ask_read_changed_file();
		say("");
	}
	buf_delete(temp);
	return user_abort ? 1 : ret;
}

// Returns 0=Monday ... 6=Sunday given day 1..31, month 1..12, year.
int day_of_week(int day, int month, int year)
{
	if (month < 3)
		month += 12, year--;
	return ((month * 13 + 3) / 5 + day + year
			+ year / 4 - year / 100 + year / 400) % 7;
}

format_date(msg, year, month, day, hour, minute, second)
char *msg;
{
	char ampm = 'a', args[24][30], *vals[24], *s, *t;
	char *months = "January February March April May June July August "
				   "September October November December ";
	char *weekdays = "Monday Tuesday Wednesday Thursday Friday Saturday Sunday ";
	int i, dow;

	sprintf(args[0], "%d", hour);
	sprintf(args[1], "%02d", hour);
	if (hour >= 12)
		ampm = 'p', hour -= 12;
	sprintf(args[2], "%d", hour ? hour : 12);
	sprintf(args[3], "%02d", hour ? hour : 12);
	sprintf(args[4], "%d", minute);
	sprintf(args[5], "%02d", minute);
	sprintf(args[6], "%d", second);
	sprintf(args[7], "%02d", second);
	sprintf(args[8], "%c", ampm);
	sprintf(args[9], "%c", toupper(ampm));
	sprintf(args[10], "%d", day);
	sprintf(args[11], "%02d", day);
	sprintf(args[12], "%d", month);
	sprintf(args[13], "%02d", month);
	sprintf(args[14], "%d", year);
	sprintf(args[15], "%02d", year % 100);
	for (i = month - 1, s = months; i > 0 && s; i--)
		s = strchr(s, ' ') + 1;
	if (month >= 1 && month <= 12 && s && (t = strchr(s, ' ')))
		sprintf(args[16], "%.*s", t - s, s);
	else
		sprintf(args[16], "???");
	strncpy(args[18], args[16], 3);
	for (i = 0; i < strlen(args[16]) + 1; i++)
		args[17][i] = toupper(args[16][i]);
	for (i = 0; i < strlen(args[18]) + 1; i++)
		args[19][i] = toupper(args[18][i]);
	dow = day_of_week(day, month, year);
	for (i = dow, s = weekdays; i > 0 && s; i--)
		s = strchr(s, ' ') + 1;
	if (dow >= 0 && dow <= 6 && s && (t = strchr(s, ' ')))
		sprintf(args[20], "%.*s", t - s, s);
	else
		sprintf(args[20], "???");
	strncpy(args[22], args[20], 3);
	for (i = 0; i < strlen(args[20]) + 1; i++)
		args[21][i] = toupper(args[20][i]);
	for (i = 0; i < strlen(args[22]) + 1; i++)
		args[23][i] = toupper(args[22][i]);
	for (i = 0; i < 24; i++)
		vals[i] = args[i];
	if (expand_string_template(msg, date_format, "iIhHnNeEaAdDmMyYfFtTwWsS", vals))
		error("Invalid date-format variable.");
}

command visit_file() on cx_tab[CTRL('V')]
{
	char tmp[FNAMELEN];

	iter = 0;
	save_var show_mouse_choices = 1;
	get_file(tmp, "Visit file", force_save_as == FSA_NEWFILE ? "" : filename);
	if (!*tmp)
		return;
	if (is_process_buffer(bufnum)
		  && !ask_yn("Confirm replacing buffer contents", "Are you sure "
					 "you want to begin executing the lines of this file? ",
					 "Yes", "No", 2))
		return;
	if (!has_arg && modified) {	/* buffer need to be saved? */
		if (ask_save_buffer())
			do_save_file(1, 1, 1);
	}
	read_file(tmp, reading_translation_type());
}

find_it(fname, strip)	/* find named file (must be in absolute form) */
char *fname;
{
	if (!*fname)
		return 2;
	if (look_file(fname)) {
		save_var _force_auto_read_changed_file = 1;
		to_buffer(bufname);
		if (!size() && !file_info.check_type)
			find_in_this_buf(fname, strip);
		return 0;
	}
	return find_in_other_buf(fname, strip);
}

find_in_other_buf(fname, transl)
char *fname;
{
	char bname[FNAMELEN+10];

	make_bname(fname, bname);
	to_buffer(bname);
	return find_in_this_buf(fname, transl);
}

find_in_this_buf(char *fname, int transl)
{
	int err;

	if (default_character_set == 2)
		file_io_converter = oem_file_converter;
	else
		file_io_converter = 0;
	err = read_file(fname, transl);
	unique_filename_identifier(fname, unique_file_identifier);
	call_mode(filename);
	make_mode();
	check_filevars();
	return err;
}

#define C_COMMENT_FILLER		"[^0-9a-zA-Z\x80-\xFF_(\n]*"
#define C_COMMENT_PREFIX		"^" C_COMMENT_FILLER

int locate_c_comment()
{
	int com, type = 1, generic = strcmp(major_mode, "C");

	save_spot narrow_start, narrow_end;
	com = in_c_comment(c_look_back);	// Comment?  If so, what type?
	if (com == IN_OLD_COMMENT || !com && !generic && parse_string(1, "[ \t]*/<*>"))
		narrow_to_block_comment();
	else if (narrow_to_new_comment())
		type = 2;
	else
		return 0;
	to_begin_line();
	if (parse_string(1, C_COMMENT_PREFIX "$"))
		point--;		// If on line right after a parag, go back.
	to_begin_line();
	if (parse_string(1, C_COMMENT_PREFIX "$")) // If no text on line,
		if (!re_search(-1, "[^ \t\n]"))	// Find a line with text.
			re_search(1, "[^ \t\n]");

	re_search(1, C_COMMENT_PREFIX "$");	// Find "blank" line w/only filler.
	if (parse_string(-1, C_COMMENT_PREFIX "$"))
		to_begin_line();	// And move to its start.
	re_search(-1, "[^ \t\n]");	// Find text of some parag before it.
	nl_forward();
	mark = point;
	point--;
	re_search(-1, C_COMMENT_PREFIX "$");	// Find "blank" line before.
	if (parse_string(1, C_COMMENT_PREFIX "\n"))
		nl_forward();
	return type;
}

// Assuming we're in (or just before) a C block comment, narrow to it.
// Optionally consider block comments with only a newline between them
// as one.

narrow_to_block_comment()
{
	save_var point;
	do {
		search(1, "*/");
		if (auto_fill_comment_rules & 8)
			break;
	} while (parse_string(1, "[ \t]*\n?[ \t]*/<*>"));
	narrow_end = size() - point;
	do {
		search(-1, "/*");
		if (auto_fill_comment_rules & 8)
			break;
	} while (parse_string(-1, "<*>/[ \t]*\n?[ \t]*"));
	narrow_start = point;
}

do_grep(buf, pat, flags, filepat, inbufs)
char *buf, *pat, *filepat;
{
	int res = 0, aborted = 0, cnt = 0, tempbuf = 0, i, skipped = 0;
	int orig, before_head, this_err;
	char *s, rel[FNAMELEN], abs[FNAMELEN], *(*finder)();
	char *months = "Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec ";
	char extra[FNAMELEN];
	spot start;

	save_var synchronous_file_io = 1;
	before_head = prepare_grep_buffer(buf, flags, pat, filepat);
	start = alloc_spot(0);
	save_var modified;
	save_var abort_file_matching = ABORT_ERROR;
	finder = give_grep_finder(inbufs, filepat, &s);	// get 1st match
	save_var readonly_warning, no_popup_errors = 1;
	if (grep_on_changed_file & 1)
		readonly_warning |= ROWARN_GREP;
	readonly_warning &= ~(ROWARN_MSG | ROWARN_BELL);
	orig = bufnum;
	for (; s; s = (*finder)(filepat, 0)) {
		if (user_abort) {
			aborted = 1;
			break;
		}
		grep_make_names(rel, abs, s, inbufs);
		if (inbufs)
			bufname = s;
		else if (!ok_grep_file_match(s, abs)) {
			skipped++;
			continue;
		}
		noteput("%d matches, searching %f...", cnt, rel);
		if (inbufs)	// skip over grep results buffer
			this_err = !strcmp(s, buf);
		else if (!look_file(abs)) {
			this_err = find_it(abs, default_translation_type);
			if (!grep_keeps_files)
				tempbuf = bufnum;
		} else {
			this_err = 0;
			if (grep_on_changed_file & 2)
				save_var auto_read_changed_file = 1;
			if (check_dates(0) == 2)
				grep_date_warning(buf, filename);
		}
		if (this_err == EREADABORT || user_abort) {
			aborted = 1;
			break;
		} else if (this_err) {
			window_bufnum = orig;
			if (tempbuf)
				buf_delete(tempbuf);
			tempbuf = 0;
			continue;
		}
		*extra = 0;
		if (grep_include_timestamp && !inbufs && file_info.check_type == CHECK_FILE)
			if (grep_include_timestamp == 2)
				sprintf(extra, "[%04d-%02d-%02d %02d:%02d:%02d] ",
						file_info.year, file_info.month, file_info.day,
						file_info.hour, file_info.minute, file_info.second);
			else
				sprintf(extra, "[%02d-%.3s-%04d %02d:%02d:%02d] ",
						file_info.day,
						(file_info.month >= 1 && file_info.month <= 12)
						? months + (file_info.month - 1) * 4 : "???",
						file_info.year,
						file_info.hour, file_info.minute, file_info.second);
		if (grep_show_absolute_path == 2)
			convert_to_8_3_filename(abs);
		res = copy_matches(name_to_bufnum(buf), flags, pat,
						   grep_show_absolute_path ? abs : rel, extra);
		window_bufnum = orig;
		if (res < 0)	/* error occurred during search */
			break;
		else if (res > 0) {
			cnt += res;
			build_first = 1;
			update_grep_title(before_head, cnt);
			refresh();
		}
		if (tempbuf)
			buf_delete(tempbuf);
		tempbuf = 0;
	}
	window_bufnum = orig;
	point = i = *start;
	free_spot(start);
	build_first = 1;
	if (tempbuf)
		buf_delete(tempbuf);
	say("");
	if (aborted)
		quick_abort();
	if (res == -DSBAD)
		error("Illegal search pattern: %s", pat);
	else if (res != -DSABORT) {
		locate_window(buf, "");
		point = i;
		if (skipped)
			say("Skipped %d file%s due to grep-ignore-file- variables",
				  skipped, (skipped == 1) ? "" : "s");
	}
}

do_grep_examine(char *cmd)
{
	int start, end, colon, old, instance, i, toolong = 0, extra;
	int lines = 1, precolon;
	char file[FNAMELEN], string[FNAMELEN], line[FNAMELEN], curdir[FNAMELEN], *p;
	char prefix[FNAMELEN];

	save_var synchronous_file_io = 1;
	_last_was_grep = 1;
	to_begin_line();
	save_var case_fold = 0;		/* turn off for grep buffer */
	if (!search(1, ": "))
		error("Can't read file name");
	colon = point;
	precolon = colon - 2;
	if (parse_string(-1, " %[[0-9]+%]: ")) {
		precolon = matchend;
		grab(precolon + 2, point, line);
		lines = numtoi(line);
	}
	to_begin_line();
	extra = parse_string(1, "%[[0-9][0-9]+-[^]]*[0-9]%] ");
	start = point;
	for (i = lines; i > 1; i--)
		nl_forward();
	to_end_line();
	end = point;
	point = start;
	grab(start, precolon, prefix);
	grab(start + extra, precolon, file);
	if (start + FNAMELEN - 3 < end)
		end = toolong = start + FNAMELEN - 3;
	grab(colon, end, string);
	line[0] = '\n';		// Put \n at start and end of line.
	grab(start, end + 1, line + 1);
	instance = count_instances(line) + 1;
	grep_get_curdir(curdir);
	if (cmd && *cmd)
		try_calling(cmd);
	if (strstr(file, MANPREFIX) == file) {
		grep_examine_man_page(file + sizeof(MANPREFIX) / sizeof(char) - 1);
		return;
	} else if (strstr(file, BUFPREFIX) != file) {
		absolute(file, curdir);
		if (find_it(file, ask_line_translate()))
			return;
	} else if (exist(p = file + sizeof(BUFPREFIX) / sizeof(char) - 1))
		to_buffer(p);
	else
		error("Can't find buffer %s.", p);
	if (strlen(line) != end - start + 2)
		error("Can't find matching lines with null characters.");
	old = point;
	save_var case_fold = 0;		/* turn off for user buffer */
	point = 0;
	for (i = 0; i < instance; i++)
		if (toolong ? !search(1, string) : !line_search(1, string)) {
			point = old;
			if (i == 0 || instance == 1)
				error("Can't find \"%s\"", string);
			else
				error("Can't find this copy of repeated line");
		} else {
			old = point;
			if (i + 1 < instance)
				nl_forward();
		}
	to_begin_line();
	if (toolong)
		say("Line too long; this may not be the correct match.");
}

do_file_replace(str, with, flags, filepat, inbufs)
char *str, *with, *filepat;	// (query-) replace across files/buffers
{
	int res, oldbuf = 1, prevbuf = 0, total = 0, cnt = 0;
	char rel[FNAMELEN], abs[FNAMELEN], *s, *(*finder)(), msg[FNAMELEN];
	int total_found = 0, total_changed = 0, total_readonly = 0, old_readonly;

	save_var synchronous_file_io = 1;
	old_readonly = readonly_warning;
	save_var readonly_warning;			// Don't show msg or ring bell
	readonly_warning &= ~(ROWARN_MSG | ROWARN_BELL); // unless matches.
	save_var abort_file_matching = ABORT_ERROR;
	save_var bufnum, window_bufnum;
	finder = give_grep_finder(inbufs, filepat, &s);	// Get 1st match
	for (; s; s = (*finder)(filepat, 0)) {
		grep_make_names(rel, abs, s, inbufs);
		if (inbufs)
			bufname = s;
		else if (!ok_grep_file_match(s, abs))
			continue;
		noteput("Searching %s...", rel);
		if (!inbufs) {
			oldbuf = look_file(abs);
			if (find_it(abs, default_translation_type) == EREADABORT)
				user_abort = 1;
		} else
			to_buffer(s);
		undo_mainloop();
		if (!user_abort) {
			res = one_file_replace(str, with, flags, oldbuf, old_readonly);
			total_found += replace_num_found;
			total_changed += replace_num_changed;
			total++;
			if (res == -2)
				total_readonly++;
			else if (res >= 0)
				cnt++;
		} else {
			res = DSABORT;
			user_abort = 0;
		}
		if (prevbuf && buf_in_window(prevbuf) == -1) {
			buf_delete(prevbuf);	// delete untouched new bufs
			prev_forget(_prev_num - 1); // fix default
		}
		prevbuf = (oldbuf || modified) ? 0 : bufnum;
		if (res == 0 && (flags & QUERY) && key == '!'
			&& ask_continue_replace("Replace without asking"))
			flags &= ~QUERY;	// no more querying
		else if (res == DSABORT || res == 1 && ungot_key != -1) {
			if (!ask_continue_replace("Continue replacing"))
				break;
		} else if (res > 0) {	// User said to stop, or bad pattern.
			if (res == DSBAD)
				show_replace(str, with, flags | BAD);
			user_abort = 0;
			return;
		}
		ungot_key = -1;
		if (user_abort)
			break;
	}
	ungot_key = -1;
	user_abort = 0;
	sprintf(msg, "Replaced %d of %d matches in %d of %d %s",
	    total_changed, total_found, cnt, total,
	    inbufs ? "buffers" : "files");
	if (total_readonly)
		sprintf(msg + strlen(msg), ", %d read-only, see %s buffer",
				total_readonly, MSG_BUFFER);
	say("%s.", msg);
	restore_vars();
	if (prevbuf && buf_in_window(prevbuf) == -1) {
		buf_delete(prevbuf);	// delete untouched new bufs
		prev_forget(_prev_num - 1); // fix default
	}
}

one_file_replace(str, with, flags, oldbuf, old_readonly)
char *str, *with;	// return 0=ok, 1=user said to stop,
{			// DSABORT=aborted, DSBAD=bad pat, -1=no matches, -2=readonly
	int res;

	replace_num_found = replace_num_changed = 0;
	save_var point = (flags & REVERSE) ? size() : 0;
	res = do_color_searching(flags, str);
	if (res != 1)
		return res ? res : -1;
	if (!oldbuf && (flags & FILEGREP))	// had to read file to search
		replace_in_readonly_hook(old_readonly);
	else
		replace_in_existing_hook(old_readonly);
	if (_buf_readonly) {
		if ((flags & FILEGREP) && file_info.check_type == CHECK_FILE
			  && (file_info.attr & ATTR_READONLY))
			show_text(0, 1, "File %r is read-only.", filename);
		else
			show_text(0, 1, "Buffer %s is read-only.", bufname);
		return -2;
	}
	point = matchstart + ((flags & REVERSE) ? 1 : -1);
	show_replace(str, with, flags);
	return string_replace(str, with, flags);
}

/* list all commands with a specifed substring in name or short description */
command apropos()
{
	char str[80], msg[300], name[80];
	int found = 0, start, old = bufnum, helpbuf;

	iter = 0;
	get_string(str, "Apropos: ");
	save_var bufnum;
	if (!*str || !get_doc(0))
		return;
	helpbuf = bufnum;
	noteput("Searching for %s...", str);
	point = 0;
	while (search(1, str)) {
		if (nl_reverse())
			point++;
		if (curchar() != '~')
			nl_forward();
		else {
			found = 1;
			point++;
			parse_string(1, "[^ \t\n]*", name);
			bufnum = old;
			descr_bindings(name, msg);
			buffer_printf(HELPBUF, "%s\n", msg);
			bufnum = helpbuf;
			start = point;
			nl_forward();
			xfer(HELPBUF, start, point);
			add_variable_info(find_index(name));
		}
	}
	if (found) {
		bufname = HELPBUF;
		point = 0;
		string_replace("^[^ \n\t]+", "\1#0\2", REGEX);
	} else
		buffer_printf(HELPBUF, "No matches.\n");
	bufnum = old;
	say("");
	sprintf(msg, " Apropos \"%s\" ", str);
	view_help(msg, 1);
}

show_cmd_help(cmd)
char *cmd;
{
	char msg[300];

	if (use_winhelp()) {
		sprintf(msg, "%s command", cmd);
		win_help_string("", msg);
	} else {
		descr_bindings(cmd, msg);
		dodesc(cmd, msg, 1);
	}
}

show_binding(fmt, cmd)	/* print msg, substituting binding of cmd at %s */
char *fmt, *cmd;
{
	char msg[300];

	give_binding(msg, cmd);
	say(fmt, msg);
}

man_follow_link() on man_tab['\n'], man_tab['e'], man_tab['\r']
{
	char buf[FNAMELEN];

	save_var show_status = 0, abort_searching = 0;
	grab_man_word_from_buffer(bufnum, point, buf,
							  "[-+_0-9a-z.:]+( *%((..?|[0-9][a-z]*)%))?");
	if (man_rewrite(buf) || _man_program && *buf)
		make_man_page(buf, _man_program);
	else if (key == MOUSE_DBL_LEFT)
		mouse_selecting();
	else
		error("Not a manual page cross-reference; must use open(2) format.");
}

// Put hex version of current buffer in bname.
buffer_to_hex(char *bname)
{
	char trans[256];
	int i;

	zap(bname);
	save_var bufnum, point = 0, narrow_start = 0, narrow_end = 0;
	for (i = 0; i < 256; i++)
		trans[i] = hex_printable(i);
	do_buffer_to_hex(bname, trans);
}

int hex_old_on_modify();
int hex_on_modify();
int color_hex_range();
char *hex_prefix();

command hex_mode()
{
	char bname[FNAMELEN];
	int pos = point;
	struct file_info saved = file_info;

	if (is_hex_buffer(bufnum)) {	// "Turn off mode" -- save back to real buf.
		hex_save_key();
		return;
	}
	if (!size())
		error("Buffer is empty.");
	sprintf(bname, "%s%s", HEX_BUFNAME_PREFIX, bufname);
	if (exist(bname)) {
		say("This buffer is in hex mode.  Press %s? for help or %sq to quit.",
			  hex_prefix(), hex_prefix());
		to_buffer(bname);
		return;
	}
	buffer_to_hex(bname);
	if (user_abort) {
		delete_buffer(bname);
		error("Canceled.");
	}
	call_on_modify = 1;		// Make the buffer read-only so only our
	buffer_on_modify = hex_old_on_modify;	// special funcs can modify it.
	file_info.check_type = 0;	// Disable file date checking during hex mode.
	to_buffer(bname);
	file_info = saved;		// Remember old file date checking info.
	call_on_modify = 1;		// Make the buffer read-only so only our
	buffer_on_modify = hex_on_modify;	// special funcs can modify it.
	mode_keys = hex_tab;		/* Use these keys. */
	major_mode = hex_overtype_mode ? _hex_overtype_mode_name : _hex_mode_name;
	recolor_range = color_hex_range;	// set up coloring rules
	recolor_from_here = recolor_by_lines;
	idle_coloring_delay = -1; // No idle coloring, it slows down apply-changes.
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	try_calling("hex-mode-hook");
	drop_all_colored_regions();
	make_mode();
	say("Entering hex mode, press %s? for help or %sq to quit.",
		  hex_prefix(), hex_prefix());
	hex_goto_place(pos, hex_overtype_mode);
}

color_html_comment()
{
	int s;

	search(-1, "<");	// Find start of comment.
	s = point;
	if (parse_string(1, "%<%!%[CDATA%[")) {	// xml cdata
		re_search(1, "%]%]%>");
		set_character_color(s, point, color_class html_value);
		return;
	} else if (parse_string(1, "%<[ \t\n]*%![ \t\n]*--")) {
		// A legal comment?
		point = matchend;
		re_search(1, "--[ \t\n]*>");	// Search for end of legal comment.
	} else if (parse_string(1, "%<%!DOCTYPE")) {
		if (re_search(1, "[[>]") && character(matchstart) == '[') {
			search(1, "]");
			search(1, ">");
		}
		set_character_color(s, point, color_class html_value);
		return;
	} else	// Some odd flavor of comment; don't be too picky about its end.
		search(1, ">");
	set_character_color(s, point, color_class html_comment);
}

do_html_indent(int manual)
{
	int region, start, i;

	recolor_buffer_range(point, point + 1);
	region = get_tagged_region("color-as-unit", point, &start);
	if (IS_ENTER_KEY(key) && !(html_auto_indent & (region == HT_JSCRIPT ? 2
		: region == HT_VBSCRIPT ? 4 : region == HT_PHP ? 8 : 1)))
		return;
	save_var narrow_start = start;
	i = (1 << (region - 1 + (IS_ENTER_KEY(key) ? 4 : 0)));
	if (region > 0 && region <= HT_PHP && !(html_indenting_rules & i))
		region = 0;
	switch (region) {
		case HT_JSCRIPT:
			html_exclude_script_head();
			save_var c_extra_keywords |= JAVA_KEYWORDS;
			do_c_indent();
			return;
		case HT_VBSCRIPT:
			html_exclude_script_head();
			do_vbasic_indent();
			return;
		case HT_PHP:
			html_exclude_script_head();
			save_var in_perl_buffer = 1;
			save_var sublanguage_code = SUBLANG_PHP;
			do_c_indent();
			return;
	}
	if (manual)
		indent_previous();
	else
		prev_indenter();
}

int html_auto_show_delimiter()
{
	int region, start, end;

	save_var point;
	recolor_buffer_range(point, point + 1);
	region = get_tagged_region("color-as-unit", point, &start, &end);
	save_var narrow_start = start, narrow_end = size() - end;
	switch (region) {
		case HT_JSCRIPT:
			if (!auto_show_c_delimiters)
				return 0;
			html_exclude_script_head();
			save_var auto_show_matching_characters = c_auto_show_delim_chars;
			save_var mode_move_level = c_move_level;
			break;
		case HT_VBSCRIPT:
			if (!auto_show_vbasic_delimiters)
				return 0;
			html_exclude_script_head();
			save_var auto_show_matching_characters = vbasic_auto_show_delim_chars;
			save_var mode_move_level = 0;
			break;
		case HT_PHP:
			if (!auto_show_php_delimiters)
				return 0;
			html_exclude_script_head();
			save_var auto_show_matching_characters = php_auto_show_delim_chars;
			save_var mode_move_level = perl_move_level;
			break;
	}
	return default_auto_show_delimiter();
}

command php_mode()
{
	html_like_mode(0);
	major_mode = _php_mode_name;
	sublanguage_code = SUBLANG_PHP;
	auto_show_matching_characters = auto_show_php_delimiters
									? php_auto_show_delim_chars : 0;
	if (php_tab_override > 0)
		tab_size = php_tab_override;
	try_calling("php-mode-hook");
}

set_mark_cmd(type)	// set mark equal to point, and begin defining
{			// the specified type of region
	int now = is_highlight_on();	// get type of current region

	if (!now || type == now && !mark_unhighlights) {
		do_set_mark(point); // start new region here
		say("Mark set.");
		if (type == REGRECT && mark_rectangle_expands)
			to_virtual_column(virtual_column() + 1);
	} else if (type == now && mark_unhighlights) {
		if (type == REGRECT && mark_rectangle_expands
			  && ABS(virtual_column() - virtual_mark_column()) == 1)
			to_virtual_column(virtual_column() - 1);
		highlight_off();	// region is same type, remove it
		return;
	}
	set_region_type(type);	// define region of this type & show it
	highlight_on();
}

custom_fix_mode_vars()
{
	char start[FNAMELEN], end[FNAMELEN], *s;
	save_var point = 0;
	*start = *end = 0;
	if (re_search(1, "^%+mode-start .*: [0-9]+ \"(.*)\"$"))
		grab(find_group(1, 1), find_group(1, 0), start);
	point = 0;
	if (re_search(1, "^%+mode-end .*: [0-9]+ \"(.*)\"$"))
		grab(find_group(1, 1), find_group(1, 0), end);
	s = strstr(start, "\\z");
	if (s)
		*s = 0;
	s = strstr(end, "\\z");
	if (s)
		*s = 0;
	point = size();
	if (*start || *end)
		bprintf("+mode-format char array variable: %d \"%s%%b%%E%%M%s\"\n",
				ptrlen(mode_format), *start ? start : " ",
				*end ? end : " %d%p %s%f");
}

command list_customizations()
{
	char cmd[FNAMELEN], *args, time[FNAMELEN];
	char ext_file[FNAMELEN], settings_file[FNAMELEN], *s;
	int tmp, cur, ext, i, max, err;
	int cust = prep_customizations(), clist;

	iter = 1;
	tiled_only();
	get_customization_directory(ext_file);
	strcat(ext_file, "extensions-list");
	get_customization_directory(settings_file);
	strcat(settings_file, "settings-list");
	delete_file(settings_file);
	i = name_match("-loaded-eel-file-", STARTMATCH);
	save_var bufnum = ext = zap("-list-customizations");
	give_binding(cmd, "exit-level");
	bprintf("Epsilon will include the following EEL extension files as part\n"
			"of your customizations.  Edit this list if needed, then\n"
			"press %s to begin.\n\n%p\n", cmd, color_class viewed_text);
	point--;
	for (; i; i = name_match("", 0))	// List new extensions by searching
		if (name_type(i) == NT_VAR) {	// for id variables.
			s = get_str_var(i);
			bprintf("%s%s\n", s, (check_file(s) == CHECK_FILE)
					? "" : " (not found)");
		}
	to_buffer_num(ext);
	point = 0;
	do {
		recursive_edit();
		if (!buf_exist(ext))
			return;
		point = 0;
		point += parse_string(1, "(.|\n.)*\n*");
		i = test_valid_files(point, size());
		if (i == 2) {
			buf_zap(ext);
			return;
		}
	} while (i);
	point = 0;		// Drop the explanatory text.
	delete(0, parse_string(1, "(.|\n.)*\n*"));
	if (err = file_write(ext_file, FILETYPE_BINARY)) {
		file_error(err, ext_file, "write error");
		return;
	}
	save_var reject_client_connections = 1;
	strcpy(cmd, argv[0]);
	convert_to_8_3_filename(cmd);
	sayput("Running %s to create original settings list...", cmd);
	args = " -s original -p0 -noinit -r build-settings-file";
	err = shell_with_args(cmd, args);
	if (err)
		error("Error running %s %s", cmd, args);
	sayput("Reading original settings list...");
	bufnum = tmp = tmp_buf();
	if (err_file_read(settings_file, FILETYPE_BINARY))
		return;
	bufnum = cur = tmp_buf();
	max = final_index();
	for (i = 1; i <= max; i++)
		list_one(bufnum, i, 3);
	use_default = 0;
	bufnum = cur;
	clist = make_difference_list(tmp, cur);
	restore_vars();
	buf_delete(tmp);
	buf_delete(cur);
	to_buffer_num(cust);
	current_time_and_date(time);
	save_spot point;
	i = point;
	bprintf("\n# List-customizations created these settings %s.\n", time);
	fix_extension_list(ext);
	grab_buffer(ext);
	grab_buffer(clist);
	buf_delete(clist);
	delete_file(settings_file);
	delete_file(ext_file);
	bprintf("\n");
	point = i;
	if (!has_arg)
		comment_old_customizations(point);
	if (!list_undefined_customizations())
		say("");
}

command import_customizations()
{
	int tmp = zap("import-customizations"), i, ver = 0, err, clean = !has_arg;
	int debug = (iter == 99);
	char buf[2000], maindir[FNAMELEN], bindir[FNAMELEN];
	char origsta[FNAMELEN], cursta[FNAMELEN], *s, *exe_ext = "", pat[FNAMELEN];
	char answer[6][2000], ranswer[6][2000], oldname[FNAMELEN];
	char origdir[FNAMELEN], *epsout = lookpath("epsout.e");

	iter = 1;
	tiled_only();
	getcd(origdir);
	on_exit chdir(origdir);
	putenv("EPSRUNS");		// Prevent error msg in old Epsilons.
	save_var bufnum;
	if (!epsout)
		error("The file epsout.e is missing.");
	for (i = EELVERSION / 10 - 1; i >= 8 ; i--) {
		if (is_win32) {
			sprintf(buf, "EPSPATH%02d", (i < 10) ? (i * 10) : i);
			if (!verenv(buf))
				continue;
		} else {
			sprintf(buf, "~/.epsilon/epsilon-v%d.sta", i);
			if (!check_file(buf))
				continue;
		}
		buf_printf(tmp, "%d\n", ver = i);
	}
	set_buf_point(tmp, 0);
	switch (count_lines_in_buf(tmp, 0)) {
		case 0:		ver = 0; break;
		case 1:		break;
		default:
			*buf = 0;
			buf_delete_final_newlines(tmp);
			if (!get_choice(tmp, buf, " Update from version ",
							"Update from which version? ",
							"OK", "Cancel", ""))
				return;
			ver = numtoi(buf);
			say("");
	}
	*maindir = *bindir = *origsta = *cursta = 0;
	if (is_win32 && ver) {
		exe_ext = ".exe";
		sprintf(buf, "EPSPATH%02d", (ver < 10) ? (ver * 10) : ver);
		if (s = verenv(buf)) {
			strcpy(maindir, s);
			absolute(maindir);
			sprintf(origsta, "%s%coriginal.sta", maindir, path_sep);
		}
		sprintf(buf, "BINPATH%02d", (ver < 10) ? (ver * 10) : ver);
		if (*maindir)
			sprintf(bindir, "%s\\bin", maindir);
		s = verenv(buf);
		if (s)
			strcpy(bindir, s);
		absolute(bindir);
		if (*maindir)
			sprintf(cursta, "%s\\epsilon.sta", maindir);
	} else if (ver) {		// Set vars for Unix.
		sprintf(buf, "/usr/local/epsilon%d*/bin/epsilon", ver);
		int max = 0;		// Find newest match of pattern.
		struct file_info fi;
		s = file_match(buf, STARTMATCH);
		*buf = 0;
		while (s && check_file(s, &fi)) {
			if (!max || fi.raw_file_date_low > max) {
				strcpy(buf, s);
				max = fi.raw_file_date_low;
			}
			s = file_match("", 0);
		}
		if (*buf) {
			get_tail(buf, 0)[-1] = 0;
			strcpy(bindir, buf);
			strcpy(origsta, buf);
			strcpy(get_tail(origsta, 0), "original.sta");
		}
		sprintf(cursta, "~/.epsilon/epsilon-v%d.sta", ver);
	}
	save_var window_bufnum;
	to_buffer_num(tmp);
	buf_zap(bufnum);
	sprintf(oldname, "Epsilon %d", ver);
	if (!ver)
		strcpy(oldname, "the old version");
	bprintf("Epsilon will try to import any customizations you saved in your\n");
	bprintf("state file in %s.  Please correct any of\n", oldname);
	give_binding(buf, "exit-level");
	bprintf("these settings as necessary, then press %s\n", buf);
	bprintf("to begin the import process.\n\n");

	if (!ver) {
		bprintf("Epsilon couldn't find any %s,\n", is_win32
				? "Epsilon registry entries"
				: "customized state file in ~/.epsilon");
		bprintf("so you'll have to fill in this information.\n\n");
	}
	bprintf("1. Your customized state file for %s is:\n%p%s\n%p\n",
			oldname, color_class viewed_text,
			*cursta ? cursta : "full path for your .sta file", -1);
	if (*cursta && check_file(cursta) != CHECK_FILE)
		bprintf("(This file doesn't exist.)\n\n");
	bprintf("2. The original pre-customization state file for %s\n", oldname);
	bprintf("is:\n%p%s\n%p\n",
			color_class viewed_text,
			*origsta ? origsta : "full path for original.sta", -1);
	if (*origsta)
		if (check_file(origsta) != CHECK_FILE)
			bprintf("(This file doesn't exist.)\n\n");
	bprintf("3. If you customized %s by loading EEL files,\n", oldname);
	bprintf("please list each .e file's pathname here, one per line:\n%p\n%p\n",
			color_class viewed_text, -1);
	sprintf(buf, "full path for epsilon%s", exe_ext);
	if (*bindir)
		sprintf(buf, "%s%c%s%s", bindir, path_sep,
				(ver < 10) ? "e32" : "epsilon", exe_ext);
	bprintf("4. The Epsilon%s executable file for %s is:\n%p%s\n%p\n",
			is_win32 ? " (GUI, not console version)" : "", oldname,
			color_class viewed_text, buf, -1);
	if (*bindir && check_file(buf) != CHECK_FILE)
		bprintf("(This file doesn't exist.)\n\n");
	sprintf(buf, "full path for eel%s", exe_ext);
	if (*bindir)
		sprintf(buf, "%s%ceel%s", bindir, path_sep, exe_ext);
	bprintf("5. The EEL compiler executable for %s is:\n%p%s\n%p\n",
			oldname, color_class viewed_text, buf, -1);
	if (*bindir && check_file(buf) != CHECK_FILE)
		bprintf("(This file doesn't exist.)\n\n");
	get_customization_directory(buf);
	strcat(buf, "upd");
	mkdir(buf);
	bprintf("6. Use this working directory for intermediate files:\n%p%s\n%p\n",
			color_class viewed_text, buf, -1);
	if (buf && check_file(buf) != CHECK_DIR)
		bprintf("(Warning: Couldn't create this directory.)\n\n");
	for (;;) {
		err = point = 0;
		recursive_edit();
		if (!buf_exist(tmp))
			return;
		bufnum = tmp;
		point = 0;		// Now retrieve the answers.
		for (i = 0; i < 6; i++) {
			sprintf(pat, "^%d%%. [^:]+:\n((.|\n.)*)\n\n", i + 1);
			if (!re_search(1, pat)) {
				say("Can't interpret response %d", i + 1);
				err = 1;
				break;
			}
			grab(find_group(1, 1), find_group(1, 0), answer[i]);	// 8.3 version.
			grab(find_group(1, 1), find_group(1, 0), ranswer[i]);	// For display.
			if (i == 2) {
				err = test_valid_files(find_group(1, 1), find_group(1, 0));
				if (err == 2)
					quick_abort();
				err = err ? 3 : 0;	// Don't prompt again.
			} else if (check_file(answer[i]) != ((i == 5)
				? CHECK_DIR : CHECK_FILE)) {
				say("Response %d: Can't read %s \"%f\".",
					  i + 1, (i == 5) ? "directory" : "file", answer[i]);
				err = 1;
				break;
			} else
				convert_to_8_3_filename(answer[i]);
		}
		if (!err)
			break;
		if (err != 3
			  && !ask_yn("Try again", "Would you like to correct your response? ",
						 "Yes", "No", 0))
			quick_abort();
	}
	if (chdir(answer[5]))
		error("Couldn't set directory to %f.", answer[5]);
	buf_zap(tmp);	// Compile epsout & extensions, then save list of them.
	compile_for_import_customizations(epsout, answer[4], answer[5], 0);
	compile_for_import_customizations(answer[2], answer[4], answer[5], tmp);
	if (err = file_write("extensions-list", FILETYPE_BINARY)) {
		file_error(err, "extensions-list", "write error");
		return;
	}

	// Run the old version with orig state file.
	save_var reject_client_connections = 1;
	sprintf(buf, " -s%s -l%s%cepsout.b -r%sbuild_settings_file",
			answer[1], answer[5], path_sep, debug ? "debug_" : "");
	show_text(-2, 1, "Run %s %s", answer[3], buf);
	shell_with_args(answer[3], buf);
	buf_zap(tmp);
	if (err = file_write("extensions-list", FILETYPE_BINARY)) {
		file_error(err, "extensions-list", "write error");
		return;
	}
	if (err_file_read("settings-list", FILETYPE_BINARY))
		return;
	// Run the old version with new state file.
	sprintf(buf, " -s%s -l%s%cepsout.b -r%sbuild_settings_file",
			answer[0], answer[5], path_sep, debug ? "debug_" : "");
	show_text(-2, 1, "Run %s %s", answer[3], buf);
	shell_with_args(answer[3], buf);
	int cur = tmp_buf();
	on_exit buf_delete(cur);
	bufnum = cur;
	if (err_file_read("settings-list", FILETYPE_BINARY))
		return;
	int clist = make_difference_list(tmp, cur);
	on_exit buf_delete(clist);	// Now format the list of differences.
	int cust = prep_customizations();
	bufnum = cust;
	current_time_and_date(buf);
	save_var point;
	i = point;
	bprintf("\n# Import-customizations created these settings %s\n", buf);
	bprintf("# from %s.\n", ranswer[0]);
	int ext = tmp_buf();
	on_exit buf_delete(ext);
	buf_printf(ext, "%s\n", ranswer[2]);
	fix_extension_list(ext);
	grab_buffer(ext);
	grab_buffer(clist);
	bprintf("\n");
	restore_vars();
	to_buffer_num(cust);
	point = i;
	if (clean)
		comment_old_customizations(point);
	if (!list_undefined_customizations())
		say("Done.");
}

// Return code for word from here to point (something with alpha or
// digits).  Return for 1 for command, 2 for comment, 0 if unknown.
is_batch_keyword(int from)
{
	char buf[500], *s;

	save_var case_fold = 1;
	if (point - from > sizeof(buf) / sizeof(char) - 10)
		save_var point = from + sizeof(buf) / sizeof(char) - 10;
	buf[0] = '|';		// get identifier, between | chars
	grab(from, point, buf + 1);
	strcpy(buf + point - from + 1, "|");
	if ((s = strchr(buf, '.')) && matches_at(from, -1, "\n[ \t]*"))
		*s = 0;
	if (strstr("|rem|", buf) && !matches_at(from, -1, "[^ \n\t@][@ \t]*"))
		return 2;
	if (strstr("|assoc|at|attrib|break|cacls|call|cd|chcp|chdir"
			   "|chkdsk|chkntfs|cls|cmd|cmdextversion|color|comp"
			   "|compact|convert|copy|date|defined|del|dir|diskcomp"
			   "|diskcopy|do|doskey|echo|else|endlocal|equ|erase"
			   "|errorlevel|exist|exit|fc|find|findstr|for|format"
			   "|ftype|geq|goto|graftabl|gtr|help|if|in|label|leq"
			   "|lss|md|mkdir|mode|more|move|neq|not|path|pause"
			   "|popd|print|prompt|pushd|rd|recover|ren|rename"
			   "|replace|rmdir|set|setlocal|shift|sort|start|subst"
			   "|time|title|tree|type|ver|verify|vol|xcopy|cmd"
			   "|command|", buf))
			return 1;
	return 0;
}

drag_drop_handler() on reg_tab[WIN_DRAG_DROP]
{
	char buf[FNAMELEN], *s, *t, dir[FNAMELEN];
	int line = -1, column = -1, j, dir_flags = directory_flags, first = 1;
	int had_action = 0;

	tiled_only();
	if (mouse_screen >= 0)
		to_mouse_window(mouse_x, mouse_y, mouse_screen);
	while (j = drag_drop_result(buf)){
		if (*buf == '+') {			// a +linenum parameter
			line = numtoi(buf + 1);
			s = strchr(buf, ':');
			if (s)
				column = numtoi(s + 1);
		} else if (*buf == '-') {
			switch (buf[1]) {
				case 'd':
					if (buf[2])
						t = buf + 2;
					else if (!drag_drop_result(t = buf)) {
						say("-d flag missing variable name.");
						break;
					}
					s = index(t, '=');
					if (!s)		// accept either VAR=VAL or VAR!VAL
						s = index(t, '!');
					if (!s)
						break;
					*s = 0;
					if (j = find_index(t)) {
						set_var_val(j, s + 1, 1);
						continue;
					} else
						say("No such variable %s.", t);
					break;
				case 'l':
					if (buf[2])
						load_commands(buf + 2);
					else if (drag_drop_result(buf))
						load_commands(buf);
					else
						say("-l flag missing file name.");
					break;
				case 'r':
					had_action = 1;
					if (buf[2])
						t = buf + 2;
					else if (!drag_drop_result(t = buf)) {
						say("-r flag missing function name.");
						break;
					}
					s = index(t, '=');
					if (!s)		// accept either CMD=ARG or CMD!ARG
						s = index(t, '!');
					if (s)
						*s = 0;
					if (j = find_index(t)) {
						if (s)
							(*(int (*)())j)(s + 1);
						else
							(*(int (*)())j)();
					} else
						say("No such function %s.", t);
					break;
				case 'w':
					dir_flags = numtoi(buf + 2);
					break;
			}
			continue;
		} else if (*buf) {
			windows_foreground();
			had_action = 1;
			absolute(buf);
			if (is_directory(buf) || is_remote_file(buf) || (is_pattern(buf) & 2))
				do_find(buf, default_translation_type);
			else
				find_it(buf, default_translation_type);
			if (line >= 0) {
				go_line(line);
				line = -1;
			}
			if (column >= 0) {
				move_to_column(column);
				column = -1;
			}
			if ((dir_flags & DIR_CWD_FROM_FILE) && first) {
				strcpy(dir, buf);
				*get_tail(dir, 1) = 0;
				chdir(dir);
			}
			first = 0;
			if (j == 2)
				show_binding("Use %s to release the Epsilon client when you finish.",
							 "resume-client");
		}
	}
	if (!had_action)			// Ensure we're not minimized, unless we got
		windows_foreground();	// a -r command but no file.
}

perl_after_expression(int p, int strict)
{
	int ch;
	if (in_shell_buffer)
		return 0;
	if (strict) {
		save_var point = p;
		if (!parse_string(-1, "}[ \t]*")) {	// If not an eval block or similar.
			// Move back over whitespace and comments.
			for (;;) {
				if (parse_string(-1, "[ \t\n]+"))
					point = matchend;
				else if (parse_string(-1, "#.*")
						 && get_character_color(matchend)
						 == color_class perl_comment)
					point = matchend;
				else
					break;
			}
			if (!parse_string(-1,  "<^word>[0-9]+([-0-9e.]+)?|[])`]")
				  // Not a number, or one of the characters ] or ) or `.
				  && !parse_string(-1, "%$%^?[^ \t\n]")) {
				// Or $, then perhaps a ^ and a nonspace.
				point -= parse_string(-1, "(%+%+|--)?");
				// Skip back over any ++ or --.
				if (!parse_string(-1, "<word>+"))	// Only other option is a word.
					return 0;
				if (parse_string(-1, "<^word>"
								 "(if|unless|package|sub|import|no|require|use|die"
								 "|do|dump|else|elsif|eval|exec|exit|for|foreach|if"
								 "|return|unless|until|while|BEGIN|END|local|my"
								 "|continue|goto|last|next|redo|qq|qx|q|qw|qr|m|s|tr|y"
								 "|format|__END__|__DATA__|split|scalar|warn)")
					  && !strchr("$%&#@", character(matchend)))
					return 0;		// That's not any Perl keyword.
			}
		}
		return 1;
	} else {
		if (!matches_at(p, -1, "([$%&#@]((<word>+(->|'))?<word>+)"
						// A variable or function with $ or similar before it.
						"(%+%+|--)?"
						// Possibly followed by ++ or --.
						"|%$%^?[^ \t\n]"
						// Or $, then perhaps a ^ and a nonspace.
						"|<^word>[0-9]+([-0-9e.]+)?|[])`])"
						// Or a number.  Or one of the characters ] or ) or `.
						"(#.*|[ \t\n])*|}[ \t]*"))
			// All followed by whitespace or a comment.
		return 0;
	}
	ch = get_character_color(matchend);
	if (ch == color_class perl_comment || ch == color_class perl_string)
		return 0;
	return 1;
}

// Match #1 has a terminating string.  See if it's a likely here document
// and color if so.
color_perl_here_doc(int start, int interpolated)
{
	char line[FNAMELEN], pat[FNAMELEN], *s;
	int p1, p2, end, type = 0;

	grab(find_group(1, 1), find_group(1, 0), line);
	p1 = give_end_line() + 1;
	if (sublanguage_code != SUBLANG_PHP
		  && get_tagged_region("color-perl-as-unit", p1) != -1) {
		get_tagged_region("color-perl-as-unit", p1, NULL, &p1);
		start = p1;
	}
	sprintf(pat, "^%q%s$", line, sublanguage_code == SUBLANG_PHP ? ";?" : "");
	if (re_search(1, pat))		// Find end of document (or eof).
		p2 = matchstart;
	else
		p2 = size();
	end = point;
	s = get_extension(line);
	if (s)
		if (!strcmp(s, ".tex"))
			type = 1;
		else if (!strcmp(s, ".html"))
			type = 2;
	set_character_color(start, point, color_class perl_constant);
	set_tagged_region("color-perl-as-unit", start, point, type);
	save_var point;
	save_var narrow_start = p1;
	save_var narrow_end = size() - p2;
	switch (type) {
		case 1: do_color_tex_range(p1, p2, interpolated); break;
		case 2: color_html_range(p1, p2); break;
	}
	restore_vars();
	if (interpolated)
		color_perl_interpolated(start, point, 0);
	point = end;
}

// Just before a delimiter character.  Find its match, then color.  If
// two, we're parsing something like s/// so repeat.  If interpolated,
// color any variable references like $var within it unless the
// delimiter is '.  (If interpolated is 2, color variable references
// even with '.)  If modifiers, also parse the modifiers after it (2
// means they're regex modifiers).

color_perl_delimiters(int two, int interpolated, int modifiers)
{
	int start = point, c, end, level = 0, rstart, rend;
	char *openers = "([{<", *closers = ")]}>", *p;
	char pat[10];

	point += parse_string(1, "[ \t\n]+");
	set_character_color(start, point, color_class perl_comment);
	rstart = start = point;
	p = strchr(openers, c = end = curchar());
	if (isalnum(c) || isspace(c))
		return;			// No valid delimiter, so stop now.
	if (c != '\'' && interpolated != 2)
		interpolated = 0;
	if (p) {			// A paired delimiter.
		end = closers[p - openers];
		sprintf(pat, "[%c%c]", end, c);
		while (re_search(1, pat))
			if (character(point - 1) == c)
				level++;
			else if (--level <= 0)
				break;
		set_character_color(start, point, color_class perl_string);
		if (interpolated)
			color_perl_interpolated(start, point, 0);
	} else {
		point++;
		color_perl_fwd_quote(c, start, interpolated);
	}
	rend = point;
	if (two)		// A command like s///, so look for 2nd set.
		if (p) {	// After a paired delimiter, whitespace and pick new char.
			start = point;
			point += parse_string(1, "([ \t\n]|#.*)+");
			set_character_color(start, point, color_class perl_comment);
			color_perl_delimiters(0, interpolated, 0);	// One more pair.
		} else		// Find 3rd instance of char.
			color_perl_fwd_quote(c, point, interpolated);
	if (modifiers) {
		start = point;
		if (modifiers == 2 && parse_string(1, "[a-z]*x"))
			color_perl_interpolated(rstart + 1, rend - 1, 1);
		point += parse_string(1, "[a-z]+");
		set_character_color(start, point, color_class perl_keyword);
	}
}

color_perl_keyword(int start)
{
	char buf[100];

	if (point - start < sizeof(buf) / sizeof(*buf) - 10
		  && (point - start > 1 || character(start - 1) != '-')) {
		buf[0] = '|';		// get identifier, between | chars
		grab(start, point, buf + 1);
		strcat(buf, "|");
	} else	// Too long to be a keyword, or else it's -s or similar.
		strcpy(buf, "not a keyword");
	if (sublanguage_code == SUBLANG_PHP) {
		if (strstr("|if|else|elseif|while|do|for|foreach|break|continue"
				   "|switch|declare|return|require|require_once|include"
				   "|include_once|endif|endwhile|endfor|endforeach|endswitch"
				   "|function|", buf))
			  set_character_color(start, point, color_class perl_keyword);
		else if (isdigit(character(start))) {		// a number
			if (matches_at(start, -1, "-?%.?"))	// number like -.25
				start = matchend;
			point += parse_string(1, "([A-Za-z0-9._]|[Ee]-)*");		// -3.0e-7
			set_character_color(start, point, color_class perl_constant);
		} else
			set_character_color(start, point, color_class perl_function);
		return;
	}
	if (strstr("|package|sub|", buf)) {
		point += parse_string(1, "[ \t]+");
		set_character_color(start, point, color_class perl_keyword);
		if (parse_string(1, "(<word>+('|::))*<word>+"))
			set_character_color(matchstart, point = matchend,
								color_class perl_function);
	} else if (strstr("|import|no|require|use|", buf)) {
		point += parse_string(1, "[ \t]+");
		set_character_color(start, point, color_class perl_keyword);
		if (parse_string(1, "(<word>+('|::))*<word>+"))
			set_character_color(matchstart, point = matchend,
								color_class perl_constant);
	} else if (strstr("|die|do|dump|else|elsif|eval|exec|exit|for|foreach|if|"
			   "return|unless|until|while|BEGIN|END|", buf)) {
		set_character_color(start, point, color_class perl_keyword);
	} else if (strstr("|local|my|", buf)) {		// special color
		set_character_color(start, point, color_class perl_constant);
	} else if (strstr("|continue|goto|last|next|redo|", buf)) {
		point += parse_string(1, "[ \t]+");
		set_character_color(start, point, color_class perl_keyword);
		if (parse_string(1, "[0-9A-Za-z_]+"))	// A label.
			color_perl_label(matchstart, matchend);
	} else if (strstr("|qq|", buf) && !isspace(curchar())) {
		set_character_color(start, point, color_class perl_keyword);
		color_perl_delimiters(0, 2, 0);	// One pattern, always interpolated.
	} else if (strstr("|qx|", buf) && !isspace(curchar())) {
		set_character_color(start, point, color_class perl_keyword);
		color_perl_delimiters(0, 1, 0);	// One pattern, interpolated.
	} else if (strstr("|m|qr|", buf) && !isspace(curchar())) {
		set_character_color(start, point, color_class perl_keyword);
		color_perl_delimiters(0, 1, 2);	// One pattern, interp, w/modifiers.
	} else if (strstr("|q|qw|", buf) && !isspace(curchar())) {
		set_character_color(start, point, color_class perl_keyword);
		color_perl_delimiters(0, 0, 0);	// One pattern, uninterpolated.
	} else if (strstr("|s|", buf) && !isspace(curchar())) {
		set_character_color(start, point, color_class perl_keyword);
		color_perl_delimiters(1, 1, 2);	// Two patterns, interpolated.
	} else if (strstr("|tr|y|", buf) && !isspace(curchar())) {
		set_character_color(start, point, color_class perl_keyword);
		color_perl_delimiters(1, 0, 1);	// Two patterns, uninterpolated.
	} else if (strstr("|format|", buf)) {
		set_character_color(start, point, color_class perl_keyword);
		color_perl_format();
	} else if (strstr("|__END__|__DATA__|", buf)) {	// Skip to the end.
		set_character_color(start, point, color_class perl_keyword);
		set_character_color(point, size(), color_class perl_constant);
		point = size();
	} else if (character(point) == ':' && matches_at(point, 1, ":[^:]")
			   && matches_at(start, -1, "^[ \t]*")) {		// a label
		set_character_color(start, point, color_class perl_constant);
	} else
		set_character_color(start, point, color_class perl_function);
}

// Found a word; see if it's decorated like a variable or subr.
color_perl_word(int start)
{
	int true_perl = !in_shell_buffer && sublanguage_code != SUBLANG_PHP;

	if (sublanguage_code == SUBLANG_PHP)
		point += parse_string(1, "(::<word>+)+");	// pkg::var
	else if (!in_shell_buffer)
		if (strchr("$*@%&<", character(start - 1))
			  || !parse_string(-1, "</word>(q|qq|qx|qw|m|qr|s|tr)"))
			point += parse_string(1, "(('|::)<word>+)+");	// pkg::var
	if (strchr(!true_perl ? "$" : "$*@%",
			   character(start - 1))) {	// $var, @var etc.
		set_character_color(start - 1, point, color_class perl_variable);
	} else if (true_perl && character(start - 1) == '&') {	// &funcname
		set_character_color(start - 1, point, color_class perl_function);
	} else if (true_perl && character(start - 1) == '<'
			   && character(point) == '>') {
		set_character_color(start - 1, point, color_class perl_constant);
	} else if (isdigit(character(start))) {		// a number
		if (matches_at(start, -1, "-?%.?"))	// number like -.25
			start = matchend;
		point += parse_string(1, "([A-Za-z0-9._]|[Ee]-)*");		// -3.0e-7
		set_character_color(start, point, color_class perl_constant);
	} else if (true_perl)
		if (parse_string(1, "[ \t\n]*=>"))
			set_character_color(start, point, color_class perl_constant);
		else if (character(start - 1) == '{' && curchar() == '}')
			set_character_color(start, point, color_class perl_constant);
		else
			color_perl_keyword(start);
}

color_this_perl_range(from, to)
{
	int t = -1, talk, c, talk_at = 0, after_here_doc = 0;
	char *langname;

	save_var point, matchstart, matchend, case_fold = 0;
	talk = (to - from > 30000);	// Show status during long delays.
	point = from;
	langname = sublanguage_code == SUBLANG_PHP ? "PHP"
			   : in_shell_buffer ? "Shell" : "Perl";

	while (point < to) {
		if (!re_search(1, "[0-9A-Za-z_]+|[\"'`/#$&<=\n]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (c = character(point - 1)) {		// check last char
			case '#':			// found comment
				to_end_line();
				set_character_color(t, point, color_class perl_comment);
				break;
			case '&':
				if (in_shell_buffer)
					break;
				if (parse_string(1, "((<word>+('|::))*|'|::)<word>+"))
					set_character_color(t, point = matchend,
										color_class perl_function);
				break;
			case '$':
				if (in_shell_buffer && curchar() == '"') {	// Translated str.
					set_character_color(t, point, color_class perl_string);
					break;
				}
				if (in_shell_buffer || sublanguage_code == SUBLANG_PHP) {
					if (parse_string(1, "<word>+|%{([^{}\n]|%{[^{}\n]*%})+%}|[^\n`]"))
						set_character_color(t, point = matchend,
											color_class perl_variable);
					break;
				}
				if (!isident(curchar()) && !strchr("#{}", curchar())) {
					if (character(point++) == '^')
						point++;	// $', $?, $^Z or similar special var
					set_character_color(t, point, color_class perl_variable);
				} else if (parse_string(1, "#?((<word>+('|::))*|'|::)<word>+"))
					set_character_color(t, point = matchend,
										color_class perl_variable);
				else if (parse_string(1, "#?{"))
					set_character_color(t, point = matchend - 1,
										color_class perl_variable);
				break;
			case '\'':
				if (in_shell_buffer ? character(t - 1) != '\\'
									: !strchr("&$@%*", character(t - 1)))
					color_perl_fwd_quote(c, t, 0);	// old syntax pkg'func
				break;
			case '"':		// found a string literal
			case '`':
				if (in_shell_buffer && character(t - 1) == '\\')
					break;
				color_perl_fwd_quote(c, t, 1);
				break;
			case '/':
				if (sublanguage_code == SUBLANG_PHP) {
					if (curchar() == '/') {
						nl_forward();
						set_character_color(t, point, color_class perl_comment);
					} else if (curchar() == '*') {
						search(1, "*/");
						set_character_color(t, point, color_class perl_comment);
					}
					break;
				}

				// if previous thing looks like an expression, skip it
				if (!in_shell_buffer && !perl_after_expression(t, 1)) {
					point--;
					color_perl_delimiters(0, 1, 2);
				}
				break;
			case '<':
				if (sublanguage_code == SUBLANG_PHP) {	// Look for <<<word
					if (parse_string(1, "%<%<(<alpha|_>+)"))
						color_perl_here_doc(t, 1);
					break;
				}
				if (perl_after_expression(t, 0) || character(t - 1) == '<')
					break; // It's a shift or lessthan op, not "here" doc or input op.
				if (curchar() == '<') {		// found <<, possible "here" document.
					point++;	// Next must be an identifier or text in quotes.
					if (parse_string(1, in_shell_buffer
									 ? "[ \t]*[-\\]?(<alpha|_>+)"
									 : "(<alpha|_>+)"))
						color_perl_here_doc(t, 1);
					else if (parse_string(1, "[ \t]*\"([^\"\n]+)\""))
						color_perl_here_doc(t, 1);
					else if (parse_string(1, "[ \t]*'([^'\n]+)'"))
						color_perl_here_doc(t, 0);
					else if (parse_string(1, "[ \t]*`([^`\n]+)`"))
						color_perl_here_doc(t, 1);
					else
						break;
					after_here_doc = 1;
					break;
				}
				if (in_shell_buffer || sublanguage_code == SUBLANG_PHP)
					break;
				point--;	// File glob operator.
				color_perl_delimiters(0, 1, 0);
				break;
			case '\n':
				if (after_here_doc)
					if (get_tagged_region("color-perl-as-unit", point) != -1)
						to_end_line();	// Already colored this line.
					else
						after_here_doc = 0;
				break;
			case '=':			// Found POD (documentation).
				if (in_shell_buffer || sublanguage_code == SUBLANG_PHP)
					break;
				if (parse_string(-1, "^=")) {
					search(1, "\n=cut");
					set_character_color(t, point, color_class perl_comment);
				}
				break;
			default:		// found identifier, kywd, or number
				color_perl_word(t);
				break;
		}
		if (talk && point > talk_at + 2000) {
			note("Coloring %s program: %d%% complete...",
				 langname, muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	if (talk)
		note("");
	return point;
}

int perl_try_space()
{
	int i, skip_unit = 0, c;

	for (;;) {
		if (index(" \t\n\f", c = character(point - 1))) {
			point--;
			if (c == '\n')
				skip_unit = 0;
		} else if (get_character_syntax_color(point - 1) == color_class perl_comment
				   && get_character_color(point - 1, &i) == color_class perl_comment
				   && strchr("#/", character(i)))
			point = i;
		else if (!skip_unit
				 && get_tagged_region("color-perl-as-unit", point - 1, &i) != -1) {
			point = i;
			skip_unit = 1;
			to_end_line();
		} else
			return point == narrow_start;
	}
}

perl_compute_indent()
{
	int ind = 0;		/* indentation to use */
	int orig = point, i;
	int first_char = curchar(); /* cur line's first char */
	int pstart;		/* start position of prev line */
	int pafter;		/* after label on prev line */
	int pend;		/* end position of prev line */
	int prev_end = ';';	/* last char on previous line */
	int norm_narrow;	/* narrowing setting for most parsing */

	save_var point;
	save_var Closeback = perl_closeback;
	save_var Topindent = perl_topindent;
	save_var c_indent = perl_indent;
	save_var c_label_indent = perl_label_indent;
	save_var c_contin_offset = perl_contin_offset;
	save_var c_align_contin_lines = perl_align_contin_lines;
	save_var c_brace_offset = perl_brace_offset;
	save_var c_top_braces = perl_top_braces;
	save_var c_top_contin = perl_top_contin;
	save_var c_top_struct = perl_top_struct;

	if (sublanguage_code != SUBLANG_PHP)
		save_var narrow_start = 0;
	recolor_buffer_range(orig, orig + 1);
	if (get_tagged_region("color-perl-as-unit", orig, &i) != -1
		  || get_character_color(orig, &i) == color_class perl_constant
		  && matches_at(i, 1, "%<%<"))		// Keep original's indentation
		return get_indentation(orig);		// if in here doc.
	if (get_character_color(orig - 1, &i) == color_class perl_string)
		return get_indentation(orig);		// or in string.

	norm_narrow = fix_narrowing(point - c_look_back);
	narrow_start = norm_narrow;
	if (first_char == '\f')
		return 0;
	else if (first_char == '}')
		return cind_close();

	to_begin_line();
	if (parse_string(1, "[ \t]*\n") && after_many_comments()) {
		re_search(-1, "[^ \t\n]");	/* find previous non-blank line */
		return get_indentation(point);
	}
	cind_space();			/* skip whitespace and comments */
	if (character(point - 1) == ':')
		move_before_labels();
	prev_end = character(point - 1); /* get last relevant char */
	pend = point;

	to_statement_start();		/* if contin line, back up */
	pstart = point;			/* now at start of biggest statement */
	past_labels();			/* containing original point */
	pafter = point;			/* just after label */
	ind = current_column();
	narrow_start = norm_narrow;	/* ok to look far back again */

	if (sublanguage_code == SUBLANG_PHP && is_case(orig))
		ind = indent_case(orig);
	else if (have_label(orig)) {
		ind = c_label_indent;
	} else if (prev_end == ':' && is_label(pstart)) {
		ind = get_column(pstart);
		if (!(Closeback && just_open(orig)))
			ind += get_c_indent();
		if (just_open(orig))
			ind += c_brace_offset;
	} else if (is_else(orig)) {
		point = orig;
		ind = indent_else(ind);
	} else if (prev_end == ';') {
		if (just_open(orig) && func_start_block(pstart))
			ind = c_top_braces;
		else
			ind = indent_after_statement(orig);
	} else if (prev_end == '}') {
		point = pend;
		ind = indent_after_block(orig);
	} else if (prev_end == '{') {
		if (ind <= c_top_braces	// special case if no previous indent
			|| get_column(pend - 1) <= c_top_braces) {
			ind += top_indent(prev_end) ? get_c_indent() : c_top_braces;
			if (just_open(orig))
				ind += c_brace_offset;
			if (ind < c_top_braces)
				ind = c_top_braces;
		} else if (Closeback || curchar() != '{')
			ind += get_c_indent();
		if (just_open(pafter))
			ind -= c_brace_offset;
	} else if (just_open(orig))
		return indent_just_open(ind);
	else if (is_contin_statement(pend, 1, orig)) {
		return indent_c_contin(ind, orig, pend);
	} else if (!index(":);{}", c_line_before()) && !c_statement_start()
			   && !(Closeback && just_open(point)))
		return ind;
	else
		ind += get_c_indent();
	return ind;
}

int compile_handler(int activity, int buf, int from, int to);

do_compile(cmd)
char *cmd;
{
	int err = 0, nerr, sep_buf_ok = 0;
	int obuf = bufnum;
	char *procbuf = PROCBUF, curdir[FNAMELEN];

	_last_was_grep = 0;
	if (maybe_save_all())
		return 3;
	if (start_process_in_buffer_directory)
		get_buffer_directory(curdir);
	else
		getcd(curdir);
	drop_final_slash(curdir);
	if (!(has_feature & FEAT_MULT_CONCUR) && !another_process())
		save_var restart_concurrent = 0;

	// If we're allowed to run non-concur, then use separate buffer.
	if (concurrent_make < 2 && (has_feature & FEAT_MULT_CONCUR)
		  && compile_in_separate_buffer)
		sep_buf_ok = 1;
	else if (concurrent_make == 2)
		create_concur("", PROCBUF);
	else if (!sep_buf_ok && concurrent_make == 0)
		try_exit_concurrent();
	if (!sep_buf_ok && clear_process_buffer)
		zap(PROCBUF);
	if (sep_buf_ok) {
		int i;
		char newname[FNAMELEN];
		bufnum = create(procbuf = "compilation-output");
		for (i = 2; is_process_buffer(bufnum); i++) {
			sprintf(procbuf = newname, "compilation-output-%d", i);
			bufnum = create(procbuf);
		}
		if (clear_process_buffer)
			zap(procbuf);
		process_mode();
		point = size();
		set_error_spot();
		_had_errors = 0;
		bprintf("Running `%s' in directory `%s'\n", cmd, curdir);
		type_point = size();
		if (compile_in_separate_buffer == 2)
			locate_window(procbuf, "");
		if (!concur_shell("", cmd, curdir, procbuf)) {
			update_mode_more_extra(0);
			when_activity = compile_handler;
			while (process_exit_status == PROC_STATUS_RUNNING) {
				if (char_avail() && getkey() != MOUSE_MOVE
					  && (!IS_WIN_KEY(key) || key == WIN_DRAG_DROP)) {
					ungot_key = key;	// Wait until user presses a key
					return 2;			// or process finishes.
				}
				delay(10, COND_KEY, bufnum);	// Delay 0.1 secs
			}
			err = process_exit_status;
		} else
			err = -1;
	} else if (concurrent_make == 4 && is_process_buffer(name_to_bufnum(PROCBUF))) {
		save_var bufname = PROCBUF;	// Run the program in the concur
		set_error_spot();			// proc buffer, but don't process
		_had_errors = 0;			// the user's keys until it finishes.
		if (*cmd) {
			point = size();
			bprintf("%s\n", cmd);
		}
		maybe_refresh();
		iter = 1;
		delay(-1, COND_PROC, bufnum);
		if (!is_process_buffer(name_to_bufnum(PROCBUF)) || !process_input())
			return 2;	/* else assume it's compiled */
	} else if (no_running(0)) {	/* goes to end of proc buffer */
		set_error_spot();
		_had_errors = 0;
		to_begin_line();
		bprintf("%pRunning `%s' in directory `%s'\n",
				color_class process_output, cmd, *process_current_directory
				? process_current_directory : curdir);
		type_point = point = size();
		if (*cmd)
			bprintf("%s\n", cmd);
		maybe_refresh();
		iter = 1;
		if (opsys != OS_UNIX)
			delay(-1, COND_PROC | COND_KEY, bufnum);
		if (!is_process_buffer(name_to_bufnum(PROCBUF)) || !process_input())
			return 2;	/* else assume it's compiled */
		if (!(has_feature & FEAT_DETECT_CONCUR_WAIT))
			return 5;
	} else {
		bufnum = create(PROCBUF);
		point = type_point = size();
		buffer_printf("process", "%p\nRunning `%s' in directory `%s'\n",
					  color_class process_output, cmd, curdir);
		bufnum = obuf;
		err = do_push(cmd, 1, 0);
	}
	maybe_restart_concurrent();
	nerr = do_next_error(1, cmd);	// scan for error messages
	if (!nerr && err) {	// no message recognized, but error code
		say("Process returned %d", err);
		locate_window(procbuf, "");
		point = size();		/* show end of process buffer */
		return err;
	} else
		bufnum = obuf;
	if (err)
		return 1;
	else if (nerr)
		return 4;
	return 0;
}

color_process_range(from, to)
{
	if (from >= to || spot_to_buffer(process_input_starts_at) <= 0
		  || !(process_coloring_rules & 1))
		return to;
	save_var point, matchstart, matchend;
	point = *process_input_starts_at;
	if (get_character_color(point - 1) != color_class process_prompt)
		return to;
	to_end_line();
	set_character_color(*process_input_starts_at, point,
						color_class process_input);
	if (curchar() == '\n')
		set_character_color(point, point + 1, -1);
	return to;
}

// The process is now ready for or has just been sent some input.
do_when_process_input(int buf, int type, int start)
{
	char hist_buf[FNAMELEN];

	save_var bufnum = buf;
	save_var point = start;
	if (type == PROCESS_INPUT_LINE) {
		if (opsys == OS_UNIX) {	// Our only opportunity to color.
			if (spot_to_buffer(process_input_starts_at) <= 0)
				process_input_starts_at = alloc_spot();
			*process_input_starts_at = start;
			if (process_coloring_rules & 1)
				set_character_color(give_begin_line(), start,
									color_class process_prompt);
		}
		sprintf(hist_buf, "-process-history-%s", bufname);
		create(hist_buf);
		process_cmds_to_end(hist_buf);
		nl_forward();
		if (point > start + 1)		// don't store empty (just \n) cmds
			xfer(hist_buf, start, point);
		process_cmds_to_end(hist_buf);
	} else if (type == PROCESS_INPUT_CHAR) {
		if (process_coloring_rules & 1)
			set_character_color(start, start + 1, color_class process_input);
	} else if (!type) {		// Got a request for line input (might now wait).
		if (spot_to_buffer(process_input_starts_at) <= 0)
			process_input_starts_at = alloc_spot();
		*process_input_starts_at = start;
		point = start;
		to_begin_line();
		if (process_coloring_rules & 1)
			set_character_color(point, start, color_class process_prompt);
		try_grab_directory_name(point, start);
	}
}

do_when_process_output(int buf, int start, int end)
{
	save_var bufnum = buf;
	set_character_color(start, end, color_class process_output);
	if (process_coloring_rules & 2)
		color_ansi_output(start);
	notice_password_prompt(start);
	save_var position_window_on_screen_line = 100;
	if (buf_in_window(buf) >= 0)
		refresh();
}

command process_mode()
{
	mode_keys = process_tab;
	major_mode = _process_mode_name;
	tab_size = process_tab_size;
	discardable_buffer = 1;
	recolor_range = color_process_range;	// set up coloring rules
	recolor_from_here = recolor_from_top;
	when_setting_want_code_coloring();	// maybe turn on coloring
	fill_mode = 0;
	try_calling("process-mode-hook");
	make_mode();
}

is_search_dir()	/* return 1 if it's a direction key inside search */
{		/* also replace with standard key value */
	if (key == CTRL('R') || key == ALT(CTRL('R'))
					|| full_key == rev_search_key || key == rev_search_key)
		key = CTRL('R');
	else if (key == CTRL('S') || key == ALT(CTRL('S'))
					|| full_key == fwd_search_key || key == fwd_search_key)
		key = CTRL('S');	/* substitute standard one */
	else
		return 0;
	return 1;
}

gen_replace(flags)	/* get input, then do replace */
{
	char *str, with[FNAMELEN];
	int restrict_region = 0, want_restrict = 0;

	save_var mark, mark_to_column, point, cursor_to_column;
	if (replace_in_region && is_highlight_on()) {
		want_restrict = 1;
		save_var search_in_region = 1, typing_deletes_highlight = 3;
	}
	if (_buf_readonly)
		check_modify(bufnum);
	if (has_arg)
		flags |= WORD;
	restore_vars();
	save_var mark, mark_to_column, point, cursor_to_column;
	if (want_restrict) {
		save_var search_in_region = 1, typing_deletes_highlight = 3;
		save_var restrict_search_type;
		restrict_region = make_restriction(flags);
		on_exit remove_region(restrict_region), restrict_region = 0;
	}
	flags = get_replace_text(flags, with);
	restore_vars();
	if (want_restrict) {
		if (restrict_region)
			remove_region(restrict_region);
		save_var restrict_search_type;
		restrict_region = make_restriction(flags);
	}
	str = default_search_string(flags);
	if (flags != -1) {
		if (!want_restrict && point > mark && is_highlight_on() && !(flags & REVERSE))
			exchange_point_and_mark();
		show_replace(str, with, flags);
		string_replace(str, with, flags | INCR);
	}
	if (restrict_region)
		remove_region(restrict_region);
	if (want_restrict && replace_in_region == 2)
		highlight_on();
}

string_replace(str, with, flags)
char *str, *with;
{
	spot orig, prev, oldmark;
	int skip = 0, show = 0, ret, evershow = 0, oldstart, i;
	int null_detect = -1, new_null_detect;
	int perform_case_replace = 0, this_case_replace;

	if (_buf_readonly)
		error("Buffer is read-only.");
	if (!*str)
		return 0;
	if (replace_by_case)
		perform_case_replace = want_case_replace(str, with, flags);
	orig = alloc_spot();
	prev = alloc_spot();
	save_var postpone_asynchronous_notify = 1;
	replace_num_found = replace_num_changed = 0;
	for (;;) {
		skip = 0;
		*prev = point;
		oldstart = matchstart;
		ret = do_color_searching(flags, str);
		if (!ret || ret == DSBAD) {
			point = *orig;
			break;
		} else if (ret == DSABORT)
			break;
		new_null_detect = (flags & REVERSE) ? point : (size() - point);
		if (null_detect == new_null_detect) {
			if ((flags & REVERSE) ? (point-- <= narrow_start)
								  : (point++ >= size() - narrow_end)) {
				point = *orig;
				break;	// Make sure replacing a nullable pattern like $
			}			// moves forward.
			continue;
		} else
			null_detect = new_null_detect;
		replace_num_found++;
ask:
		if (show) {
			show_replace(str, with, (flags & ~INCR));
			evershow = 1;
			show = 0;
		}
		if (!(flags & QUERY)) goto do_replace;
		if (!get_replace_key(flags, str, skip)) { // show match, then get key
			point = *orig;		// match went away (maybe flags changed)
			break;
		}
		if (key == '?' || key == 'h'
			  || binding_of(full_key, root_keys) == (short) help) {
			help_on_command("query-replace");
			show = 1;
			point = *prev;
			null_detect = -1;
			continue;
		} else if (user_abort || key == abort_key) {
			point = *orig;
			ret = DSABORT;
			goto done;
		}
		switch(tolower(key)) {
			case '^': point = *prev; matchstart = oldstart;
					  null_detect = -1;
			case '\b':
			case GREYBACK:
			case 'n': skip = 1;	/* go to the next match */
			case ' ':
			case 'y':		/* do the replace */
			case '!':
			case ',':
			case '.': break;
			case CTRL('R'):
				say("");
				null_detect = -1;
				oldmark = alloc_spot();
				*oldmark = mark;	/* save mark */
				mark = matchstart;
					/* set temporary mark before text */
				recursive_edit();
				if (spot_to_buffer(oldmark) < 0) {
					ret = DSABORT;	// user deleted buf
					goto done;
				}
				mark = *oldmark;	/* restore mark */
				free_spot(oldmark);
				show = 1;
				continue;
			default:	if (set_flag_modes(&flags)) {
						show = 1;
						goto ask;
					}
					if (!IS_ENTER_KEY(key))
						ungot_key = key;
			case ESC:	ret = 1; goto done;
		}
do_replace:
		if (!skip) {
			save_spot matchstart;
			if (perform_case_replace)
				this_case_replace = (perform_case_replace == 1)
									? check_for_case_replace(matchstart, point)
									: check_alt_case_replace(matchstart, point);
			if (flags & REGEX)
				fromorig(with);
			else {
				delete(matchstart, point);
				stuff(with);
			}
			if (perform_case_replace == 1)
				switch (this_case_replace) {
					case 1: do_uppercase_region(matchstart, point); break;
					case 2: do_capitalize_region(matchstart, point); break;
				}
			else if (perform_case_replace == 2) {
				i = MIN(matchstart, point);
				switch (this_case_replace) {
					case 1: do_uppercase_region(matchstart, point); break;
					case 2: do_lowercase_region(matchstart, point); break;
					case 3: do_uppercase_region(i, i + 1); break;
					case 4: do_lowercase_region(i, i + 1); break;
				}
			}
			replace_num_changed++;
			restore_vars();
		}
		if (flags & QUERY)
			switch (key) {
			case ',': case '^':	skip = 1; goto ask;
			case '.':		ret = 1; goto done;
			case '!':		flags &= ~QUERY; show = 1;
			}
	}
done:
	if (ret == DSABORT)
		say("Replacement canceled");
	else if (ret == DSBAD)
		show_replace(str, with, (flags & ~INCR) | BAD);
	else if (!in_macro(1))
		if (evershow || (flags & INCR))
			say("Replaced %d of %d.", replace_num_changed,
			    replace_num_found);
		else if (evershow)
			say("");
	free_spot(orig);
	free_spot(prev);
	return ret;
}

save_buffer_list(char *sdir, int b)
{
	int i, k = 0, d = 0, tmp = tmp_buf(), nums[3];

	save_var bufnum;
	minimize_buf_accessed();
	// All buffers should now have a buf_accessed setting from 0 to buf_accessed_clock.
	for (i = buf_list(0, 0); i; i = buf_list(1, 1)) {
		bufnum = i;
		if (buf_accessed < 0 || buf_accessed > buf_accessed_clock)
			error("Internal error: buffer %s clock %d, not 0..%d",
				  bufname, buf_accessed, buf_accessed_clock);
		if (i == b || !remember_this_buffer())
			continue;
		buf_printf(tmp, "%09d %d %d\n", buf_accessed, is_dired_buf(), bufnum);
	}
	bufnum = tmp;
	do_sort_region(0, size(), 0, 1);
	point = 0;
	while (grab_numbers(tmp, nums) == 3) {
		if (nums[1] ? (session_restore_max_directories >= 0
					   && ++k > session_restore_max_directories)
					: (session_restore_max_files >= 0
					   && ++d > session_restore_max_files))
			continue;
		bufnum = nums[2];
		session_save_buffer(b, sdir);
	}
	restore_vars();
	buf_delete(tmp);
}

int session_restore_one_file(int b, char *file, char *sdir, int *readerrs)
{
	int i, trans = FILETYPE_AUTO;
	int num1[10], num1cnt, num2[10], num2cnt, num3[10], num3cnt;
	char firstline[150];
	jmp_buf this_level;

	grab_line(b, firstline);
	if ((num1cnt = break_into_numbers(firstline, num1)) < 3)
		return 2;
	if ((num2cnt = grab_numbers(b, num2)) < 4)
		return 3;
	if ((num3cnt = grab_numbers(b, num3)) < 5)
		return 4;
	if (session_restore_files)
		filename = file;
	save_var top_level = &this_level;
	if (setjmp(top_level))
		return -1;
	if (num1cnt >= 4 && num1[3] && session_restore_files) {	// restore dired buf
		filename = "";
		if (num1cnt >= 5)
			save_var file_pattern_wildcards = num1[4];
		save_var show_status = 0;
		if (num1cnt >= 6)
			dired_sorts_files = num1[5];
		create_dired_listing(file, 0);
		if (is_remote_file(get_dired_buffer_pattern()))
			dired_orig_position = -1;
	} else if (*file && session_restore_files) {
		if (!is_remote_file(file))
			absolute(file, sdir);
		if (has_feature & FEAT_UNICODE) {
			if (num1cnt >= 7) {		 // If explicit trans, force.
				char *p;
				if (p = strchr(firstline, '@')) {
					i = encoding_from_name(p + 1);
					if (i)
						trans = MAKE_TRANSLATE(i, num1[6], 1);
				}
			}
		} else {
			if (num3cnt > 8 && num3[7])
				trans = num1[2];	// If explicit trans, or common unicode, force.
		}
		i = do_file_read(file, trans);
		force_save_as = 0;
		if (user_abort || i == EREADABORT)
			return 0;
		if (!i || i == 2) {
			if (!i)
				set_readonly_from_file();
			window_bufnum = bufnum;		// make suffix- kbd macros work
			call_mode(filename);
			make_mode();
		} else if (!*readerrs++) {
			file_error(i, file, "read error");
			maybe_ding(bell_on_read_error);
		} else
			show_text(0, -1, "%d files could not be read", *readerrs);
	}
	if (session_restore_files) {
		if (bufnum != b)	/* same tmp name as session buffer */
			point = num1[0];
		mark = num1[1];
		if (is_dired_buf())
			move_to_column(goal_column);
	}
	if (session_restore_files) {
		session_restore_one_file_1(num2, num2cnt);
		session_restore_one_file_2(num3, num3cnt);
	}
	return -1;
}

do_set_variable(i)
{
	char msg[80], res[2048], *tname, *p, tmp[2048];
	int def;

	tname = variable_type_name(vartype(i));
	if (!(p = get_var_val(i, tmp, 0)))
		error("Can't interpret type of %s", name_name(i));
	sprintf(msg, "Set %s variable %s to", tname, name_name(i));
	if (run_by_mouse && (has_feature & FEAT_GUI_DIALOGS)) {
		get_strpopup(res, msg, p, name_name(i));
		def = 1;			// set both
	} else {
		strcat(msg, ": ");
		get_strnone(res, msg, p);
		def = ask_default("Set", name_name(i), 1);
	}
	if (def)
		set_var_val(i, res, 1);		/* set default */
	if (def >= 0)
		set_var_val(i, res, 0);		/* set current */
	do_show_variable(name_name(i), def);
	full_redraw = 1;	/* might have set a display-type variable */
}

command create_file_associations()
{
	int buf = zap("File Associations");
	char ext[FNAMELEN], cmd[FNAMELEN], dir[FNAMELEN], tmp[FNAMELEN];

	tiled_only();
	save_var window_bufnum;
	to_buffer_num(buf);
	stuff("# Associate Epsilon with the following extensions:\n\n");
	stuff("# Delete or add lines as required, then type Ctrl-X Ctrl-Z\n");
	stuff("# (the exit-level command) to carry out the change.  This\n");
	stuff("# command cannot remove existing Epsilon associations, only\n");
	stuff("# add new ones.  Changes you make here cannot be undone by\n");
	stuff("# uninstalling; make new associations using Windows Explorer\n");
	stuff("# or by running this command again.\n");
	stuff("\n.c\n.cpp\n.h\n.e\n.txt\n");
	conf_mode();
	save_var case_fold = 1;
	for (;;) {
		recursive_edit();
		if (strcmp(bufname, "File Associations"))
			error("Canceled.");
		point = 0;
		if (!re_search(1, "^%.(exe|bat|cmd)$"))
			break;
		say("Don't modify the association for .exe, .bat, or .cmd files.");
	}
	make_registration_entry("Epsilon", "Epsilon");
	get_executable_directory(dir);
	strcpy(tmp, argv[0]);
	absolute(tmp, dir);
	if (is_gui == IS_WIN31)
		sprintf(cmd, "%s %%1", tmp);
	else
		sprintf(cmd, "\"%s\" \"%%1\"", tmp);
	make_registration_entry("Epsilon\\shell\\open\\command", cmd);
	make_registration_entry("Epsilon\\shell\\open\\ddeexec", "\"%1\"");
	make_registration_entry("Epsilon\\shell\\open\\ddeexec\\application",
							"Epsilon");
	make_registration_entry("Epsilon\\shell\\open\\ddeexec\\topic",
							"Open");
	point = 0;
	while (point < size() - narrow_end) {
		grab_line(bufnum, ext);
		if (*ext == '.')
			make_registration_entry(ext, "Epsilon");
		else if (*ext && *ext != '#')
			say("Ignoring invalid extension \"%s\"", ext);
	}
	restore_vars();
	buf_delete(buf);
	say("File associations created.");
}

sort_region_with_settings(from, to, col, rev, buf)
{	/* sort a region in place, using case-folding from buf */
	char *temp1 = temp_buf(), *temp2 = temp_buf();
	int this_tab_size, this_case_fold, this_sort_case_fold;
	int res = ABORT_JUMP, orig = bufnum, sz;
	jmp_buf this_level;

	bufnum = buf;
	this_case_fold = case_fold;
	this_sort_case_fold = sort_case_fold;
	this_tab_size = tab_size;

	bufnum = orig;
	save_var top_level = &this_level;
	if (!setjmp(top_level)) {
		point = from;
		if (curchar() == '\n' && character(point - 1) != '\n')
			point++;
		else
			to_begin_line();
		from = point;
		point = to;
		if (curchar() == '\n' && character(point - 1) != '\n')
			point++;
		else
			to_begin_line();
		to = point;
		xfer(temp1, from, to);
		bufname = temp1;
		case_fold = this_case_fold;
		sort_case_fold = this_sort_case_fold;
		tab_size = this_tab_size;
		res = do_buffer_sort(temp2, col, rev);
		if (!res) {
			bufnum = orig;
			point = MIN(from, to);
			sz = buffer_size(temp2);
			grab_buffer(name_to_bufnum(temp2));
			delete(fix_narrowing(from) + sz, fix_narrowing(to) + sz);
		}
	}
	bufnum = orig;
	delete_buffer(temp1);
	delete_buffer(temp2);
	return res;
}

load_var()	/* variable definition */
{
	int i, new_type, vtype, hasnum;
	char name[80], junk[80], num[20], *p, *tmp;

	point++;			/* skip initial + */
	parse_string(1, "[^ \t\n:]+", name);
	i = find_index(name);
	new_type = parse_string(1, " buffer", junk)
			? NT_BUFVAR : parse_string(1, " window", junk)
			? NT_WINVAR : NT_VAR;
	if (parse_string(1, " (char ptr )?variable: %<complex>", junk)) {
		if (i <= 0 || name_type(i) != new_type)
			init_warning("Variable %s too complex to define.", name);
		nl_forward();
		return;
	} else if (parse_string(1, " char array ", junk))
		vtype = TYPE_CARRAY;
	else if (parse_string(1, " int ", junk))
		vtype = TYPE_INT;
	else if (parse_string(1, " short ", junk))
		vtype = TYPE_SHORT;
	else if (parse_string(1, " char ", junk))
		vtype = TYPE_CHAR;
	else
bad:		init_error("Bad definition for variable %s.", name);
	if (!parse_string(1, "variable: ", junk))
		goto bad;
	hasnum = parse_string(1, "[-0-9][a-zA-Z0-9]*", num);
	if (vtype == TYPE_CARRAY) {
		if (i <= 0 && hasnum) {
			i = new_variable(name, new_type, TYPE_CARRAY, numtoi(num));
			set_name_user(i, 1);		// Make it a user variable.
		}
		if (!parse_string(1, " *<dquote>", junk)
				|| name_type(i) != new_type
				|| vartype(i) != TYPE_CARRAY)
			goto bad;
		use_default = 1;
		tmp = p = varptr(i);
		use_default = 0;
		for (; curchar() != '"'; point++) {
			if (ptrlen(p) < 1)
				init_error("Definition too long for variable %s.", name);
			if (curchar() != '\\') {
				*p++ = curchar();
				continue;
			}
			point++;
			switch (curchar()) {
				case 't': *p++ = '\t'; break;
				case 'r': *p++ = '\r'; break;
				case 'n': *p++ = '\n'; break;
				case 'z': *p++ = 0; break;
				default: *p++ = curchar();
			}
		}
		while (p - tmp < ptrlen(tmp))
			*p++ = 0;
		if (!parse_string(1, "<dquote>$", junk))
			goto bad;
	} else if (vtype) {
		if (i <= 0) {
			i = new_variable(name, new_type, vtype, 1);
			set_name_user(i, 1);		// Make it a user variable.
		}
		if (!hasnum && parse_string(1, ".*", junk))
			sprintf(num, "%d", find_index(junk));
		set_var_val(i, num, 2);
	}
}

// Move past a series of comma-separated identifiers with
// whitespace/comments between.  Return how many.
int skip_identifier_list()
{
	int cnt = 0;
	for (;; point++) {
		c_fwd_space();
		if (!parse_string(1, "[a-zA-Z0-9_.]+"))
			break;
		point = matchend;
		cnt++;
		c_fwd_space();
		if (curchar() != ',')
			break;
	}
	return cnt;
}

// Return a regex pattern that matches an identifier, including :: or .
// as apppropriate.
char *tag_c_ident_pattern()
{
	return (c_extra_keywords == CSHARP_KEYWORDS) ?
			"[a-zA-Z0-9_.]+" : "[a-zA-Z0-9_:]+";
}

// Tag the following structure/class/enum if we're defining a tag for it.
// Flavor: 0=struct/union, 1=class/interface, 2=enum, 3=namespace.
tag_c_struct(int flavor)
{
	char func[TAGLEN], fullname[TAGLEN], *s;
	int start;

	c_fwd_space();
	start = point;
	if (!parse_string(1, tag_c_ident_pattern(), func)) {
		if (flavor != 1)
			*func = 0;	// all but class don't require a name
		else
			return 1;
	}
	c_fwd_space();	// skip compiler-specific junk betw "class" & name
	if (flavor == 1 && !is_java_buffer()
		  && parse_string(1, tag_c_ident_pattern())) {
		parse_string(1, tag_c_ident_pattern(), func);
		c_fwd_space();
	}
	while (strchr(func, '.')) {
		strcpy(fullname, func);
		s = strchr(fullname, '.');
		sprintf(s, "::%s", func + 1 + (s - fullname));
		strcpy(func, fullname);
	}
	if (curchar() != '{' && (flavor != 1 || !parse_other_class()))
		return 1;	// classes can have other stuff before {
	if (flavor == 1 || flavor == 3 || flavor == 0) {
		point++;	// skip over { so we do stuff within
		if (*c_tagging_class)
			strcat(c_tagging_class, "::");
		strcat(c_tagging_class, func);	// remember, now in class
	} else if (flavor == 2) {
		tag_c_enum();
		strcpy(fullname, func);
		if (*c_tagging_class)
			sprintf(fullname, "%s::%s", c_tagging_class, func);
		if (*func && (tag_which_items & 32))
			add_tag(fullname, start);
		return 1;
	}
	if (*func && (tag_which_items & 16))
		add_tag(c_tagging_class, start);
	return 1;
}

/* We've parsed "class foo" but there's no { character right after.
 * See if there's some inheritance information we can skip past to find
 * the { character.
 * Return nonzero if we found ": foo {" or similar (C++).
 * Return nonzero if we found Java's "extends foo implements bar {" or similar.
 * In either case, move before the final { character.  Otherwise return
 * zero.
 */
parse_other_class()
{
	if (curchar() == ':') {
		point++;
		for (;;) {
			c_fwd_space();
			if (!parse_string(1, "[a-zA-Z0-9_]+|,|:"))
				break;
			point = matchend;
			if (curchar() == '<')
				point += parse_string(1, "%<[a-zA-Z0-9_, :\n\t]+%>");
		}
		c_fwd_space();
	} else if (is_java_buffer()) {
		while (strchr("ei", curchar())
			   && parse_string(1, "(extends|implements)</word>")) {
			point = matchend;
			skip_identifier_list();
		}
	}
	return curchar() == '{';
}

// Color the comment whose % is at start and move to its end.
tex_color_comment(int start)
{
	to_end_line();
	set_character_color(start, point, color_class tex_comment);
}

color_tex_math(int from, int to, int color)
{
	int s;

	set_character_color(from, to, color);
	set_tagged_region("tex-math-mode", from, to, 1);
	save_var narrow_start = from, narrow_end = size() - to, point = from;
	while (re_search(1, "[\\%&]")) {
		s = point - 1;		// Color % and & specially in math mode unless
		switch (character(s)) {	// protected by \ character.
			case '\\':
				point++;
				break;
			case '%':
				tex_color_comment(s);
				break;
			case '&':
				set_character_color(s, point, color_class tex_section);
				break;
		}
	}
}

// Recolor just this section.
// Last colored region may go past to.
int do_color_tex_range(int from, int to, int perl_interpolation)
{
	int s, talk, cnt = 0;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	point = from;
	from = point = tex_start_construct();

	talk = (to - from > tex_look_back + 1000); // Show status during long delays.
	if (point < to)
		set_character_color(point, to, color_class tex_text);
	set_tagged_region("tex-math-mode", point, to, -1);
	save_var case_fold = 1;
	for (s = point; ; s = point) {
		if (point >= to) {
			if (get_tagged_region("tex-math-mode", point, NULL, &to) == -1)
				break;
			set_character_color(point, to, color_class tex_text);
			set_tagged_region("tex-math-mode", point, to, -1);
		}
		if (!re_search(1, "[][{}%\\$&]"))
			break;
		if (point > to)
			set_character_color(s, point - 1, color_class tex_text);
		s = point - 1;
		switch (character(s)) {
			case '\\':
				color_tex_cmd(s);
				if (talk)
					if (!(cnt++ % 500))
						note("Coloring TeX: %d%% complete...",
							 muldiv(point - from, 100, to - from));
				break;
			case '$':
				if (!perl_interpolation)
					color_tex_dollar(s);
				break;
			case '%':
				tex_color_comment(s);
				break;
			case '{': case '}':
				set_character_color(s, point, color_class tex_curly_brace);
				break;
			case '&':
				set_character_color(s, point, color_class tex_section);
				break;
			case '[': case ']':
				set_character_color(s, point,
									color_class tex_square_bracket);
				break;
		}
	}
	if (talk)
		note("");
	return point;
}

int tex_maybe_break_line(int mode);
int color_tex_range(int from, int to);
int recolor_tex_from_here(int safe);

command tex_mode()
{
	mode_default_settings();
	mode_keys = tex_tab;		/* Use these keys. */
	compile_buffer_cmd = compile_tex_cmd;
	tex_paragraphs = 1;
	buffer_maybe_break_line = tex_maybe_break_line;
	if (tex_force_latex == 2){		// Determine if it's LaTeX.
		save_var point = 0;
		tex_force_latex = search(1, "\\begin{document}");
	}
	major_mode = tex_force_latex ? _latex_mode_name : _tex_mode_name;
	strcpy(comment_start, "([^\\\n]|^)%%[ \t\f]*");
	strcpy(comment_pattern, "[^\\]%%.*$");
	strcpy(comment_begin, "% ");
	strcpy(comment_end, "");
	recolor_range = color_tex_range;	// set up coloring rules
	recolor_from_here = recolor_tex_from_here;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	idle_coloring_size = 3000;	// Do bigger bunches.
	if (auto_show_tex_delimiters)
		auto_show_matching_characters = tex_auto_show_delim_chars;
	try_calling("tex-mode-hook");
	drop_all_colored_regions();
	make_mode();
}

tex_fill_paragraph() on tex_tab[ALT('q')]
{
	if (get_character_syntax_color(point) == color_class tex_comment)
		fill_comment();
	else
		fill_paragraph();
}

jump_to_dvi_by_command()
{
	char dvi[FNAMELEN], *args[4], lnum[20], cmd[FNAMELEN * 3], buf[FNAMELEN];
	char basename[FNAMELEN];
	int i = 1;

	*dvi = 0;
	if (!has_arg) {
		if (*jump_to_dvi_main_file)
			strcpy(dvi, jump_to_dvi_main_file);
		if ((!*dvi || check_file(dvi) != CHECK_FILE) && *filename) {
			strcpy(dvi, filename);
			strcpy(get_extension(dvi), ".dvi");
		}
		if (*dvi && check_file(dvi) != CHECK_FILE)
			*dvi = 0;
	}
	save_var only_file_extensions = "|.dvi|";
	if (!*dvi) {
		get_file_dir(dvi, "DVI file to view: ");
		strcpy(jump_to_dvi_main_file, dvi);
	}
	if (!*dvi) {
		strcpy(dvi, filename);
		strcpy(get_extension(dvi), ".dvi");
	}
	if (check_file(dvi) != CHECK_FILE)
		error("Can't locate file %s.", dvi);
	sprintf(lnum, "%d", lines_between(0, point, 1) + 1);
	strcpy(basename, get_tail(filename, 1));
	*get_extension(basename) = 0;
	args[0] = lnum;
	args[1] = filename;
	args[2] = dvi;
	args[3] = basename;
	if (expand_string_template(cmd, jump_to_dvi_command, "lfdb", args))
		error("Invalid jump-to-dvi-command setting.");
	do {
		sprintf(buf, "-jump-to-dvi-output-%d", i++);
	} while (exist(buf) && is_process_buffer(name_to_bufnum(buf)));
	pipe_text(NULL, buf, cmd, NULL, 0, 0);
	iter = 0;
}

do_uniq(incl_uniq, incl_dups, talk)
{
	int old = 0, ocnt, ncnt;

	if (talk)
		ocnt = count_lines_in_buf(bufnum, 0);
	save_spot point;
	for (;;) {
		point = old;
		if (!nl_forward())
			break;
		if (uniq_count_dups(old, point) ? incl_dups : incl_uniq)
			old = point;
		else
			delete(old, point);
		if (talk && size())
			noteput("%d%% complete...", muldiv(point, 100, size()));
	}
	if (!incl_uniq)
		delete(old, point);
	if (talk) {
		ncnt = count_lines_in_buf(bufnum, 0);
		say("Deleted %d lines leaving %d.", ocnt - ncnt,
			  ncnt + (size() && character(size() - 1) != '\n'));
	}
}
