/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2004 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 12.01 to 12.02.

#include "eel.h"
#include "epsnet.h"
#include "proc.h"
#include "c.h"
#include "colcode.h"
#include "perl.h"
#include "tags.h"
#include "dired.h"
#include "python.h"

int find_template_end()
{
	int opos = point, level = 1, parens = 0, bad = 0;

	if (curchar() != '<')
		return 0;
	point++;
	for (;; point++) {
		if (parse_string(1, "[ \n\t,a-zA-Z0-9_:]+|//.*"))
			point = matchend - 1;
		else
			switch (curchar()) {
				case '/':
					if (character(point + 1) != '*')
						bad = 1;
					point += 2;
					if (!search(1, "*/"))
						bad = 1;
					break;
				case '<':
					if (parens == 0)
						level++;
					break;
				case '>':
					if (--level <= 0) {
						point++;
						return 1;
					}
					break;
				case '(':
					parens++;
					break;
				case ')':
					if (--parens <= 0)
						bad = 1;
					break;
				default:
					bad = 1;
					break;
			}
		if (bad) {
			point = opos;
			return 0;
		}
	}
}

int have_csharp_attribute(int pos)
{
	save_var point = pos, narrow_start = 0;
	if (cind_try_space())
		return 1;
	return parse_string(-1, "[;{},]");
}

indent_c_contin(ind, orig, lastl)	/* indent continuation line */
{
	int c, plevel = 0, qlevel = 0, findc, alevel = 0;

	if (!c_align_contin_lines)	// never line up
		return c_default_indent(ind, orig, 1);
 	if (c_align_inherit >= 0 && matches_at(orig, -1, ":\n[ \t]*")
 		  && !c_matches_in(give_begin_line(), lastl, "%?"))
 		return get_indentation(point) + c_align_inherit;
	to_begin_line();
	save_var narrow_start = point;
	point = orig;
	findc = character(orig);
	if (c_align_open_paren && parse_string(1, "%(([ \t]|//|/%*)*$"))
		return ind;
	while (re_search(-1, "[()?:\"'\n<>]|[=!<>]?=")) {
		c = curchar();
		if (in_perl_buffer && (in_perl_comment(point) || strchr("\"'\n", c)))
			continue;	// Skip over comments differently.
		switch (c) {
			case '(':	/* indent after unmatched lparen */
				if (plevel-- > 0)
					break;
				if (c_align_open_paren && findc != ')')
					return ind + get_c_indent();
indhere:		if (findc != ')')
					point++;
				if (!parse_string(1, "[ \t\f]*/[*/]"))
					point += parse_string(1, "[ \t\f]*![^ \t\f\n]");
				return limit_c_contin_indent(ind);
			case ')':
				plevel++;
				break;
			case '?':	/* indent : to matching ? */
				if (qlevel-- > 0 || plevel > 0 || findc != ':')
					break;
				point--;
				goto indhere;
			case ':':
				qlevel++;
				break;
			case '=': case '!':
				if (character(point + 1) == '=' || plevel > 0
				    || character(lastl - 1) == ',' || findc == ')')
					break;
				goto indhere;
			case '\n':		 // skip over comments
				cind_try_space(); // at ends of lines
				break;
			case '"': case '\'':
				point++;
				c_skip_item(c, RE_REVERSE);
				break;
			case '<':
				if (character(point + 1) == '=')
					break;
				if (alevel-- > 0 || !(c_extra_keywords & (CPP_KEYWORDS
					| JAVA_KEYWORDS | CSHARP_KEYWORDS)))
					break;
				// Someday, test if there's an identifier before the <
				// that's the name of a template, not just if it's a
				// template keyword.
				if (parse_string(-1, "</word>template[ \t\n]*")) {
					point++;
					point += parse_string(1, "[ \t]*!<word>");
					return limit_c_contin_indent(ind);
				}
				break;
			case '>':
				if (character(point + 1) != '=')
					alevel++;
				break;
		}
	}
	if (c_align_inherit >= 0 && c_matches_in(point, orig, ":[ \t]*($|//|/%*)")
		  && !c_matches_in(point, orig, "%?"))
		return get_indentation(point) + c_align_inherit;

	if (character(lastl - 1) == ',') {
		if (character(lastl - 2) != ')') {
			point = lastl;
			to_begin_line();
		}
		point += parse_string(1, "(.*[ \t\f]on)?[ \t\f]+");
		if (in_c_comment(c_look_back))
			to_indentation();
		return limit_c_contin_indent(ind);
	}
	if ((c_extra_keywords & CSHARP_KEYWORDS) && parse_string(1, "[ \t]*%[")
		  && have_csharp_attribute(point))
		return get_indentation(point);
	return c_default_indent(ind, orig, 2);
}

do_decompression(char *prog, char *cmdline)
{
	char fname[FNAMELEN], *s;

	if (translation_type != FILETYPE_BINARY)
		return;
	if (is_remote_file(filename) && is_retrieving_remote_file(bufnum))
		return;
	if (prog && *prog) {
		sprintf(fname, "%s%s", prog, is_unix ? "" : ".exe");
		if (!(s = getenv("PATH")) || !look_on_path(fname, 0, s))
			return;
	}
	save_var _buf_readonly = 0;
	force_save_as = FSA_READONLY;
	if (!pipe_text(bufname, bufname, cmdline, NULL,
				   PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0))
		error("Could not uncompress %s", filename);
	point = modified = 0;
	strcpy(fname, filename);
	if (s = get_extension(fname)) {
		*s = 0;
		call_mode(fname);
	}
}

decompress_after_reading()
{
	char *ext = get_extension(filename);
	if (!strcmp(".gz", ext))
		suffix_gz();
	else if (!strcmp(".bz2", ext))
		suffix_bz2();
}

ftp_activity(int activity, int buf, int from, int to)
{
	char msg[FNAMELEN];
	int show_log = 0, orig_buf = bufnum, op;

	if (!buf_exist(buf))
		return;
	save_var iter, has_arg;
	save_var bufnum = buf;
	if (!ftp_job)
		return;
	switch (activity) {
		case NET_DONE:
			op = ftp_job->operation;
			if (op == FTP_SEND)
				show_text(0, 1, "%s written.", filename);
			if (translation_type == FILETYPE_AUTO && op == FTP_RECV)
				translation_type = new_buffer_translation_type;
			if (!modified && !point && op == FTP_RECV)
				go_line(ftp_go_to_line);
			free(ftp_job);
			ftp_job = 0;
			set_mode_message(NULL);
			check_filevars();
			if (op == FTP_RECV)
				decompress_after_reading();
			break;
		case NET_LOG_DONE:
			ftp_job = 0;
			set_mode_message(NULL);
			check_filevars();
			break;
		case NET_RECV:
			if (!from) {		// First block.
				position_in_all_windows(buf, 0);
				if (mode_keys == buf_tab)		// Switch out of Fundamental mode.
					call_mode(filename);
			}
			show_ftp_percent("got", from);
			break;
		case NET_SEND:
//			ftp_job->so_far = from;
			ftp_job->file_size = to;
			show_ftp_percent("sent", from);
			break;
		case NET_LOG_WRITE:
			save_var point = from;	// Get file size from msg.
			if (re_search(1, "^150 ") && re_search(1, "[0-9]+ bytes")) {
				char filesize[100];
				grab(matchstart, matchend, filesize);
				if (numtoi(filesize) && !got_bad_number)
					ftp_job->file_size = numtoi(filesize);
			}
			point = from;
			if (found_ftp_status("^220 "))	// Welcome msg.
				set_both_mode_message("[FTP: Sending user name] ");
			if (found_ftp_status("^331 "))	// Host wants password.
				set_both_mode_message("[FTP: Sending password] ");
			if (found_ftp_status("^230 "))	// User logged in.
				set_both_mode_message("[FTP: Working] ");
			if (found_ftp_status("^530 "))	// Bad password.
				discard_password_data();
			if (re_search(1, "^[45][0-9][0-9] ") && !parse_string(-1, "PASV\n... ")) {
				parse_string(1, ".*", msg);		// Some error code.
				say("FTP error: %s", msg);
				// Don't go to log if we're just creating a new file.
				if (ftp_job->operation != FTP_RECV
					|| !parse_string(-1, "^550 .*"))
					show_log = 1;
			}
			break;
	}
	if (buffer_ftp_activity)	// Do some buffer-specific FTP work.
		buffer_ftp_activity(activity, buf, from, to);
	switch (activity) {		// These don't automatically refresh later.
		case NET_DONE:
		case NET_LOG_DONE:
		case NET_SEND:
			refresh();
			break;
	}
	if (show_log) {
		if (num_of_popups() > 0 || orig_buf != ftp_job->buf) {
			say("FTP errors occurred, see buffer %s.",
				buf_exist(ftp_job->log_buf) ?
				bufnum_to_name(ftp_job->log_buf) : "<deleted>");
			if (buf_exist(ftp_job->buf) && ftp_job->operation == FTP_SEND) {
				save_var bufnum = ftp_job->buf;
				modified = 1;
			}
		} else {
			restore_vars();
			save_var iter, has_arg;
			save_var bufnum = buf;
			if (buf_exist(ftp_job->buf)) {
				to_buffer_num(ftp_job->buf);
				if (ftp_job->operation == FTP_SEND)
					modified = 1;
			}
			if (buf_exist(ftp_job->log_buf))
				to_buffer_num(ftp_job->log_buf);
			point = size();
			refresh();
			restore_vars();
			bufnum = window_bufnum;
		}
	}
	restore_vars();
	if (num_of_popups() > 0 && !strcmp(bufname, CONNECT_LIST_BUF)) {
		int line = lines_between(0, point, 0) + 1, col = current_column();
		make_connect_list();
		go_line(line);
		move_to_column(col);
		refresh();
	}
}

struct scp_job *do_scp(struct url_parts *parts, int op);

scp_get_response(int code, char *url, struct url_parts *parts, int b)
{
	char cmd[FNAMELEN], *s;
	struct scp_job *p;

	save_var bufnum = b;
	buf_zap(bufnum);
	p = do_scp(parts, FTP_MISC);
	strcpy(p->file, url);
	*get_url_file_part(p->file, 0) = 0;
	if (!(p->flags & SCP_CLIENT_MODE)) {
		s = (p->flags & SCP_OLD_SFTP) ? "ls \"%s*\"" : "ls -l \"%s*\"";
		switch (code) {
			case 1:	s = "";
					break;
			case 2:	p->flags |= SCP_CHECK_IS_DIR;
					s = "cd \"%s\"";
					break;
			case 3:	s = (p->flags & SCP_OLD_SFTP) ? "ls \"%s\"" : "ls -l \"%s\"";
					break;
		}
		scp_send_some_command(parts, p, s);
	} else {
		sprintf(cmd, (code == 1 || code == 2) ? "prdir %s\n" : "comp %s\n",
				parts->fname);
		p->cmd = p->cmdlist = strsave(cmd);
		if (p->busy != 2)
			scp_next_cmd(p);
	}
	delayed_say(0, 100, 30, "Working...");
	while (p->busy && !user_abort)
		delay(1, COND_KEY);
	set_buf_mode_message(bufnum, NULL);
	user_abort = 0;
	delayed_say(0, 0, 0, "");
	say("");
	point = 0;
	if (!(scp_flags(parts) & SCP_CLIENT_MODE)) {
		switch (code) {
			case 0:		// Discard the extra info.
			case 3:
				delete_final_newlines();
				simple_re_replace(1, "^.*" UNIX_DATE " *", "");
				strcpy(cmd, url);
				*get_tail(get_url_file_part(cmd, 0), 1) = 0;
				if (*cmd && !is_path_separator(cmd[strlen(cmd) - 1]))
					strcat(cmd, "/");
				if (size())
					simple_re_replace(1, "^", cmd);
				break;
			case 1:
				buf_printf(b, "%s%s", p->file, p->lastdir);
				set_buf_point(b, 0);
				break;
		}
	}
}

is_retrieving_remote_file(int b)
{
	save_var bufnum = b;
	return ftp_job || is_busy_doing_scp(b);
}

fix_extension_list(int buf)
{
	save_var bufnum = buf;
	point = 0;
	do_drop_matching_lines(REGEX, "[bv][0-9]+patch%.e$", 1);
	string_replace("[\\\"]", "\\#0", REGEX);
	string_replace("^.+$", "(load-eel-from-path \"#0\" 2)", REGEX);
}

int perl_try_space()
{
	int i, skip_unit = 0, c;

	for (;;) {
		if (index(" \t\n\f", c = character(point - 1))) {
			point--;
			if (c == '\n')
				skip_unit = 0;
		} else if (get_character_syntax_color(point - 1) == color_class perl_comment
				   && get_character_color(point - 1, &i) == color_class perl_comment
				   && strchr("#/", character(i)) && i < point)
			point = i;
		else if (!skip_unit
				 && get_tagged_region("color-perl-as-unit", point - 1, &i) != -1
				 && i < point) {
			point = i;
			skip_unit = 1;
			to_end_line();
		} else
			return point == narrow_start;
	}
}

int do_python_indent();
int color_python_range();

command python_mode()
{
	mode_default_settings();
	mode_keys = python_tab;		/* Use these keys. */
	major_mode = _python_mode_name;
	compile_buffer_cmd = compile_python_cmd;	// can compile this
	auto_indent = 1;
	if (python_tab_override > 0)
		tab_size = python_tab_override;
	indent_with_tabs = python_indent_with_tabs;
	indenter = do_python_indent;
	soft_tab_size = python_indent;
	strcpy(comment_start, "#[ \t]*");
	strcpy(comment_pattern, "#.*$");
	strcpy(comment_begin, "# ");
	strcpy(comment_end, "");
	recolor_range = color_python_range;	// set up coloring rules
	recolor_from_here = recolor_from_top;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	if (auto_show_python_delimiters)
		auto_show_matching_characters = python_auto_show_delim_chars;
	buffer_maybe_break_line = generic_maybe_break_line;
	fill_mode = misc_language_fill_mode;
	try_calling("python-mode-hook");
	drop_all_colored_regions();
	make_mode();
}

session_save_buffer(int b, char *sdir)
{
	char file[FNAMELEN];
	int transl = GET_LINE_TRANSLATE(translation_type);
	int uni_compat = 0;

	if (translation_type & FORCED_TRANS)
		transl = GET_LINE_TRANSLATE(translation_type);
	if (has_feature & FEAT_UNICODE) {
		char *s = encoding_to_name(GET_ENCODING(translation_type));
		if (s)
			uni_compat = strstr(s, "UTF-16") != 0;
		if (uni_compat)
			transl = FILETYPE_BINARY;
	}
	buf_printf(b, "%s\n", bufname);
	if (is_dired_buf()) {
		insert_char_in_buf(b, 0);	// Hide file pattern from old versions.
		buf_printf(b, "%s\n", get_dired_buffer_pattern());
	} else {
		relative(filename, file, sdir);
		buf_printf(b, "%s\n", file);
	}
	buf_printf(b, "%d %d %d %d %d %d",
			   point, mark, transl,
			   is_dired_buf(), dired_buffer_pattern_wildcards,
			   dired_sorts_files);
	if ((has_feature & FEAT_UNICODE) && (translation_type & FORCED_TRANS))
		buf_printf(b, " %d @%s", GET_LINE_TRANSLATE(translation_type),
				   encoding_to_name(GET_ENCODING(translation_type)));
	buf_printf(b, "\n");
	buf_printf(b, "%d %d %d %d\n", over_mode, fill_mode,
			   indent_with_tabs, margin_right);
	transl = FILETYPE_AUTO;
	if (uni_compat || (translation_type & FORCED_TRANS))
		transl = GET_LINE_TRANSLATE(translation_type);
	// This flavor of the translation type is used by non-Unicode/
	// pre-v12 versions only.
	buf_printf(b, "%d %d %d %d %d %d %d %d %d %d\n", tab_size,
			   case_fold, 0, want_warn,
			   _buf_readonly, buf_accessed,
			   (file_io_converter == oem_file_converter),
			   uni_compat,
			   transl, soft_tab_size);
}

session_restore_one_file_2(int *nums, int i)
{
	tab_size = nums[0];
	case_fold = nums[1]; // nums[2] var was removed
	want_warn = nums[3];
	if (nums[4] == 1)
		set_read_only(1);
	buf_accessed = i > 5 ? nums[5] : 0;
	if (i >= 10)
		soft_tab_size = nums[9];
	undo_mainloop();
}

char *tag_c_ident_pattern()
{
	return (c_extra_keywords == CSHARP_KEYWORDS) ?
			"[a-zA-Z0-9_.]+" : "[a-zA-Z0-9_]+";
}

tag_c_struct(int flavor)
{
	char func[TAGLEN], fullname[TAGLEN], *s;
	int start, i;

	c_fwd_space();
	start = i = point;
	if (!parse_string(1, tag_c_ident_pattern(), func)) {
		if (flavor != 1)
			*func = 0;	// all but class don't require a name
		else
			return 1;
	}
	if (flavor == 0 && (c_extra_keywords & CPP_KEYWORDS))
		flavor = 1;	// In C++, treat structs just like classes.
	c_fwd_space();	// skip compiler-specific junk betw "class" & name
	if (flavor == 1 && !is_java_buffer()
		  && parse_string(1, tag_c_ident_pattern())) {
		i = point;
		parse_string(1, tag_c_ident_pattern(), func);
		c_fwd_space();
	}
	while (strchr(func, '.')) {
		strcpy(fullname, func);
		s = strchr(fullname, '.');
		sprintf(s, "::%s", func + 1 + (s - fullname));
		strcpy(func, fullname);
	}
	if (strchr(";[,", curchar()) && i > start) {
		point = i;	// A plain definition.
		return 1;
	}
	if (curchar() != '{' && (flavor != 1 || !parse_other_class()))
		return 1;	// classes can have other stuff before {
	if (flavor == 1 || flavor == 3 || flavor == 0) {
		point++;	// skip over { so we do stuff within
		if (*c_tagging_class)
			strcat(c_tagging_class, "::");
		strcat(c_tagging_class, func);	// remember, now in class
	} else if (flavor == 2) {
		tag_c_enum();
		strcpy(fullname, func);
		if (*c_tagging_class)
			sprintf(fullname, "%s::%s", c_tagging_class, func);
		if (*func && (tag_which_items & 32))
			add_tag(fullname, start);
		return 1;
	}
	if (*func && (tag_which_items & 16))
		add_tag(c_tagging_class, start);
	return 1;
}

parse_other_class()
{
	if (curchar() == ':') {
		point++;
		for (;;) {
			c_fwd_space();
			if (parse_string(1, tag_c_ident_pattern())
				  || parse_string(1, "[,:]"))
				point = matchend;
			else if (!find_template_end())
				break;
		}
		c_fwd_space();
	} else if (is_java_buffer()) {
		while (strchr("ei", curchar())
			   && parse_string(1, "(extends|implements)</word>")) {
			point = matchend;
			skip_identifier_list();
		}
	}
	return curchar() == '{';
}
