/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2004 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 12.02 to 12.03.

#include "eel.h"
#include "epsnet.h"
#include "proc.h"
#include "c.h"
#include "colcode.h"
#include "perl.h"
#include "tags.h"
#include "dired.h"
#include "python.h"
#include "browse.h"
#include "bufed.h"
#include "session.h"
#include "filter.h"
#include "grep.h"
#include "html.h"
#include "vbasic.h"
#include "lowlevel.h"

do_set_bookmark(b)
{
	if (spot_to_buffer(_bookmark_spots[b]) > 0)
		free_spot(_bookmark_spots[b]);
	_bookmark_spots[b] = alloc_spot(0);
	_bookmark_saved_pos[b] = 0;
	_last_bookmark = b;
}

do_jump_to_bookmark(b)
{
	int i;
	spot m = _bookmark_spots[b];

	if (!m)
		error("Bookmark %c is not in use", key);
	i = spot_to_buffer(m);
	if (i == -1)
		error("Bookmark %c's buffer has been deleted", key);
	else if (i == -2)
		error("Can't find bookmark %c", key);
	_last_bookmark = b;
	to_buffer_num(i);
	if (_bookmark_saved_pos[b] && !*m) {
		*m = _bookmark_saved_pos[b];
		_bookmark_saved_pos[b] = 0;
	}
	point = *m;
	say("Jumped to bookmark %c.", bookmark_to_key(b));
}

drop_bookmark(b)
{
	if (is_good_bookmark(b)) {
		free_spot(_bookmark_spots[b]);
		_bookmark_spots[b] = 0;
		_bookmark_saved_pos[b] = 0;
	}
}

list_a_bookmark(list, i)	/* maybe add a line to list for mark i */
{				/* return 1 if we did, else 0 */
	int b, opt;
	spot m = _bookmark_spots[i];

	if (!m || (b = spot_to_buffer(m)) <= 0)
		return 0;
	bufnum = b;
	buf_printf(list, "%c %s: ", bookmark_to_key(i), bufname);
	opt = point;
	if (_bookmark_saved_pos[i] && !*m) {
		*m = _bookmark_saved_pos[i];
		_bookmark_saved_pos[i] = 0;
	}
	point = *m;
	buf_xfer(list, give_begin_line(), give_end_line());
	point = opt;
	buf_printf(list, "\n");
	return 1;
}

#define BROWSER_BUF "Symbols"

#define BSC_DLL "bsc.dll" /* Compiler-specific DLL file from Microsoft. */

// This takes the address of a pointer, and makes the pointer point
// to a BSC handle.
#define BSCOpen(browsefile, ppbsc) \
	((int) call_dll(BSC_DLL, \
					"BSCOpen", "c", "sr", browsefile, ppbsc))

// This takes a pointer to a BSC handle.
#define BSCClose(pbsc) ((int) call_dll(BSC_DLL, "BSCClose", "c", "p", pbsc))


#define BSCGetOverloadArray(pbsc, name, mbf, ppinst, pnum) \
	((int) call_dll(BSC_DLL, \
					"BSCGetOverloadArray", "c", "psirr", \
					pbsc, name, mbf, ppinst, pnum))

#define BSCGetDefArray(pbsc, inst, ppidef, pnum) \
	  ((int) call_dll(BSC_DLL, \
						"BSCGetDefArray", "c", "pirp", \
						pbsc, inst, ppidef, pnum))

#define BSCIdefInfo(pbsc, idef, pmodname, plineno) \
						((int) call_dll(BSC_DLL, \
						"BSCIdefInfo", "c", "pirp", \
						pbsc, idef, pmodname, plineno))

#define BSCDisposeArray(pbsc, any_array) \
	((int) call_dll(BSC_DLL, \
					"BSCDisposeArray", "c", "pp", \
					pbsc, any_array))

int dump_defs(int *pbsc, char *bsc_filename, char *symbol, int *insts, int num_insts)
{
	char sym_name[512];
	char sym_descr[512];
	int r = 0;
	int i, j;

	bufnum = zap(BROWSER_BUF);

	use_black(); bprintf("Browser File: ");
	use_blue();  bprintf("%s\n", bsc_filename);

	use_black(); bprintf("Symbol: ");
	use_blue();  bprintf("%s\n", symbol);

	use_black();bprintf("\n%s:\n", "Definitions");

	for (j = 0; j < num_insts; j++){
		int *defs, num_defs;
		int inst = insts[j];

		get_inst_data(pbsc, inst, sym_name, sym_descr);
		BSCGetDefArray(pbsc, inst, &defs, &num_defs);
		for (i = 0; i < num_defs; i++){
			char modname[1024];
			byte *temp;
			short line_num[1];
			BSCIdefInfo(pbsc, defs[i], &temp, line_num);
			r++;
			bytes_to_chars(modname, temp);

			// The file name.
			use_black();
			bprintf("    %s(%d): ", modname, *line_num);

			// The symbol name.
			use_blue();
			bprintf("%s", sym_name);

			// The symbol description.
			use_green();
			bprintf(" (%s)", sym_descr);
			bprintf("\n");
			use_black(); // flush
		}
	}
	point = 0;
	return r;
}
int browser_query(char *bsc_filename, char *symbol, int flags)
{
	int *bsc;
	int *Inst, num_insts;
	int r = 0;
	byte temp[FNAMELEN];

	save_var bufnum;
	UNREFERENCED_PARAMETER(flags);
	chars_to_bytes(temp, bsc_filename);
	if (BSCOpen(temp, &bsc) == 0) {
		say("Can't use %s to read %s.", BSC_DLL, bsc_filename);
		return -1;
	}
	chars_to_bytes(temp, symbol);
	BSCGetOverloadArray(bsc, temp, MBFALL, &Inst, &num_insts);	

	// flags will indicate defs or refs.
	r = dump_defs(bsc, bsc_filename, symbol, Inst, num_insts);

	if (num_insts)
		BSCDisposeArray(bsc, Inst);
	BSCClose(bsc);
		// We unload the DLL to be sure that all the resources are released.
	call_dll(BSC_DLL);
	return r;
}

// Return the line number with position pos on it, in buffer buf.
buf_position_to_line_number(int buf, int pos)
{
	save_var bufnum = buf;
	return lines_between(0, pos, 0) + 1;
}

do_switch_buffers(int orig_key)
{
	char bname[FNAMELEN];
	int other_key = -1;

	// If the key that invoked us is a function key, numpad or cursor
	// pad, or "gray key", find its shifted equivalent and look for that
	// also.  Usually Shift+orig key will give us the original key code,
	// if the original included Ctrl, but sometimes it might give us a
	// different one.
	if (!IS_EXT_ASCII_KEY(orig_key) && IS_TRUE_KEY(orig_key))
		other_key = NUMSHIFT(NUMPLAIN(orig_key));
	error_if_input(1);
	tiled_only();
	save_var _bufed_sort = 'a', bufed_grouping = 0;
	do_bufed(BUFED_SWITCH | (has_arg ? BUFED_ALL : 0));
	move_in_switch_buffers(shift_pressed() & KB_SHIFT_DN);
	while ((shift_pressed() & KB_CTRL_DN)) {
		maybe_refresh();
		if (char_avail()) {
			getkey();
			if (key == GETFOCUS || key == LOSEFOCUS) {
				do_topkey();
				continue;
			} else if (key == orig_key || key == other_key)
				move_in_switch_buffers(shift_pressed() & KB_SHIFT_DN);
			else if (binding_of(key, root_keys) == (short) down_line)
				move_in_switch_buffers(0);
			else if (binding_of(key, root_keys) == (short) up_line)
				move_in_switch_buffers(1);
		} else
			delay(10, COND_KEY | COND_RETURN_ABORT);
		if (user_abort || key == abort_key || key == ESC
			  || key == NUMCTRL(GREYESC)) {
			while (char_avail())	// Discard keys so we don't get a delayed
				getkey();			// Canceled msg.
			user_abort = 0;
			remove_bufed();
			return;
		}
	}
	if (!bufed_getname(bname))
		*bname = 0;
	remove_bufed();
	if (*bname && exist(bname))
		to_buffer(bname);
}

c_ignore_declspec()
{
	int start = point;
	if (character(point - 1) == ')' && c_rev_skip_level('(', ')')
		  && parse_string(-1, "__declspec")) {
		point = matchend;
		return 1;
	}
	point = start;
	return 0;
}

int find_c_class_info(int start, char *type, char *class, char *func)
{
	char this[FNAMELEN], last[FNAMELEN];
	int class_start = point, i;

	point--;	// See if "class foo" is before this.
	strcpy(last, "");
	for (i = 20; i-- > 0; ) {	// Look through some words before the {.
		if (cind_try_space() || give_begin_line() > start)
			return 0;
		if (c_ignore_declspec())
			continue;
		if (strchr("<>:,", character(point - 1)))
			point--;
		else {
			if (!parse_string(-1, (c_extra_keywords
								   & (CSHARP_KEYWORDS | JAVA_KEYWORDS))
							  ? "[a-zA-Z0-9_.]+" : "[a-zA-Z0-9_:]+", this))
				return 0;
			if (!strcmp(this, "class") || !strcmp(this, "struct")
				  || !strcmp(this, "union") || !strcmp(this, "namespace")
				  || !strcmp(this, "enum")
				  || (c_extra_keywords == CSHARP_KEYWORDS) && !strcmp(this, "interface")) {
				if (class) {
					if (*class && ((display_c_options & 1)
								   || strcmp(type, "namespace"))) {
						strcat(class, (c_extra_keywords == CSHARP_KEYWORDS) ? "." : "::");
						class += strlen(class);
					}
					strcpy(class, last);	// Is place where we started within func?
				}
				if (type)
					strcpy(type, this);		// Word to the right of type is class.
				if (find_inner_func_name(start, class_start, func))
					return CF_INFO_TYPE | CF_INFO_CLASS | CF_INFO_FUNC;
				if (func)
					*func = 0;
				return CF_INFO_TYPE | CF_INFO_CLASS;
			}
			strcpy(last, this);		// Go on to the next word.
		}
	}
	return 0;		// Couldn't recognize anything here.
}

// Return nonzero if the left paren at pos has a match.
check_unterminated_parens(int pos)
{
	save_var point = pos;
	return c_skip_level('(', ')');
}

// Check if this line ending in ; is inside the parens of a for ()
// statement.  If so, and orig isn't also within its bounds, move to its
// start and return nonzero.
move_before_for_statement(int orig)
{
	int cnt = 0, opos = point;
	while (cnt++ < 2) {		// Check this line and previous one only.
		to_begin_line();
		if (parse_string(1, "[ \t]*for[ \t]*%(")) {
			if (for_unclosed(point, opos) && !for_unclosed(point, orig)) {
				to_begin_line();
				return 1;
			}
		}
		nl_reverse();
	}
	point = opos;
	return 0;
}

indent_after_statement(orig) // orig is place needing indentation, for toplevel
{
	int lastcont, i, c, atparen = -1;

	if ((i = parse_string(1, "[ \t\f]*for[ \t\f]*%([ \t\f]*"))
		  && for_unclosed(point, orig))
		return get_column(point + i); // part of "for" statement
	for (;;) { /* was only the end of a larger statement, find start */
		do {	/* look back for line that ends with ;{}: */
			lastcont = point;
			to_begin_line();
			if (cind_try_space())
				break;
			if ((c = character(point - 1)) == ')') {
				atparen = lastcont;
				if (!c_rev_skip_level('(', ')'))
					break;
				to_begin_line();
			} else if (c == ';')
				if (move_before_for_statement(orig))
					c = 'x';	// Keep looking.
		} while (!index(";{}:", c));
		if (c == ':' && parse_string(-1, "%)[ \t]*:")
			  && (c_extra_keywords & (CPP_KEYWORDS | JAVA_KEYWORDS))
			  && !matches_at(give_begin_line(), 1, "[ \t]*case</word>"))
			continue;	// Not a label but C++ inheritance or similar.
		point = lastcont;	/* use indentation of line */
		to_c_indentation();	/* just after it */
		if (parse_string(RE_FORWARD, "(}[ \t\f]*)?e!lse[^a-zA-Z0-9_]")) {
			point = matchend;
			if (!to_matching_if())
				break;
		} else if (parse_string(RE_FORWARD,
								"(}[ \t\f]*)?!(catch|finally)[^a-zA-Z0-9_]")) {
				point = matchend;
				if (!to_matching_try())
					break;
			} else
			break;
	}
	i = current_column();
	if (!i && atparen != -1 && (Topindent || !c_statement_start()))
		i = get_indentation(atparen);	// 2nd func param decl
	if ((atparen = parse_string(1, "for[ \t\f]*%([ \t\f]*"))
	    && for_unclosed(point, orig))
		i = get_column(point + atparen); // part of "for" statement
	if (i < c_top_braces)
		return 0;	// 1st line of new func
	return i;
}

indent_c_contin(ind, orig, lastl)	/* indent continuation line */
{
	int c, plevel = 0, qlevel = 0, findc, alevel = 0;

	if (!c_align_contin_lines)	// never line up
		return c_default_indent(ind, orig, 1);
 	if (c_align_inherit >= 0 && matches_at(orig, -1, ":\n[ \t]*")
 		  && !c_matches_in(give_begin_line(), lastl, "%?"))
 		return get_indentation(point) + c_align_inherit;
	to_begin_line();
	save_var narrow_start = point;
	point = orig;
	findc = character(orig);
	if (c_align_open_paren && parse_string(1, "%(([ \t]|//|/%*)*$"))
		return ind;
	while (re_search(-1, "[()?:\"'\n<>]|[=!<>]?=")) {
		c = curchar();
		if (in_perl_buffer && (in_perl_comment(point) || strchr("\"'\n", c)))
			continue;	// Skip over comments differently.
		switch (c) {
			case '(':	/* indent after unmatched lparen */
				if (plevel-- > 0)
					break;
				if (c_align_open_paren && findc != ')')
					return ind + get_c_indent();
indhere:		if (findc != ')')
					point++;
				if (!parse_string(1, "[ \t\f]*/[*/]"))
					point += parse_string(1, "[ \t\f]*![^ \t\f\n]");
				return limit_c_contin_indent(ind);
			case ')':
				plevel++;
				break;
			case '?':	/* indent : to matching ? */
				if (qlevel-- > 0 || plevel > 0 || findc != ':')
					break;
				point--;
				goto indhere;
			case ':':
				qlevel++;
				break;
			case '=': case '!':
				if (character(point + 1) == '=' || plevel > 0
				    || character(lastl - 1) == ',' || findc == ')')
					break;
				goto indhere;
			case '\n':		 // skip over comments
				cind_try_space(); // at ends of lines
				break;
			case '"': case '\'':
				point++;
				c_skip_item(c, RE_REVERSE);
				break;
			case '<':
				if (character(point + 1) == '=')
					break;
				if (alevel-- > 0 || !(c_extra_keywords & (CPP_KEYWORDS
					| JAVA_KEYWORDS | CSHARP_KEYWORDS)))
					break;
				// Someday, test if there's an identifier before the <
				// that's the name of a template, not just if it's a
				// template keyword.
				if (parse_string(-1, "</word>template[ \t\n]*")) {
					point++;
					point += parse_string(1, "[ \t]*!<word>");
					return limit_c_contin_indent(ind);
				}
				break;
			case '>':
				if (character(point + 1) != '=')
					alevel++;
				break;
		}
	}
	if (c_align_inherit >= 0 && c_matches_in(point, orig, ":[ \t]*($|//|/%*)")
		  && !c_matches_in(point, orig, "%?"))
		return get_indentation(point) + c_align_inherit;

	if (character(lastl - 1) == ',') {
		if (character(lastl - 2) != ')') {
			point = lastl;
			to_begin_line();
		}
		point += parse_string(1, "(.*[ \t\f]on)?[ \t\f]+");
		if (in_c_comment(c_look_back))
			to_indentation();
		if (curchar() == '#')
			past_initial_preprocessor();
		return limit_c_contin_indent(ind);
	}
	if ((c_extra_keywords & CSHARP_KEYWORDS) && parse_string(1, "[ \t]*%[")
		  && have_csharp_attribute(point))
		return get_indentation(point);
	return c_default_indent(ind, orig, 2);
}

parse_command_line_flags(int flavor)
{
	int i, j;
	char *s, *flag, *arg, buf[1024];

	for (i = 1; i < argc && argv[i][0] == '-'; i++) {
		flag = argv[i];
		arg = (flag[1] && flag[2] || i + 1 >= argc) ? (flag + 2) : argv[++i];
		if (!strchr((flavor == 3 ? "lr" : flavor == 2 ? "pd"  : "d"), flag[1]))
			continue;
		switch (flag[1]) {
			case 'p':
				if (!flag[2])
					preserve_session_once = preserve_session ? 1 : 2;
				else if (isdigit(flag[2]))
					preserve_session_once = numtoi(flag + 2) + 1;
				else
					cmd_line_session_file = flag + 2;
				break;
			case 'l':
				load_commands(arg); break;
			case 'd':
			case 'r': 
				strcpy(buf, arg);
				s = index(buf, '=');
				if (!s)		// accept either VAR=VAL or VAR!VAL
					s = index(buf, '!');
				if (s)
					*s = 0;
				if (j = find_index(buf)) {
					if (flag[1] == 'd') {
						if (!s)
							bad_flag(flag);
						if ((flavor == 1) != !strcmp(buf, "einit-file-name"))
							continue;	// Do this var only if flavor is 1.
						now_parsing_cmds = 1;	// Don't record this.
						safe_set_var_val(j, s + 1, 1);
						now_parsing_cmds = 0;
					} else {
						if (s)
							(*(int (*)())j)(s + 1);
						else
							(*(int (*)())j)();
					}
					break;
				}
				if (flag[1] == 'd') {
					bprintf("No such variable %s.\n", buf);
					init_help_abort();
				} else
					say("No command %s, %s ignored.", buf, flag);
				break;
		}
	}
}

when_restoring()
{
	int tflag = 0, i, splitmode = HORIZONTAL;
	char *flag, *arg;

	early_restoring();
	early_init();

	for (i = 1; i < argc && argv[i][0] == '-'; i++) {
		if (!stricmp(argv[i], "-noinit")) {
			noinit_flag = 1;
			continue;
		}
		flag = argv[i];
		switch (flag[1]) {
		case 'f':
			arg = (flag[2] && flag[3] || i + 1 >= argc) ? (flag + 3) : argv[++i];
			switch (flag[2]) {
			case 's':
				set_swapname(arg);
				break;
			case 'd':
				strcpy(_docname, arg);
				break;
			default:
				bad_flag(flag);
			}
			break;
		case 't':
			tflag = 1;
			break;
		case 'd':
		case 'r':
		case 'l': if (!flag[2] && i + 1 < argc)
					  i++;
				  break;
		case 'p': break;
		case 'h': case '?': usage(); break;
		case 'v':
			switch (flag[2]) {
				case 'v':
					splitmode = VERTICAL;
					break;
				default:
					bad_flag(flag);
			}
			break;
		default:
			bad_flag(flag);
		}			/* i is 1st file argument */
	}
	term_init();
	try_calling("key-setup");
	middle_init();
	init_menu();
	term_cmd_line();
	delete_buffer("-customizations");	// Don't record any settings before now.
	parse_command_line_flags(1);
	if (!noinit_flag && load_from_state)
		init_from_file();
	parse_command_line_flags(2);
	if (tflag) {
		preserve_session_once = 1;
		tutorial();
	} else {
		if (load_from_state && (i >= argc || session_always_restore)) {
			if (session_restore_files == 2 && i < argc)	// Have cmdline files,
				save_var session_restore_files = 0;	// skip session files.
			load_session();
		}
		if (i < argc)
			load_cmd_line_files(i, splitmode);
		delete_dull_buffer("startup");
	}
	switching_to_buffer();
	set_toolbar();
	restore_check_version();
	show_text(0, 1, "Epsilon Programmer's Editor %s, EEL version %s",
			  version, eel_version);
	try_calling("start-up");
	parse_command_line_flags(3);
}

init_from_file()
{
	char *p, file[FNAMELEN];
	int res;
	jmp_buf this_level;

	save_var top_level = &this_level;	// Don't let errors abort,
	if (setjmp(top_level)) {		// so session file can still load.
		show_text(0, -1, "Error loading %f, see %s buffer.",
				  einit_file_name, MSG_BUFFER);
		return;
	}
	save_var _abort_load_on_error = 2;
	if ((load_customizations & 1) && (p = lookpath(einit_file_name))) {
		strcpy(file, p);
		save_var bufnum = zap("-customizations");
		do_file_read(file, default_translation_type);
		ini_mode();
		res = parse_cmds();
		tab_size = tab_size.default;
		restore_vars();
		if (res)
			show_text(0, -1, "Error loading %f, see %s buffer.", file, MSG_BUFFER);
	}
}

error_if_input(abort)	/* disallow recursive cmd line editing */
{
	switch (_doing_input) {
		case DI_SEARCH_NOINCR:
		case DI_SEARCH:		quick_abort(); break;
		case DI_LINEINPUT:	leave_recursion(abort ? INP_ABORT : 0);
					break;
		case DI_VIEWLAST:	remove_final_view(); break;
		case DI_LISTBOX:	break;
	}
}

dired_isdir()
{
	int i = get_dired_item("dired-isdir-", 1);
	if (i)
		return (*(int (*)())i)();
	i = get_dired_item("dired-dirsearch-", 0);
	if (i) {
		save_var point;
		to_begin_line();
		return parse_string(1, get_str_var(i)) != 0;
	}
	return matches_at(column_to_pos(ISDIR_START), 1, DIRSEARCH) != 0;
}

color_dired_range(from, to)
{
	int color, namecolor, old, namestart, nameend, isdir;
	char buf[FNAMELEN];

	if (from >= to)
		return to;
	detect_dired_format();
	save_var point, matchstart, matchend;
	point = from;
	save_var narrow_start = give_begin_line();
	point = to;
	nl_forward();
	save_var narrow_end = size() - point;
	point = 0;
	while (point < size() - narrow_end) {
		switch (curchar()) {
			case 'D': color = color_class dired_deleted; break;
			case 'U': case 'L': case 'R':
			case 'M': color = color_class dired_moved; break;
			case 'C': color = color_class dired_copied; break;
			default: color = color_class dired_normal; break;
		}
		old = point;
		if (parse_string(1, SKIPLINEPAT) != 0) {
			color = color_class dired_header;
			namestart = matchend;
			nameend = give_end_line();
			namecolor = color_class dired_dir_name;
		} else {
			dired_getindex(buf, goal_column);
			namestart = matchstart;
			nameend = matchend;
			namecolor = color_class dired_file_name;
			if (dired_format == DF_VMS)
				nameend = point + parse_string(1, "[^\n;]*;?[0-9]*");
			isdir = dired_isdir();
			if (isdir) {
				namecolor = color_class dired_dir_name;
				if (!dired_format)
					namestart = column_to_pos(ISDIR_START);
				if (isdir == 3 && parse_string(1, ".* -> "))
					nameend = matchend;
			}
			if (!is_remote_file(get_dired_buffer_pattern())
				  && character(namestart - 1) == ' ')
				namestart--;
		}
		nl_forward();
		set_character_color(old, point - 1, color);
		set_character_color(namestart, nameend, namecolor);
	}
	return point;
}

dired_sorter(start, end)
{
	int tmp, next;
	char *pat = DIRSEARCH;
	int i = get_dired_item("dired-dirsearch-", 0);

	if (i)
		pat = get_str_var(i);
	save_var case_fold = 1, sort_case_fold = (dired_format != DF_UNIX);
	save_var narrow_start = start;
	save_var narrow_end = size() - end;
	if (dired_groups_dirs) {
		save_var point = 0;
		tmp = tmp_buf();
		while (re_search(1, pat) && !user_abort) {
			to_begin_line();
			next = give_end_line() + 1;
			buf_xfer(tmp, point, next);
			delete(point, next);
		}
		dired_group_sorter(start, point);
		point = 0;
		grab_buffer(tmp);
		buf_delete(tmp);
		dired_group_sorter(mark, point);
	} else
		dired_group_sorter(start, end);
}

make_registration_entry(name, val)
char *name, *val;
{
	RegSetValue(HKEY_CLASSES_ROOT, name, REG_SZ, val, strlen(val));
}

make_show_spaces()	// set show-spaces versions of two arrays
{
	int i, ansi_font;
	char *p;

	ansi_font = !using_oem_font(window_to_screen(window_handle));
	if (_show_disp_class) {
		for (i = 0; i < 256; i++)
			_show_disp_class[i] = _std_disp_class[i];
		if (!(show_spaces & 2))
			_show_disp_class[' '] = ansi_font ? 183 : 250;
	}
	if (_show_disp_chars) {
		for (i = 0; i < NUMDISPCHARS + 1; i++)
			_show_disp_chars[i] = _display_characters[i];
		p = _show_disp_chars + NUMLINECHARS + NUMCORNERCHARS;
		if (!(show_spaces & 4))
			p[3] = ansi_font ? 187 : 0x1a;	// tab
		if (!(show_spaces & 8))
			p[4] = ansi_font ? 182 : 0x11;	// newline
	}
}

command change_show_spaces() on reg_tab[FSHIFT(6)]
{
	show_spaces = has_arg ? iter : (show_spaces ^ 1);
	iter = 0;
	when_setting_show_spaces();
	say("%s spaces in this buffer.", (show_spaces & 1) ? "Showing" : "Hiding");
}

when_setting_show_spaces()
{
	last_show_spaces = show_spaces;
	if (show_spaces) {
		if (!_show_disp_class)
			_show_disp_class = malloc(256);
		if (!_show_disp_chars)
			_show_disp_chars = malloc(NUMDISPCHARS + 1);
		make_show_spaces();
	}
	if ((show_spaces & 1) && _display_class == _std_disp_class)
		_display_class = _show_disp_class;
	else if (!(show_spaces & 1) && _display_class == _show_disp_class)
		_display_class = _std_disp_class;
	buffer_display_characters = (show_spaces & 1) ? _show_disp_chars : 0;
	full_redraw = 1;
}

prepare_dlg_filter()
{
	char pats[FNAMELEN], *s;

	save_var bufnum = create("-dlg-filter");
	if (!size()) {
		if (s = lookpath("filter.txt"))
			file_read(s, FILETYPE_AUTO);
		else
			stuff("All Files *.*\n");
		point = 0;
		simple_re_replace(1, "^#.*", "");
		while (search(1, ";")) {
			replace(point - 1, ' ');
			grab(point, give_end_line(), pats);
			to_end_line();
			delete(point, point + 1);
			insert(0);
			stuff(pats);
			insert(0);
		}
		insert(0);
		int i = new_variable("dlg-filter-str", NT_VAR, TYPE_CARRAY, size() + 1);
		s = varptr(i);
		grab(0, size(), varptr(i));
	}
}

use_common_open_file_dlg(fname, title, flags, save)
char *fname, *title;
int *flags;
{
	char dir[FNAMELEN];
	int ret;

	if (common_open_use_directory & 1) {
		getcd(dir);
		common_open_curdir = dir;
	} else if (common_open_use_directory & 12) {
		if (get_buffer_directory(dir) || (common_open_use_directory & 8))
			common_open_curdir = dir;
	}
	prepare_dlg_filter();
	ret = common_file_dlg(fname, title, flags, save,
						  varptr(find_index("dlg-filter-str")),
						  custom_filter, &filter_index);
	if (ret) {
		absolute(fname);
		if (common_open_use_directory & 2)
			chdir(common_open_curdir);
	}
	return ret;
}

comm_break_line_here(int c_like, int comm, int starts_with_comm, int starts_in_comm,
					 int comm_start, spot next_ends, spot after, char *fill)
{
	int offset, generic = strcmp(major_mode, "C");
	char rest[FNAMELEN], *new_comm;

	delete_horizontal_space();
	insert('\n');
	if (display_column > 0)	// Horizontally scroll back to left margin.
		display_column = 0;
	if (key == '\n')	// User might be trying to break line elsewhere as
		key = ' ';		// we're breaking it here.  Prevent a mess.

	if (comm == IN_NEW_COMMENT || !c_like) {		// A //-style C++ comment.
		new_comm = generic ? get_comment_begin() : "// ";
		if (starts_with_comm)	// This line began with a comment, so just
			stuff(fill);		// repeat its prefix (indentation and such).
		else if (next_ends && get_column(comm_start) > get_column(*next_ends)) {
			grab(point, give_end_line(), rest);	// Continue it on the next line
			offset = give_end_line() - *after;	// since there's no comment
			delete(point - 1, give_end_line());	// there.
			point = *next_ends;
			to_column(get_column(comm_start) - 2);
			bprintf("%s%s%s", isspace(character(point - 1)) ? "" : " ",
					new_comm, rest);
			*after = point - offset;	// Point might be in the middle of moved text.
		} else {			// Make a new line after the commented code line we
			stuff(new_comm);	// broke.
			point -= strlen(new_comm);
			if (c_like)
				indent_to_column(c_compute_indent());
			else if (indenter)
				(*indenter)();
		}
	} else {			/* Old-style K&R comment. */
		if (strstr(fill, "*/"))
			*fill = 0;
		if (starts_in_comm)		// Just repeat the *'s or whatever at start of
			stuff(fill);		// old line.
		else if (starts_with_comm) {
			if (!strcmp(fill, "/**"))	// A Javadoc comment.
				fill = " * ";
			replace_in_string(fill, '/', ' ');
			if (fill_c_comment_plain)	// Like above, but convert that first /*.
				replace_in_string(fill, '*', ' ');
			stuff(fill);
		} else {
			delete(point - 1, point);	// Delete the \n we just inserted.
			stuff(" */\n/* ");			// And put this instead.
			to_begin_line();
			indent_to_column(c_compute_indent());
		}
	}
}

do_grep(buf, pat, flags, filepat, inbufs)
char *buf, *pat, *filepat;
{
	int res = 0, aborted = 0, cnt = 0, tempbuf = 0, i, skipped = 0;
	int orig, before_head, this_err;
	char *s, rel[FNAMELEN], abs[FNAMELEN], *(*finder)();
	char *months = "Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec ";
	char extra[FNAMELEN];
	spot start;

	save_var synchronous_file_io = 1;
	before_head = prepare_grep_buffer(buf, flags, pat, filepat);
	start = alloc_spot(0);
	save_var modified;
	save_var abort_file_matching = ABORT_ERROR;
	finder = give_grep_finder(inbufs, filepat, &s);	// get 1st match
	save_var readonly_warning, no_popup_errors = 1;
	if (grep_on_changed_file & 1)
		readonly_warning |= ROWARN_GREP;
	readonly_warning &= ~(ROWARN_MSG | ROWARN_BELL);
	orig = bufnum;
	for (; s; s = (*finder)(filepat, 0)) {
		if (user_abort) {
			aborted = 1;
			break;
		}
		grep_make_names(rel, abs, s, inbufs);
		if (inbufs)
			bufname = s;
		else if (!ok_grep_file_match(s, abs)) {
			skipped++;
			continue;
		}
		noteput("%d matches, searching %f...", cnt, rel);
		if (inbufs)	// skip over grep results buffer
			this_err = !strcmp(s, buf);
		else if (!look_file(abs)) {
			this_err = find_it(abs, default_translation_type);
			if (!grep_keeps_files)
				tempbuf = bufnum;
		} else {
			this_err = 0;
			if (grep_on_changed_file & 2)
				save_var auto_read_changed_file = 1;
			if (check_dates(0) == 2)
				grep_date_warning(buf, filename);
		}
		if (this_err == EREADABORT || user_abort) {
			aborted = 1;
			break;
		} else if (this_err) {
			window_bufnum = orig;
			if (tempbuf)
				buf_delete(tempbuf);
			tempbuf = 0;
			continue;
		}
		*extra = 0;
		if (grep_include_timestamp && !inbufs && file_info.check_type == CHECK_FILE)
			if (grep_include_timestamp == 2)
				sprintf(extra, "[%04d-%02d-%02d %02d:%02d:%02d] ",
						file_info.year, file_info.month, file_info.day,
						file_info.hour, file_info.minute, file_info.second);
			else
				sprintf(extra, "[%02d-%.3s-%04d %02d:%02d:%02d] ",
						file_info.day,
						(file_info.month >= 1 && file_info.month <= 12)
						? months + (file_info.month - 1) * 4 : "???",
						file_info.year,
						file_info.hour, file_info.minute, file_info.second);
		if (grep_show_absolute_path == 2)
			convert_to_8_3_filename(abs);
		res = copy_matches(name_to_bufnum(buf), flags, pat,
						   grep_show_absolute_path ? abs : rel, extra);
		window_bufnum = orig;
		if (res < 0)	/* error occurred during search */
			break;
		else if (res > 0) {
			cnt += res;
			build_first = 1;
			update_grep_title(before_head, cnt);
			refresh();
		}
		if (tempbuf)
			buf_delete(tempbuf);
		tempbuf = 0;
	}
	window_bufnum = orig;
	point = i = *start;
	free_spot(start);
	build_first = 1;
	if (tempbuf)
		buf_delete(tempbuf);
	say("");
	if (aborted)
		quick_abort();
	if (res == -DSBAD)
		error("Illegal search pattern: %s", pat);
	else if (res != -DSABORT) {
		locate_window(buf, "");
		point = i;
		if (skipped) {
			say("Skipped %d file%s due to grep-ignore-file- variables",
				  skipped, (skipped == 1) ? "" : "s");
			bprintf("%pSkipped %d file%s due to grep-ignore-file- variables\n",
					color_class grep_header,
					skipped, (skipped == 1) ? "" : "s");
		}
	}
}

color_html_include()
{
	int s, t, c, lang, orig = point;
	char *end, *name;

	search(-1, "<");	// Find start of include.
	c = character(point + 1);
	s = point;
	switch (c) {
		case '%':	if (parse_string(1, "%<%%--")) {
						search(1, "--%>");	// Found a JSP "hidden comment".
						set_character_color(s, point, color_class html_comment);
						return 1;
					}
					point += parse_string(1, "%<%%[a-z0-9]*");
					end = "%>";
					lang = get_html_asp_language(sublanguage_code
						== SUBLANG_XML ? xml_asp_coloring : html_asp_coloring);
					break;
		case '?':	if (sublanguage_code != SUBLANG_XML
						  || parse_string(1, "%<%?php"))
						lang = html_php_coloring;
					else
						lang = xml_php_coloring;
					point += parse_string(1, "%<%?[a-z0-9]*");
					end = "?>";
					break;
		default:	point++;
					end = ">";
					lang = get_html_asp_language(html_asp_coloring);
					break;
	}
	if (lang == HT_IGNORE) {
		point = orig;
		return 0;
	}
	t = point;
	// First mark whole section embedded.
	set_character_color(s, point, color_class html_embedded);
	set_tagged_region("color-as-unit", s, point, 99);
	for (;;) {
		c = search(1, end) ? matchstart : point;	// Find the end marker.
		color_embedded_code(t, point, lang);	// Color it too for now.
		// If it wound up being colored as a comment or string, that's
		// not really the end, so find the next marker.
		name = name_color_class(get_character_color(c, &t));
		if (point <= c || !name || !strstr(name, "comment")
			  && !strstr(name, "string"))
			break;
		// Force recoloring of that entire comment/string.
		set_tagged_region("needs-color", t, s = point, -1);
	}
	// Color the marker ending the embedded code.
	set_character_color(c, point, color_class html_embedded);
	set_tagged_region("color-as-unit", c, point, 99);
	return 1;
}

color_html_move_back()
{
	int start = point, p;

	save_var narrow_start = point - color_html_look_back;
	if (re_search(-1, "%<%!%[CDATA%[|%]%]%>")
		  && !parse_string(-1, "%<script[^<]*%><wspace>*"))
		if (curchar() == '<')
			return 1;
		else {
			point += parse_string(1, "%]%]%>");
			return -1;
		}
	point = start;
	if (re_search(-1, "<Langle>[ \t\n]*/?script")
		&& parse_string(1, "<Langle>[ \t\n]*script"))
		return 2;				// Found a <script>, not a </script>.
	point = start;
	if (re_search(-1, "<Langle>[%?]|[%?]<Rangle>")
		&& curchar() == '<')
		return 3;				// Found a <% or <?, not a %> or ?>.
	point = start;
	do {			// start coloring from the previous tag
		if (!search(-1, "<"))
			break;	// but ignore <'s before " or ' characters
		if (get_character_color(point, &p, NULL) == color_class html_comment) {
			point = p + 2;	// If we're in a comment,
			search(-1, "<!");	// go back to its start.
		}
	} while (html_in_quoted_attribute());
	return 0;
}

// Color html from point to end.  Move past the part we colored, which
// may extend past end.

color_html_loop(int end)
{
	int s, t, i, script;

	for (s = point; point < end; ) {
		if (curchar() == '&') {
			set_character_color(s, point, color_class html_text);
			s = point++;	 // Color an entity like &amp; or &#251;
			if (parse_string(1, "([a-z]+|#[0-9]+|#x[0-9a-f]+);?"))
				point = matchend;
			set_character_color(s, point, color_class html_entity);
			s = point;
			continue;
		}
		if (curchar() != '<') {
			point++;
			continue;
		}
		set_character_color(s, point, color_class html_text);
		i = t = point++;
		while (strchr(" \t\n", curchar()))
			point++;	// Parse < followed by whitespace.
		set_character_color(t, point, color_class html_punctuation);
		if (curchar() == '!')
			color_html_comment();
		else if (strchr("%?", curchar()) && color_html_include())
			;
		else if (s = parse_string(1, "/?[-_.a-z0-9]*")) { // Parse the tag.
			set_character_color(point, point + s, color_class html_tag);
			script = (s == 6 && parse_string(1, "script"));
			if (s == 3 && parse_string(1, "jsp:(declaration|expression"
									   "|scriptlet)</word>"))
				script = 2;
			point += s;			// Look for javascript tag.
			if (curchar() == ':' && parse_string(1, ":[-_.a-z0-9]*")) {
				set_character_color(point - s, point + 1,
									color_class html_namespace);
				set_character_color(point + 1, matchend,
									color_class html_tag);
				point = matchend;
			}
			if (parse_string(1, "[ \t\n]*/[ \t\n]*>")) {
				set_character_color(point, matchend, color_class html_tag);
				point = matchend;
			} else {
				for (i = point; strchr(" \t\n", curchar()); )
					point++;	// Parse whitespace.
				set_character_color(i, point, color_class html_punctuation);
				color_html_attributes(script, t);
			}
		}
		s = point;
	}
	re_search(1, "![<&]");
	if (s < point)
		set_character_color(s, point, color_class html_text);
}

color_html_range(from, to) // recolor just this section
{			// last colored region may go past to
	int i;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	if (from < to) {
		set_character_color(from, to, color_class html_text);
		set_tagged_region("needs-color", from, to, -1);	// Say it needs coloring.
	}
	point = from;
	save_var case_fold = 1;
	color_html_move_back();
	while (point > narrow_start
		   && get_tagged_region("color-as-unit", point - 1, &from) != -1)
		point = from;		// Move before script areas.
	if (to > size() - narrow_end)
		to = size() - narrow_end;
	set_tagged_region("color-as-unit", from, to, -1);
	from = point;
	for (;;) {
		color_html_loop(to);
		set_tagged_region("needs-color", from, point, 0);	// Say it's colored.
		i = get_tagged_region("color-as-unit", point, NULL, &to);
		if (i == -1 || i == 99 || point >= size() - narrow_end || to <= point)
			break;
		// The next text was colored as scripting, but it shouldn't be
		// any more, since we never stop coloring inside scripting.
		// Continue coloring it as HTML, and mark the rest of the buffer
		// to be recolored later.
		set_tagged_region("needs-color", point, size(), -1);
	}
	return point;
}

html_redirect_active_key() on html_tab['{'], html_tab['}'], html_tab[')'],
							  html_tab[']'], html_tab[':'], html_tab['#']
{
	int region, start, end, f = 0;
	short *tab = 0;

	recolor_buffer_range(point, point + 1);
	region = get_tagged_region("color-as-unit", point, &start, &end);
	switch (region) {
		case 1:
			save_var auto_show_matching_characters = c_auto_show_delim_chars;
			save_var mode_move_level = c_move_level;
			tab = c_tab;
			break;
		case 2:
			save_var auto_show_matching_characters = vbasic_auto_show_delim_chars;
			save_var mode_move_level = 0;
			tab = vbasic_tab;
			break;
		case 3:
			save_var auto_show_matching_characters = perl_auto_show_delim_chars;
			save_var mode_move_level = c_move_level;
			save_var in_perl_buffer = 1;
			save_var sublanguage_code = SUBLANG_PHP;
			tab = c_tab;
			break;
	}
	if (tab)
		f = binding_of(key, tab);
	if (f > 0) {
		save_var narrow_start = start, narrow_end = size() - end;
		html_exclude_script_head();
		do_command(f);
		return;
	}
	normal_character();
}

char *html_comment_begin_override()
{
	save_var narrow_start = 0, narrow_end = 0;
	recolor_buffer_range(point, point + 1);
	switch (get_tagged_region("color-as-unit", point)) {
		case HT_PHP:
			switch (php_comment_style) {
				case 1: return "# ";
				case 2: return "// ";
			}
			return "/* ";
		case HT_JSCRIPT:
			return new_c_comments ? "// " : "/* ";
		case HT_VBSCRIPT:
			return "' ";
	}
	return 0;
}

char *html_comment_end_override()
{
	save_var narrow_start = 0, narrow_end = 0;
	recolor_buffer_range(point, point + 1);
	switch (get_tagged_region("color-as-unit", point)) {
		case HT_PHP:
			switch (php_comment_style) {
				case 1:
				case 2: return "";
			}
			return " */";
		case HT_JSCRIPT:
			return new_c_comments ? "" : " */";
		case HT_VBSCRIPT:
			return "";
	}
	return 0;
}

indent_like_tab()	/* indent like tab char, but maybe use spaces */
{
	int cur = current_column(), tab = get_soft_tab_size();

	to_column(cur + tab - cur % tab);
}

maybe_indent_rigidly(int rev)
{
	switch (rev) {
		case 0: rev = IS_SHIFT_KEY(full_key) != 0; break;
		case 2: rev = 0; break;
	}
	if (!is_highlight_on())
		return 0;
	if (resize_rectangle_on_tab && is_highlight_on() == REGRECT) {
		int tab = get_soft_tab_size();
		to_virtual_column((virtual_column() + (rev ? -1 : tab)) / tab * tab);
		iter = 0;
		return 1;
	}
	if (rev) {
		if (!has_arg)
			iter = -get_soft_tab_size();
		else if (iter > 0)
			iter = -iter;
		has_arg = 1;
	}
	indent_rigidly();
	highlight_on();
	shift_selecting = 2;
	return 1;
}

prep_customizations()
{
	int b;
	char file[FNAMELEN], *p = 0;

	if (!exist("-customizations"))	// Read the file 1st time, in case we
		p = lookpath(einit_file_name);		// didn't at start.
	save_var bufnum = b = create("-customizations");
	save_var use_default = 0;
	if (p) {
		do_file_read(p, default_translation_type);
		ini_mode();
	}
	if (!size()) {
		bprintf("# Epsilon customization file. -*- Mode: Ini -*-\n");
		bprintf("# This file lists your Epsilon customizations, such as\n");
		bprintf("# setting variables, colors, bindings and so forth.\n");
		bprintf("# Epsilon reads it at startup.\n\n");
		ini_mode();
	}
	if (!*filename) {
		get_customization_directory(file);
		sprintf(file + strlen(file), einit_file_name);
		filename = file;
	}
	want_backups = 1;
	if (record_customizations == 2)
		save_without_prompt = 1;
	auto_read_changed_file = 1;
	if (!modified)
		check_dates(0);
	point = size();
	return b;
}

insert_key_unbindings(int src)
{
	char keyspec[FNAMELEN], pat[FNAMELEN];
	int orig = point, obuf = bufnum;

	save_var bufnum = src, point = 0;
	while (re_search(1, "^(~<^wspace>+ \".*\": )")) {
		grab(matchstart, matchend, keyspec);
		bufnum = obuf;
		point = orig;
		sprintf(pat, "^%q", keyspec);
		if (!re_search(1, pat))
			bprintf("%s0\n", keyspec);
		bufnum = src;
	}
}

int make_difference_list(int orig, int cur)
{
	save_var show_status = 0;
	save_var bufnum = orig;
	save_var case_fold = 1;
	do_sort_region(0, size(), 0, 0);
	save_var bufnum = cur;
	save_var case_fold = 1;
	do_sort_region(0, size(), 0, 0);
	int out = tmp_buf();
	do_compare_sorted(orig, cur, "original-settings",
					  bufnum_to_name(out), NULL);
	bufnum = out;
	point = 0;
	save_var case_fold = 1;
	custom_apply_immediately();
	// Drop vars starting with -, but not color scheme ones.
	do_drop_matching_lines(REGEX, "^%+-[^o][^u][^r][^-]", 1);
	drop_boring_change_lines();
	custom_fix_mode_vars();
	insert_key_unbindings(name_to_bufnum("original-settings"));
	do_drop_matching_lines(REGEX, "^%+(push-cmd|echo-line|permanent-menu"
						   "|avoid-(top|bottom)-lines|bufed-width"
						   "|mode-start|mode-end) ", 1);
	do_drop_matching_lines(REGEX, "^&.* color class for after-exiting: .*$", 1);
	return out;		// Don't count vars saved in session file, others.
}

perl_compute_indent()
{
	int ind = 0;		/* indentation to use */
	int orig = point, i;
	int first_char = curchar(); /* cur line's first char */
	int pstart;		/* start position of prev line */
	int pafter;		/* after label on prev line */
	int pend;		/* end position of prev line */
	int prev_end = ';';	/* last char on previous line */
	int norm_narrow;	/* narrowing setting for most parsing */

	save_var point;
	save_var Closeback = perl_closeback;
	save_var Topindent = perl_topindent;
	save_var c_indent = perl_indent;
	save_var c_label_indent = perl_label_indent;
	save_var c_contin_offset = perl_contin_offset;
	save_var c_align_contin_lines = perl_align_contin_lines;
	save_var c_brace_offset = perl_brace_offset;
	save_var c_top_braces = perl_top_braces;
	save_var c_top_contin = perl_top_contin;
	save_var c_top_struct = perl_top_struct;

	if (sublanguage_code != SUBLANG_PHP)
		save_var narrow_start = 0;
	recolor_buffer_range(orig, orig + 1);
	if (get_tagged_region("color-perl-as-unit", orig, &i) != -1
		  || get_character_color(orig, &i) == color_class perl_constant
		  && matches_at(i, 1, "%<%<"))		// Keep original's indentation
		return get_indentation(orig);		// if in here doc.
	if (get_character_color(orig - 1, &i) == color_class perl_string)
		return get_indentation(orig);		// or in string.
	if (get_character_color(orig - 1, &i) == color_class perl_comment
		  && matches_at(i, 1, "="))
		return get_indentation(orig);		// or in POD.

	norm_narrow = fix_narrowing(point - c_look_back);
	narrow_start = norm_narrow;
	if (first_char == '\f')
		return 0;
	else if (first_char == '}')
		return cind_close();

	to_begin_line();
	if (parse_string(1, "[ \t]*\n") && after_many_comments()) {
		re_search(-1, "[^ \t\n]");	/* find previous non-blank line */
		return get_indentation(point);
	}
	cind_space();			/* skip whitespace and comments */
	if (character(point - 1) == ':')
		move_before_labels();
	prev_end = character(point - 1); /* get last relevant char */
	pend = point;

	to_statement_start();		/* if contin line, back up */
	pstart = point;			/* now at start of biggest statement */
	past_labels();			/* containing original point */
	pafter = point;			/* just after label */
	ind = current_column();
	narrow_start = norm_narrow;	/* ok to look far back again */

	if (sublanguage_code == SUBLANG_PHP && is_case(orig))
		ind = indent_case(orig);
	else if (have_label(orig)) {
		ind = c_label_indent;
	} else if (prev_end == ':' && is_label(pstart)) {
		ind = get_column(pstart);
		if (!(Closeback && just_open(orig)))
			ind += get_c_indent();
		if (just_open(orig))
			ind += c_brace_offset;
	} else if (is_else(orig)) {
		point = orig;
		ind = indent_else(ind);
	} else if (prev_end == ';') {
		if (just_open(orig) && func_start_block(pstart))
			ind = c_top_braces;
		else
			ind = indent_after_statement(orig);
	} else if (prev_end == '}') {
		point = pend;
		ind = indent_after_block(orig);
	} else if (prev_end == '{') {
		if (ind <= c_top_braces	// special case if no previous indent
			|| get_column(pend - 1) <= c_top_braces) {
			ind += top_indent(prev_end) ? get_c_indent() : c_top_braces;
			if (just_open(orig))
				ind += c_brace_offset;
			if (ind < c_top_braces)
				ind = c_top_braces;
		} else if (Closeback || curchar() != '{')
			ind += get_c_indent();
		if (just_open(pafter))
			ind -= c_brace_offset;
	} else if (just_open(orig))
		return indent_just_open(ind);
	else if (is_contin_statement(pend, 1, orig)) {
		return indent_c_contin(ind, orig, pend);
	} else if (!index(":);{}", c_line_before()) && !c_statement_start()
			   && !(Closeback && just_open(point)))
		return ind;
	else
		ind += get_c_indent();
	return ind;
}

int compile_handler(int activity, int buf, int from, int to);

do_compile(cmd)
char *cmd;
{
	int err = 0, nerr, sep_buf_ok = 0;
	int obuf = bufnum;
	char *procbuf = PROCBUF, curdir[FNAMELEN];

	_last_was_grep = 0;
	if (maybe_save_all())
		return 3;
	if (start_process_in_buffer_directory)
		get_buffer_directory(curdir);
	else
		getcd(curdir);
	drop_final_slash(curdir);
	if (!(has_feature & FEAT_MULT_CONCUR) && !another_process())
		save_var restart_concurrent = 0;

	// If we're allowed to run non-concur, then use separate buffer.
	if (concurrent_make < 2 && (has_feature & FEAT_MULT_CONCUR)
		  && compile_in_separate_buffer)
		sep_buf_ok = 1;
	else if (concurrent_make == 2)
		create_concur("", PROCBUF);
	else if (!sep_buf_ok && concurrent_make == 0)
		try_exit_concurrent();
	if (!sep_buf_ok && clear_process_buffer)
		zap(PROCBUF);
	if (sep_buf_ok) {
		int i;
		char newname[FNAMELEN];
		bufnum = create(procbuf = "compilation-output");
		for (i = 2; is_process_buffer(bufnum); i++) {
			sprintf(procbuf = newname, "compilation-output-%d", i);
			bufnum = create(procbuf);
		}
		if (clear_process_buffer)
			zap(procbuf);
        process_mode();
		point = size();
		set_error_spot();
		_had_errors = 0;
		bprintf("Running `%s' in directory `%s'\n", cmd, curdir);
		type_point = size();
		if (compile_in_separate_buffer == 2)
			locate_window(procbuf, "");
		if (!concur_shell("", cmd, curdir, procbuf)) {
			update_mode_more_extra(0);
			when_activity = compile_handler;
			while (process_exit_status == PROC_STATUS_RUNNING) {
				if (char_avail() && getkey() != MOUSE_MOVE
					  && (!IS_WIN_KEY(key) || key == WIN_DRAG_DROP)) {
					ungot_key = key;	// Wait until user presses a key
					return 2;			// or process finishes.
				}
				delay(10, COND_KEY, bufnum);	// Delay 0.1 secs
			}
			err = process_exit_status;
			bufnum = obuf;
			check_dates(0);
		} else
			err = -1;
	} else if (concurrent_make == 4 && is_process_buffer(name_to_bufnum(PROCBUF))) {
		save_var bufname = PROCBUF;	// Run the program in the concur
		set_error_spot();			// proc buffer, but don't process
		_had_errors = 0;			// the user's keys until it finishes.
		if (*cmd) {
			point = size();
			bprintf("%s\n", cmd);
		}
		maybe_refresh();
		iter = 1;
		delay(-1, COND_PROC, bufnum);
		if (!is_process_buffer(name_to_bufnum(PROCBUF)) || !process_input())
			return 2;	/* else assume it's compiled */
	} else if (no_running(0)) {	/* goes to end of proc buffer */
		set_error_spot();
		_had_errors = 0;
		to_begin_line();
		bprintf("%pRunning `%s' in directory `%s'\n",
				color_class process_output, cmd, *process_current_directory
				? process_current_directory : curdir);
		type_point = point = size();
		if (*cmd)
			bprintf("%s\n", cmd);
		maybe_refresh();
		iter = 1;
		if (opsys != OS_UNIX)
			delay(-1, COND_PROC | COND_KEY, bufnum);
		if (!is_process_buffer(name_to_bufnum(PROCBUF)) || !process_input())
			return 2;	/* else assume it's compiled */
		if (!(has_feature & FEAT_DETECT_CONCUR_WAIT))
			return 5;
	} else {
		bufnum = create(PROCBUF);
		point = type_point = size();
		buffer_printf("process", "%p\nRunning `%s' in directory `%s'\n",
				color_class process_output, cmd, curdir);
		bufnum = obuf;
		err = do_push(cmd, 1, 0);
	}
	maybe_restart_concurrent();
	nerr = do_next_error(1, cmd);	// scan for error messages
	if (!nerr && err) {	// no message recognized, but error code
		say("Process returned %d", err);
		locate_window(procbuf, "");
		point = size();		/* show end of process buffer */
		return err;
	} else
		bufnum = obuf;
	if (err)
		return 1;
	else if (nerr)
		return 4;
	return 0;
}

try_grab_directory_name(int from, int to)
{
	char dir[FNAMELEN];

	if (!is_win32 || !use_process_current_directory
		  || to - from > 200 || can_get_process_directory == 1)
		return;					// Too long for a likely dir name.
	if (*process_prompt_pattern) {
		save_var narrow_end = size() - to;
		if (!matches_at(from, 1, process_prompt_pattern)
			  && !matches_at(from, -1, process_prompt_pattern))
			return;
		grab(find_group(1, 1), find_group(1, 0), dir);
	} else {
		if (character(--to) != '>')
			return;
		grab(from, to, dir);
	}
	convert_cygwin_filename(dir, 1);
	if (is_directory(dir)) {
		process_current_directory = dir;
		if (use_process_current_directory == 2)
			chdir(dir);
		can_get_process_directory = 2;
	}
}

int proc_complete_list_matches(char *pat, int initial, int skip_dots)
{
	int flags = 0, proc_tab_size = tab_size;
	char *pat2 = 0, file[FNAMELEN], *s, *t;

	if (initial && (s = getenv("PATH")) && !strchr(pat, '/') &&
		  (opsys == OS_UNIX || !strchr(pat, '\\') && !strchr(pat, ':'))) {
		pat2 = malloc(strlen(s) + strlen(pat) + 100);
		sprintf(pat2, (opsys == OS_UNIX) ? "{%s}/%s*" :
				"{%s,.}/%s*.{exe,bat,cmd,com}", s, pat);
		while (s = strchr(pat2, path_list_char))
			*s = ',';
		flags = EXACTONLY;
		pat = pat2;
	}
	zap(_MATCH_BUF);
	if (s = do_file_match(pat, STARTMATCH | flags))
		do {
			if (opsys == OS_UNIX && initial && lowaccess(s, LOWACC_X))
				continue;
			if (flags) {
				strcpy(file, s);
				s = get_tail(file, 0);
				if (opsys != OS_UNIX)
					*get_extension(s) = 0;
			}
			if (skip_dots && *get_tail(s, 0) == '.'
				  && string_matches_pattern(get_tail(s, 0), "%.%.?[/\\]?<eof>"))
				continue;
			if (cygwin_filenames & 64)
				while (t = strchr(s, '\\'))
					*t = '/';
			buffer_printf(_MATCH_BUF, "%s\n", s);
		} while (s = do_file_match(pat, 0));
	free(pat2);	// Free if we allocated it.
	save_var bufname = _MATCH_BUF;
	do_sort_region(0, size(), 0, 0);
	do_uniq(1, 1, 0);
	tab_size = proc_tab_size;
	return lines_between(0, size(), 0);
}

show_line()		/* display point for a moment, then restore */
{
	int oldstart = window_start;
	int time;

	if (in_macro())
		return;
	save_var narrow_start = 0, narrow_end = 0;
	refresh();
	time = (window_start == oldstart) ? near_pause : far_pause;
	delay(time, COND_TRUE_KEY);
	window_start = oldstart;
	build_first = 1;
}

struct {
	char prefix[100];		// Find words that start with this.
	char found[200];		// The current candidate replacement.
	char last[200];			// Previous replacement we found.
	int len;				// Length of the current candidate replacement.

	// These apply to the destination buffer.
	int savepos;		// Buffer position of the start of the prefix.

	// The rest apply to the source buffer we're searching for matches.
	int window_src;		// Get matches from this window.
	int pullsrc;		// Region we're highlighting.
	spot start;			// Its start and end.
	spot end;

	int tag_position;	// Position in tag buffer we're searching.

	spot ignore_start;	// Ignore matches in this range.
	spot ignore_end;

	int match_current_word;	// At start, match word we're inside.

	int orig;			// Position in search buffer where we started.
} pull;

void alloc_pl()
{
	/* Get the word prefix, if any, that occurs
	 * immediately before point.
	 */
	pull.len = 0;
	pull.prefix[0] = 0;
	pull.ignore_end = alloc_spot(1);
	parse_string(-1, word_pattern, pull.prefix);
	pull.ignore_start = alloc_spot(0);	// Ignore original word.

	pull.savepos = point;
	pull.len = strlen(pull.prefix);

	if (_doing_input == DI_LINEINPUT || _doing_input == DI_SEARCH
		  || _doing_input == DI_SEARCH_NOINCR) {
		pull.window_src = prompt_orig_win;
		save_var window_handle = prompt_orig_win;
		free_spot(pull.ignore_start);	// Don't restrict matches.
		pull.ignore_start = 0;
	} else
		pull.window_src = 0;
	pull.start = alloc_spot(0);
	pull.end = alloc_spot(0);
	pull.orig = point;
	pull.pullsrc = add_region(pull.start, pull.end, color_class pull_highlight, REGNORM);
	*pull.last = 0;
	pull.tag_position = -1;
	pull.match_current_word = 0;
	if (_doing_input == DI_LINEINPUT || _doing_input == DI_SEARCH
		  || _doing_input == DI_SEARCH_NOINCR)
		pull.match_current_word = 1;
}

search_read(trystr, prmpt, flags)
char *trystr, *prmpt;
{
	int lastok, goodchars, oldflags, origwin = window_handle;
	int editbuf, pbuf, pwin, editwin, s_code, ostart, cnt;
	int oreg_type, old_mark = -1, search_contin = search_continuation != 0;
	int restrict_region = 0, recall_cnt = 0;
	char tmpbuf[FNAMELEN];
	old_place orig, mstart, okpoint;

	if (_doing_input == DI_SEARCH || _doing_input == DI_SEARCH_NOINCR)
		error_if_input(0);
	save_var _highlight_control;
	if (search_in_region && is_highlight_on()) {
		save_var restrict_search_type;
		restrict_region = make_restriction(flags);
	}
	if (is_highlight_on() && point != mark && ABS(point - mark) < 200
		  && !*trystr && typing_deletes_highlight && typing_deletes_highlight != 3
		  && !in_macro(1) && _len_def_mac < 1 && search_defaults_from == 1) {
		old_mark = mark;	// Use this as the initial search string.
		highlight_off();
	}
	save_var iter, has_arg, this_cmd, prev_cmd;
	save_var _doing_input = (flags & INCR) ? DI_SEARCH : DI_SEARCH_NOINCR;
	save_var prompt_orig_buf = bufnum;
	save_var prompt_orig_win = window_handle;
	save_var last_index;
	save_var mark_to_column = -1, mark = point;
	save_var cursor_to_column = -1;
	save_var postpone_asynchronous_notify = 1;
	save_var macro_runs_immediately = 0;	// Make special keys work in macros.
	oreg_type = region_type();
	set_region_type(REGNORM);
	flags &= ~SEARCH_EMPTY | SEARCH_CONTIN;
	remember_old_place(&orig, point);
	remember_old_place(&mstart, point);
	remember_old_place(&okpoint, point);
	ostart = window_start;
	goodchars = lastok = 0;	/* 2 versions of len of last good string */
				/* (only lastok is set by getting default) */

	editbuf = tmp_buf();		/* make buffer for cmd line editing */
	pwin = add_popup(0, echo_line, screen_cols - 1, 1, 0,
			pbuf = tmp_buf());	/* prompt in window */
	stuff(trystr);
	trystr = malloc(FNAMELEN);
	display_column = 0;
	system_window = 1;	/* don't let user switch to it */
	text_color = color_class echo_area;

	editwin = add_popup(1, echo_line, screen_cols - 1, 1,
			0, editbuf);	/* window on top of prompt window */
	text_color = color_class echo_area;
	_display_class = _echo_disp_class;
	display_column = 0;

	if (old_mark != -1) { // Start out with highlighted region.
		window_handle = origwin;
		buf_xfer(editbuf, old_mark, point);
		window_handle = editwin;
		point = 0;
		if (flags & REGEX)	// Quote special chars.
			string_replace("[[%.!#()*+|?^$<]", "%#0", REGEX);
		point = size();
		highlight_on();		// Mark it so user can delete by typing.
		shift_selecting = 2;
	} else if (search_defaults_from == 2 && !size()) {
		if (flags & UNKNOWN)
			flags = known_flags(flags);
		recall_cnt++;
		stuff(default_search_string(flags));
		point = size();
		highlight_on();		// Mark it so user can delete by typing.
		shift_selecting = 2;
	}

	for (;;) {
		do {
			show_search_prompt(prmpt, flags, origwin,
					   pwin, editwin);
			getkey();
		} while (key == MOUSE_MOVE);
		if (IS_MOUSE_KEY(key))
			if (mouse_y == echo_line)
				flags &= ~INCR;
			else {
				flags |= SONCE;
				ungot_key = key;
				break;
			}
		cnt = 1;
		if (key == CTRL('^'))
			full_getkey("", CTRLIFY_KEY);
		window_handle = editwin;
		s_code = 0;
		oldflags = flags;
		flags &= ~(BAD | FAILING | SONCE);	/* assume ok */
		if (is_search_dir()) {
			window_handle = origwin;
			to_old_place(&okpoint);
			window_handle = editwin;
			if ((key == CTRL('R')) != ((flags & REVERSE) != 0)) {
				flags ^= REVERSE;	/* switch dir */
				if (goodchars)
					flags |= SONCE;
				lastok = size();
			} else if (!size()) {
				if (flags & UNKNOWN)
					flags = known_flags(flags);
get_defaults:
				stuff(default_search_string(flags));
				lastok = size();
			} else if (oldflags & FAILING) {
				if (search_contin) {
					flags |= SEARCH_CONTIN;
					s_code = FHERE;			/* Try other records. */
				} else if (search_wraps) {
					flags |= (flags & SEARCH_WRAPPED) ? SEARCH_OVERWRAPPED
						: SEARCH_WRAPPED;
					window_handle = origwin; /* Wrap to other end of buffer. */
					point = (flags & REVERSE) ? size() : 0;
					window_handle = editwin;
					s_code = FHERE;			/* And search again. */
				} else
					flags |= FAILING; /* no more matches, stop */
			} else
				flags |= SONCE;
			if ((flags & (INCR | SONCE)) && !(flags & FAILING))
				flags |= SONCE, s_code = FHERE;
		} else if (set_flag_modes(&flags)) {
			if ((flags & INCR) && size())
				s_code = FMATCH; /* if modes changed, search again */
		} else if (key == CTRL('O')) {
			flags ^= INCR;
			point = size();
			if (flags & INCR)
				s_code = FMATCH;
		} else if (binding_of(full_key, reg_tab) == (short) help)
			help_on_command(help_on_search(flags));
		else if (IS_ENTER_KEY(key)) {
			if (oldflags & SONCE)
				flags |= SONCE;
			break;
		} else if (binding_of(full_key, reg_tab)
				   == (short) backward_delete_character && (flags & INCR)) {
			run_topkey();
			window_handle = editwin;
			if (!(oldflags & FAILING) || size() <= lastok)
				s_code = FORIG;
			else		/* don't bother searching */
				flags |= FAILING;
		} else if (key == abort_key && (oldflags &
					((flags & INCR) ? FAILING : SONCE))) {
			if (flags & INCR)
				delete(goodchars, size());
		} else if (is_recall_key(flags)) {
			if ((recall_prior_response_options & 2) && !IS_ALT_KEY(key))
				recall_cnt = pick_prev_search_from_end(&flags,
					recall_cnt + ((is_recall_key(flags))));
			else
				flags = pick_prev_search(flags) & ~INCR;
			if (user_abort)
				break;
		} else if (binding_of(full_key, reg_tab) == (short) pull_word
				  || binding_of(full_key, reg_tab) == (short) pull_word_fwd) {
			flags &= ~UNKNOWN;
			window_handle = origwin;
			_highlight_control = 0;
			window_handle = editwin;
			pull_word();
		} else if (user_abort || (flags & INCR) && exit_incr())
			break;
		else {
			flags &= ~UNKNOWN;
			if ((binding_of(full_key, comp_tab) == (short) inp_grab_word
				 || binding_of(full_key, comp_tab) == (short) inp_grab_line)
				  && (IS_ALT_KEY(key) /*|| !(flags & INCR) */ )) {
				if (!(oldflags & FAILING)) {
					if (grab_search_word(flags, origwin,
						(binding_of(full_key, comp_tab) == (short) inp_grab_line))) {
						window_handle = origwin;	// Move origin
						remember_old_place(&orig, point);
						remember_old_place(&mstart, point);	// to start of word.
						window_handle = editwin;
						s_code = FORIG;
					} else
						s_code = (flags & REVERSE) ? FHERE : FMREGEX;
				}
			} else if (key == '\t')
				normal_character();
			else
				run_topkey();
			window_handle = editwin;
			if (flags & INCR)
				if (flags & (REGEX | WORD))
					s_code = FMREGEX;
				else if (oldflags & FAILING)
					flags |= FAILING;
				else
					s_code = FMADD;
		}
		_doing_input = (flags & INCR) ? DI_SEARCH : DI_SEARCH_NOINCR;
		if (!is_window(origwin) || !is_window(editwin)
		    || !is_window(pwin)) {	// user deleted our window
			user_abort = 1;
			goto end;
		}
		user_abort = 0;
		grab_expanding(0, size(), &trystr, 256);
		if (!s_code)
			continue;
		window_handle = origwin;
		if (s_code == FORIG || s_code == FMREGEX)
			to_old_place(&orig);
		else if (s_code >= FMATCH) {
			to_old_place(&mstart);
			if (flags & REVERSE)
				point++;
		}
		_highlight_control = 0;
search_again:
		switch (do_color_searching(flags, trystr)) {
			case DSBAD:
				flags = BAD | (flags & ~INCR);
				flags |= FAILING;
				to_old_place(&okpoint);
				if (in_macro(1))
					goto end;
				else if (!(oldflags & FAILING))
					maybe_ding(bell_on_search);
				break;
			case 0:
				if ((flags & SEARCH_CONTIN) && search_contin
					  && (*search_continuation)(cnt++, flags, trystr))
					goto search_again;
				flags |= FAILING;
				to_old_place(&okpoint);
				if (in_macro(1))
					goto end;
				else if (!(oldflags & FAILING))
					maybe_ding(bell_on_search);
				break;
			case 1:
				remember_old_place(&okpoint, point);
				remember_old_place(&mstart, matchstart);
				mark = matchstart;
				highlight_on();
				if ((flags & SEARCH_WRAPPED)
					  && past_old_place(&orig) == !(flags & REVERSE))
					flags |= SEARCH_OVERWRAPPED;
				window_handle = editwin;
				goodchars = lastok = size();
				break;
			case DSABORT:
				to_old_place(&okpoint);
				window_handle = editwin;
				if ((flags & INCR) && s_code == FMADD)
					delete(size() - 1, size());
				while (ungot_key == -1 && char_avail()
					   && getkey() == abort_key || !IS_TRUE_KEY(key))
					;
				if (key != abort_key && IS_TRUE_KEY(key))
					ungot_key = full_key;
				drop_pending_says();
				break;
		}
	}
	window_handle = origwin;
	if (key == abort_key || user_abort) {
		to_old_place(&orig);
		window_start = ostart;
	} else {
		window_handle = editwin;
		if (size()) {
			_last_was_regex = ((flags & REGEX) != 0);
			*(_last_was_regex ? &_default_regex_flags
					  : &_default_flags) = flags;
			grab_expanding(0, size(), &trystr, 256);
			if (_last_was_regex)
				grab_expanding(0, size(), &_default_regex_search, 256);
			else
				grab_expanding(0, size(), &_default_search, 256);
			save_search_recall(flags, trystr);
		} else
			flags |= SEARCH_EMPTY;
		if ((flags & INCR) && !IS_ENTER_KEY(key) && key != ESC)
			ungot_key = full_key;
	}
end:
	strncpy(tmpbuf, trystr, FNAMELEN - 1);
	free(trystr);
	if (is_window(origwin))
		window_handle = origwin;
	set_region_type(oreg_type);
	remove_window(editwin);
	remove_window(pwin);
	buf_delete(editbuf);
	buf_delete(pbuf);
	if (restrict_region)
		remove_region(restrict_region);
	if ((flags & FAILING) && in_macro(1))
		error("%s: not found", tmpbuf);
	check_abort();
	say("");
	restore_vars();
	maybe_refresh();	/* put cursor back in prev window */
	return flags;
}

do_searching(flags, str)	/* search according to flags */
char *str;	/* return 0=not found, 1=found, 2=aborted, 3=bad pattern */
{
	int ret, dir = (flags & REVERSE) ? -1 : 1;
	int null_detect = -1;

	if (flags & MODFOLD)
		save_var case_fold = (flags & FOLD);
	save_var abort_searching = ABORT_ERROR;
	for (;;) {
		if (flags & REGEX)
			ret = do_regex_searching(flags, str);
		else if (flags & WORD)
			ret = word_search(dir, str);
		else
			ret = search(dir, str);
		if (ret != 1 || !restrict_search_type || restricted_search_ok(flags))
			break;

		if (null_detect == point)	// Don't get stuck on nullable pat.
			if ((flags & REVERSE) ? (point-- <= narrow_start)
								  : (point++ >= size() - narrow_end))
				return 0;
		null_detect = point;
	}
	if (ret == ABORT_ERROR)
		ret = DSABORT;
	return ret;
}

restore_bookmark_session(b)
{
	int nums[10], i;
	char buf[FNAMELEN];

	if (grab_numbers(b, nums) < 2)
		return 1;
	_bookmark_slot = nums[0];
	_last_bookmark = nums[1];
	for (i = 0; i < NUM_BOOKMARKS; i++) {
		grab_line(b, buf);
		if (grab_numbers(b, nums) < 1)
			return 1;
		if (*buf && exist(buf)) {
			bufname = buf;
			_bookmark_spots[i] = alloc_spot(0);
			*_bookmark_spots[i] = nums[0];
			if (!size() && is_remote_file(filename))
				_bookmark_saved_pos[i] = nums[0];
		}
	}
	return 0;
}

do_set_variable(i)
{
	char msg[FNAMELEN], res[2048], *tname, *p, tmp[2048], prompt[FNAMELEN];
	int def;

	tname = variable_type_name(vartype(i));
	if (!(p = get_var_val(i, tmp, 0)))
		error("Can't interpret type of %s", name_name(i));
	sprintf(msg, "Set %s variable %s to", tname, name_name(i));
	if (run_by_mouse && (has_feature & FEAT_GUI_DIALOGS)) {
		get_strpopup(res, msg, p, name_name(i));
		def = 1;			// set both
	} else {
		build_prompt(prompt, msg, p, insert_default_response, 0);
		comp_read(res, prompt, (char *(*)()) 0, 0, p);
		def = ask_default("Set", name_name(i), 1);
	}
	if (def)
		set_var_val(i, res, 1);		/* set default */
	if (def >= 0)
		set_var_val(i, res, 0);		/* set current */
	do_show_variable(name_name(i), def);
	full_redraw = 1;	/* might have set a display-type variable */
}

command create_file_associations()
{
	int buf = zap("File Associations");
	char ext[FNAMELEN], cmd[FNAMELEN], dir[FNAMELEN], tmp[FNAMELEN];

	tiled_only();
	save_var window_bufnum;
	to_buffer_num(buf);
	stuff("# Associate Epsilon with the following extensions:\n\n");
	stuff("# Delete or add lines as required, then type Ctrl-X Ctrl-Z\n");
	stuff("# (the exit-level command) to carry out the change.  This\n");
	stuff("# command cannot remove existing Epsilon associations, only\n");
	stuff("# add new ones.  Changes you make here cannot be undone by\n");
	stuff("# uninstalling; make new associations using Windows Explorer\n");
	stuff("# or by running this command again.\n");
	stuff("\n.c\n.cpp\n.h\n.e\n.txt\n");
	conf_mode();
	save_var case_fold = 1;
	for (;;) {
		recursive_edit();
		if (strcmp(bufname, "File Associations"))
			error("Canceled.");
		point = 0;
		if (!re_search(1, "^%.(exe|bat|cmd)$"))
			break;
		say("Don't modify the association for .exe, .bat, or .cmd files.");
	}
	make_registration_entry("Epsilon", "Epsilon");
	get_executable_directory(dir);
	strcpy(tmp, argv[0]);
	absolute(tmp, dir);
	sprintf(cmd, "\"%s\" -add \"%%1\"", tmp);
	make_registration_entry("Epsilon\\shell\\open\\command", cmd);
	make_registration_entry("Epsilon\\shell\\open\\ddeexec", "\"%1\"");
	make_registration_entry("Epsilon\\shell\\open\\ddeexec\\application",
							"Epsilon");
	make_registration_entry("Epsilon\\shell\\open\\ddeexec\\topic",
							"Open");
	point = 0;
	while (point < size() - narrow_end) {
		grab_line(bufnum, ext);
		if (*ext == '.')
			make_registration_entry(ext, "Epsilon");
		else if (*ext && *ext != '#')
			say("Ignoring invalid extension \"%s\"", ext);
	}
	restore_vars();
	buf_delete(buf);
	say("File associations created.");
}

// In a list-all buffer, delete some variable references that aren't
// preservable customizations.
drop_boring_change_lines()
{
	point = 0;
	string_replace("%+(maxmem|minmem|m?colors|normal-cursor"
				   "|-display-characters|buf-accessed-clock|c-look-back"
				   "|-bookmark-.*|-num-kill|-cur-kill-buf|noinit-flag"
				   "|-errors-in-cmd-file|-prev-kbd-extended|eel-version"
				   "|when-loading) .*\n",
				   "", REGEX);
}

init_warning(msg, s)		/* display error and return */
char *msg, *s;
{
	_errors_in_cmd_file++;
	if (_abort_load_on_error == 1)
		say(msg, s);
	else if (_abort_load_on_error == 2) {
		buffer_printf(MSG_BUFFER, "%s %d: ", einit_file_name,
					  lines_between(0, point, 0) + 1);
		buffer_printf(MSG_BUFFER, msg, s);
		buffer_printf(MSG_BUFFER, "\n");
	} else {
		if (!nl_forward())
			insert('\n');
		stuff("# COMMAND FILE ERROR: ");
		bprintf(msg, s);
		stuff("\n");
	}
}

char *tag_c_ident_pattern();

// Tag the following structure/class/enum if we're defining a tag for it.
// Flavor: 0=struct/union, 1=class/interface, 2=enum, 3=namespace.
tag_c_struct(int flavor)
{
	char func[TAGLEN], fullname[TAGLEN], *s;
	int start, i;

	c_fwd_space();
	start = i = point;
	if (parse_string(1, "__declspec</word>")) {
		point = matchend;
		c_fwd_space();
		if (curchar() == '(')
			c_skip_level('(', ')');
		c_fwd_space();
	}

	if (!parse_string(1, tag_c_ident_pattern(), func)) {
		if (flavor != 1)
			*func = 0;	// all but class don't require a name
		else
			return 1;
	}
	if (flavor == 0 && (c_extra_keywords & CPP_KEYWORDS))
		flavor = 1;	// In C++, treat structs just like classes.
	c_fwd_space();	// skip compiler-specific junk betw "class" & name
	if (flavor == 1 && !is_java_buffer()
		  && parse_string(1, tag_c_ident_pattern())) {
		i = point;
		parse_string(1, tag_c_ident_pattern(), func);
		c_fwd_space();
	}
	while (strchr(func, '.')) {
		strcpy(fullname, func);
		s = strchr(fullname, '.');
		sprintf(s, "::%s", func + 1 + (s - fullname));
		strcpy(func, fullname);
	}
	if (strchr(";[,", curchar()) && i > start) {
		point = i;	// A plain definition.
		return 1;
	}
	if (curchar() != '{' && (flavor != 1 || !parse_other_class()))
		return 1;	// classes can have other stuff before {
	if (flavor == 1 || flavor == 3 || flavor == 0) {
		point++;	// skip over { so we do stuff within
		if (*c_tagging_class)
			strcat(c_tagging_class, "::");
		strcat(c_tagging_class, func);	// remember, now in class
	} else if (flavor == 2) {
		tag_c_enum();
		strcpy(fullname, func);
		if (*c_tagging_class)
			sprintf(fullname, "%s::%s", c_tagging_class, func);
		if (*func && (tag_which_items & 32))
			add_tag(fullname, start);
		return 1;
	}
	if (*func && (tag_which_items & 16))
		add_tag(c_tagging_class, start);
	return 1;
}

list_definitions_when_displaying(from, to)
{
	char tagname[FNAMELEN], msg[FNAMELEN];
	int dlg = window_handle;

	from = from;		// not used
	if (1) {
		set_wattrib(window_handle, PVERTBORDCOLOR, color_class text);
		set_wattrib(window_handle, PHORIZBORDCOLOR, color_class text);
//		set_wattrib(window_handle, PTITLECOLOR, color_class text);
		text_color = color_class text;
	}
	to = tag_line_to_position(lines_between(0, point, 0) + 1, tagname);
	save_var window_handle = _bufed_orig_window;
	if (!(list_definitions_live_update & 1))
		save_var point;
	point = to;
	if (*tagname) {
		strcpy(msg, tag_declarations ? " Select a declaration or definition - "
									 : " Select a definition - ");
		if (list_definitions_live_update & 2)
			sprintf(msg + strlen(msg), "Line %d ",
					buf_position_to_line_number(bufnum, to));
		sprintf(msg + strlen(msg), "%q", tagname);
		window_title(dlg, BTOP, TITLELEFT(2), msg);
		set_window_caption(dlg, msg);
	}
	if (list_definitions_live_update & 1)
		position_to_tag_line();
}

when_loading()
{
	drop_name("filter-str");
}
