/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2004 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 12.05 to 12.06.

#include "eel.h"

prepare_dlg_filter()
{
	char pats[FNAMELEN], *s;

	save_var bufnum = create("-dlg-filter");
	if (!size()) {
		if (s = lookpath("filter.txt"))
			file_read(s, FILETYPE_AUTO);
		else
			stuff("All Files *.*\n");
		point = 0;
		simple_re_replace(1, "^#.*", "");
		while (search(1, ";")) {
			replace(point - 1, ' ');
			grab(point, give_end_line(), pats);
			to_end_line();
			delete(point, point + 1);
			insert(0);
			stuff(pats);
			insert(0);
		}
		insert(0);
		point = 0;
		simple_re_replace(1, "\n", "");
		int i = new_variable("dlg-filter-str", NT_VAR, TYPE_CARRAY, size() + 1);
		s = varptr(i);
		grab(0, size(), varptr(i));
	}
}

custom_apply_immediately()
{
	int tmp = tmp_buf();
	on_exit buf_delete(tmp);
	buf_xfer(tmp, 0, size());
	save_var bufnum = tmp, show_status = 0;
	point = 0;
	do_drop_matching_lines(REGEX, "^%+(push-cmd|_region_file|_grep_files) ", 0);
	save_var _abort_load_on_error = 2;
	parse_cmds();
}

// Each key binding in src that doesn't match any in current buffer
// indicates the user unbound that key.  Insert a setting for that in
// current buffer.
insert_key_unbindings(int src)
{
	char keyspec[FNAMELEN], pat[FNAMELEN];
	int orig = point, obuf = bufnum;

	save_var point = 0, bufnum = src, point = 0;
	while (re_search(1, "^(~<^wspace>+ \".*\": )")) {
		grab(matchstart, matchend, keyspec);
		bufnum = obuf;
		point = orig;
		sprintf(pat, "^%q", keyspec);
		if (!re_search(1, pat))
			bprintf("%s0\n", keyspec);
		bufnum = src;
	}
}

// Position pos in buf is the start of a binding in .ecm format.  Fill
// start, end, text with its settings and return 1, or 0 if bad format.
int get_binding_range(int buf, int pos, int *start, int *end, char *text)
{
	save_var bufnum = buf, point = pos;
	if (!parse_string(1, "~[^ \t\n:]+ "))
		return 0;
	point = matchend;
	int *kcode = get_keycode();
	if (!kcode || kcode[0] != 2)
		return 0;
	*start = *end = kcode[1];
	if (parse_string(1, " \"")) {	// It uses a range spec.
		point++;
		kcode = get_keycode();
		if (!kcode || kcode[0] != 2)
			return 0;
		*end = kcode[1];
	}
	return parse_string(1, ": .*", text) != 0;
}

// Search buffer buf for an .ecm-format binding line that starts with
// init, skipping over those binding only keys < start.  Locate the one
// with the lowest-numbered range of keys.  Fill in its brange, erange,
// and def, and return 1.  If none, return 0.
find_lowest_binding_range(int buf, char *init, int start,
						  int *brange, int *erange, char *def)
{
	int bpos, epos, found = 0;
	save_var bufnum = buf, point = 0;
	char pat[FNAMELEN], text[FNAMELEN];
	sprintf(pat, "^%q", init);
	while (re_search(1, pat)) {
		if (get_binding_range(bufnum, give_begin_line(), &bpos, &epos, text)) {
			if (epos > start && (!found || bpos < *brange)) {
				found = 1;
				*brange = bpos;
				*erange = epos;
				strcpy(def, text);
			}
		}
	}
	return found;
}

// Insert in buffer buf a new binding line, binding keys from start to
// end, with text before and after.  If s1 > e1, ignore.
insert_new_binding_range(int buf, int start, int end,
						 char *before, char *after)
{
	int b[10];
	if (start > end)
		return;
	save_var bufnum = buf;
	bprintf("%s", before);
	b[0] = 2;
	b[1] = start;
	stuff_macro(b, 1);
	if (end > start) {
		bprintf(" ");
		b[1] = end;
		stuff_macro(b, 1);
	}
	bprintf("%s\n", after);
}

// Look through all multiple-key bindings in .ecm-format buffer b1.  For
// each, see if b2 has any corresponding binding lines that intersect
// b1's range but aren't identical.  If so, split up b1's range into a
// series of ranges, so each range is completely inside or outside b2's
// ranges.  This rewriting lets us compare binding changes line-by-line.
scan_binding_ranges(int b1, int b2)
{
	char init[FNAMELEN], def1[FNAMELEN], def2[FNAMELEN];
	int s1, e1, s2, e2;
	save_var bufnum = b1;
	point = 0;
	while (re_search(1, "~[^ \t\n:]+ !.*\" \".*\": ")) {
		grab(matchstart, matchend, init);
		int *kcode = get_keycode();
		if (!kcode || kcode[0] != 2)
			continue;
		s1 = kcode[1];
		if (!parse_string(1, " \""))
			continue;
		point++;
		kcode = get_keycode();
		if (!kcode || kcode[0] != 2)
			continue;
		e1 = kcode[1];
		if (!parse_string(1, ": .*", def1))
			continue;
		if (!find_lowest_binding_range(b2, init, s1, &s2, &e2, def2))
			continue;
		if ((s1 != s2 || e1 != e2) && s2 <= e1) {
			delete(give_begin_line(), give_end_line() + 1);
			int pos = point;
			insert_new_binding_range(b1, s1, s2 - 1, init, def1);
			insert_new_binding_range(b1, MAX(s1, s2), MIN(e1, e2), init, def1);
			insert_new_binding_range(b1, e2 + 1, e1, init, def1);
			point = pos;
		}
	}
}

// Return a buffer listing the differences in settings between orig and
// cur, two buffers with lists of settings.  Put the original settings
// corresponding to the modified settings into a buffer
// "original-settings".
int make_difference_list(int orig, int cur)
{
	noteput("Comparing...");
	save_var show_status = 0;
	scan_binding_ranges(orig, cur);
	scan_binding_ranges(cur, orig);
	save_var bufnum = orig;
	save_var case_fold = 1;
	do_sort_region(0, size(), 0, 0);
	save_var bufnum = cur;
	save_var case_fold = 1;
	do_sort_region(0, size(), 0, 0);
	int out = tmp_buf();
	do_compare_sorted(orig, cur, "original-settings",
					  bufnum_to_name(out), NULL);
	bufnum = out;
	point = 0;
	save_var case_fold = 1;
	custom_apply_immediately();
	// Drop vars starting with -, but not color scheme ones.
	do_drop_matching_lines(REGEX, "^%+-[^o][^u][^r][^-]", 1);
	drop_boring_change_lines();
	custom_fix_mode_vars();
	insert_key_unbindings(name_to_bufnum("original-settings"));
	do_drop_matching_lines(REGEX, "^%+(push-cmd|echo-line|permanent-menu"
						   "|avoid-(top|bottom)-lines|bufed-width"
						   "|mode-start|mode-end) ", 1);
	do_drop_matching_lines(REGEX, "^&.* color class for after-exiting: .*$", 1);
	do_drop_matching_lines(REGEX, "^&.* color class for .*: "
						   "%[0x4000000 on 0x0%]$", 1);
	return out;		// Don't count vars saved in session file, others.
}

command list_customizations()
{
	char cmd[FNAMELEN], *args, time[FNAMELEN];
	char ext_file[FNAMELEN], settings_file[FNAMELEN], *s;
	int tmp, cur, ext, i, max, err;
	int cust = prep_customizations(), clist;

	iter = 1;
	tiled_only();
	get_customization_directory(ext_file);
	strcat(ext_file, "extensions-list");
	get_customization_directory(settings_file);
	strcat(settings_file, "settings-list");
	delete_file(settings_file);
	i = name_match("-loaded-eel-file-", STARTMATCH);
	save_var bufnum = ext = zap("-list-customizations");
	give_binding(cmd, "exit-level");
	bprintf("Epsilon will include the following EEL extension files as part\n"
			"of your customizations.  Edit this list if needed, then\n"
			"press %s to begin.\n\n%p\n", cmd, color_class viewed_text);
	point--;
	for (; i; i = name_match("", 0))	// List new extensions by searching
		if (name_type(i) == NT_VAR) {	// for id variables.
			s = get_str_var(i);
			bprintf("%s%s\n", s, (check_file(s) == CHECK_FILE)
					? "" : " (not found)");
		}
	to_buffer_num(ext);
	point = 0;
	do {
		recursive_edit();
		if (!buf_exist(ext))
			return;
		point = 0;
		point += parse_string(1, "(.|\n.)*\n*");
		i = test_valid_files(point, size());
		if (i == 2) {
			buf_zap(ext);
			return;
		}
	} while (i);
	point = 0;		// Drop the explanatory text.
	delete(0, parse_string(1, "(.|\n.)*\n*"));
	simple_re_replace(1, " *%(not found%)", "");
	if (err = file_write(ext_file, FILETYPE_BINARY)) {
		file_error(err, ext_file, "write error");
		return;
	}
	save_var reject_client_connections = 1;
	strcpy(cmd, argv[0]);
	convert_to_8_3_filename(cmd);
	sayput("Running %s to create original settings list...", cmd);
	args = " -s original -p0 -noinit -r build-settings-file";
	err = shell_with_args(cmd, args);
	if (err)
		error("Error running %s %s", cmd, args);
	sayput("Reading original settings list...");
	bufnum = tmp = tmp_buf();
	if (err_file_read(settings_file, FILETYPE_BINARY))
		return;
	bufnum = cur = tmp_buf();
	max = final_index();
	for (i = 1; i <= max; i++)
		list_one(bufnum, i, 3);
	use_default = 0;
	bufnum = cur;
	clist = make_difference_list(tmp, cur);
	restore_vars();
	buf_delete(tmp);
	buf_delete(cur);
	to_buffer_num(cust);
	current_time_and_date(time);
	save_spot point;
	i = point;
	bprintf("\n# List-customizations created these settings %s.\n", time);
	fix_extension_list(ext);
	grab_buffer(ext);
	grab_buffer(clist);
	buf_delete(clist);
	delete_file(settings_file);
	delete_file(ext_file);
	bprintf("\n");
	point = i;
	if (!has_arg)
		comment_old_customizations(point);
	if (!list_undefined_customizations())
		say("");
}

compile_for_import_customizations(char *src, char *eel, char *dest, int list)
{
	char buf[2000], *s, cmd[FNAMELEN], args[FNAMELEN], rdest[FNAMELEN];

	for (;;) {
		strcpy(rdest, dest);
		convert_to_8_3_filename(rdest);
		strcpy(buf, src);
		if (!*buf)
			return;
		if (s = strchr(buf, '\n'))
			*s = 0;
		if (s = strstr(buf, " (not found)"))
			*s = 0;
		convert_to_8_3_filename(buf);
		if (check_file(buf) != CHECK_FILE)
			error("Can't find %s", buf);
		strcpy(cmd, eel);
		convert_to_8_3_filename(cmd);
		sprintf(rdest + strlen(rdest), "%c%s", path_sep, get_tail(buf, 0));
		strcpy(get_extension(rdest), byte_extension);
		delete_file(rdest);
		sprintf(args, " -o %s %s", rdest, buf);
		show_text(-2, 1, "Run %s %s", cmd, args);
		if (shell_with_args(cmd, args) || check_file(rdest) != CHECK_FILE)
			error("Error compiling %s with cmd line %s %s", buf, cmd, args);
		if (list)
			buf_printf(list, "%s\n", rdest);
		if (s = strchr(src, '\n'))
			src = s + 1;
		else
			break;
	}
}

is_batch_keyword(int from)
{
	char buf[500], *s;

	save_var case_fold = 1;
	if (point - from > sizeof(buf) / sizeof(char) - 10)
		save_var point = from + sizeof(buf) / sizeof(char) - 10;
	buf[0] = '|';		// get identifier, between | chars
	grab(from, point, buf + 1);
	if ((s = strchr(buf + 1, '.')) && matches_at(from, -1, "\n[ \t]*"))
		*s = 0;
	strcat(buf, "|");
	if (strstr("|rem|", buf) && !matches_at(from, -1, "[^ \n\t@][@ \t]*"))
		return 2;
	if (strstr("|assoc|at|attrib|break|cacls|call|cd|chcp|chdir"
			   "|chkdsk|chkntfs|cls|cmd|cmdextversion|color|comp"
			   "|compact|convert|copy|date|defined|del|dir|diskcomp"
			   "|diskcopy|do|doskey|echo|else|endlocal|equ|erase"
			   "|errorlevel|exist|exit|fc|find|findstr|for|format"
			   "|ftype|geq|goto|graftabl|gtr|help|if|in|label|leq"
			   "|lss|md|mkdir|mode|more|move|neq|not|path|pause"
			   "|popd|print|prompt|pushd|rd|recover|ren|rename"
			   "|replace|rmdir|set|setlocal|shift|sort|start|subst"
			   "|time|title|tree|type|ver|verify|vol|xcopy|cmd"
			   "|command|", buf))
			return 1;
	return 0;
}

suffix_btm()
{
	batch_mode();
}

drop_boring_change_lines()
{
	point = 0;
	string_replace("%+(maxmem|minmem|m?colors|normal-cursor"
				   "|-display-characters|buf-accessed-clock|c-look-back"
				   "|-bookmark-.*|-num-kill|-cur-kill-buf|noinit-flag"
				   "|-errors-in-cmd-file|-prev-kbd-extended|eel-version"
				   ") .*\n", "", REGEX);
	string_replace("when-loading: .*\n", "", REGEX);
}
