/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2004 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 12.06 to 12.07.

#include "eel.h"
#include "c.h"
#include "dired.h"
#include "colcode.h"
#include "bufed.h"
#include "epsnet.h"
#include "grep.h"
#include "html.h"
#include "misclang.h"
#include "perl.h"
#include "python.h"
#include "vbasic.h"

when_loading()
{
	drop_name("default-oem-word");
}

color_asm_range(from, to) // recolor just this section
{			// last colored region may go past to
	int t = -1, talk, s;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	while (from > 0 && get_tagged_region("color-as-unit", from, &s) != -1
		   && s <= from)
		from = s - 1;
	c_init_color(from, to);
	point = from;
	to_begin_line();
	talk = (to - from > 2000);	// show status during long delays
	while (point < to) {
		if (!re_search(1, "[0-9A-Za-z_.@#%?$]+|[\"';]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(point - 1)) {		// check last char
			case ';':			// found comment
				nl_forward();
				set_character_color(t, point, color_class c_comment);
				break;
			case '"':		// found a string literal
				point = t;
				re_search(1, "\"([^\"\\\n]|\\(.|\n))*[\"\n]");
				set_character_color(t, point, color_class c_string);
				if (get_character_color(point, (int *) 0, &s) == 
					color_class c_string && s > to)  // fix up after
					c_init_color(point, to = s); // quoted "'s
				break;
			case '\'':		// found a char const
				point = t;
				re_search(1, "\'([^\'\\\n]|\\(.|\n))*[\'\n]");
				set_character_color(t, point, color_class c_charconst);
				break;
			default:		// found identifier, kywd, or number
				set_character_color(t, point, asm_keyword_color(t));
				if (asm_title_line()) {
					t = point;
					point++;
					nl_forward();
					set_character_color(t, point, color_class c_string);
				}
				asm_comment_line();
				break;
		}
		if (talk)
			note("Coloring ASM program: %d%% complete...",
				 muldiv(point - from, 100, to - from));
	}
	c_init_color(to, t);
	if (talk)
		note("");
	return point;
}

#define FIRSTGROUP	(*filename && !buffer_not_saveable)
#define GOODNAME	(*bufname != '-' && strcmp(bufname, inbuffer))
#define SECONDGROUP	(GOODNAME || bufnum == old)

make_buffer_list(flags, inbuffer)
char *inbuffer;		/* put buffer list in specified buffer */
{
	int old = bufnum, i, opos;

	if (exist(inbuffer)) {
		bufname = inbuffer;
		set_read_only(0);
		bufnum = old;
	}
	zap(inbuffer);
	i = buf_list(0, 2);	/* start at end of list (most recent) */
	do {
		bufnum = i;
		if (FIRSTGROUP && !(flags & BUFED_ONLY_UNSAVED)
			|| is_unsaved_buffer())
			list_line(inbuffer);
	} while (i = buf_list(-1, 1));
	if (bufed_grouping)
		opos = sort_buffer_list(0, inbuffer);
	i = buf_list(0, 2);
	do {
		bufnum = i;
		if (!FIRSTGROUP && SECONDGROUP
			&& (!(flags & BUFED_ONLY_UNSAVED) || is_unsaved_buffer()))
			list_line(inbuffer);
	} while (i = buf_list(-1, 1));
	if (bufed_grouping)
		opos = sort_buffer_list(opos, inbuffer);
	if (flags & BUFED_ALL) {
		i = buf_list(0, 2);
		do {
			bufnum = i;
			if (!FIRSTGROUP && !SECONDGROUP
				&& (!(flags & BUFED_ONLY_UNSAVED) || is_unsaved_buffer()))
				list_line(inbuffer);
		} while (i = buf_list(-1, 1));
		if (bufed_grouping)
			opos = sort_buffer_list(opos, inbuffer);
	}
	if (!bufed_grouping)
		sort_buffer_list(0, inbuffer);
	bufname = inbuffer;
	point = 0;
	delete_final_newlines();
	set_read_only(1);
	case_fold = (preserve_filename_case & 8) != 0;
	bufnum = old;
}

do_make_bname(char *fname, char *bname, char *dup_ok)
{
	int uniq = 1, ok;
	char *to, *oldbuf;

	strcpy(bname, get_tail(fname, 0));
	if (!(preserve_filename_case & 2)
		  && (filename_rules(fname) & FSYS_CASE_MASK) != FSYS_CASE_SENSITIVE)
		for (to = bname; *to; to++)
			*to = tolower(*to);
	for (to = bname; *to; to++)
		if (*to < ' ')
			*to = '!';
	to = bname + strlen(bname);
	while (exist(bname) && (!dup_ok || strcmp(bname, dup_ok))) {
		oldbuf = bufname;
		bufname = bname;
		ok = (size() == 0);
		bufname = oldbuf;
		if (ok) return;
		sprintf(to, "<%d>", uniq++);
	}
}

c_compute_indent()
{
	int ind = 0;		/* indentation to use */
	int orig = point;
	int first_char = curchar(); /* cur line's first char */
	int pstart;		/* start position of prev line */
	int pafter;		/* after label on prev line */
	int pend;		/* end position of prev line */
	int prev_end = ';';	/* last char on previous line */
	int norm_narrow;	/* narrowing setting for most parsing */

	save_var point;
	if (setjmp(&c_indent_giveup))	/* if we hit end & gave up, */
		return ind;		/* use best guess so far */
	if (in_perl_buffer)
		return perl_compute_indent();
	norm_narrow = fix_narrowing(point - c_look_back);
	save_var narrow_start = norm_narrow;
	if (!reindent_c_comments && parse_string(1, "[ \t\f]*/<*>"))
		return get_indentation(point);
	if (!reindent_one_line_c_comments && parse_string(1, "[ \t\f]*//"))
		return get_indentation(point);
	if (!reindent_c_preprocessor_lines && parse_string(1, "[ \t\f]*#"))
		return get_indentation(point);
	if (first_char == '#' || first_char == '\f')
		return 0;
	else if (first_char == '}')
		return cind_close();

	to_begin_line();
	if (character(point - 2) == '\\')
		if (!indent_preprocessor_contin)
			return indent_after_backslash();
		else if (!prior_contin_code(orig))
			return indent_preprocessor_contin;
		else {	// Note we're indenting preprocessor continuation line now.
			save_var indent_preprocessor_contin = 255;
			save_var narrow_start = point - parse_string(-1, "(\n.*\\)*\n");
		}
	if (parse_string(1, "[ \t]*\n") && after_many_comments()) {
		re_search(-1, "[^ \t\n]");	/* find previous non-blank line */
		return get_indentation(point);
	}
	cind_space();			/* skip whitespace and comments */
	if (character(point - 1) == ':')
		move_before_labels();
	prev_end = character(point - 1); /* get last relevant char */
	pend = point;

	to_statement_start();		/* if contin line, back up */
	pstart = point;			/* now at start of biggest statement */
	past_labels();			/* containing original point */
	pafter = point;			/* just after label */
	ind = current_column();
	narrow_start = norm_narrow;	/* ok to look far back again */

	if (is_case(orig)) {
		ind = indent_case(orig);
	} else if (is_dispatch_table(point)) {
		// Found MFC dispatch table, leave alone.
	} else if (have_label(orig)) {
		char type[FNAMELEN];		// Could it be a bitfield, not a label?
		find_c_func_info(type, NULL, NULL, 0);
		if (!strcmp(type, "struct") || !strcmp(type, "union"))	// Yes, a bitfield.
			ind += get_c_indent();
		else					// It's a label.
			ind = c_label_indent;
	} else if (prev_end == ':' && is_label(pstart)) {
		ind = get_column(pstart);
		if (!(Closeback && just_open(orig)))
			ind += get_c_indent();
		if (just_open(orig))
			ind += c_brace_offset;
	} else if (is_else(orig)) {
		point = orig;
		ind = indent_else(ind);
	} else if (prev_end == ';') {
		if (just_open(orig) && func_start_block(pstart))
			ind = c_top_braces;
		else
			ind = indent_after_statement(orig);
	} else if (prev_end == '}') {
		point = pend;
		ind = indent_after_block(orig);
	} else if (prev_end == '{') {
		if (!c_indent_after_extern_c
			  && parse_string(1, "[ \t]*extern[ \t]*\"C\"[ \t\n]*{"))
			return ind;
		if (!c_indent_after_namespace
			  && parse_string(1, "[ \t]*namespace</word>.*{"))
			return ind;
		if (ind <= c_top_braces	// special case if no previous indent
		    || get_column(pend - 1) <= c_top_braces) {
			ind += top_indent(prev_end) ? get_c_indent() : c_top_braces;
			if (just_open(orig))
				ind += c_brace_offset;
			if (ind < c_top_braces)
				ind = c_top_braces;
		} else if (have_struct_init(pend - 1)) {
			ind = indent_struct_init(pend - 1) + get_c_indent();
		} else if (Closeback || curchar() != '{')
			ind += get_c_indent();
		if (just_open(pafter))
			ind -= c_brace_offset;
	} else if (just_open(orig) || matches_at(orig, 1, "[ \t\f]*{[ \t\f]*}")) {
		if (have_struct_init(orig))
			return indent_struct_init(orig);
		else if (have_other_init(orig))
			return ind;
		else
			return indent_just_open(ind);
	} else if (in_c_initializer(orig)) {
		return ind;
	} else if (is_contin_statement(pend, 1, orig)) {
		return indent_c_contin(ind, orig, pend);
	} else if (!index(":);{}", c_line_before()) && !c_statement_start()
				&& !(Closeback && just_open(point)))
		return ind;
	else
		ind += get_c_indent();
	return ind;
}

is_dispatch_table(int p)
{
	save_var point = p, case_fold = 0;
	if (!parse_string(1, "ON_[A-Z0-9_]+%("))
		return 0;
	return re_search(-1, "(BEGIN_MESSAGE_MAP|AFX_MSG_MAP)%(") != 0;
}

early_restoring()
{
	char fname[FNAMELEN];
	int i;

	after_loading(); /* call early, so we can display errors, etc. */
	putenv("EPSRUNS=Y");	/* for catching nested Epsilons */
	if (opsys == OS_DOS)
		putenv("PKNOFASTCHAR=Y"); //make PKZIP, UNZIP output capturable
	root_keys = reg_tab;
		/* now set buffer-specific variables */
	word_pattern.default = default_word;
	_display_class.default = _std_disp_class;
	_echo_display_class = _echo_disp_class;
	i = *get_color_scheme_variable();
	if (name_type(i) == NT_COLSCHEME)
		selected_color_scheme = i;
	else
		get_standard_color_scheme();
	use_default = 1;		/* default is fundamental mode */
	mode_default_settings();
	use_default = 0;
	apply_defaults();	/* apply defaults to init buffer & window */

	force_save_as = FSA_NEWFILE;	// make startup buffer prompt
	strcpy(fname, "startup.txt");	// for filename
	absolute(fname);
	filename = fname;
	strcpy(_docname, "edoc");
}

recolor_buffer_range(start, end)
{
	if (want_code_coloring && recolor_range) {
		save_var point;
		shrink_colored_area();
		if (!(coloring_flags & COLOR_RETAIN_NARROWING))
			save_var narrow_start = 0, narrow_end = 0;
		color_for_area(start, end);
	}
}

color_for_narrowed_area(from, to)
{
	int i, j = 0, give_up = 0, last_j;

	save_var point, abort_searching = 0;
	init_color_code();
	to = MIN(to, size() - narrow_end);
	for (;;) {
		last_j = j;
		if (get_tagged_region("needs-color", from, &i, &j)) {
			if (i < from) {
				point = from;		// From isn't colored, so start before it.
				color_go_back_lines();	// Color a few extra lines before
										// to speed up scrolling.
				if (point > i) // Next, find starting place known to be on 
					i = recolor_from_here(i); // the edge of a colored region.
			}				// (Or use known safe place if it's closer.)
		} else {
			if (j >= to) {		// From is colored, see if good region extends
				return;			// far enough.  If not, get following bad reg.
			} else if (!get_tagged_region("needs-color", j, &i, &j)) {
				int a, b, res;

				say("Coloring internal error 1");
				buffer_printf("debug-color", "Coloring internal error 1:\n");
				buffer_printf("debug-color", "bad color 0 from %d to %d in %s, size %d\n",
							  i, j, bufname, size());
				res = get_tagged_region("needs-color", from, &a, &b);
				buffer_printf("debug-color", "color %d at %d (from %d to %d) in %s, size %d\n",
							  res, from, a, b, bufname, size());
				res = get_tagged_region("needs-color", j - 1, &a, &b);
				buffer_printf("debug-color", "color %d at %d (from %d to %d) in %s, size %d\n",
							  res, j - 1, a, b, bufname, size());
				return;
			}
		}
		j = MIN(j, to);		// Color only some of it.
		if (i >= j)
			return;
		j = recolor_range(i, j);
		set_tagged_region("needs-color", i, j, 0);
		if (j <= last_j && give_up++ > 100) {
			say("Coloring internal error 2");
			when_displaying = 0;
			return;
		}
	}
}

color_c_range(from, to) // recolor just this section
{			// last colored region may go past to
	int t = -1, talk, s, talk_at = 0;
	char pat[200];

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	c_init_color(from, to);
	point = from;
	talk = (to - from > 20000);	// show status during long delays
	save_var case_fold = 0;
	strcpy(pat, "/<*>|//|^[ \t]*#|[\"']");
	if (!minimal_coloring)
		strcat(pat, "|[A-Za-z_][A-Za-z0-9_]*"
		       "|-?%.?[0-9]([A-Za-z0-9._]|[Ee]-)*");
	while (point < to) {
		if (!re_search(1, pat)) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(point - 1)) {		// check last char
		case '/':			// found // one-line comment
			while (nl_forward() && character(point - 2) == '\\')
				;		// Continued // comment.
			set_character_color(t, point, color_class c_comment);
			break;
		case '*':			// found /* starting comment
			search(1, "*/");
			set_character_color(t, point, color_class c_comment);
			if (get_character_color(point, (int *) 0, &s)
				  == color_class c_comment)
				to = MAX(s, to);	// Comment extended further before; rescan it.
			break;
		case '#':			// found preproc line
			c_preproc_color();	// move past correct part
			set_character_color(t, point,
					    color_class c_preprocessor);
			break;
		case '"':		// found a string literal
			point = t;
			if ((c_extra_keywords & CSHARP_KEYWORDS)
				  && character(t - 1) == '@')
				t--, re_search(1, "\"([^\"]|\"\")*\"![^\"]");
			else
				re_search(1, "\"([^\"\\\n]|\\(.|\n))*[\"\n]");
			set_character_color(t, point, color_class c_string);
			if (get_character_color(point, (int *) 0, &s) == 
			    color_class c_string && s > to)  // fix up after
				c_init_color(point, to = s); // quoted "'s
			break;
		case '\'':		// found a char const
			point = t;
			re_search(1, "\'([^\'\\\n]|\\(.|\n))*[\'\n]");
			set_character_color(t, point, color_class c_charconst);
			break;
		default:		// found identifier, kywd, or number
			set_character_color(t, point, c_keyword_color(t));
			if ((c_extra_keywords & IDL_KEYWORDS)	// Special rule for
				&& (point - t == 4)			// uuid's in IDL files.
				&& parse_string(-1, "uuid")
				&& parse_string(1, "%([0-9A-Fa-f]"))
				maybe_color_uuid();
			break;
		}
		if (talk && point > talk_at + 2000) {
			note("Coloring C program: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	c_init_color(to, t);
	if (talk)
		note("");
	return point;
}

int buf_lister();

get_buf(res, pr, def)
	char *res, *pr, *def;
{
	char *s;

	save_var completion_lister = buf_lister;
	save_var completion_column_marker = '\t';
	save_var resize_menu_list = 1;
	prompt_comp_read(res, pr, b_match, CAUTIOUS
					 | ((preserve_filename_case & 4) ? FM_FOLD : 0), def);
	s = strchr(res, '\t');
	if (s)
		*s = 0;
}

dired_delete_hierarchy(char *dir)
{
	char pat[FNAMELEN], tail[FNAMELEN], *s;
	int cnt = -1, last;

	strcpy(pat, dir);
	add_final_slash(pat);
	save_var file_pattern_rules |= FPAT_SKIP_DIR_SYMLINKS
								   | FPAT_SKIP_RECUR_SYMLINKS;
	save_var file_pattern_wildcards = 0;
	strcat(pat, "**/*");
	do {	// Delete everything that matches our pattern.
		s = do_file_match(pat, STARTMATCH | EXACTONLY);
		last = cnt;
		for (cnt = 0; s; s = do_file_match(pat, EXACTONLY), cnt++) {
			check_abort();
			strcpy(tail, get_tail(s, 0));
			drop_final_slash(tail);
			if (!strcmp(tail, ".") || !strcmp(tail, ".."))
				continue;
			if (get_file_read_only(s))
				set_file_read_only(s, 0);
			if (cnt % 10 == 0)
				noteput("Deleting %f ...", s);
			if (is_directory(s))
				rmdir(s);
			else
				delete_file(s);
		}	// Repeat until the number of matches stops decreasing.
	} while (cnt < last || last == -1);
	return rmdir(dir);
}

dired_ftp_activity(int activity, int buf, int from, int to)
{
	to = to;
	save_var bufnum = buf, iter, has_arg;
	switch (activity) {
		case NET_DONE:
			detect_dired_format();
			if (dired_format == DF_VMS)
				dired_vms_fix_lines();
			else if (dired_format == DF_UNIX)
				dired_unix_fix_lines();
			if (!modified && (point == dired_orig_position || dired_orig_position == -1))
				sort_for_dired();
			else
				say("Sorting omitted, press S to sort.");
			dired_strip_dotfiles();
			break;
		case NET_RECV:
			if (dired_orig_position == -1)
				dired_orig_position = point = from;
			save_spot point = from;
			save_var modified;
			while (re_search(1, "^.")) {
				point--;
				stuff("  ");
				point++;
			}
			if (!want_code_coloring)
				detect_dired_format();
			break;
		case NET_LOG_WRITE:
			break;
	}
}

int scp_client_handle_list(struct scp_job *p)
{
	int i, nextcmd = 0;
	char msg[FNAMELEN];

	while (nl_forward()) {
		i = point;
		point--;
		to_begin_line();
		if (character(point) == 'd' && character(point + 1) == ' ') {
			if (buf_exist(p->destbuf)) {	// Got some data, copy where it belongs.
				if (*p->file)
					buf_printf(p->destbuf, "%s", p->file);
				buf_xfer(p->destbuf, point + 2, i);
			}
		} else if (character(point) == 's' && character(point + 1) == ' ') {
			if (buf_exist(p->destbuf) && (!p->cmd || !*p->cmd))
				set_buf_mode_message(p->destbuf, NULL);
			p->busy = 0;			// Got a status msg.
			nextcmd = 1;			// Send a pending command if any.
		} else if (parse_string(1, "Epsilon-xfer-helper version ")) {
			if (character(matchend) < '1')
				say("Bad epsilon-xfer-helper version: %c", character(matchend));
			p->busy = 0;
			nextcmd = 1;
		} else if (!parse_string(1, "(## |Exited)")) {
			grab(point, MIN(give_end_line(),
							point + sizeof(msg) / sizeof(*msg) - 1), msg);
			if (*msg)		// Any other line is an error msg.
				say("%s", msg);
		}
		point = i;
	}
	return nextcmd;
}

int scp_sftp_handle_list(struct scp_job *p)
{
	int i, isdata;
	char msg[FNAMELEN];

	save_var case_fold = 1;
	while (nl_forward()) {
		i = point;
		point--;
		to_begin_line();
		isdata = 0;
		if (parse_string(1, "[-dcblsp][-r][-w][-sxt][-r][-w][-sxt]"
						 "[-r][-w][-sxt] .*" UNIX_DATE))
			isdata = 1;
		if ((p->flags & SCP_CHECK_IS_DIR) && parse_string(1, ".* not a directory"))
			isdata = 1;
		if (parse_string(1, ".* No such file or directory")
			  && p->cmd && !strncmp(p->cmd, "ls ", 3)) {
			if (buf_exist(p->destbuf))	// Error part way through, so don't
				set_buf_mode_message(p->destbuf, NULL);	// copy any more.
			p->flags |= SCP_CMD_ERROR;
		}
		if (isdata) { // Got some data, copy where it belongs.
			if (buf_exist(p->destbuf) && !(p->flags & SCP_CMD_ERROR)) {
				if (*p->file && !(p->flags & SCP_CHECK_IS_DIR))
					buf_printf(p->destbuf, "%s", p->file);
				buf_xfer(p->destbuf, point, i);
			}
		} else if (parse_string(1, "Remote (working )?directory(:| is) ")) {
			grab(matchend, give_end_line(), msg);
			if (p->lastdir)
				free(p->lastdir);
			p->lastdir = strsave(msg);
			if (!p->homedir)
				p->homedir = strsave(msg);
		} else if ((p->flags & SCP_IS_XFER)	// Ignore this msg.
				   && parse_string(1, "((Fetching|Uploading) .* to |"
								   "(remote|local):.* => (remote|local):|"
								   ".*--:-- ETA)")) {
		} else if (!parse_string(1, "(Exited|Connecting to "
								 "|Connection closed|Listing directory "
								 "|Using username \")")) {
			grab(point, MIN(give_end_line(),
							point + sizeof(msg) / sizeof(*msg) - 1), msg);
			if (*msg)		// Any other line is an error msg.
				say("%s", msg);
			if ((p->flags & SCP_IS_XFER)
				  && parse_string(1, ".*: (Permission denied|No such file"
								  "|Device out of space|Read-only file system)"))
				p->flags |= SCP_CMD_ERROR;
		}
		point = i;
	}
	if (parse_string(-1, "p?sftp%> *")) {
		if (!p->cmd || !*p->cmd) {
			if (buf_exist(p->destbuf))
				set_buf_mode_message(p->destbuf, NULL);
			if (p->flags & SCP_IS_XFER)
				scp_handle_done(p);
		}
		p->busy = 0;			// Got a prompt for another cmd.
		return 1;			// Send a pending command if any.
	}
	return 0;
}

struct scp_job *prep_scp_buffer();
char *scp_variable();

net_send_file(char *from, char *to)
{
	struct url_parts parts;
	struct scp_job *p;

	wait_for_remote_file(bufnum);
	prepare_url_operation(to, FTP_SEND, &parts);
	if (!*parts.fname)
		strcpy(parts.fname, ".");
	p = prep_scp_buffer(&parts, FTP_SEND);
	strcpy(p->file, from);
	do_scp_xfer(&parts, p, scp_variable(&parts, "write-file"), 0, 0);
}

command refresh_files()
{
	TIMER talk_now;
	int i, cnt = 0, res;

	time_begin(&talk_now, 100);		// Wait a second before chatter.
	save_var readonly_warning;
	save_var bufnum, window_bufnum;
	i = buf_list(0, 0);
	do {
		bufnum = i;
		readonly_warning |= ROWARN_REFRESH;
		save_var buf_accessed;	// Don't update due to checking.
		res = check_dates(0);
		if (res)
			cnt++;
		if (res == 2) {
			readonly_warning &= ~ROWARN_REFRESH;
			to_buffer_num(i);
			check_dates(0);
			time_begin(&talk_now, 100);
		}
		check_abort();
		if (*filename && time_done(&talk_now)) {
			note("%r: refreshing.", filename);
			time_begin(&talk_now, 100);
		}
	} while (i = buf_list(1, 1));
	if (!cnt)
		say("No files have been modified on disk since they were read.");
	else if (time_done(&talk_now))
		say("");
}

#define URL_PATTERN	"|((ftp|http|scp|telnet|ssh):[/\\][/\\][^\n\t >]+)"


default_find_linked_file(char *file)
{
	char *pat;

	save_var case_fold = 1;
	if (opsys == OS_UNIX)
		pat = "[+-9A-z~]+[/.][+-9A-z]*[--9+A-z]" URL_PATTERN;
	else
		pat = "([a-zA-Z]:)?[+-9A-z~]+<.>[+-9A-z]*[--9+A-z]" URL_PATTERN;
	point -= parse_string(-1, "[^ \t\n!-*]*");
	if (!re_search(1, pat)) {
		to_begin_line();
		if (!re_search(1, pat))
			error("Can't find any file name on this line.");
	}
	if (strchr("\"'`<", character(matchstart)))
		matchstart++;
	if (strchr("\"'`.,>", character(matchend - 1)))
		matchend--;
	grab(matchstart, matchend, file);
}

guess_mode_without_extension(char *res, char *pat)
{
	if (!strfcmp(get_tail(filename), "makefile")
		  || !strnfcmp(get_tail(filename), "makefile.", 9)) {
		sprintf(res, pat, "mak");	// Recognize special file names
		return 1;					// w/o .mak extension.
	}
	save_var point = 0;
	if (parse_string(1, "#%!.*/perl")) {
		sprintf(res, pat, "perl");
		return 1;
	}
	if (parse_string(1, "#%!.*/python")) {
		sprintf(res, pat, "python");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(%!DOCTYPE[ \t]+html|html|head|%!--)")) {
		sprintf(res, pat, "html");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(%!DOCTYPE|xml|xst)")) {
		sprintf(res, pat, "xml");
		return 1;
	}
	if (parse_string(1, "(:.*\n)?#%!.*/(sh|csh|bash|ksh|tcsh|zsh)</word>|:\n")
		  || !strnfcmp(get_tail(filename), ".profile", 8)) {
		sprintf(res, pat, "sh");
		return 1;
	}
	if (parse_string(1, "%$title ")) {
		sprintf(res, pat, "gms");
		return 1;
	}
	if (parse_string(1, "%%%!PS-Adobe-")) {
		sprintf(res, pat, "ps");
		return 1;
	}
	if (size() > 1000)		// Look in 1st 1000 chars only.
		save_var narrow_end = size() - 1000;
	if (parse_string(1, "[ \t\n]*(//|/%*|#)")
		&& re_search(1, "^[ \t]*(#define|#include)[ \t]+")) {
		// It starts with a comment or preprocessor line, and it's
		// got a #define or #include, so assume it's a C++ header
		// file.
		sprintf(res, pat, "hpp");
		return 1;
	}
	point = 0;
	if (GET_LINE_TRANSLATE(translation_type) == FILETYPE_UNIX
		  && parse_string(1, "([ \t]*\n)*#")) {
		sprintf(res, pat, "conf");
		return 1;
	}
	return 0;
}

char *(*give_grep_finder)()(inbufs, filepat, ps)
char *filepat, **ps;	// put 1st match in ps & return finder for next matches
{
	char *(*finder)();

	finder = inbufs ? b_match : do_file_match;
	*ps = (*finder)(filepat, STARTMATCH | EXACTONLY
					| (inbufs ? ((preserve_filename_case & 4) ? FM_FOLD : 0)
							  : FM_NO_DIRS));
	return finder;
}

get_grep_file(inbufs, filepat)
char *filepat;
{
	char pr[FNAMELEN], dir[FNAMELEN];
	int show_rel = 1;		// Show the relative path of the default.

	if (!inbufs) {
		if (is_dired_buf() && !inp_get_other_name)
			save_var inp_get_other_name = get_dired_buffer_pattern();
		if (grep_default_directory == 1 && !grep_prompt_with_buffer_directory
			  || grep_default_directory == 3)
			show_rel = 0;
		build_prompt(pr, "In file(s)", _grep_files, 0, show_rel);
		save_var prompt_with_buffer_directory = grep_prompt_with_buffer_directory;
		get_file_dir(filepat, pr);
		if (!*filepat && grep_default_directory) {
			save_var _help_on_cmd = last_index;
			if (grep_default_directory == 3)
				strcpy(filepat, _grep_files);
			else
				get_last_recall(f_match, filepat);
			if (grep_default_directory >= 2 && *filepat
				  && get_buffer_directory(dir)) {
				if (grep_default_directory != 3)
					absolute(filepat, dir);
				return 1;
			}
		}
		if (*filepat) {
			if (is_directory(filepat))
				sprintf(filepat + strlen(filepat), "%c*", path_sep);
			strcpy(_grep_files, filepat);
		}
		if (grep_default_directory == 3)
			strcpy(filepat, _grep_files);
		else
			relative(_grep_files, filepat);
		if (!*filepat)
			return 0;
	} else
		comp_read(filepat, "In buffers [pattern]: ", b_match,
				  ((preserve_filename_case & 4) ? FM_FOLD : 0), bufname);
	return 1;
}

prepare_grep_buffer(char *buf, int flags, char *pat, char *filepat)
{
	char search_type[FNAMELEN], dir[FNAMELEN];
	int before_head;

	create(buf);
	if (grep_empties_buffer)
		zap(buf);
	locate_window(buf, "");
	point = size();
	if (grep_empties_buffer)
		modified = 0;
	save_var modified;
	if (size())
		bprintf("\n");	/* separate from previous query */
	make_search_prompt(search_type, flags & SMODES);
	before_head = point;
	getcd(dir);
	bprintf("Found 0 lines with %s\"%s\" in %s%s (cur dir \"%s\")\n",
			search_type, pat, (flags & BUFGREP) ? "(buffers) " : "",
			filepat, dir);
	set_character_color(before_head, point, color_class grep_header);
	if (display_column.default != -1)
		display_column = 0;	/* horizontally scroll this window */
	grep_mode();
	return before_head;
}

do_grep(buf, pat, flags, filepat, inbufs)
char *buf, *pat, *filepat;
{
	int res = 0, aborted = 0, cnt = 0, tempbuf = 0, i, skipped = 0;
	int orig, before_head, this_err;
	char *s, rel[FNAMELEN], abs[FNAMELEN], *(*finder)();
	char *months = "Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec ";
	char extra[FNAMELEN];
	spot start;

	flags |= (inbufs ? BUFGREP : FILEGREP);
	save_var synchronous_file_io = 1;
	before_head = prepare_grep_buffer(buf, flags, pat, filepat);
	start = alloc_spot(0);
	save_var modified;
	save_var abort_file_matching = ABORT_ERROR;
	finder = give_grep_finder(inbufs, filepat, &s);	// get 1st match
	save_var readonly_warning, no_popup_errors = 1;
	if (grep_on_changed_file & 1)
		readonly_warning |= ROWARN_GREP;
	readonly_warning &= ~(ROWARN_MSG | ROWARN_BELL);
	orig = bufnum;
	for (; s; s = (*finder)(filepat, 0)) {
		if (user_abort) {
			aborted = 1;
			break;
		}
		grep_make_names(rel, abs, s, inbufs);
		if (inbufs)
			bufname = s;
		else if (!ok_grep_file_match(s, abs)) {
			skipped++;
			continue;
		}
		noteput("%d matches, searching %f...", cnt, rel);
		if (inbufs)	// skip over grep results buffer
			this_err = !strcmp(s, buf);
		else if (!look_file(abs)) {
			this_err = find_it(abs, default_translation_type);
			if (!grep_keeps_files)
				tempbuf = bufnum;
		} else {
			this_err = 0;
			if (grep_on_changed_file & 2)
				save_var auto_read_changed_file = 1;
			if (check_dates(0) == 2)
				grep_date_warning(buf, filename);
		}
		if (this_err == EREADABORT || user_abort) {
			aborted = 1;
			break;
		} else if (this_err) {
			window_bufnum = orig;
			if (tempbuf)
				buf_delete(tempbuf);
			tempbuf = 0;
			continue;
		}
		*extra = 0;
		if (grep_include_timestamp && !inbufs && file_info.check_type == CHECK_FILE)
			if (grep_include_timestamp == 2)
				sprintf(extra, "[%04d-%02d-%02d %02d:%02d:%02d] ",
						file_info.year, file_info.month, file_info.day,
						file_info.hour, file_info.minute, file_info.second);
			else
				sprintf(extra, "[%02d-%.3s-%04d %02d:%02d:%02d] ",
						file_info.day,
						(file_info.month >= 1 && file_info.month <= 12)
						? months + (file_info.month - 1) * 4 : "???",
						file_info.year,
						file_info.hour, file_info.minute, file_info.second);
		if (grep_show_absolute_path == 2)
			convert_to_8_3_filename(abs);
		res = copy_matches(name_to_bufnum(buf), flags, pat,
						   grep_show_absolute_path ? abs : rel, extra);
		window_bufnum = orig;
		if (res < 0)	/* error occurred during search */
			break;
		else if (res > 0) {
			cnt += res;
			build_first = 1;
			update_grep_title(before_head, cnt);
			refresh();
		}
		if (tempbuf)
			buf_delete(tempbuf);
		tempbuf = 0;
	}
	window_bufnum = orig;
	point = i = *start;
	free_spot(start);
	build_first = 1;
	if (tempbuf)
		buf_delete(tempbuf);
	say("");
	if (aborted)
		quick_abort();
	if (res == -DSBAD)
		error("Illegal search pattern: %s", pat);
	else if (res != -DSABORT) {
		locate_window(buf, "");
		point = i;
		if (skipped) {
			say("Skipped %d file%s due to grep-ignore-file- variables",
				  skipped, (skipped == 1) ? "" : "s");
			bprintf("%pSkipped %d file%s due to grep-ignore-file- variables\n",
					color_class grep_header,
					skipped, (skipped == 1) ? "" : "s");
		}
	}
}

char *man_match(s, flags)
char *s;
{
	char pat[FNAMELEN], match[FNAMELEN], *path, *t, *f, *q;
	char man_path[FNAMELEN * 2];

	if (strchr(s, '/') || strchr(s, '\\'))
		return do_file_match(s, flags);
	save_var cygwin_filenames |= 32;
	if ((flags & STARTMATCH) && *s && (!isalpha(*s) || s[1] == ' ')) {
		zap(_MATCH_BUF);	// Something like "-k foo" or "2 bar".
		buffer_printf(_MATCH_BUF, "%s%s\n", s, (flags & LISTMATCH) ? "" : " ");
		// Just return text itself as the match.
	} else if (flags & STARTMATCH) {
		show_text(-1, 20, "Listing matches...");	// Show for at least .2 secs.
		zap(_MATCH_BUF);
		build_man_path(path = man_path);
		while (!user_abort) {
			t = strchr(path, ':');	// Assume MANPATH uses Unix sep, not DOS.
			if (t)
				strncpy(pat, path, t - path);
			else
				strcpy(pat, path);
			convert_cygwin_filename(pat, 0);
			if (*pat && pat[strlen(pat) - 1] != '/')
				strcat(pat, "/");
			sprintf(pat + strlen(pat), "man*/%s*", s);
			f = do_file_match(pat, STARTMATCH);
			for (; f && !user_abort; f = do_file_match(pat, 0)) {
				strcpy(match, get_tail(f, 0));
				q = get_extension(match);
				if (q && !strcmp(q, ".gz"))
					*q = 0;
				q = get_extension(match);
				if (!q || !*q)
					continue;
				*q = 0;
				buffer_printf(_MATCH_BUF, "%s\n", match);
			}
			if (!t)
				break;
			path = t + 1;
		}
		save_var bufname = _MATCH_BUF;
		point = 0;
		do_sort_region(0, size(), 0, 0);
		do_uniq(1, 1, 0);
		say("");
		restore_vars();
	}
	return general_matcher(s, flags);
}

do_html_indent(int manual)
{
	int region, start, i;

	recolor_buffer_range(point, point + 1);
	region = get_tagged_region("color-as-unit", point, &start);
	if (IS_ENTER_KEY(key) && !(html_auto_indent & (region == HT_JSCRIPT ? 2
		: region == HT_VBSCRIPT ? 4 : region == HT_PHP ? 8 : 1)))
		return;
	save_var narrow_start = start;
	i = (1 << (region - 1 + (IS_ENTER_KEY(key) ? 4 : 0)));
	if (region > 0 && region <= HT_PHP && !(html_indenting_rules & i))
		region = 0;
	switch (region) {
		case HT_JSCRIPT:
			html_exclude_script_head();
			save_var c_extra_keywords |= JAVA_KEYWORDS;
			do_c_indent();
			return;
		case HT_VBSCRIPT:
			html_exclude_script_head();
			do_vbasic_indent();
			return;
		case HT_PHP:
			html_exclude_script_head();
			save_var in_perl_buffer = 1;
			save_var sublanguage_code = SUBLANG_PHP;
			do_c_indent();
			return;
	}
	narrow_start = 0;
	if (manual)
		indent_previous();
	else
		prev_indenter();
}

suffix_xmp()
{
	xml_mode();
}

suffix_rdf()
{
	xml_mode();
}

command list_customizations()
{
	char cmd[FNAMELEN], *args, time[FNAMELEN];
	char ext_file[FNAMELEN], settings_file[FNAMELEN], *s;
	int tmp, cur, ext, i, err;
	int cust = prep_customizations(), clist;

	iter = 1;
	tiled_only();
	get_customization_directory(ext_file);
	strcat(ext_file, "extensions-list");
	get_customization_directory(settings_file);
	strcat(settings_file, "settings-list");
	delete_file(settings_file);
	i = name_match("-loaded-eel-file-", STARTMATCH);
	save_var bufnum = ext = zap("-list-customizations");
	give_binding(cmd, "exit-level");
	bprintf("Epsilon will include the following EEL extension files as part\n"
			"of your customizations.  Edit this list if needed, then\n"
			"press %s to begin.\n\n%p\n", cmd, color_class viewed_text);
	point--;
	for (; i; i = name_match("", 0))	// List new extensions by searching
		if (name_type(i) == NT_VAR) {	// for id variables.
			s = get_str_var(i);
			bprintf("%s%s\n", s, (check_file(s) == CHECK_FILE)
					? "" : " (not found)");
		}
	to_buffer_num(ext);
	point = 0;
	do {
		recursive_edit();
		if (!buf_exist(ext))
			return;
		point = 0;
		point += parse_string(1, "(.|\n.)*\n*");
		i = test_valid_files(point, size());
		if (i == 2) {
			buf_zap(ext);
			return;
		}
	} while (i);
	point = 0;		// Drop the explanatory text.
	delete(0, parse_string(1, "(.|\n.)*\n*"));
	simple_re_replace(1, " *%(not found%)", "");
	if (err = file_write(ext_file, FILETYPE_BINARY)) {
		file_error(err, ext_file, "write error");
		return;
	}
	save_var reject_client_connections = 1;
	strcpy(cmd, argv[0]);
	convert_to_8_3_filename(cmd);
	sayput("Running %s to create original settings list...", cmd);
	args = " -s original -p0 -noinit -r build-settings-file";
	err = shell_with_args(cmd, args);
	if (err)
		error("Error running %s %s", cmd, args);
	sayput("Reading original settings list...");
	bufnum = tmp = tmp_buf();
	if (err_file_read(settings_file, FILETYPE_BINARY))
		return;
	bufnum = cur = tmp_buf();
	list_all_to_buffer(cur, 3);
	clist = make_difference_list(tmp, cur);
	restore_vars();
	buf_delete(tmp);
	buf_delete(cur);
	to_buffer_num(cust);
	current_time_and_date(time);
	save_spot point;
	i = point;
	bprintf("\n# List-customizations created these settings %s.\n", time);
	fix_extension_list(ext);
	grab_buffer(ext);
	grab_buffer(clist);
	buf_delete(clist);
	delete_file(settings_file);
	delete_file(ext_file);
	bprintf("\n");
	point = i;
	if (!has_arg)
		comment_old_customizations(point);
	if (!list_undefined_customizations())
		say("");
}

build_settings_file()
{
	char ext_file[FNAMELEN], settings_file[FNAMELEN], eel_file[FNAMELEN];
	int tmp = tmp_buf(), err;

	get_customization_directory(ext_file);
	strcat(ext_file, "extensions-list");
	get_customization_directory(settings_file);
	strcat(settings_file, "settings-list");
	save_var bufnum = tmp;
	if (err_file_read(ext_file, FILETYPE_BINARY)) {
		show_text(0, 300, "Can't read extensions list file, aborting.");
		leave(1);
	}
	point = 0;
	while (grab_line(bufnum, eel_file)) {
		sayput("Loading %f...", eel_file);
		if (load_eel_from_path(eel_file, 2))
			return;
	}
	buf_zap(bufnum = tmp);
	sayput("Listing original settings...");
	list_all_to_buffer(tmp, 3);
	if (err = file_write(settings_file, FILETYPE_BINARY)) {
		file_error(err, settings_file, "write error");
		delete_file(settings_file);	// Ensure caller notices failure.
	}
	say("");
	leave(0);
}

command edit_customizations()
{
	iter = 0;
	tiled_only();
	to_buffer_num(prep_customizations());
}

command clean_customizations()
{
	int dup;

	iter = 0;
	tiled_only();
	to_buffer_num(prep_customizations());
	point = 0;		// First delete previous batch of cleanings.
	simple_re_replace(0, "###.*\n", "");
	point = 0;
	while (point < size() - narrow_end) {
		dup = 0;
		if (parse_string(1, "^(%+<word|dash>+ )"))
			dup = check_dup_customization(1, "^<>");
		else if (parse_string(1, "^(%&<word|dash>+ color class "
							  "for <word|dash>+: )"))
			dup = check_dup_customization(1, "^<>");
		else if (parse_string(1, "^%(bind-to-key \"[^\"\n]+\" (\"[^\"\n]+\")%)"))
			dup = check_dup_customization(1, "^%(bind-to-key \"[^\"\n]+\" <>%)");
		else if (parse_string(1, "^(<word|dash>+: macro )"))
			dup = check_dup_customization(1, "^<>");
		else if (parse_string(1, "^(%(load-eel-from-path \"[^\"\n]+\" 2%)$)"))
			dup = check_dup_customization(1, "^<>");
		else if (parse_string(1, "^(~<^wspace>+ \".*\": )"))
			dup = check_dup_customization(1, "^<>");
		if (dup) {
			stuff("### ");
			nl_forward();
			continue;
		}
		switch (curchar()) {
			case '(':	skip_paren_cmd(); break;
			case ';':
			case '#':	nl_forward(); break;
			default:
				nl_forward();
				break;
		}
	}
}

color_batch_range(from, to)
{
	int t = -1;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend, case_fold = 1;
	point = to;				// Color entire lines.
	nl_forward();
	to = point;
	save_var narrow_end = size() - to;
	set_character_color(from, to, -1);
	point = from;

	while (point < to) {
		if (!re_search(1, "[-a-z0-9_.%/\\]+|[<>|]|^::")) {
			t = size();
			break;
		}
		t = matchstart;
		if (strchr("<>|", character(point - 1))) {
			set_character_color(t, point, color_class batch_punctuation);
			continue;
		} else if (character(point - 1) == ':') {
			nl_forward();
			set_character_color(t, point, color_class batch_comment);
			continue;
		}
		switch (is_batch_keyword(t)) {
			case 1:
				set_character_color(t, point, color_class batch_keyword);
				break;
			case 2:
				nl_forward();
				set_character_color(t, point, color_class batch_comment);
				break;
			default:
				set_character_color(t, point, -1);
				break;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	return point;
}

perl_after_expression(int p, int strict)
{
	int ch;
	if (in_shell_buffer)
		return 0;
	if (strict) {
		save_var point = p;
		if (!parse_string(-1, "}[ \t]*")) {	// If not an eval block or similar.
			// Move back over whitespace and comments.
			for (;;) {
				if (parse_string(-1, "[ \t\n]+"))
					point = matchend;
				else if (parse_string(-1, "#.*")
						 && get_character_color(matchend)
						 == color_class perl_comment)
					point = matchend;
				else
					break;
			}
			if (!parse_string(-1,  "<^word>[0-9]+([-0-9e.]+)?|[])'\"`]")
				  // Not a number, or one of the characters ])'" or `.
				  && !parse_string(-1, "%$%^?[^ \t\n]")) {
				// Or $, then perhaps a ^ and a nonspace.
				point -= parse_string(-1, "(%+%+|--)?");
				// Skip back over any ++ or --.
				if (!parse_string(-1, "<word>+"))	// Only other option is a word.
					return 0;
				if (parse_string(-1, "<^word>"
								 "(if|unless|package|sub|import|no|require|use|die"
								 "|do|dump|else|elsif|eval|exec|exit|for|foreach|if"
								 "|return|unless|until|while|BEGIN|END|local|my"
								 "|continue|goto|last|next|redo|qq|qx|q|qw|qr|m|s|tr|y"
								 "|format|__END__|__DATA__|split|scalar|warn)")
					  && !strchr("$%&#@", character(matchend)))
					return 0;		// That's not any Perl keyword.
			}
		}
		return 1;
	} else {
		if (!matches_at(p, -1, "([$%&#@]((<word>+(->|'))?<word>+)"
						// A variable or function with $ or similar before it.
						"(%+%+|--)?"
						// Possibly followed by ++ or --.
						"|%$%^?[^ \t\n]"
						// Or $, then perhaps a ^ and a nonspace.
						"|<^word>[0-9]+([-0-9e.]+)?|[])'\"`])"
						// Or a number.  Or one of the characters ])'" or `.
						"(#.*|[ \t\n])*|}[ \t]*"))
			// All followed by whitespace or a comment.
		return 0;
	}
	ch = get_character_color(matchend);
	if (ch == color_class perl_comment || ch == color_class perl_string
		  && !matches_at(matchend, 1, "[\"']"))
		return 0;
	return 1;
}

int in_perl_comment(int pos)
{
	int ch = get_character_syntax_color(pos);

	if (ch == color_class perl_comment
		  || get_character_syntax_color(pos - 1) == color_class perl_comment
		  && character(pos - 1) != '\n')
		return IN_NEW_COMMENT;
	else if (ch == color_class perl_string)
		return IN_STR_CONST;
	return 0;
}

suffix_bash()
{
	shell_mode();
}

suffix_tcsh()
{
	shell_mode();
}

suffix_zsh()
{
	shell_mode();
}

int run_helper(char *prog, char *file, char *moreargs)
{
	char cmd[FNAMELEN * 2];

	sprintf(cmd, "~/.epsilon/%s \"%s\" %s >/dev/null 2>&1",
			prog, file, moreargs);
	if (!shell("", cmd, ""))	// Run user's version; fail silently.
		return 1;		// If we failed, look on path.
	sprintf(cmd, "%s \"%s\" %s >/dev/null 2>&1", prog, file, moreargs);
	if (!shell("", cmd, ""))	// Search along the path; fail silently.
		return 1;		// If we failed, try again with std copy we provide.
	get_executable_directory(cmd);
	sprintf(cmd + strlen(cmd), "%s \"%s\" %s", prog, file, moreargs);
	if (!shell("", cmd, ""))
		return 1;
	return 0;
}

safe_copy_buffer_variables(to, from)
{
	save_var bufnum = to, buf_accessed;
	copy_buffer_variables(to, from);
}

do_python_indent()
{
	int ind;

	save_var point;
	do {		// Find prev non-comment or non-blank line.
		if (!nl_reverse())		// First line in the file, no indent.
			return 0;
		to_begin_line();
	} while (parse_string(1, python_indent_to_comment ? "[ \t]*\n"
						  : "[ \t]*[#\n]"));
	to_indentation();
	ind = current_column();
	// Does this (previous) line end with a : character (plus maybe a
	// comment)?  If so, indent one level more.
	if (parse_string(1, ".*:[ \t\n]*(#.*)?$"))
		ind += python_indent;
	restore_vars();
	indent_to_column(ind);
}

fromorig(with)
char *with;
{
	int group, fold_as = 0, from, to;
	char *p, str[1000], *q = str, *fold_from = 0, *owith = with;

	if (strlen(with) > 900) {
		fromorig_bigrepl(owith);
		return;
	}
	while (*with) {
		p = with;
		if (p)
			p = index(p, '#');
		if (!p) break;
		strncpy(q, with, p - with);	/* add till just before # */

		q += strlen(q);
		if (fold_from)
			fold_replacement_text(fold_as, fold_from, q);
		if (isdigit(*++p)) {
			group = *p - '0';
			from = find_group(group, 1);
			to = find_group(group, 0);
			if (q - str + (to - from) > 900) {
				fromorig_bigrepl(owith);
				return;
			}
			grab(from, to, q);
			if (to - from != strlen(q)) {	// Null chars in substitution text.
				fromorig_bigrepl(owith);
				return;
			}
			q += strlen(q);		/* add variable part */
		} else
			switch (*p) {
				case 'U':
				case 'L':
				case 'C':
				case 'E':
					fold_as = *p;
					fold_from = q;
					break;
				case '<':
					int val = get_angled_value_from_string(&p);
					if (val >= 0 && IS_ASCII_KEY(val))
						*q++ = val;
					p--;
					break;
				default:
					*q++ = *p;		/* #x gives x */
					break;
			}
		with = p + 1;
	}
	strcpy(q, with);			/* add trailing part */
	if (fold_from)
		fold_replacement_text(fold_as, fold_from, q + strlen(q));
	delete(matchstart, point);		/* do replacement */
	stuff(str);
}

list_all_to_buffer(int buf, int flags)
{
	save_var bufnum = buf, use_default;
	int i, max = final_index(), percent, opercent = -1;
	for (i = 1; i <= max; i++) {
		if (flags & 32)
			check_abort();
		if ((flags & 64) && (percent = muldiv(i, 100, max)) != opercent) {
			noteput("Describing state: %d%%...", percent);
			opercent = percent;
		}
		list_one(bufnum, i, flags);
	}
	point = size();
	buf_printf(bufnum, "+path-sep char variable: %d\n", path_sep);
	buf_printf(bufnum, "+abort-key int variable: %d\n", abort_key);
}

command list_all()
{
	int flags = 32 | 64;
	tiled_only();
	zap("list-all");
	to_buffer("list-all");
	save_var use_default;
	if (has_arg)
		flags |= iter & 31;
	iter = 1;
	list_all_to_buffer(bufnum, flags);
	point = 0;
	say("");
}

tex_find_linked_file(char *file)
{
	char rel[FNAMELEN], curdir[FNAMELEN], *p, *include;

	to_begin_line();
	if (parse_string(1, ".*\\(include|input|usepackage"
					 "|RequirePackage)[ \t{]*([^{} \t\n]+)"))
		grab(find_group(2, 1), find_group(2, 0), rel);
	else
		error("Can't find any link here.");
	if (!*get_extension(rel))
		strcpy(get_extension(rel),
			   character(find_group(1, 1)) == 'u' ? ".sty" : ".tex");
	strcpy(file, rel);
	get_buffer_directory(curdir);
	absolute(file, curdir);
	if (check_file(file))
		return;
	include = getenv("TEXINPUTS");
	p = 0;
	if (!p && include)
		p = look_on_path(rel, 0, include);
	if (!p && *include_directories)
		p = look_on_path(rel, 0, include_directories);
	if (!p)
		error("%s not found in include-directories path", rel);
	strcpy(file, p);
}

do_unicode_convert(int flags)
{
	char msg[FNAMELEN], *enc;
	int err = 0, from = 0, to = size(), high, scol, ecol;

	high = is_highlight_on();
	check_modify(bufnum);
	sprintf(msg, "Convert %s %s", high ? "region" : "buffer",
			(flags & CONV_TO_16) ? "to Unicode from encoding"
								 : "from Unicode to encoding");
	if (!(enc = encoding_to_name(ask_encoding(msg, 1))))
		return;
	if (high) {
		save_spot point, mark;
		fix_region();
		from = point;
		to = mark;
	} else
		save_spot point;
	if (high == REGRECT) {
		scol = get_column(point);
		ecol = get_column(mark);
		for (; point < mark && !err; nl_forward())
			err = perform_unicode_conversion(bufnum, column_to_pos(scol),
											 column_to_pos(ecol), flags, enc);
	} else
		err = perform_unicode_conversion(bufnum, from, to, flags, enc);
	if (err == ETOOBIG)
		error("This is not an 8-bit encoding; a character is already wide.");
	if (!high && (flags & CONV_TO_16)) {
		int trans = GET_LINE_TRANSLATE(translation_type);
		int conv = encoding_from_name(enc);
		translation_type = MAKE_TRANSLATE(conv, trans,
										  (trans != FILETYPE_AUTO || conv > 0));
	}
}

int vbasic_indent_more()
{
	char kwd[FNAMELEN], pat[FNAMELEN];

	if (vbasic_indent_subroutines
		  && parse_string(1, "[ \t]*(%<[^\n>]+%>)?[ \t]*"
						  "(</word>(Private|Public|Default|Friend"
						  "|Static|Protected|Shadows|MustInherit"
						  "|NotInheritable|Overloads|Overrides|Overridable"
						  "|NotOverridable|MustOverride|Shared|ReadOnly"
						  "|WriteOnly)</word>[ \t]*)*"
						  "</word>(Sub|Function|Property|Class"
						  "|Module|Interface|Namespace|Structure)</word>"))
		return 1;
	if (parse_string(1, "[ \t]*</word>(If)</word>.*"
					 "</word>Then</word>[ \t]*[^ \t\n':]"))
		return 0;		// Old one-line form of If statement.
	if (parse_string(1, "[ \t]*</word>((private|public)[ \t]+)?"
					 "(Type|Enum|Else|With|Select|Case|Try|Catch|Finally"
					 "|SyncLock|If|Elseif)</word>")) {
		grab(find_group(3, 1), find_group(3, 0), kwd);
		sprintf(pat, ".*:[ \t]*</word>End[ \t]+%s</word>", kwd);
		if (!parse_string(1, pat))	// Multiple statements: ignore block.
			return 1;
	}
	if (parse_string(1, "[ \t]*</word>(For)</word>")
		  && !parse_string(1, ".*:[ \t]*</word>Next</word>"))
		return 1;
	if (parse_string(1, "[ \t]*</word>(Do)</word>")
		  && !parse_string(1, ".*:[ \t]*</word>Loop</word>"))
		return 1;
	if (parse_string(1, "[ \t]*</word>(While)</word>")
		  && !parse_string(1, ".*:[ \t]*</word>Wend</word>"))
		return 1;
	if (parse_string(1, "[ \t]*(Get|Set)[ \t]*(%([ \t]*%))?[ \t]*[\n(':]"))
		return 1;	// Get/set keyword with nothing after it, or (, starts block.
	return 0;
}

vbasic_indenter()
{
	int orig = point, ind;

	if (matches_at(give_begin_line(), 1, "[ \t]*#")
		  || point <= narrow_start) {
		indent_to_column(0);
		return;
	}
	if (vbasic_is_continuation()) {
		vbasic_indent_continuation(orig);
		return;
	}
	do {
		to_begin_line();
		if (!re_search(-1, "[^ \t\n]"))	/* Find previous non-blank line */
			break;
		to_indentation();
	} while (parse_string(1, "['#]"));	// that's not a comment or preproc.
	vbasic_before_continuation();
	ind = get_indentation(point);
	if (vbasic_indent_more())
		ind += vbasic_step();
	point = orig;
	if (vbasic_indent_less())
		ind -= vbasic_step();
	to_indentation();		/* go to current line's indent */
	indent_to_column(ind);
}
