/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2005 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 12.07 to 12.08.

#include "eel.h"
#include "c.h"
#include "dired.h"
#include "colcode.h"
#include "bufed.h"
#include "epsnet.h"
#include "grep.h"
#include "menu.h"
#include "misclang.h"
#include "perl.h"
#include "proc.h"
#include "tags.h"
#include "vbasic.h"
#include "vhdl.h"


color_asm_range(from, to) // recolor just this section
{			// last colored region may go past to
	int t = -1, talk, s, talk_at = 0;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	while (from > 0 && get_tagged_region("color-as-unit", from, &s) != -1
		   && s <= from)
		from = s - 1;
	c_init_color(from, to);
	point = from;
	to_begin_line();
	talk = (to - from > 20000);	// show status during long delays
	while (point < to) {
		if (!re_search(1, "[0-9A-Za-z_.@#%?$]+|[\"';]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(point - 1)) {		// check last char
			case ';':			// found comment
				nl_forward();
				set_character_color(t, point, color_class c_comment);
				break;
			case '"':		// found a string literal
				point = t;
				re_search(1, "\"([^\"\\\n]|\\(.|\n))*[\"\n]");
				set_character_color(t, point, color_class c_string);
				if (get_character_color(point, (int *) 0, &s) == 
					color_class c_string && s > to)  // fix up after
					c_init_color(point, to = s); // quoted "'s
				break;
			case '\'':		// found a char const
				point = t;
				re_search(1, "\'([^\'\\\n]|\\(.|\n))*[\'\n]");
				set_character_color(t, point, color_class c_charconst);
				break;
			default:		// found identifier, kywd, or number
				set_character_color(t, point, asm_keyword_color(t));
				if (asm_title_line()) {
					t = point;
					point++;
					nl_forward();
					set_character_color(t, point, color_class c_string);
				}
				asm_comment_line();
				break;
		}
		if (talk && point > talk_at + 5000) {
			note("Coloring ASM program: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	c_init_color(to, t);
	if (talk_at)
		note("");
	return point;
}

is_asm_preproc(p)	// is text in p (must be surrounded by |'s) a preproc kwd?
char *p;
{
	if (strstr(
			   "|%out|.186|.286|.286c|.286p|.287|.386|.386c|.386p|.387|.486|.486p|.8086"
			   "|.8087|.alpha|.break|.code|.const|.continue|.cref|.data|.data?|.dosseg"
			   "|.else|.elseif|.endif|.endw|.err|.err1|.err2|.errb|.errdef|.errdif|.errdifi"
			   "|.erre|.erridn|.erridni|.errnb|.errndef|.errnz|.exit|.fardata|.fardata?"
			   "|.if|.lall|.lfcond|.list|.listall|.listif|.listmacro|.listmacroall|.model"
			   "|.msfloat|.no87|.nocref|.nolist|.nolistif|.nolistmacro|.radix|.repeat|.sall"
			   "|.seq|.sfcond|.stack|.startup|.tfcond|.type|.until|.untilcxz|.while|.xall"
			   "|.xcref|.xlist|alias|align|assume|catstr|comm|comment|db|dd|df|dosseg|dq"
			   "|dt|dup|dw|echo|else|elseif|elseif1|elseif2|elseifb|elseifdef|elseifdif"
			   "|elseifdifi|elseife|elseifidn|elseifidni|elseifnb|elseifndef|end|endif"
			   "|endm|endp|ends|eq|equ|even|exitm|extern|externdef|extrn|for|forc|ge|goto"
			   "|group|gt|high|highword|if|if1|if2|ifb|ifdef|ifdif|ifdifi|ife|ifidn|ifidni"
			   "|ifnb|ifndef|include|includelib|instr|invoke|irp|irpc|label|le|length|lengthof"
			   "|local|low|lowword|lroffset|lt|macro|mask|mod|name|ne|offset|opattr|option"
			   "|org|page|popcontext|proc|proto|ptr|public|purge|pushcontext|record|repeat"
			   "|rept|seg|segment|short|size|sizeof|sizestr|struc|struct|substr|subtitle"
			   "|subttl|textequ|this|title|type|typedef|union|while|width|.586|.586p|.686"
			   "|.686p|.k3d|.mmx|.xmm|%depth|%incl|%linum|%noincl|%nomacs|%notrunc|.386r"
			   "|.common|.prot|.twocase|.xlisti|ideal|locals|model|$|?|@b|@f|abs|addr|all"
			   "|assumes|at|basic|byte|c|carry?|casemap|common|compact|cpu|dotname|dword"
			   "|emulator|epilogue|error|export|expr16|expr32|far|far16|far32|farstack"
			   "|flat|forceframe|fortran|fword|huge|language|large|listing|ljmp|loadds"
			   "|m510|medium|memory|near|near16|near32|nearstack|nodotname|noemulator|nokeyword"
			   "|noljmp|nom510|none|nonunique|nooldmacros|nooldstructs|noreadonly|noscoped"
			   "|nosignextend|nothing|notpublic|oldmacros|oldstructs|os_dos|overflow?|para"
			   "|parity?|pascal|private|prologue|qword|radix|readonly|real10|real4|real8"
			   "|req|sbyte|scoped|sdword|setif2|sign?|small|stack|stdcall|sword|syscall"
			   "|tbyte|tiny|use16|use32|uses|vararg|word|zero?|@@:|@CatStr|@code|@CodeSize"
			   "|@Cpu|@CurSeg|@data|@DataSize|@Date|@Environ|@fardata|@fardata?|@FileCur"
			   "|@FileName|@InStr|@Interface|@Line|@Model|@SizeStr|@stack|@SubStr|@Time"
			   "|@Version|@WordSize|err|masm51|nowarn|pword|quirks|show|warn|"
			   , p))
			return 1;
	return 0;
}

is_asm_keyword(p)	// is text in p (must be surrounded by |'s) a keyword?
char *p;
{
	if (strstr(
			   "|aaa|aad|aam|aas|adc|add|and|arpl|bound|bsf|bsr|bswap|bt|btc|btr|bts|call"
			   "|cbw|cdq|clc|cld|cli|clts|cmc|cmp|cmps|cmpsb|cmpsd|cmpsw|cmpxchg|cwd|cwde"
			   "|daa|das|dec|div|enter|esc|f2xm1|fabs|fadd|faddp|fbld|fbstp|fchs|fclex"
			   "|fcom|fcomp|fcompp|fcos|fdecstp|fdisi|fdiv|fdivp|fdivr|fdivrp|feni|ffree"
			   "|fiadd|ficom|ficomp|fidiv|fidivr|fild|fimul|fincstp|finit|fist|fistp|fisub"
			   "|fisubr|fld|fld1|fldcw|fldenv|fldenvd|fldenvw|fldl2e|fldl2t|fldlg2|fldln2"
			   "|fldpi|fldz|fmul|fmulp|fnclex|fndisi|fneni|fninit|fnop|fnsave|fnsaved|fnsavew"
			   "|fnstcw|fnstenv|fnstenvd|fnstenvw|fnstsw|fpatan|fprem|fprem1|fptan|frndint"
			   "|frstor|frstord|frstorw|fsave|fsaved|fsavew|fscale|fsetpm|fsin|fsincos"
			   "|fsqrt|fst|fstcw|fstenv|fstenvd|fstenvw|fstp|fstsw|fsub|fsubp|fsubr|fsubrp"
			   "|ftst|fucom|fucomp|fucompp|fwait|fxam|fxch|fxtract|fyl2x|fyl2xp1|hlt|idiv"
			   "|imul|in|inc|ins|insb|insd|insw|int|into|invd|invlpg|iret|iretd|iretdf"
			   "|iretf|ja|jae|jb|jbe|jc|jcxz|je|jecxz|jg|jge|jl|jle|jmp|jna|jnae|jnb|jnbe"
			   "|jnc|jne|jng|jnge|jnl|jnle|jno|jnp|jns|jnz|jo|jp|jpe|jpo|js|jz|lahf|lar"
			   "|lds|lea|leave|les|lfs|lgdt|lgs|lidt|lldt|lmsw|lock|lods|lodsb|lodsd|lodsw"
			   "|loop|loopd|loope|looped|loopew|loopne|loopned|loopnew|loopnz|loopnzd|loopnzw"
			   "|loopw|loopz|loopzd|loopzw|lsl|lss|ltr|mov|movs|movsb|movsd|movsw|movsx"
			   "|movzx|mul|neg|nop|not|or|out|outs|outsb|outsd|outsw|pop|popa|popad|popf"
			   "|popfd|push|pusha|pushad|pushd|pushf|pushfd|pushw|rcl|rcr|rep|repe|repne"
			   "|repnz|repz|ret|retf|retn|rol|ror|sahf|sal|sar|sbb|scas|scasb|scasd|scasw"
			   "|seta|setae|setb|setbe|setc|sete|setg|setge|setl|setle|setna|setnae|setnb"
			   "|setnbe|setnc|setne|setng|setnge|setnl|setnle|setno|setnp|setns|setnz|seto"
			   "|setp|setpe|setpo|sets|setz|sgdt|shl|shld|shr|shrd|sidt|sldt|smsw|stc|std"
			   "|sti|stos|stosb|stosd|stosw|str|sub|test|verr|verw|wait|wbinvd|xadd|xchg"
			   "|xlat|xlatb|xor|addpd|addps|addsd|addss|addsubpd|addsubps|andnpd|andnps"
			   "|andpd|andps|clflush|cmova|cmovae|cmovb|cmovbe|cmovc|cmove|cmovg|cmovge"
			   "|cmovl|cmovle|cmovna|cmovnae|cmovnb|cmovnbe|cmovnc|cmovne|cmovng|cmovnge"
			   "|cmovnl|cmovnle|cmovno|cmovnp|cmovns|cmovnz|cmovo|cmovp|cmovpe|cmovpo|cmovs"
			   "|cmovz|cmppd|cmpps|cmpss|cmpxchg8b|comisd|comiss|cpuid|cvtdq2pd|cvtdq2ps"
			   "|cvtpd2dq|cvtpd2pi|cvtpd2ps|cvtpi2pd|cvtpi2ps|cvtps2dq|cvtps2pd|cvtps2pi"
			   "|cvtsd2si|cvtsd2ss|cvtsi2sd|cvtsi2ss|cvtss2sd|cvtss2si|cvttpd2dq|cvttpd2pi"
			   "|cvttps2dq|cvttps2pi|cvttsd2si|cvttss2si|divpd|divps|divsd|divss|emms|fcmovb"
			   "|fcmovbe|fcmove|fcmovnb|fcmovnbe|fcmovne|fcmovnu|fcmovu|fcomi|fcomip|fisttp"
			   "|fxrstor|fxsave|haddpd|haddps|hsubpd|hsubps|lddqu|ldmxcsr|lfence|loopde"
			   "|loopdne|loopwe|loopwne|maskmovdqu|maskmovq|maxpd|maxps|maxsd|maxss|mfence"
			   "|minpd|minps|minsd|minss|monitor|movapd|movaps|movd|movddup|movdq2q|movdqa"
			   "|movdqu|movhlps|movhpd|movhps|movlhps|movlpd|movlps|movmskpd|movmskps|movntdq"
			   "|movnti|movntpd|movntps|movntq|movq|movq2dq|movshdup|movsldup|movss|movupd"
			   "|movups|mulpd|mulps|mulsd|mulss|mwait|orpd|orps|packssdw|packsswb|packuswb"
			   "|paddb|paddd|paddq|paddsb|paddsw|paddusb|paddusw|paddw|pand|pandn|pause"
			   "|pavgb|pavgw|pcmpeqb|pcmpeqd|pcmpeqw|pcmpgtb|pcmpgtd|pcmpgtw|pextrw|pinsrw"
			   "|pmaddwd|pmaxsw|pmaxub|pminsw|pminub|pmovmskb|pmulhuw|pmulhw|pmullw|pmuludq"
			   "|por|prefetchh|psadbw|pshufd|pshufhw|pshuflw|pshufw|pslld|pslldq|psllq"
			   "|psllw|psrad|psraw|psrld|psrldq|psrlq|psrlw|psubb|psubd|psubq|psubsb|psubsw"
			   "|psubusb|psubusw|psubw|punpckhbw|punpckhdq|punpckhqdq|punpckhwd|punpcklbw"
			   "|punpckldq|punpcklqdq|punpcklwd|pxor|rcpps|rcpss|rdmsr|rdpmc|rdtsc|rsm"
			   "|rsqrtps|rsqrtss|sfence|shufpd|shufps|sqrtpd|sqrtps|sqrtsd|sqrtss|stmxcsr"
			   "|subpd|subps|subsd|subss|sysenter|sysexit|ucomisd|ucomiss|ud2|unpckhpd"
			   "|unpckhps|unpcklpd|unpcklps|wrmsr|xorpd|xorps|lgdte|lidte|sidte|"
			   , p))
			return 1;
	return 0;
}

is_asm_register(p)	// is text in p (must be surrounded by |'s) a reg name?
char *p;
{
	if (strstr(
			   "|ah|al|ax|bh|bl|bp|bx|ch|cl|cr0|cr2|cr3|cs|cx|dh|di|dl|dr0|dr1|dr2|dr3"
			   "|dr6|dr7|ds|dx|eax|ebp|ebx|ecx|edi|edx|es|esi|esp|fs|gs|si|sp|ss|st|tr3"
			   "|tr4|tr5|tr6|tr7|cr4|mm|mm0|mm1|mm2|mm3|mm4|mm5|mm6|mm7|st0|st1|st2|st3"
			   "|st4|st5|st6|st7|xmm|xmm0|xmm1|xmm2|xmm3|xmm4|xmm5|xmm6|xmm7|"
			   , p))
		return 1;
	return 0;
}

bufed_ask_exit()
{
	char save_it[FNAMELEN];

	if (windows_state() & WINSTATE_MINIMIZED)
		windows_restore();
	if (is_gui && want_gui_prompts) {
		tiled_only();
		save_var _doing_input = DI_VIEW;
		save_var bufnum, window_handle;
		save_var return_raw_buttons = 1;
		for (;;) {
			do_bufed(BUFED_EXIT | BUFED_ONLY_UNSAVED);
			go_to_link = 0;
			view_loop(_bufed_window);
			bufed_getname(save_it);
			remove_bufed();
			if (user_abort) {
				user_abort = 0;
				return 0;
			}
			switch (key_is_button) {
				case 1:	key = 'Y'; break;
				case 2: key = 'N'; break;
				case 5: key = 'S'; break;
				}
			switch (toupper(key)) {
				case 'X':
				case 'Y':	return 1;
				case 'E':
				case '\n':
				case '\r':	if (exist(save_it))
								to_buffer(save_it);
							return 0;
				case 'C':
				case 'N':	return 0;
				case 'S':	if (*save_it && exist(save_it))
								bufed_save_one(save_it);
							if (!unsaved_buffers())
								return 1;
							break;
			}
		}
	} else {
		do_bufed(BUFED_ONLY_UNSAVED);
		maybe_refresh();
		sayput("Buffers not saved.  Exit? [N] ");
		do {
			maybe_refresh();
			getkey();
		} while (!IS_TRUE_KEY(key));
		say("");
		switch (toupper(key)) {
			case 'X':
			case 'Y':	break;
			case 'N':
			case 'C':
			case '\n':
			case '\r':	bufed_msg();
					return 0;
			default:	check_abort();
					ungot_key = full_key;
					return 0;
		}
		remove_bufed();
		maybe_refresh();
	}
	return 1;
}

bufed_ask_save()
{
	if (is_gui && want_gui_prompts) {
		save_var _doing_input = DI_VIEW;
		save_var bufnum, window_handle;
		save_var return_raw_buttons = 1;
		for (;;) {
			do_bufed(BUFED_SAVE | BUFED_ONLY_UNSAVED);
			go_to_link = 0;
			view_loop(_bufed_window);
			remove_bufed();
			switch (key_is_button) {
				case 1:	key = 'S'; break;
				case 2: key = 'N'; break;
				case 3:	key = 'C'; break;
				default: check_abort(); break;
			}
			switch (toupper(key)) {
				case '!':	save_when_making = 0;
				case 'N':	return 1;
				case '\n':
				case '\r':
				case 'S':
				case 'Y':	return 2;
				case 'C':	return 0;
			}
		}
	} else {
		do_bufed(0);
		sayput("Save unsaved buffers first? [N] (! to never ask) ");
		do {
			maybe_refresh();
			getkey();
		} while (key == MOUSE_MOVE);
		switch (toupper(key)) {
			case '!':	save_when_making = 0;
			case 'N':
			case '\n':
			case '\r':	remove_bufed();
						return 1;
			case 'Y':	break;
			default:	check_abort();
						say("");
						ungot_key = full_key;
						return 0;
		}
		remove_bufed();
		return 2;
	}
}

// See if text at pos matches pattern.
matches_at(int pos, int dir, char *pat)
{
	save_var point = pos;
	if (re_compile(dir, pat))
		return 0;
	return (re_match() != -1);
}

// See if text at pos matches pattern.
matches_at_length(int pos, int dir, char *pat)
{
	save_var point = pos;
	return parse_string(dir, pat);
}

suffix_java()
{
	c_mode();
	compile_buffer_cmd = compile_java_cmd;	// can compile this
	c_extra_keywords = JAVA_KEYWORDS;	// recognize Java keywords
	if (java_indent > 0)
		c_indent = java_indent;
}

c_compute_indent()
{
	int ind = 0;		/* indentation to use */
	int orig = point;
	int first_char = curchar(); /* cur line's first char */
	int pstart;		/* start position of prev line */
	int pafter;		/* after label on prev line */
	int pend;		/* end position of prev line */
	int prev_end = ';';	/* last char on previous line */
	int norm_narrow;	/* narrowing setting for most parsing */
	int in_initializer;

	save_var point;
	if (setjmp(&c_indent_giveup))	/* if we hit end & gave up, */
		return ind;		/* use best guess so far */
	if (in_perl_buffer)
		return perl_compute_indent();
	norm_narrow = fix_narrowing(point - c_look_back);
	save_var narrow_start = norm_narrow;
	if (!reindent_c_comments && parse_string(1, "[ \t\f]*/<*>"))
		return get_indentation(point);
	if (!reindent_one_line_c_comments && parse_string(1, "[ \t\f]*//"))
		return get_indentation(point);
	if (!reindent_c_preprocessor_lines && parse_string(1, "[ \t\f]*#"))
		return get_indentation(point);
	if (first_char == '#' || first_char == '\f')
		return 0;
	else if (first_char == '}')
		return cind_close();

	to_begin_line();
	if (parse_string(-1, PREPROC_ELSE ".*\n")) {
		point = matchend;
		locate_by_ifdef(-1);
	}
	if (character(point - 2) == '\\')
		if (!indent_preprocessor_contin)
			return indent_after_backslash();
		else if (!prior_contin_code(orig))
			return indent_preprocessor_contin;
		else {	// Note we're indenting preprocessor continuation line now.
			save_var indent_preprocessor_contin = 255;
			save_var narrow_start = point - parse_string(-1, "(\n.*\\)*\n");
		}
	if (parse_string(1, "[ \t]*\n") && after_many_comments()) {
		re_search(-1, "[^ \t\n]");	/* find previous non-blank line */
		return get_indentation(point);
	}
	cind_space();			/* skip whitespace and comments */
	if (character(point - 1) == ':')
		move_before_labels();
	prev_end = character(point - 1); /* get last relevant char */
	pend = point;

	to_statement_start();		/* if contin line, back up */
	pstart = point;			/* now at start of biggest statement */
	past_labels();			/* containing original point */
	pafter = point;			/* just after label */
	ind = current_column();
	narrow_start = norm_narrow;	/* ok to look far back again */

	if (is_case(orig)) {
		ind = indent_case(orig);
	} else if (is_dispatch_table(point)) {
		// Found MFC dispatch table, leave alone.
	} else if (have_label(orig)) {
		char type[FNAMELEN];		// Could it be a bitfield, not a label?
		find_c_func_info(type, NULL, NULL, 0);
		if (!strcmp(type, "struct") || !strcmp(type, "union"))	// Yes, a bitfield.
			ind += get_c_indent();
		else					// It's a label.
			ind = c_label_indent;
	} else if (prev_end == ':' && is_label(pstart)) {
		ind = get_column(pstart);
		if (!(Closeback && just_open(orig)))
			ind += get_c_indent();
		if (just_open(orig))
			ind += c_brace_offset;
	} else if (is_else(orig)) {
		point = orig;
		ind = indent_else(ind);
	} else if (prev_end == ';') {
		if (just_open(orig) && func_start_block(pstart))
			ind = c_top_braces;
		else
			ind = indent_after_statement(orig);
	} else if (prev_end == '}') {
		point = pend;
		ind = indent_after_block(orig);
	} else if (prev_end == '{') {
		if (!c_indent_after_extern_c
			  && parse_string(1, "[ \t]*extern[ \t]*\"C\"[ \t\n]*{"))
			return ind;
		if (!c_indent_after_namespace
			  && parse_string(1, "[ \t]*namespace</word>.*{"))
			return ind;
		if (ind <= c_top_braces	// special case if no previous indent
		    || get_column(pend - 1) <= c_top_braces) {
			ind += top_indent(prev_end) ? get_c_indent() : c_top_braces;
			if (just_open(orig))
				ind += c_brace_offset;
			if (ind < c_top_braces)
				ind = c_top_braces;
		} else if (have_struct_init(pend - 1)) {
			ind = indent_struct_init(pend - 1) + get_c_indent();
		} else if (Closeback || curchar() != '{')
			ind += get_c_indent();
		if (just_open(pafter))
			ind -= c_brace_offset;
	} else if (just_open(orig) || matches_at(orig, 1, "[ \t\f]*{[ \t\f]*}")) {
		if (have_struct_init(orig))
			return indent_struct_init(orig);
		else if (have_other_init(orig))
			return ind;
		else
			return indent_just_open(ind);
	} else if (in_initializer = in_c_initializer(orig)) {
		return ind;
	} else if (is_contin_statement(pend, 1, orig)) {
		return indent_c_contin(ind, orig, pend, in_initializer);
	} else if (!index(":);{}", c_line_before()) && !c_statement_start()
				&& !(Closeback && just_open(point)))
		return ind;
	else
		ind += get_c_indent();
	return ind;
}

/* Indent continuation line. */
indent_c_contin(ind, orig, lastl, in_initializer)
{
	int c, plevel = 0, qlevel = 0, findc, alevel = 0;

 	if (c_align_inherit >= 0 && matches_at(orig, -1, ":\n[ \t]*")
 		  && !c_matches_in(give_begin_line(), lastl, "%?"))
 		return get_indentation(point) + c_align_inherit;
	if (!c_align_contin_lines)	// never line up
		if (!in_initializer)
			return c_default_indent(ind, orig, 1);
	to_begin_line();
	save_var narrow_start = point;
	point = orig;
	findc = character(orig);
	if (c_align_open_paren && parse_string(1, "%(([ \t]|//|/%*)*$"))
		return ind;
	while (re_search(-1, "[()?:\"'\n<>]|[=!<>]?=")) {
		c = curchar();
		if (in_perl_buffer && (in_perl_comment(point) || strchr("\"'\n", c)))
			continue;	// Skip over comments differently.
		switch (c) {
			case '(':	/* indent after unmatched lparen */
				if (plevel-- > 0)
					break;
				if (c_align_open_paren && findc != ')')
					return ind + get_c_indent();
indhere:		if (findc != ')')
					point++;
				if (!parse_string(1, "[ \t\f]*/[*/]"))
					point += parse_string(1, "[ \t\f]*![^ \t\f\n]");
				return limit_c_contin_indent(ind);
			case ')':
				plevel++;
				break;
			case '?':	/* indent : to matching ? */
				if (qlevel-- > 0 || plevel > 0 || findc != ':')
					break;
				point--;
				goto indhere;
			case ':':
				qlevel++;
				break;
			case '=': case '!':
				if (character(point + 1) == '=' || plevel > 0
				    || character(lastl - 1) == ',' || findc == ')')
					break;
				goto indhere;
			case '\n':		 // skip over comments
				cind_try_space(); // at ends of lines
				break;
			case '"': case '\'':
				point++;
				c_skip_item(c, RE_REVERSE);
				break;
			case '<':
				if (character(point + 1) == '=')
					break;
				if (alevel-- > 0 || !(c_extra_keywords & (CPP_KEYWORDS
					| JAVA_KEYWORDS | CSHARP_KEYWORDS)))
					break;
				// Someday, test if there's an identifier before the <
				// that's the name of a template, not just if it's a
				// template keyword.
				if (parse_string(-1, "</word>template[ \t\n]*")) {
					point++;
					point += parse_string(1, "[ \t]*!<word>");
					return limit_c_contin_indent(ind);
				}
				break;
			case '>':
				if (character(point + 1) != '=')
					alevel++;
				break;
		}
	}
	if (c_align_inherit >= 0 && c_matches_in(point, orig, ":[ \t]*($|//|/%*)")
		  && !c_matches_in(point, orig, "%?"))
		return get_indentation(point) + c_align_inherit;

	if (character(lastl - 1) == ',') {
		if (character(lastl - 2) != ')') {
			point = lastl;
			to_begin_line();
		}
		point += parse_string(1, "(.*[ \t\f]on)?[ \t\f]+");
		if (in_c_comment(c_look_back))
			to_indentation();
		if (curchar() == '#')
			past_initial_preprocessor();
		return limit_c_contin_indent(ind);
	}
	return c_default_indent(ind, orig, 2);
}

c_default_indent(ind, orig, defmult)
{
	if ((c_extra_keywords & CSHARP_KEYWORDS) && parse_string(1, "[ \t]*%[")
		  && have_csharp_attribute(point))
		return get_indentation(point);
	if (matches_at(orig, 1, "[ \t]*//"))
		return ind;
	if ((c_extra_keywords & (CPP_KEYWORDS | JAVA_KEYWORDS | CSHARP_KEYWORDS))
		  && matches_at(orig, 1, "[ \t]*class</word>"))
		return ind;		// Put class keyword in continuation back on left.
	if (ind)
		return ind + defmult * get_c_indent() + c_contin_offset;
	if (parse_string(1, "[ \t\f]*(struct|union|class|namespace)[ \t\f\n]+"
			 "[a-zA-Z0-9_]*[ \t\f\n]*{[ \t\f\n]*[\n/]"))
		return ind + c_top_struct;
	if (parse_string(1, "[ \t\f]*(template)[ \t\f\n]*%<"))
		return ind + c_top_template;
	if (just_open(orig))
		return c_top_braces;
	point = orig;
	cind_space();
	return is_contin_statement(point, 0, orig)
			? c_top_contin : c_param_decl;
}

int c_rev_skip_level(left_d, right_d)
char left_d, right_d;
{
	char pattern[40], ch;
	int level = 0;

	if (in_perl_buffer)
		return perl_skip_level(-1, left_d, right_d);
	sprintf(pattern, "<*>/|[%c%c'\"]|//.*\n", right_d, left_d);

bgn:	do {
		if (!re_search(RE_REVERSE, pattern))
			return 0;
		switch (ch = curchar()) {
			case '/':
				if (at_c_comment(point + 2, 100) & IN_STRING)
					to_end_line();
				break;
			case '\'': case '"': case '*':
				point = matchstart;
				if (ch != '*' && (in_c_comment(100) & IN_COMMENT)) {
					  point--;
					  break;
				}
				c_skip_item(character(point - 1), RE_REVERSE);
				goto bgn;
			default:
				if (ch == right_d)
					level++;
				else if (ch == left_d)
					level--;
		}
	} while (level > 0);
	return 1;
}

position_on_ifdef()
{
	to_begin_line();
	if (point < window_start || point > window_end) {
		int line = text_height() - show_tag_line;
		if (parse_string(1, PREPROC_IF))
			line = show_tag_line;
		else if (parse_string(1, PREPROC_ELSE))
			line /= 2;
		window_start = prev_screen_line(line);
	}
}

// Move to next/previous #if/#else/#endif or similar directive.  Return
// 1 if we didn't start adjacent to a preprocessor line, and simply
// moved to the next/previous one.  Return 2 if we started on a
// preprocessor line and have now moved to a matching one, or 0 if
// there was no matching one.
int locate_by_ifdef(int dir)
{
	int level = 0;

	to_begin_line();
	if (!parse_string(1, PREPROC_LINE) ||
		  parse_string(1, (dir > 0) ? PREPROC_ENDIF : PREPROC_IF)) {

// If we're not starting from a preprocessor line, or we're moving out of
// the scope of the current line's directive, just move to the next.

		if (dir > 0)
			to_end_line();
		re_search(dir, PREPROC_LINE);
		return 1;
	}
	if (dir > 0)
		to_end_line();
	for (;;) {		// Look for a preprocessor line at the right level.
		if (!re_search(dir, PREPROC_LINE))
			return 0;
		if (parse_string(-dir, PREPROC_IF)) {
			if ((dir > 0) ? (++level <= 0) : (++level > 0))
				break;
		} else if (parse_string(-dir, PREPROC_ENDIF)) {
			if ((dir > 0) ? (--level < 0) : (--level >= 0))
				break;
		} else if (parse_string(-dir, PREPROC_ELSE)) {
			if ((dir > 0) ? (level <= 0) : (level >= 0))
				break;
		}
	}
	return 2;
}

// Move forward or backward stopping at #if, #else, or #endif (or similar)
// but skipping over inner nested preprocessor lines.

move_by_ifdef(dir)
{
	if (!locate_by_ifdef(dir))
		error("Unmatched preprocessor line.");
	position_on_ifdef();
}

select_menu_item(resbuf, menuwin, owin, dir)
{	// let user select line from menu window, maybe copy line to resbuf
	int res; // dir tells how to srch on norm keys, or 0 if they're cmds
	char str[FNAMELEN], pat[FNAMELEN * 2];

	save_var mode_keys;
	cursor_to_column = -1;
	do {
		mode_keys = menu_tab;
		res = do_recursion();
		if (!is_window(menuwin))
			res = INP_ABORT;
		else if (res == INP_ADD && search_in_menu && dir) {
			window_handle = owin;
			run_topkey();
			window_handle = owin;
			grab(0, FNAMELEN - 1, str);
			sprintf(pat, "^%q", str);
			window_handle = menuwin;
			point = (dir == 1) ? 0 : size();
			re_search(dir, pat);
			to_begin_line();
			res = 0;
		} else if (res == INP_ADD && !dir)
			switch (isalpha(key) ? toupper(key) : 0) {
				case 'N':
				case 'P':	ungot_key = CTRL(key);
				default:	res = 0;	// fall through
				case 'Q':	break;
			}
	} while (!res);
	switch (res) {		// try all other chars
		case INP_ADD:	ungot_key = full_key;
				break;
		case INP_GET_EXIT:
		case INP_GET:	if (resbuf)
					get_menu_line(resbuf);
				break;
	}
	if (res == INP_ABORT && !user_abort)	// user deleted edit window?
		user_abort = 1;
	remove_line_highlight();
	window_handle = owin;
	remove_window(menuwin);
	return res;
}

view_linked_buf(buf, last, linker)
int (*linker)();
{
	int win, res, height, prev;

	remove_final_view();
	prev = window_handle;
	if (!last)
		save_var _doing_input = DI_VIEW, bufnum, window_handle;
	height = screen_lines - _view_top - _view_bottom;
	win = add_popup(_view_left, _view_top,
					screen_cols - _view_left - _view_right, height,
					_view_border, buf);
	go_to_link = linker;
	if (linker) {
		create_links();
		mouse_dbl_selects = 1;
	}
	if (is_gui && want_gui_prompts)
		if (!linker)
			one_window_to_dialog(_view_title ? _view_title : "No Title", win,
								 "OK", "Cancel", "");
		else
			one_window_to_dialog(_view_title ? _view_title : "No Title", win,
								 "", "Done", "");

	text_color = (is_gui && want_gui_prompts) ? color_class viewed_gui_text
											  : color_class viewed_text;
	set_wattrib(win, PVERTBORDCOLOR, color_class viewed_border);
	set_wattrib(win, PHORIZBORDCOLOR, color_class viewed_border);
	if (_view_title)
		window_title(win, BTOP, TITLECENTER, _view_title);
	point = 0;
	window_to_fit(height);
	if (last) {
		_view_prev_input = _doing_input;
		_view_prev_window = prev;
		_view_prev_mode = mode_keys;
		mode_keys = view_tab;		/* switch buf to view mode */
		_doing_input = DI_VIEWLAST;
		_window_flags &= ~FORCE_MODE_LINE;
		display_more_msg(win);
	} else {
		res = view_loop(win);
		if (res == INP_EXIT && !IS_ENTER_KEY(key))
			ungot_key = full_key;
		remove_window(win);
	}
	user_abort = 0;
}

// Check for program prog (if nonnull), then run the command line to
// uncompress stdin to stdout.
do_decompression(char *prog, char *cmdline)
{
	char fname[FNAMELEN], *s;

	if (GET_LINE_TRANSLATE(translation_type) != FILETYPE_BINARY)
		return;
	if (is_remote_file(filename) && is_retrieving_remote_file(bufnum))
		return;
	if (prog && *prog) {
		sprintf(fname, "%s%s", prog, is_unix ? "" : ".exe");
		if (!(s = getenv("PATH")) || !look_on_path(fname, 0, s))
			return;
	}
	save_var _buf_readonly = 0;
	force_save_as = FSA_READONLY;
	if (!pipe_text(bufname, bufname, cmdline, NULL,
				   PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0))
		error("Could not uncompress %s", filename);
	point = modified = 0;
	strcpy(fname, filename);
	if (s = get_extension(fname)) {
		*s = 0;
		call_mode(fname);
	}
}

dired_do_delete(name, to, in_dired)
char *name, *to;
{
	int isdir = in_dired ? dired_isdir() : 0;

	to = to;
	if (is_remote_file(name)) {
		buffer_printf(REMOTE_FTP_OP_BUFFER,
					  isdir == 1 ? "RMD %s\n" : "DELE %s\n",
					  get_url_file_part(name, 0));
	} else {
		if (in_dired)
			noteput("Deleting %f.", name);
		if ((isdir && dired_isdir() == 1) ? dired_warn_rmdir(name)
			  : dired_warn_delete(name))
			return 1;
	}
	if (in_dired) {
		save_var modified;	// leave modify bit alone
		to_begin_line();	// but delete line
		delete(point - 1, give_end_line());
	}
	return 0;
}

dired_do(op, tag, count, func, to)
char *op, *tag, *to;
int (*func)();
{
	spot final;
	int bad = 0, tried = 0;
	char name[FNAMELEN];

	final = alloc_spot();
	point = 0;
	save_var abort_searching = 0;
	while (search(1, tag)) {
		if (user_abort) {
			if (!bad)
				*final = point;
			break;
		}
		dired_getname(name, 3);
		tried++;
		if ((*func)(name, to, 1))
			if (!bad++)		/* error occurred */
				*final = point;	/* 1st error */
	}
	point = *final;
	free_spot(final);
	move_to_column(goal_column);
	if (bad)
		error("%d of %d file%s not %s", bad, tried,
			tried != 1 ? "s" : "", op);
	say("%d file%s %s", count, count != 1 ? "s" : "", op);
}

do_shell_execute(char *fname, char *op, char *descr)
{
	char dirname[FNAMELEN * 3];

	strcpy(dirname, fname);
	*get_tail(dirname, 1) = 0;
	if ( do_run_viewer(fname, op, dirname) == 0) {
		say("Could not %s for \"%f\"", descr, fname);
		get_executable_directory(dirname);
		strcat(dirname, "epsilon-viewer");
		if (is_unix && !check_file(dirname)) {
			int b = zap("-shell-execute-error");
			buf_printf(b, "Epsilon runs an external program named "
					   "epsilon-viewer to provide\n");
			buf_printf(b, "this feature.  It requires Gnome shared "
					   "library files, which may\nbe missing.\n\n");
			buf_printf(b, "Read the source file epsilon-viewer.c in"
					   " Epsilon's bin subdirectory\n");
			buf_printf(b, "for more details.\n");
			save_var _view_title = " Can't run epsilon-viewer ";
			view_buf(b, 1);
		}
	}
}

dired_run_viewer()
{
	char fname[FNAMELEN];

	dired_getname(fname, 1);
	do_shell_execute(fname, "", "run viewer"); // Def. action might not be Open.
}

command quick_dired_command()
{
	char file[FNAMELEN], rel[FNAMELEN], msg[FNAMELEN], dest[FNAMELEN];
	char dir[FNAMELEN];

	if (is_dired_buf()) {
		dired_getname(file, 3);
		dired_try_getdir(dir, 1);
		save_var inp_get_other_name = file;
	} else {
		strcpy(file, filename);
		strcpy(dir, filename);
		*get_tail(dir, 0) = 0;
	}
	if (!*file)
		error("This buffer has no associated file.");
	strcpy(rel, get_tail(file, 0));
again:
	do {
		sayput("%s: D=Delete, Move, Copy, View, T=Props, .=Dired, "
			   "G=Go dir, ?=Show more: ", rel);
		getkey();
		check_abort();
	} while (!IS_TRUE_KEY(key));
	say("");
	if (is_remote_file(file) && exist(REMOTE_FTP_OP_BUFFER))
		zap(REMOTE_FTP_OP_BUFFER);
	switch (toupper(key)) {
		case '.':
			dired_one(file);
			break;
		case 'D':
			sprintf(msg, "Delete %s? ", rel);
			if (!ask_yn("Delete file", msg, "Delete", "Cancel", 1))
				break;
			save_var dired_confirmation;
			if (dired_do_delete(file, NULL, 0)) {
				check_abort();
				file_error(errno, file, "Could not delete");
			} else if (!is_remote_file(file))
				say("%f deleted.", file);
			break;
		case 'C':
			if (no_remote_copy(file))
				error("Copy operation not implemented for remote files.");
			if (is_directory(file) == 1)
				error("Can't copy directories.");
			sprintf(msg, "Copy %s to: ", rel);
			get_file_dir(dest, msg);
			if (*dest) {
				switch (dired_do_copy(file, dest, 0)) {
					case 1:
						file_error(errno, dest, "Could not copy");
						break;
					case 0:
						say("%f copied.", file);
						break;
				}
			}
			break;
		case 'M':
			sprintf(msg, "Move %s to: ", rel);
			get_file_dir(dest, msg);
			if (*dest) {
				if (dired_do_move(file, dest, 0))
					file_error(errno, dest, "Could not move");
				else if (!is_remote_file(file))
					say("%f moved.", file);
			}
			break;
		case 'V':
			do_shell_execute(file, "Open", "run viewer");
			break;
		case 'T':
			do_shell_execute(file,  "Properties", "show properties");
			break;
		case 'F':
			do_shell_execute(dir, "Open", "run viewer");
			break;
		case 'G':
			if (chdir(dir))
				file_error(errno, dir, "unknown error");
			else {
				getcd(dir);
				say("Current directory: %f", dir);
			}
			break;
		case '+':
			do_create_directory(0);
			break;
		case '!':
			do_shell_command(file, dir);
			break;
		case 'A':
			dired_do_set_attributes(file, dir);
			check_dates(0);
			break;
		default:
			if (key == '?' || key == HELPKEY || key == FKEY(1))
				help_on_command("quick-dired-command");
			else
				say("Invalid key.");
			goto again;
	}
	if (is_remote_file(file))
		send_remote_ftp_op(file);
}

dired_try_getdir(fullname, showerr)	// grab directory, add / or \ to end
char *fullname;		// if we can't find, set fullname "" if !showerr
{			// return nonzero if ok
	int i;
	char *p, path_char;

	*fullname = 0;
	save_var point;
	to_end_line();
	if (!re_search(-1, DIRLINEPAT))	// find prev "Directory of "
		if (showerr)		// bad dired buf
			error("Can't determine directory of files");
		else
			return 0;
	point = matchstart;
	re_search(1, " *");		// skip initial spaces
	parse_string(1, FILECHARS, fullname);
	path_char = is_remote_file(fullname) ? '/' : path_sep;
	i = strlen(fullname);
	if (is_remote_file(fullname))
		while (i > 0 && !is_path_separator(fullname[i - 1]))
			fullname[--i] = 0;
	p = fullname + i;
	if (i > 0 && !is_path_separator(fullname[i - 1])
		  && (strcmp(fullname, "~") || !is_unix))
		p[0] = path_char, p[1] = 0; // add final / or \ if none
	return 1;
}

scp_next_cmd(struct scp_job *p)
{
	char *s;

	if (p->flags & SCP_CMD_ERROR) {
		p->cmd = 0;
		if (p->destbuf)
			set_buf_mode_message(p->destbuf, NULL);
	}
	if (p->cmd && *p->cmd) {
		p->busy = 1;
		if (p->flags & SCP_CLIENT_MODE)
			buf_printf(p->buf, "## ");
		if (s = strchr(p->cmd, '\n')) {
			buf_stuff(p->buf, p->cmd, s++ - p->cmd + 1);
		} else {
			buf_stuff(p->buf, p->cmd, strlen(p->cmd));
			buf_stuff(p->buf, "\n", 1);
		}
		p->cmd = s;
	}
}

int scp_sftp_handle_list(struct scp_job *p)
{
	int i, isdata;
	char msg[FNAMELEN];

	save_var case_fold = 1;
	while (nl_forward()) {
		i = point;
		point--;
		to_begin_line();
		isdata = 0;
		if (parse_string(1, "[-dcblsp][-r][-w][-sxt][-r][-w][-sxt]"
						 "[-r][-w][-sxt] .*" UNIX_DATE))
			isdata = 1;
		if ((p->flags & SCP_CHECK_IS_DIR) && parse_string(1, ".* not a directory"))
			isdata = 1;
		if (parse_string(1, ".* No such file or directory")
			  && p->cmd && !strncmp(p->cmd, "ls ", 3)) {
			if (buf_exist(p->destbuf))	// Error part way through, so don't
				set_buf_mode_message(p->destbuf, NULL);	// copy any more.
			p->flags |= SCP_CMD_ERROR;
		}
		if (isdata) { // Got some data, copy where it belongs.
			if (buf_exist(p->destbuf) && !(p->flags & SCP_CMD_ERROR)) {
				if (*p->file && !(p->flags & SCP_CHECK_IS_DIR))
					buf_printf(p->destbuf, "%s", p->file);
				buf_xfer(p->destbuf, point, i);
			}
		} else if (parse_string(1, "Remote (working )?directory(:| is| is now) ")) {
			grab(matchend, give_end_line(), msg);
			if (p->lastdir)
				free(p->lastdir);
			p->lastdir = strsave(msg);
			if (!p->homedir)
				p->homedir = strsave(msg);
		} else if ((p->flags & SCP_IS_XFER)	// Ignore this msg.
				   && parse_string(1, "((Fetching|Uploading) .* to |"
								   "(remote|local):.* => (remote|local):|"
								   ".*--:-- ETA)")) {
		} else if (!parse_string(1, "(Exited|Connecting to "
								 "|Connection closed|Listing directory "
								 "|Using username \")")) {
			grab(point, MIN(give_end_line(),
							point + sizeof(msg) / sizeof(*msg) - 1), msg);
			if (*msg)		// Any other line is an error msg.
				say("%s", msg);
			if ((p->flags & SCP_IS_XFER)
				  && parse_string(1, ".*: (Permission denied|No such file"
								  "|Device out of space|Read-only file system"
								  "|No such file)"))
				p->flags |= SCP_CMD_ERROR;
		}
		point = i;
	}
	if (parse_string(-1, "p?sftp%> *")) {
		if (!p->cmd || !*p->cmd) {
			if (buf_exist(p->destbuf))
				set_buf_mode_message(p->destbuf, NULL);
			if (p->flags & SCP_IS_XFER)
				scp_handle_done(p);
		}
		p->busy = 0;			// Got a prompt for another cmd.
		return 1;			// Send a pending command if any.
	}
	return 0;
}

// Handle activity in scp buffer.
scp_handler(int activity, int buf, int from, int to)
{
	struct scp_job *p;

	save_var bufnum = buf;
	for (p = scp_job; p; p = p->next)	// Find the data for this buffer.
		if (bufnum == p->buf)
			break;
	if (!p || !p->destbuf)
		return;
	if (activity == NET_DONE)
		scp_handle_done(p);
	else if (activity == NET_RECV) {
		if (p->op == FTP_LIST || p->op == FTP_MISC || (p->flags & SCP_IS_XFER))
			scp_handle_list(p, from, to);
		else
			scp_handle_xfer(p, from, to);
		if (matches_at(to, -1, "( |^)pass(word|phrase): ?"))
			say("Password requested in buffer %s", bufname);
	}
	refresh();
}

maybe_auto_read(struct file_info *now)
{
	int pos;

	// Only auto-read if size on disk is > 90% of current size, or
	// old file is very small, and new file's not huge.
	if (!modified && (now->fsize / 9 >= size() / 10 || size() < 300)
		  && now->fsize < 10000000) {
		pos = point;
		if (!read_file(filename, reading_translation_type()))
			say("%f: Modified file read from disk.", filename);
		else
			auto_read_changed_file = _force_auto_read_changed_file = 0;
		point = pos;
		to_begin_line();
		maybe_add_readonly_once(now);
		return 1;
	}
	return 0;
}

check_dates(save)
{
	struct file_info now, *then = &file_info;
	int ret = 1, temp, orig = bufnum, pos;

	if (num_of_popups() > 0)	// Don't pop up over bufed or similar.
		return 0;
	if (file_date_skip_drives && check_dates_skip_drive())
		return 0;
	ignoring_file_change = 0;	// Resume checking on getting focus.
	if (!want_warn || !filename || then->check_type != CHECK_FILE
			|| check_file(filename, &now) != CHECK_FILE)
		return 0;
	update_readonly_warning(&now);
	if ((readonly_warning & ROWARN_REFRESH) && !save
		  && (file_info.attr ^ now.attr) & ATTR_READONLY)
		return 2;

	if (!compare_dates(&now, then)) {
		maybe_add_readonly_once(&now);
		return 0;
	}
	switch_to_buffer(bufnum);
	maybe_add_readonly_once(&now);

	if ((auto_read_changed_file || _force_auto_read_changed_file)
		&& !save && maybe_auto_read(&now))
		return 1;
	if ((readonly_warning & (ROWARN_GREP | ROWARN_REFRESH)) && !save)
		return 2;
	tiled_only();
	switch_to_buffer(bufnum);
	maybe_ding(bell_on_date_warning);
	temp = tmp_buf();
	save_var _view_right = 0, _view_bottom = screen_lines - 30;
	save_var return_raw_buttons = 1;
	while (!user_abort) {
		show_date_warning(temp, orig, then, &now, save);
		bufnum = orig;
		pos = point;
		if (is_gui && want_gui_prompts) {
			int win, owin = window_handle;

			remove_final_view();
			save_var _doing_input = DI_VIEW;
			win = add_popup(0, 0, 75, 15, _view_border, temp);
			go_to_link = 0;
			display_dialog_box("FileDateBox2", "File has changed on disk",
							   win, 0, 0,
							   "Compare", "Read", save ? "Save" : "Ignore");
			text_color = (is_gui && want_gui_prompts) ? color_class text
						 : color_class viewed_text;
			point = 0;
			view_loop(win);
			switch (key_is_button) {
				case 1:	key = 'C'; break;
				case 2:	key = 'R'; break;
				case 3:	key = save ? 'S' : 'I'; break;
			}
			window_handle = owin;
			remove_window(win);
			if (user_abort)
				break;
		} else {
			view_buf(temp, 0);
			if (key == abort_key)
				user_abort = 1;
			if (user_abort)
				break;
			getkey();
		}
		char junk[FNAMELEN];
		if (key == WIN_DRAG_DROP)	// Ignore dropped files here.
			while(drag_drop_result(junk))
				;
		key = toupper(key);
		if (key == 'C') {
			ignoring_file_change = 2;	// Don't check for a while.
			refresh();
			split_window_vertically();
			window_number--;	/* to 1st window */
			find_in_other_buf(filename, default_translation_type);
			point = pos;
			to_begin_line();
			set_buffer_filename("");
			error("This buffer has disk version.");
			break;		/* abort out of calling func */
		} else if (key == 'R') {
			refresh();
			if (!read_file(filename, reading_translation_type()))
				say("%s: Read from disk.", filename);
			point = pos;
			to_begin_line();	// Don't check immediately after dialog ends.
			ignoring_file_change = is_gui ? 2 : 0;
			break;
		} else if (key == (save ? 'S' : 'I')) {
			if (key == 'I')
				ignoring_file_change = 100;	// Don't check on getting focus.
			else
				ignoring_file_change = 2;	// Don't check immediately after dialog ends.
			refresh();
			ret = 0;
			break;
		} else if (key == '!')
			want_warn = !want_warn;
		else if (key == 'A')
			ask_read_changed_file();
		say("");
	}
	buf_delete(temp);
	return user_abort ? 1 : ret;
}

format_file_date(char *s, struct file_info *p)
{
	format_date(s, p->year, p->month, p->day, p->hour, p->minute, p->second);
}

prepare_save_file()
{
	int i = get_mode_variable("%s-delete-trailing-spaces");

	if (got_bad_number)
		i = default_delete_trailing_spaces;
	if (i && GET_LINE_TRANSLATE(translation_type) != FILETYPE_BINARY)
		simple_re_replace(1, "[ \t]+$", "");
}

do_save_file(backup, checkdate, getdate) /* 0=ok, 1=err, 2=user canceled */
{				/* always prints status in echo area */
	struct file_info *finfo;
	char othername[FNAMELEN], msg[FNAMELEN];
	int err;

	iter = 0;
	if (!*filename || force_save_as)
		return write_file();
	if (checkdate)
		if (check_dates(1))
			return 2;
	if ((buffer_flags(0) & BF_UNICODE) && !GET_ENCODING(translation_type)
		  && (has_feature & FEAT_UNICODE)) {
		if (!buffer_has_pattern("<#256-#65535>"))
			save_var translation_type = SET_ENCODING(translation_type, 1);
		else		// Force to raw temporarily.
			pick_write_encoding();	// Have 16-bit chars, get an encoding for them.
		if (!GET_ENCODING(translation_type))
			error("Invalid encoding.");
	}
	if (_read_aborted) {
		relative(filename, othername);
		sprintf(msg, "Only part of %s was read from disk.  "
			"Save anyway? ", othername);
		if (!ask_yn("Confirm save", msg, "Save", "Cancel", 0))
			return 2;
	}
	if (is_retrieving_remote_file(bufnum)) {
		relative(filename, othername);
		sprintf(msg, "%s is still being downloaded.  "
				"Save partial file now? ", othername);
		if (!ask_yn("Confirm save", msg, "Save", "Cancel", 0))
			return 2;
	}
	if (backup)
		do_write_backup();
	prepare_save_file();
	finfo = getdate ? &file_info : 0;
	if (err = file_convert_write(filename, translation_type, finfo)) {
		auto_read_changed_file = 0;
		file_info.check_type = 0;	// Prevent checking date.
		file_error(err, filename, "write error");
		maybe_ding(bell_on_write_error);
	} else {
		if (!save_without_prompt && backup != 2)
			note_written(filename);
		undo_set_flag(UNDO_WRITE_FILE);
		if (finfo)
			update_readonly_warning(finfo);
	}
	return (err != 0);
}

do_save_all_buffers(int newfile_bufs, int *count, int *failed)
{
	int i;

	save_var bufnum, window_bufnum;
	i = buf_list(0, 0);
	do {
		bufnum = i;
		if (!*filename || !modified || buffer_not_saveable
			  || is_retrieving_remote_file(i))
			continue;
		if (newfile_bufs ? (force_save_as != FSA_NEWFILE)
						 : (force_save_as == FSA_NEWFILE))
			continue;	// Only do one group.
		if (newfile_bufs) {
			if (save_all_without_asking) {
				buffer_printf(MSG_BUFFER, "save-all-buffers: Could not"
							  " save buffer %s, no file name\n", bufname);
				(*failed)++;
				continue;
			}
			to_buffer_num(i);	// Show the buffer we'll be prompting for.
		}
		switch (do_save_file(1, 1, 1)) {
			case 1:
				buffer_printf(MSG_BUFFER, "save-all-buffers: Could not save %s\n",
							  filename);
				(*failed)++;
				break;
			case 0:
				show_text(-2, 1, "save-all-buffers: Saved %s", filename);
				if (!save_without_prompt)
					(*count)++;
				break;
		}
	} while (i = buf_list(1, 1));
}

get_grep_file(inbufs, filepat)
char *filepat;
{
	char pr[FNAMELEN], dir[FNAMELEN];
	int show_rel = 1;		// Show the relative path of the default.

	if (!inbufs) {
		if (is_dired_buf() && !inp_get_other_name)
			save_var inp_get_other_name = get_dired_buffer_pattern();
		if (grep_default_directory == 1 && !grep_prompt_with_buffer_directory
			  || grep_default_directory == 3)
			show_rel = 0;
		build_prompt(pr, "In file(s)", _grep_files, 0, show_rel);
		save_var prompt_with_buffer_directory = grep_prompt_with_buffer_directory;
		get_file_dir(filepat, pr);
		if (!*filepat && grep_default_directory) {
			save_var _help_on_cmd = last_index;
			if (grep_default_directory == 3)
				strcpy(filepat, _grep_files);
			else
				get_last_recall(f_match, filepat);
			if (grep_default_directory >= 2 && *filepat
				  && get_buffer_directory(dir)) {
				if (grep_default_directory != 3)
					absolute(filepat, dir);
				return 1;
			}
		}
		if (*filepat) {
			if (is_directory(filepat))
				sprintf(filepat + strlen(filepat), "%c*", path_sep);
			strcpy(_grep_files, filepat);
		}
		if (grep_default_directory == 3)
			strcpy(filepat, _grep_files);
		else
			relative(_grep_files, filepat);
		if (grep_default_directory == 1 && !grep_prompt_with_buffer_directory)
			strcpy(_grep_files, filepat);
		if (!*filepat)
			return 0;
	} else
		comp_read(filepat, "In buffers [pattern]: ", b_match,
				  ((preserve_filename_case & 4) ? FM_FOLD : 0), bufname);
	return 1;
}

maybe_indent_rigidly(int rev)
{
	if (!is_highlight_on()) {
		if (!rev && full_key == NUMSHIFT(GREYTAB)) {
			back_to_tab_stop();
			return 1;
		}
		return 0;
	}
	switch (rev) {
		case 0: rev = IS_SHIFT_KEY(full_key) != 0; break;
		case 2: rev = 0; break;
	}
	if (resize_rectangle_on_tab && is_highlight_on() == REGRECT) {
		int tab = get_soft_tab_size();
		to_virtual_column((virtual_column() + (rev ? -1 : tab)) / tab * tab);
		iter = 0;
		return 1;
	}
	if (rev) {
		if (!has_arg)
			iter = -get_soft_tab_size();
		else if (iter > 0)
			iter = -iter;
		has_arg = 1;
	}
	indent_rigidly();
	highlight_on();
	shift_selecting = 2;
	return 1;
}

await_menu_click()	// show menu, return 1 if user clicked on it
{
	save_var mouse_cursor = &std_pointer;
	load_menu();
	prep_menu_bar();
	save_var window_handle = menu_window;
	do {
		move_to_column(mouse_x);
		maybe_refresh();
		getkey();
	} while (key == MOUSE_MOVE && still_on_menu_bar());
	remove_window(menu_window);
	if (IS_MOUSE_DOWN(key) && still_on_menu_bar())
		return 1;
	ungot_key = full_key;
	return 0;
}

show_menu_loop(num, cmd)
char *cmd;
{
	int sub = 0, line = 0, max, isdir, oldsub, orignum;

	max = max_submenus();
	if (IS_MOUSE_KEY(key))
		ungot_key = full_key;
	do {
		oldsub = sub;
		sub = to_sub_line(sub, num, line);
		if (getkey() == MOUSE_MOVE) {
			refresh();
			while (getkey() == MOUSE_MOVE)
				;
		}
		isdir = 1;		// assume it's a direction key
		switch (get_direction()) {
			case BTOP:	line--; break;
			case BBOTTOM:	line++; break;
			case BLEFT:	num--, line = 0; break;
			case BRIGHT:	num++, line = 0; break;
			default:	isdir = 0; break;
		}
		if (isalpha(key))
			line = do_menu_search(key);
		else if (IS_MOUSE_DOWN(key)) {
			if (oldsub && mouse_y <= 0)	// maybe drop menu
				orignum = count_submenus(mouse_x) - 1;
			else
				orignum = -2;	// upclick on bar won't drop
			sub = while_mouse_pressed(sub);
			num = after_mouse_pressed(sub, num, orignum);
			line = 0;
		} else if (binding_of(key, reg_tab) == (short) help) {
			get_menu_cmd(cmd, sub);
			if (*cmd && find_index(cmd))
				help_on_command(cmd);
			isdir = 1;
		}
		num = (num + max) % max;
		line = (line + window_height) % window_height;
	} while (isdir || isalpha(key) || IS_MOUSE_KEY(key));
	if (!exit_menu_doit() && !exit_menu_cancel()) {
		ungot_key = full_key;
		key = ESC;
	}
	return sub;
}

menu_invoked()
{
	int i;

	if (_doing_input == DI_LINEINPUT && is_gui) {
		ungot_key = full_key;
		error_if_input(1);
	}
	if (*menu_command == '$') {		// help file
		char old[FNAMELEN], dir[FNAMELEN];
// MSVC 2 wants the current directory to be \msvc20\bin, when using the help
// file \msvc20\help\xxx.hlp.  The other help files we know about don't care.
// So temporarily set the current directory based on the help file name.
		getcd(old);
		strcpy(dir, menu_command + 1);
		*get_tail(dir, 0) = 0;
		sprintf(get_tail(dir, 0), "bin");
		chdir(dir);
		i = windows_help_from(menu_command + 1, winhelp_display_contents);
		chdir(old);
		if (!i)
			error("Could not display help from %s", menu_command + 1);
		return;
	} else if (*menu_command == '!') {
		run_by_pattern(menu_command + 1);
		return;
	}
	save_var run_by_mouse = 1;
	i = find_index(menu_command);
	switch (name_type(i)) {
		case NT_MACRO:
			run_by_mouse = 0;
		case NT_COMMAND:
		case NT_SUBR:
		case NT_AUTOLOAD:
		case NT_AUTOSUBR:
			add_key_to_macro(CMD_INDEX_KEY + i);
			if (shift_pressed() & (KB_CTRL_DN | KB_SHIFT_DN))
				has_arg = iter = 1;
			do_command(i);
			break;
		case NT_WINVAR:
		case NT_BUFVAR:
		case NT_VAR:
		case NT_BUILTVAR:
			do_set_variable(i);
			break;
		default:
			error("Couldn't call `%s' from menu.", menu_command);
	}
}

color_postscript_range(from, to)
{
	int t = -1, talk, r, talk_at = 0;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend, case_fold = 1;
	set_character_color(from, to, -1);
	point = from;
	to_begin_line();	// Move back over continued lines.
	point -= parse_string(-1, "(.*\\\n)+");
	while (point > narrow_start
		   && get_tagged_region("color-as-unit", point - 1, &r) != -1)
		point = r;		// Move before binary blocks.
	talk = (to - point > 1000000);	// Show status during long delays.
	while (point < to) {
		if (!re_search(1, "[%(]|/[-_a-z0-9]+|</word>(doclutimage|doNimage)</word>")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(t)) {		// check first char
			case '%':
				nl_forward();
				set_character_color(t, point, color_class postscript_comment);
				if (parse_string(-1, "^%%%%BeginData:?\n")) {
					r = point;
					if (search(1, "\n%%EndData"))
						point = matchstart;
					set_character_color(r, point, color_class postscript_text);
					set_tagged_region("color-as-unit", t, point, 1);
				}
				break;
			case '(':		// A text string.
				point--;
				fwd_postscript_string();
				set_character_color(t, point, color_class postscript_text);
				break;
			case '/':		// Found /literalname.
				set_character_color(t, point, color_class postscript_literal);
				break;
			default:		// Found start of Windows PS driver binary code.
				re_search(1, "~>[ \n]*(Z|endimage)");	// Find end.
				set_character_color(t, point, color_class postscript_text);
				set_tagged_region("color-as-unit", t, point, 1);
				break;
		}
		if (talk && point > talk_at + 20000) {
			note("Coloring PostScript: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	if (talk_at)
		note("");
	return point;
}

show_scroll_bar()	// display scroll bar while mouse is over it
{
	int obord, origx = mouse_x;

	save_var mouse_cursor = &right_pointer;
	save_var window_handle = window_to_scroll();
	obord = get_wattrib(window_handle, PBORDERS);
	if (!bar_on_edge() && !delay_scroll_bar())
		return;
	save_var _window_flags |= FIXED_BORDERS;  // no auto setting
	save_var display_scroll_bar = 1;
	if (!GET_BORD(BRIGHT, obord))	// make right border if not one now
		set_wattrib(window_handle, PBORDERS, obord
			    | (BORD(BRIGHT, -1) & tiled_border));
	for (;;) {
		refresh();
		getkey();
		if (key == MOUSE_MOVE && over_scroll_bar() != window_handle) {
			mouse_cursor = &std_pointer;
			while (over_scroll_bar() == -1)
				if (char_avail() && getkey() != MOUSE_MOVE
				    || ABS(mouse_x - origx) > 4)
					break;
			mouse_cursor = &right_pointer;
		}
		if (over_scroll_bar() != window_handle)
			break;		// still not on this scroll bar
		if (IS_MOUSE_LEFT(key) || IS_MOUSE_CENTER(key) ||
		    IS_MOUSE_RIGHT(key) && mouse_x >= screen_cols - 1) {
			mouse_scroll_bar();
			ungot_key = MOUSE_MOVE;
		} else if (key != MOUSE_MOVE)
			break;
	}
	set_wattrib(window_handle, PBORDERS, obord);
	restore_vars();
	refresh();
	ungot_key = full_key;
}

delay_scroll_bar()	// did mouse stay over scroll bar for a while?
{
	TIMER ts;

	time_begin(&ts, 10);
	while (!time_done(&ts))
		if (char_avail() && (getkey() != MOUSE_MOVE
				     || over_scroll_bar() != window_handle)) {
			ungot_key = full_key;
			return 0;
		}
	return 1;
}

perl_after_expression(int p, int strict)
{
	int ch;
	if (in_shell_buffer)
		return 0;
	if (strict) {
		save_var point = p;
		if (!parse_string(-1, "}[ \t]*")) {	// If not an eval block or similar.
			// Move back over whitespace and comments.
			for (;;) {
				if (parse_string(-1, "[ \t\n]+"))
					point = matchend;
				else if (parse_string(-1, "#.*")
						 && get_character_color(matchend)
						 == color_class perl_comment)
					point = matchend;
				else
					break;
			}
			if (!parse_string(-1,  "<^word>[0-9]+([-0-9e.]+)?|[])'\"`]")
				  // Not a number, or one of the characters ])'" or `.
				  && !parse_string(-1, "%$%^?[^ \t\n]")) {
				// Or $, then perhaps a ^ and a nonspace.
				point -= parse_string(-1, "(%+%+|--)?");
				// Skip back over any ++ or --.
				if (!parse_string(-1, "<word>+"))	// Only other option is a word.
					return 0;
				if (parse_string(-1, perl_detect_expression_pattern)
					  && !strchr("$%&#@", character(matchend)))
					return 0;		// That's not any Perl keyword.
			}
		}
		return 1;
	} else {
		if (!matches_at(p, -1, "([$%&#@]((<word>+(->|'))?<word>+)"
						// A variable or function with $ or similar before it.
						"(%+%+|--)?"
						// Possibly followed by ++ or --.
						"|%$%^?[^ \t\n]"
						// Or $, then perhaps a ^ and a nonspace.
						"|<^word>[0-9]+([-0-9e.]+)?|[])'\"`])"
						// Or a number.  Or one of the characters ])'" or `.
						"(#.*|[ \t\n])*|}[ \t]*"))
			// All followed by whitespace or a comment.
		return 0;
	}
	ch = get_character_color(matchend);
	if (ch == color_class perl_comment || ch == color_class perl_string
		  && !matches_at(matchend, 1, "[\"']"))
		return 0;
	return 1;
}

color_perl_delimiters(int two, int interpolated, int modifiers)
{
	int start = point, c, end, level = 0, rstart, rend;
	char *openers = "([{<", *closers = ")]}>", *p;
	char pat[10];

	point += parse_string(1, "[ \t\n]+");
	set_character_color(start, point, color_class perl_comment);
	rstart = start = point;
	p = strchr(openers, c = end = curchar());
	if (isalnum(c) || isspace(c))
		return;			// No valid delimiter, so stop now.
	if (c == '\'' && interpolated != 2)
		interpolated = 0;
	if (p) {			// A paired delimiter.
		end = closers[p - openers];
		sprintf(pat, "[%c%c]", end, c);
		while (re_search(1, pat))
			if (character(point - 1) == c)
				level++;
			else if (--level <= 0)
				break;
		set_character_color(start, point, color_class perl_string);
		if (interpolated)
			color_perl_interpolated(start, point, 0);
	} else {
		point++;
		color_perl_fwd_quote(c, start, interpolated);
	}
	rend = point;
	if (two)		// A command like s///, so look for 2nd set.
		if (p) {	// After a paired delimiter, whitespace and pick new char.
			start = point;
			point += parse_string(1, "([ \t\n]|#.*)+");
			set_character_color(start, point, color_class perl_comment);
			color_perl_delimiters(0, interpolated, 0);	// One more pair.
		} else		// Find 3rd instance of char.
			color_perl_fwd_quote(c, point, interpolated);
	if (modifiers) {
		start = point;
		if (modifiers == 2 && parse_string(1, "[a-z]*x"))
			color_perl_interpolated(rstart + 1, rend - 1, 1);
		point += parse_string(1, "[a-z]+");
		set_character_color(start, point, color_class perl_keyword);
	}
}

// Found a word; see if it's decorated like a variable or subr.
color_perl_word(int start)
{
	int true_perl = !in_shell_buffer && sublanguage_code != SUBLANG_PHP;

	if (isdigit(character(start))) {		// a number
		if (matches_at(start, -1, "-?%.?"))	// number like -.25
			start = matchend;
		point += parse_string(1, "([A-Za-z0-9._]|[Ee]-)*");		// -3.0e-7
		set_character_color(start, point, color_class perl_constant);
		return;
	}
	if (sublanguage_code == SUBLANG_PHP)
		point += parse_string(1, "(::<word>+)+");	// pkg::var
	else if (!in_shell_buffer)
		if (strchr("$*@%&<", character(start - 1))
			  || !parse_string(-1, "</word>(q|qq|qx|qw|m|qr|s|tr)"))
			point += parse_string(1, "(('|::)<word>+)+");	// pkg::var
	if (strchr(!true_perl ? "$" : "$*@%",
			   character(start - 1))) {	// $var, @var etc.
		set_character_color(start - 1, point, color_class perl_variable);
	} else if (true_perl && character(start - 1) == '&') {	// &funcname
		set_character_color(start - 1, point, color_class perl_function);
	} else if (true_perl && character(start - 1) == '<'
			   && character(point) == '>') {
		set_character_color(start - 1, point, color_class perl_constant);
	} else if (true_perl)
		if (parse_string(1, "[ \t\n]*=>"))
			set_character_color(start, point, color_class perl_constant);
		else if (character(start - 1) == '{' && curchar() == '}')
			set_character_color(start, point, color_class perl_constant);
		else
			color_perl_keyword(start);
}

perl_compute_indent()
{
	int ind = 0;		/* indentation to use */
	int orig = point, i, j;
	int first_char = curchar(); /* cur line's first char */
	int pstart;		/* start position of prev line */
	int pafter;		/* after label on prev line */
	int pend;		/* end position of prev line */
	int prev_end = ';';	/* last char on previous line */
	int norm_narrow;	/* narrowing setting for most parsing */

	save_var point;
	save_var Closeback = perl_closeback;
	save_var Topindent = perl_topindent;
	save_var c_indent = perl_indent;
	save_var c_label_indent = perl_label_indent;
	save_var c_contin_offset = perl_contin_offset;
	save_var c_align_contin_lines = perl_align_contin_lines;
	save_var c_brace_offset = perl_brace_offset;
	save_var c_top_braces = perl_top_braces;
	save_var c_top_contin = perl_top_contin;
	save_var c_top_struct = perl_top_struct;

	if (sublanguage_code != SUBLANG_PHP)
		save_var narrow_start = 0;
	recolor_buffer_range(orig, orig + 1);
	if (get_tagged_region("color-perl-as-unit", orig, &i) != -1
		  || get_character_color(orig, &i) == color_class perl_constant
		  && matches_at(i, 1, "%<%<"))		// Keep original's indentation
		return get_indentation(orig);		// if in here doc.
	if (get_character_color(orig - 1, &i) == color_class perl_string)
		return get_indentation(orig);		// or in string.
	if (get_character_color(orig - 1, &i) == color_class perl_comment
		  && matches_at(i, 1, "="))
		return get_indentation(orig);		// or in POD.

	norm_narrow = fix_narrowing(point - c_look_back);
	narrow_start = norm_narrow;
	if (first_char == '\f')
		return 0;
	else if (first_char == '}')
		return cind_close();

	to_begin_line();
	if (parse_string(1, "[ \t]*\n") && after_many_comments()) {
		re_search(-1, "[^ \t\n]");	/* find previous non-blank line */
		return get_indentation(point);
	}
	cind_space();			/* skip whitespace and comments */
	if (character(point - 1) == ':')
		move_before_labels();
	prev_end = character(point - 1); /* get last relevant char */
	pend = point;

	to_statement_start();		/* if contin line, back up */
	pstart = point;			/* now at start of biggest statement */
	past_labels();			/* containing original point */
	pafter = point;			/* just after label */
	ind = current_column();
	narrow_start = norm_narrow;	/* ok to look far back again */

	if (sublanguage_code == SUBLANG_PHP		// Line following here doc.
		  && (get_tagged_region("color-perl-as-unit", give_end_line(), &i, &j) != -1
			  || get_character_color(orig, &i) == color_class perl_constant
			  && matches_at(i, 1, "%<%<"))) {
		point = j;		// We might have "EOF;" after the here doc.
		point += parse_string(1, "[ \t]*");
		prev_end = character((pend = point) - 1);
	}

	if (sublanguage_code == SUBLANG_PHP && is_case(orig))
		ind = indent_case(orig);
	else if (have_label(orig)) {
		ind = c_label_indent;
	} else if (prev_end == ':' && is_label(pstart)) {
		ind = get_column(pstart);
		if (!(Closeback && just_open(orig)))
			ind += get_c_indent();
		if (just_open(orig))
			ind += c_brace_offset;
	} else if (is_else(orig)) {
		point = orig;
		ind = indent_else(ind);
	} else if (prev_end == ';') {
		if (just_open(orig) && func_start_block(pstart))
			ind = c_top_braces;
		else
			ind = indent_after_statement(orig);
	} else if (prev_end == '}') {
		point = pend;
		ind = indent_after_block(orig);
	} else if (prev_end == '{') {
		if (ind <= c_top_braces	// special case if no previous indent
			|| get_column(pend - 1) <= c_top_braces) {
			ind += top_indent(prev_end) ? get_c_indent() : c_top_braces;
			if (just_open(orig))
				ind += c_brace_offset;
			if (ind < c_top_braces)
				ind = c_top_braces;
		} else if (Closeback || curchar() != '{')
			ind += get_c_indent();
		if (just_open(pafter))
			ind -= c_brace_offset;
	} else if (just_open(orig))
		return indent_just_open(ind);
	else if (is_contin_statement(pend, 1, orig)) {
		return indent_c_contin(ind, orig, pend, 0);
	} else if (!index(":);{}", c_line_before()) && !c_statement_start()
			   && !(Closeback && just_open(point)))
		return ind;
	else
		ind += get_c_indent();
	return ind;
}

int compile_handler();

do_compile(cmd)
char *cmd;
{
	int err = 0, nerr, sep_buf_ok = 0;
	int obuf = bufnum;
	char *procbuf = PROCBUF, curdir[FNAMELEN];

	_last_was_grep = 0;
	if (maybe_save_all())
		return 3;
	if (start_process_in_buffer_directory)
		get_buffer_directory(curdir);
	else
		getcd(curdir);
	drop_final_slash(curdir);
	if (!(has_feature & FEAT_MULT_CONCUR) && !another_process())
		save_var restart_concurrent = 0;

	// If we're allowed to run non-concur, then use separate buffer.
	if (concurrent_make < 2 && (has_feature & FEAT_MULT_CONCUR)
		  && compile_in_separate_buffer)
		sep_buf_ok = 1;
	else if (concurrent_make == 2)
		create_concur("", PROCBUF);
	else if (!sep_buf_ok && concurrent_make == 0)
		try_exit_concurrent();
	if (!sep_buf_ok && clear_process_buffer)
		zap(PROCBUF);
	if (sep_buf_ok) {
		int i;
		char newname[FNAMELEN];
		bufnum = create(procbuf = "compilation-output");
		for (i = 2; is_process_buffer(bufnum); i++) {
			sprintf(procbuf = newname, "compilation-output-%d", i);
			bufnum = create(procbuf);
		}
		if (clear_process_buffer)
			zap(procbuf);
        process_mode();
		point = size();
		set_error_spot();
		_had_errors = 0;
		bprintf("Running `%s' in directory `%s'\n", cmd, curdir);
		type_point = size();
		if (compile_in_separate_buffer == 2)
			locate_window(procbuf, "");
		if (!concur_shell("", cmd, curdir, procbuf)) {
			update_mode_more_extra(0);
			when_activity = compile_handler;
			while (process_exit_status == PROC_STATUS_RUNNING) {
				if (char_avail() && getkey() != MOUSE_MOVE
					  && (!IS_WIN_KEY(key) || key == WIN_DRAG_DROP)) {
					ungot_key = full_key;	// Wait until user presses a key
					return 2;			// or process finishes.
				}
				delay(10, COND_KEY, bufnum);	// Delay 0.1 secs
			}
			err = process_exit_status;
			bufnum = obuf;
			check_dates(0);
		} else
			err = -1;
	} else if (concurrent_make == 4 && is_process_buffer(name_to_bufnum(PROCBUF))) {
		save_var bufname = PROCBUF;	// Run the program in the concur
		set_error_spot();			// proc buffer, but don't process
		_had_errors = 0;			// the user's keys until it finishes.
		if (*cmd) {
			point = size();
			bprintf("%s\n", cmd);
		}
		maybe_refresh();
		iter = 1;
		delay(-1, COND_PROC, bufnum);
		if (!is_process_buffer(name_to_bufnum(PROCBUF)) || !process_input())
			return 2;	/* else assume it's compiled */
	} else if (no_running(0)) {	/* goes to end of proc buffer */
		set_error_spot();
		_had_errors = 0;
		to_begin_line();
		bprintf("%pRunning `%s' in directory `%s'\n",
				color_class process_output, cmd, *process_current_directory
				? process_current_directory : curdir);
		type_point = point = size();
		if (*cmd)
			bprintf("%s\n", cmd);
		maybe_refresh();
		iter = 1;
		if (opsys != OS_UNIX)
			delay(-1, COND_PROC | COND_KEY, bufnum);
		if (!is_process_buffer(name_to_bufnum(PROCBUF)) || !process_input())
			return 2;	/* else assume it's compiled */
		if (!(has_feature & FEAT_DETECT_CONCUR_WAIT))
			return 5;
	} else {
		bufnum = create(PROCBUF);
		point = type_point = size();
		buffer_printf("process", "%p\nRunning `%s' in directory `%s'\n",
				color_class process_output, cmd, curdir);
		bufnum = obuf;
		err = do_push(cmd, 1, 0);
	}
	maybe_restart_concurrent();
	nerr = do_next_error(1, cmd);	// scan for error messages
	if (!nerr && err) {	// no message recognized, but error code
		say("Process returned %d", err);
		locate_window(procbuf, "");
		point = size();		/* show end of process buffer */
		return err;
	} else
		bufnum = obuf;
	if (err)
		return 1;
	else if (nerr)
		return 4;
	return 0;
}

search_read(trystr, prmpt, flags)
char *trystr, *prmpt;
{
	int lastok, goodchars, oldflags, origwin = window_handle;
	int editbuf, pbuf, pwin, editwin, s_code, ostart, cnt;
	int oreg_type, old_mark = -1, search_contin = search_continuation != 0;
	int restrict_region = 0, recall_cnt = 0;
	char tmpbuf[FNAMELEN];
	old_place orig, mstart, okpoint;

	if (_doing_input == DI_SEARCH || _doing_input == DI_SEARCH_NOINCR)
		error_if_input(0);
	if (search_in_region && is_highlight_on()) {
		save_var _highlight_control;
		save_var restrict_search_type;
		restrict_region = make_restriction(flags);
	} else if (is_highlight_on() && point != mark && ABS(point - mark) < 200
		  && !*trystr && typing_deletes_highlight && typing_deletes_highlight != 3
		  && !in_macro(1) && _len_def_mac < 1 && search_defaults_from == 1) {
		old_mark = mark;	// Use this as the initial search string.
		highlight_off();
	}
	save_var _highlight_control;
	save_var iter, has_arg, this_cmd, prev_cmd;
	save_var _doing_input = (flags & INCR) ? DI_SEARCH : DI_SEARCH_NOINCR;
	save_var prompt_orig_buf = bufnum;
	save_var prompt_orig_win = window_handle;
	save_var last_index;
	save_var mark_to_column = -1, mark = point;
	save_var cursor_to_column = -1;
	save_var postpone_asynchronous_notify = 1;
	save_var macro_runs_immediately = 0;	// Make special keys work in macros.
	oreg_type = region_type();
	set_region_type(REGNORM);
	flags &= ~SEARCH_EMPTY | SEARCH_CONTIN;
	remember_old_place(&orig, point);
	remember_old_place(&mstart, point);
	remember_old_place(&okpoint, point);
	ostart = window_start;
	goodchars = lastok = 0;	/* 2 versions of len of last good string */
				/* (only lastok is set by getting default) */

	editbuf = tmp_buf();		/* make buffer for cmd line editing */
	pwin = add_popup(0, echo_line, screen_cols - 1, 1, 0,
			pbuf = tmp_buf());	/* prompt in window */
	stuff(trystr);
	trystr = malloc(FNAMELEN);
	display_column = 0;
	system_window = 1;	/* don't let user switch to it */
	text_color = color_class echo_area;

	editwin = add_popup(1, echo_line, screen_cols - 1, 1,
			0, editbuf);	/* window on top of prompt window */
	text_color = color_class echo_area;
	_display_class = _echo_disp_class;
	display_column = 0;

	if (old_mark != -1) { // Start out with highlighted region.
		window_handle = origwin;
		buf_xfer(editbuf, old_mark, point);
		window_handle = editwin;
		point = 0;
		if (flags & REGEX)	// Quote special chars.
			string_replace("[[%.!#()*+|?^$<]", "%#0", REGEX);
		point = size();
		highlight_on();		// Mark it so user can delete by typing.
		shift_selecting = 2;
	} else if (search_defaults_from == 2 && !size()) {
		if (flags & UNKNOWN)
			flags = known_flags(flags);
		recall_cnt++;
		stuff(default_search_string(flags));
		point = size();
		highlight_on();		// Mark it so user can delete by typing.
		shift_selecting = 2;
	}

	for (;;) {
		do {
			show_search_prompt(prmpt, flags, origwin,
					   pwin, editwin);
			getkey();
		} while (key == MOUSE_MOVE);
		if (IS_MOUSE_KEY(key))
			if (mouse_y == echo_line)
				flags &= ~INCR;
			else {
				flags |= SONCE;
				ungot_key = full_key;
				break;
			}
		cnt = 1;
		if (key == CTRL('^'))
			full_getkey("", CTRLIFY_KEY);
		window_handle = editwin;
		s_code = 0;
		oldflags = flags;
		flags &= ~(BAD | FAILING | SONCE);	/* assume ok */
		if (is_search_dir()) {
			window_handle = origwin;
			to_old_place(&okpoint);
			window_handle = editwin;
			if ((key == CTRL('R')) != ((flags & REVERSE) != 0)) {
				flags ^= REVERSE;	/* switch dir */
				if (goodchars)
					flags |= SONCE;
				lastok = size();
			} else if (!size()) {
				if (flags & UNKNOWN)
					flags = known_flags(flags);
get_defaults:
				stuff(default_search_string(flags));
				lastok = size();
			} else if (oldflags & FAILING) {
				if (search_contin) {
					flags |= SEARCH_CONTIN;
					s_code = FHERE;			/* Try other records. */
				} else if (search_wraps) {
					flags |= (flags & SEARCH_WRAPPED) ? SEARCH_OVERWRAPPED
						: SEARCH_WRAPPED;
					window_handle = origwin; /* Wrap to other end of buffer. */
					point = (flags & REVERSE) ? size() : 0;
					window_handle = editwin;
					s_code = FHERE;			/* And search again. */
				} else
					flags |= FAILING; /* no more matches, stop */
			} else
				flags |= SONCE;
			if ((flags & (INCR | SONCE)) && !(flags & FAILING))
				flags |= SONCE, s_code = FHERE;
		} else if (set_flag_modes(&flags)) {
			if ((flags & INCR) && size())
				s_code = FMATCH; /* if modes changed, search again */
		} else if (key == CTRL('O')) {
			flags ^= INCR;
			point = size();
			if (flags & INCR)
				s_code = FMATCH;
		} else if (binding_of(full_key, reg_tab) == (short) help)
			help_on_command(help_on_search(flags));
		else if (IS_ENTER_KEY(key)) {
			if (oldflags & SONCE)
				flags |= SONCE;
			break;
		} else if (binding_of(full_key, reg_tab)
				   == (short) backward_delete_character && (flags & INCR)) {
			run_topkey();
			window_handle = editwin;
			if (!(oldflags & FAILING) || size() <= lastok)
				s_code = FORIG;
			else		/* don't bother searching */
				flags |= FAILING;
		} else if (key == abort_key && (oldflags &
					((flags & INCR) ? FAILING : SONCE))) {
			if (flags & INCR)
				delete(goodchars, size());
		} else if (is_recall_key(flags)) {
			if ((recall_prior_response_options & 2) && !IS_ALT_KEY(key))
				recall_cnt = pick_prev_search_from_end(&flags,
					recall_cnt + ((is_recall_key(flags))));
			else
				flags = pick_prev_search(flags) & ~INCR;
			if (user_abort)
				break;
		} else if (binding_of(full_key, reg_tab) == (short) pull_word
				  || binding_of(full_key, reg_tab) == (short) pull_word_fwd) {
			flags &= ~UNKNOWN;
			window_handle = origwin;
			_highlight_control = 0;
			window_handle = editwin;
			pull_word();
		} else if (user_abort || (flags & INCR) && exit_incr())
			break;
		else {
			flags &= ~UNKNOWN;
			if ((binding_of(full_key, comp_tab) == (short) inp_grab_word
				 || binding_of(full_key, comp_tab) == (short) inp_grab_line)
				  && (IS_ALT_KEY(key) /*|| !(flags & INCR) */ )) {
				if (!(oldflags & FAILING)) {
					if (grab_search_word(flags, origwin,
						(binding_of(full_key, comp_tab) == (short) inp_grab_line))) {
						window_handle = origwin;	// Move origin
						remember_old_place(&orig, point);
						remember_old_place(&mstart, point);	// to start of word.
						window_handle = editwin;
						s_code = FORIG;
					} else
						s_code = (flags & REVERSE) ? FHERE : FMREGEX;
				}
			} else if (key == '\t')
				normal_character();
			else
				run_topkey();
			window_handle = editwin;
			if (flags & INCR)
				if (flags & (REGEX | WORD))
					s_code = FMREGEX;
				else if (oldflags & FAILING)
					flags |= FAILING;
				else
					s_code = FMADD;
		}
		_doing_input = (flags & INCR) ? DI_SEARCH : DI_SEARCH_NOINCR;
		if (!is_window(origwin) || !is_window(editwin)
		    || !is_window(pwin)) {	// user deleted our window
			user_abort = 1;
			goto end;
		}
		user_abort = 0;
		grab_expanding(0, size(), &trystr, 256);
		if (!s_code)
			continue;
		window_handle = origwin;
		if (s_code == FORIG || s_code == FMREGEX)
			to_old_place(&orig);
		else if (s_code >= FMATCH) {
			to_old_place(&mstart);
			if (flags & REVERSE)
				point++;
		}
		_highlight_control = 0;
search_again:
		switch (do_color_searching(flags, trystr)) {
			case DSBAD:
				flags = BAD | (flags & ~INCR);
				flags |= FAILING;
				to_old_place(&okpoint);
				if (in_macro(1))
					goto end;
				else if (!(oldflags & FAILING))
					maybe_ding(bell_on_search);
				break;
			case 0:
				if ((flags & SEARCH_CONTIN) && search_contin
					  && (*search_continuation)(cnt++, flags, trystr))
					goto search_again;
				flags |= FAILING;
				to_old_place(&okpoint);
				if (in_macro(1))
					goto end;
				else if (!(oldflags & FAILING))
					maybe_ding(bell_on_search);
				break;
			case 1:
				remember_old_place(&okpoint, point);
				remember_old_place(&mstart, matchstart);
				mark = matchstart;
				highlight_on();
				if ((flags & SEARCH_WRAPPED)
					  && past_old_place(&orig) == !(flags & REVERSE))
					flags |= SEARCH_OVERWRAPPED;
				window_handle = editwin;
				goodchars = lastok = size();
				break;
			case DSABORT:
				to_old_place(&okpoint);
				window_handle = editwin;
				if ((flags & INCR) && s_code == FMADD)
					delete(size() - 1, size());
				while (ungot_key == -1 && char_avail()
					   && getkey() == abort_key || !IS_TRUE_KEY(key))
					;
				if (key != abort_key && IS_TRUE_KEY(key))
					ungot_key = full_key;
				drop_pending_says();
				break;
		}
	}
	window_handle = origwin;
	if (key == abort_key || user_abort) {
		to_old_place(&orig);
		window_start = ostart;
	} else {
		window_handle = editwin;
		if (size()) {
			_last_was_regex = ((flags & REGEX) != 0);
			*(_last_was_regex ? &_default_regex_flags
					  : &_default_flags) = flags;
			grab_expanding(0, size(), &trystr, 256);
			if (_last_was_regex)
				grab_expanding(0, size(), &_default_regex_search, 256);
			else
				grab_expanding(0, size(), &_default_search, 256);
			save_search_recall(flags, trystr);
		} else
			flags |= SEARCH_EMPTY;
		if ((flags & INCR) && !IS_ENTER_KEY(key) && key != ESC)
			ungot_key = full_key;
	}
end:
	strncpy(tmpbuf, trystr, FNAMELEN - 1);
	free(trystr);
	if (is_window(origwin))
		window_handle = origwin;
	set_region_type(oreg_type);
	remove_window(editwin);
	remove_window(pwin);
	buf_delete(editbuf);
	buf_delete(pbuf);
	if (restrict_region)
		remove_region(restrict_region);
	if ((flags & FAILING) && in_macro(1))
		error("%s: not found", tmpbuf);
	check_abort();
	say("");
	restore_vars();
	maybe_refresh();	/* put cursor back in prev window */
	return flags;
}

string_replace(str, with, flags)
char *str, *with;
{
	spot orig, prev, oldmark;
	int skip = 0, show = 0, ret, evershow = 0, oldstart, i;
	int null_detect = -1, new_null_detect;
	int perform_case_replace = 0, this_case_replace;

	if (_buf_readonly)
		error("Buffer is read-only.");
	if (!*str)
		return 0;
	if (replace_by_case)
		perform_case_replace = want_case_replace(str, with, flags);
	orig = alloc_spot();
	prev = alloc_spot();
	save_var postpone_asynchronous_notify = 1;
	replace_num_found = replace_num_changed = 0;
	for (;;) {
		skip = 0;
		*prev = point;
		oldstart = matchstart;
		ret = do_color_searching(flags, str);
		if (!ret || ret == DSBAD) {
			point = *orig;
			break;
		} else if (ret == DSABORT)
			break;
		new_null_detect = (flags & REVERSE) ? point : (size() - point);
		if (null_detect == new_null_detect) {
			if ((flags & REVERSE) ? (point-- <= narrow_start)
								  : (point++ >= size() - narrow_end)) {
				point = *orig;
				break;	// Make sure replacing a nullable pattern like $
			}			// moves forward.
			continue;
		} else
			null_detect = new_null_detect;
		replace_num_found++;
ask:
		if (show) {
			show_replace(str, with, (flags & ~INCR));
			evershow = 1;
			show = 0;
		}
		if (!(flags & QUERY)) goto do_replace;
		if (!get_replace_key(flags, str, skip)) { // show match, then get key
			point = *orig;		// match went away (maybe flags changed)
			break;
		}
		if (key == '?' || key == 'h'
			  || binding_of(full_key, root_keys) == (short) help) {
			help_on_command("query-replace");
			show = 1;
			point = *prev;
			null_detect = -1;
			continue;
		} else if (user_abort || key == abort_key) {
			point = *orig;
			ret = DSABORT;
			goto done;
		}
		switch(tolower(key)) {
			case '^': point = *prev; matchstart = oldstart;
					  null_detect = -1;
			case '\b':
			case GREYBACK:
			case 'n': skip = 1;	/* go to the next match */
			case ' ':
			case 'y':		/* do the replace */
			case '!':
			case ',':
			case '.': break;
			case CTRL('R'):
				say("");
				null_detect = -1;
				oldmark = alloc_spot();
				*oldmark = mark;	/* save mark */
				mark = matchstart;
					/* set temporary mark before text */
				recursive_edit();
				if (spot_to_buffer(oldmark) < 0) {
					ret = DSABORT;	// user deleted buf
					goto done;
				}
				mark = *oldmark;	/* restore mark */
				free_spot(oldmark);
				show = 1;
				continue;
			default:	if (set_flag_modes(&flags)) {
						show = 1;
						goto ask;
					}
					if (!IS_ENTER_KEY(key))
						ungot_key = full_key;
			case ESC:	ret = 1; goto done;
		}
do_replace:
		if (!skip) {
			save_spot matchstart;
			if (perform_case_replace)
				this_case_replace = (perform_case_replace == 1)
									? check_for_case_replace(matchstart, point)
									: check_alt_case_replace(matchstart, point);
			if (flags & REGEX)
				fromorig(with);
			else {
				delete(matchstart, point);
				stuff(with);
			}
			if (perform_case_replace == 1)
				switch (this_case_replace) {
					case 1: do_uppercase_region(matchstart, point); break;
					case 2: do_capitalize_region(matchstart, point); break;
				}
			else if (perform_case_replace == 2) {
				i = MIN(matchstart, point);
				switch (this_case_replace) {
					case 1: do_uppercase_region(matchstart, point); break;
					case 2: do_lowercase_region(matchstart, point); break;
					case 3: do_uppercase_region(i, i + 1); break;
					case 4: do_lowercase_region(i, i + 1); break;
				}
			}
			replace_num_changed++;
			restore_vars();
		}
		if (flags & QUERY)
			switch (key) {
			case ',': case '^':	skip = 1; goto ask;
			case '.':		ret = 1; goto done;
			case '!':		flags &= ~QUERY; show = 1;
			}
	}
done:
	if (ret == DSABORT)
		say("Replacement canceled");
	else if (ret == DSBAD)
		show_replace(str, with, (flags & ~INCR) | BAD);
	else if (!in_macro(1))
		if (evershow || (flags & INCR))
			say("Replaced %d of %d.", replace_num_changed,
			    replace_num_found);
		else if (evershow)
			say("");
	free_spot(orig);
	free_spot(prev);
	return ret;
}

tag_c_variable()	// is this a var definition?
{
	if (parse_string(1, "([) \t\n]|<[>[^]\n]*<]>|//.*)*[,;=]")) {
		point = matchend - 1;	// ordinary variable definition
		if (character(point) == '=')
			point++, skip_initializer();
	} else if (parse_string(1, "[ \t\n]*<)>+[ \t\n]*!"
							"(_ANSI_ARGS_|__P)?<(>"))
		re_search(1, "![;{]");	// func ptr definition
	else
		return 0;
	return 1;
}

suffix_sty()
{
	suffix_ltx();
}

suffix_cls()
{
	if (matches_at(0, 1, "[ \t\n]*VERSION .* CLASS"))
		vbasic_mode();
	else
		suffix_ltx();
}

int do_vhdl_indent();
int color_vhdl_range();

command vhdl_mode()
{
	mode_default_settings();
	mode_keys = vhdl_tab;		/* Use these keys. */
	major_mode = _vhdl_mode_name;
	compile_buffer_cmd = compile_vhdl_cmd;	// can compile this
	indenter = do_vhdl_indent;
	strcpy(comment_start, "--[ \t]*");
	strcpy(comment_pattern, "--.*$");
	strcpy(comment_begin, "-- ");
	strcpy(comment_end, "");
	recolor_range = color_vhdl_range;	// set up coloring rules
	recolor_from_here = recolor_by_lines;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	if (auto_show_vhdl_delimiters)
		auto_show_matching_characters = vhdl_auto_show_delim_chars;
	buffer_maybe_break_line = generic_maybe_break_line;
	fill_mode = misc_language_fill_mode;
	try_calling("vhdl-mode-hook");
	drop_all_colored_regions();
	make_mode();
}

vhdl_func_name_finder()
{
	int from, to, pos, orig = point;
	char pat[FNAMELEN * 2];

	save_var point, case_fold = 1;
	to_end_line();
	while (do_color_searching(REVERSE | REGEX, "^[ \t]*<^c:vhdl-comment"
							  "|vhdl-string>(procedure|function)</word>")) {
		pos = point;
		if (!parse_string(1, "^[ \t]*(procedure|function)[ \t]+([a-zA-Z_]<word>*)"))
			continue;
		from = find_group(2, 1);
		to = find_group(2, 0);
		to = MIN(to, from + ptrlen(display_func_name) - 1);
		grab(from, to, display_func_name);
		point = matchend;
		point += parse_string(1, "[ \t]*");
		if (curchar() == '(')
			move_level(1, "(", ")", 0, 0);
		if (!parse_string(1, "[ \t]*(return[ \t\n]+[a-z0-9_]+[ \t]*)?is</word>")) {
			point = pos;
			continue;
		}
		if (matchend - matchstart < FNAMELEN) {
			sprintf(pat, "^[ \t]*end[ \t]+%s</word>",
					display_func_name);
			if (re_search(1, pat) && point < orig)
				*display_func_name = 0;
		}
		return 1;
	}
	*display_func_name = 0;
	return 1;
}

tag_mode_vhdl()
{
	char func[TAGLEN];
	int start;

	save_var point = 0, case_fold = 1;
	while (do_color_searching(REGEX, "^[ \t]*<^c:vhdl-comment"
							  "|vhdl-string>(procedure|function|entity)"
							  "[ \t]+([a-zA-Z_]<word>*)")) {
		grab(start = find_group(2, 1), find_group(2, 0), func);
		add_tag(func, start);
	}
}
