/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2005 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 12.08 to 12.09.

#include "eel.h"
#include "c.h"
#include "dired.h"
#include "colcode.h"
#include "bufed.h"
#include "epsnet.h"
#include "grep.h"
#include "menu.h"
#include "misclang.h"
#include "perl.h"
#include "proc.h"
#include "tags.h"
#include "vbasic.h"
#include "vhdl.h"

int scp_sftp_handle_list(struct scp_job *p)
{
	int i, isdata;
	char msg[FNAMELEN];

	save_var case_fold = 1;
	while (nl_forward()) {
		i = point;
		point--;
		to_begin_line();
		isdata = 0;
		if (parse_string(1, "[-dcblsp][-r][-w][-sxt][-r][-w][-sxt]"
						 "[-r][-w][-sxt] .*" UNIX_DATE))
			isdata = 1;
		if ((p->flags & SCP_CHECK_IS_DIR) && parse_string(1, ".* not a directory"))
			isdata = 1;
		if (parse_string(1, ".* No such file or directory")
			  && p->cmd && !strncmp(p->cmd, "ls ", 3)) {
			if (buf_exist(p->destbuf))	// Error part way through, so don't
				set_buf_mode_message(p->destbuf, NULL);	// copy any more.
			p->flags |= SCP_CMD_ERROR;
		}
		if (isdata) { // Got some data, copy where it belongs.
			if (buf_exist(p->destbuf) && !(p->flags & SCP_CMD_ERROR)) {
				if (*p->file && !(p->flags & SCP_CHECK_IS_DIR))
					buf_printf(p->destbuf, "%s", p->file);
				buf_xfer(p->destbuf, point, i);
			}
		} else if (parse_string(1, "Remote (working )?directory(:| is| is now) ")) {
			grab(matchend, give_end_line(), msg);
			if (p->lastdir)
				free(p->lastdir);
			p->lastdir = strsave(msg);
			if (!p->homedir)
				p->homedir = strsave(msg);
		} else if ((p->flags & SCP_IS_XFER)	// Ignore this msg.
				   && parse_string(1, "((Fetching|Uploading) .* to |"
								   "(remote|local):.* => (remote|local):|"
								   ".*--:-- ETA)")) {
		} else if (!parse_string(1, "(Exited|Connecting to "
								 "|Connection closed|Listing directory "
								 "|Using username \")")) {
			grab(point, MIN(give_end_line(),
							point + sizeof(msg) / sizeof(*msg) - 1), msg);
			if (*msg)		// Any other line is an error msg.
				say("%s", msg);
			if ((p->flags & SCP_IS_XFER)
				  && parse_string(1, ".*: (Permission denied|No such file"
								  "|Device out of space|Read-only file system)"))
				p->flags |= SCP_CMD_ERROR;
		}
		point = i;
	}
	if (parse_string(-1, "p?sftp%> *")) {
		if (!p->cmd || !*p->cmd) {
			if (buf_exist(p->destbuf))
				set_buf_mode_message(p->destbuf, NULL);
			if (p->flags & SCP_IS_XFER)
				scp_handle_done(p);
		}
		p->busy = 0;			// Got a prompt for another cmd.
		return 1;			// Send a pending command if any.
	}
	return 0;
}

auto_save_one()
{
	char name[FNAMELEN];
	int err, origfile = !strcmp(auto_save_name, "%f");

	build_filename(name, auto_save_name, filename);
	if (_read_aborted)		// we didn't read whole file in,
		return;				// so don't save it
	else if (!origfile)			// auto-save to a different file
		err = file_convert_write(name, translation_type, NULL);
	else if (check_dates(1))	// make sure other prog hasn't changed
		return;
	else				// get new time info
		err = file_convert_write(filename, translation_type, &file_info);
	if (err) {
		if (is_gui)		// Don't pop up dialog when autosave fails.
			say("Error autosaving %f", name);
		else
			file_error(err, name, "couldn't autosave");
		maybe_ding(bell_on_autosave_error);
	} else
		show_text(0, 1, "%f autosaved.", name);
	if (!origfile)
		modified = 1;
}

line_search(dir, s)	/* search for line with only s on it */
char *s;		/* go to its end (if reverse: start) */
{
	int start, end;

	if (!*s)
		return re_search(dir, "^$");
	if (dir < 0)
		to_end_line();
	while (search(dir, s)) {
		start = (dir > 0) ? matchstart : point;
		end = (dir > 0) ? point : matchstart;
		if ((end >= size() - narrow_end || character(end) == '\n')
			  && (start <= narrow_start
				  || character(start - 1) == '\n'))
			return 1;
	}
	return 0;
}

#define MANCONF_PATH	"^(MANDATORY_|OPTIONAL_)?MANPATH" \
						"(_MAP[ \t]+[^ \t\n]+)?[ \t]+([^ \t\n]+)$"

build_man_path(char *dest)
{
	char *path = getenv("MANPATH"), out[FNAMELEN * 4];
	int tmp;

	if (path && *path) {
		strcpy(dest, path);
		return;
	}
	save_var bufnum = tmp = tmp_buf();
	if (!file_read("/etc/manpath.config", FILETYPE_AUTO)
		  || !file_read("/etc/man.config", FILETYPE_AUTO)) {
		do_drop_matching_lines(REGEX, MANCONF_PATH, 0);
		string_replace(MANCONF_PATH, "#3", REGEX);
	}
	stuff("/usr/man\n/usr/share/man\n");
	*out = 0;
	replace_file_with_output(out, "manpath %s");
	while (path = strchr(out, ':'))
		*path = '\n';
	stuff(out);
	do_sort_region(0, size(), 0, 0);
	do_uniq(1, 1, 0);
	simple_re_replace(0, "\n", ":");
	simple_re_replace(0, ":<eof>", "");
	grab(0, size(), dest);
	restore_vars();
	buf_delete(tmp);
}
