/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2005 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 12.09 to 12.10.

#include "eel.h"
#include "c.h"
#include "dired.h"
#include "colcode.h"
#include "info.h"
#include "epsnet.h"
#include "perl.h"
#include "proc.h"
#include "tags.h"
#include "tex.h"

suffix_java()
{
	c_mode();
	compile_buffer_cmd = compile_java_cmd;	// can compile this
	c_extra_keywords = JAVA_KEYWORDS;	// recognize Java keywords
}

c_default_indent(ind, orig, defmult)
{
	if ((c_extra_keywords & CSHARP_KEYWORDS) && parse_string(1, "[ \t]*%[")
		  && have_csharp_attribute(point))
		return get_indentation(point);
	if (matches_at(orig, 1, "[ \t]*//"))
		return ind;
	if ((c_extra_keywords & (CPP_KEYWORDS | JAVA_KEYWORDS | CSHARP_KEYWORDS))
		  && matches_at(orig, 1, "[ \t]*class</word>"))
		return ind;		// Put class keyword in continuation back on left.
	if (ind && indent_preprocessor_contin == 255)
		return ind;
	if (ind)
		return ind + defmult * get_c_indent() + c_contin_offset;
	if (parse_string(1, "[ \t\f]*(struct|union|class|namespace)[ \t\f\n]+"
			 "[a-zA-Z0-9_]*[ \t\f\n]*{[ \t\f\n]*[\n/]"))
		return ind + c_top_struct;
	if (parse_string(1, "[ \t\f]*(template)[ \t\f\n]*%<"))
		return ind + c_top_template;
	if (just_open(orig))
		return c_top_braces;
	point = orig;
	cind_space();
	return is_contin_statement(point, 0, orig)
			? c_top_contin : c_param_decl;
}

get_c_indent()	/* indent this many spaces for each level */
{			/* if c_indent/java_indent is 0, use tab_size instead */
	if ((c_extra_keywords & JAVA_KEYWORDS) && java_indent > 0)
		return java_indent;
	return (c_indent > 0) ? c_indent : tab_size;
}

command quick_dired_command()
{
	char file[FNAMELEN], rel[FNAMELEN], msg[FNAMELEN], dest[FNAMELEN];
	char dir[FNAMELEN];

	if (is_dired_buf()) {
		dired_getname(file, 3);
		dired_try_getdir(dir, 1);
		save_var inp_get_other_name = file;
	} else {
		strcpy(file, filename);
		strcpy(dir, filename);
		*get_tail(dir, 0) = 0;
	}
	if (!*file)
		error("This buffer has no associated file.");
	strcpy(rel, get_tail(file, 0));
again:
	do {
		sayput("%s: D=Delete, Move, Copy, View, T=Props, .=Dired, "
			   "G=Go dir, ?=Show more: ", rel);
		getkey();
		check_abort();
	} while (!IS_TRUE_KEY(key));
	say("");
	if (is_remote_file(file) && exist(REMOTE_FTP_OP_BUFFER))
		zap(REMOTE_FTP_OP_BUFFER);
	switch (toupper(key)) {
		case '.':
			dired_one(file);
			break;
		case 'D':
			sprintf(msg, "Delete %s? ", rel);
			if (!ask_yn("Delete file", msg, "Delete", "Cancel", 1))
				break;
			save_var dired_confirmation;
			if (dired_do_delete(file, NULL, 0)) {
				check_abort();
				file_error(errno, file, "Could not delete");
			} else if (!is_remote_file(file))
				say("%f deleted.", file);
			break;
		case 'C':
			if (no_remote_copy(file))
				error("Copy operation not implemented for remote files.");
			if (is_directory(file) == 1)
				error("Can't copy directories.");
			sprintf(msg, "Copy %s to: ", rel);
			get_file_dir(dest, msg);
			if (*dest) {
				switch (dired_do_copy(file, dest, 0)) {
					case 1:
						file_error(errno, dest, "Could not copy");
						break;
					case 0:
						say("%f copied.", file);
						break;
				}
			}
			break;
		case 'M':
			sprintf(msg, "Move %s to: ", rel);
			get_file_dir(dest, msg);
			if (*dest) {
				if (dired_do_move(file, dest, 0))
					file_error(errno, dest, "Could not move");
				else if (!is_remote_file(file))
					say("%f moved.", file);
			}
			break;
		case 'V':
			do_shell_execute(file, "", "run viewer");
			break;
		case 'T':
			do_shell_execute(file,  "Properties", "show properties");
			break;
		case 'F':
			do_shell_execute(dir, "Open", "run viewer");
			break;
		case 'G':
			if (chdir(dir))
				file_error(errno, dir, "unknown error");
			else {
				getcd(dir);
				say("Current directory: %f", dir);
			}
			break;
		case '+':
			do_create_directory(0);
			break;
		case '!':
			do_shell_command(file, dir);
			break;
		case 'A':
			dired_do_set_attributes(file, dir);
			check_dates(0);
			break;
		default:
			if (key == '?' || key == HELPKEY || key == FKEY(1))
				help_on_command("quick-dired-command");
			else
				say("Invalid key.");
			goto again;
	}
	if (is_remote_file(file))
		send_remote_ftp_op(file);
}

int dired_ftp_activity();

do_remote_dired(char *dir)
{
	char plain[FNAMELEN], *p;

	strcpy(plain, dir);
	p = get_tail(plain, 1);	// If name contains wildcards, strip them.
	if (index(p, '*') || index(p, '?'))
		*p = 0;
	buffer_ftp_activity = dired_ftp_activity;
	bprintf("     Directory of %s\n", plain);
	point = modified = 0;
	dired_orig_position = point;
	if (index(dir, '*') || index(dir, '?'))	// Make FTP listing
		save_var ftp_compatible_dirs |= 1;	// show just the basename.
	return url_operation(dir, FTP_LIST);
}

kill_ftp_job(int buf)
{
	save_var bufnum = buf;
	if (ftp_job)
		say("Disconnecting buffer from ftp job in %s.", bufname);
	else if (telnet_id)
		telnet_send(telnet_id, NULL);
	ftp_job = 0;
}

int ftp_activity(), color_ftp_log_range();

do_ftp_op(char *dest, char *host, char *port, char *usr, char *pwd,
		  char *fname, int cmd)
{
	char log[FNAMELEN], *p;
	int id, i = 1, portnum;
	struct ftp_job *job;

	portnum = numtoi(port);
	if (got_bad_number)
		portnum = 21;
	save_var bufnum;
	bufname = dest;
	if (ftp_job)
		error("This buffer's FTP job has not yet completed.");
	strcpy(log, "FTP Log");
	p = log + strlen(log);
	for (;;) {
		if (!exist(log))
			break;
		bufname = log;
		if (!ftp_job)
			break;
		sprintf(p, " %d", i++);
	}
	if (ftp_compatible_dirs & 1)
		cmd |= FTP_USE_CWD;
	if (ftp_compatible_dirs & 2)
		cmd |= FTP_PLAIN_LIST;
	id = ftp_op(dest, log, host, portnum, usr, pwd, fname, cmd);
	if (id) {

		bufname = dest;
		job = (struct ftp_job *) malloc(sizeof(struct ftp_job));
		job->id = id;
		job->operation = cmd & FTP_OP_MASK;
		job->buf = bufnum;
		job->log_buf = name_to_bufnum(log);
		job->file_size = 0;
		job->time_started = time_ms();
		ftp_job = job;
		when_net_activity = ftp_activity;

		bufname = log;
		ftp_job = job;
		discardable_buffer = 1;
		when_net_activity = ftp_activity;
		translation_type = FILETYPE_MSDOS;
		recolor_range = color_ftp_log_range;	// set up coloring rules
		recolor_from_here = recolor_by_lines;
		when_setting_want_code_coloring();	// maybe turn on coloring
		set_both_mode_message("[FTP: Connecting] ");
	}
}

int literal_if_args();
char *man_match(char *s, int flags);
char *man_use_program();

show_man_page(int type)
{
	char topic[FNAMELEN], buf[FNAMELEN], pr[FNAMELEN], *prog = man_use_program(type);

	tiled_only();
	save_var show_status = 0, abort_searching = 0;
	grab_man_word_from_buffer(bufnum, point, buf,
							  "[-+_0-9a-z.:]+(%(..?%))?");
	if (type == 1 && !strcmp(buf, "use")) {
		save_var point;
		point += parse_string(1, "<word>*<space>*");
		grab_man_word_from_buffer(bufnum, point, buf,
								  "[-+_0-9a-z.:]+(%(..?%))?");
	}
	man_rewrite(buf);
	save_var completion_literal_char = literal_if_args;
	sprintf(pr, "Show %s entry for", type ? prog : "manual");
	prompt_comp_read(topic, pr, prog ? man_match : 0,
					 COMP_FOLD | SPACE_VALID | CAUTIOUS, buf);
	if (!*topic)
		return;
	make_man_page(topic, type);
}

make_man_page(char *topic, int type)
{
	char cmdline[FNAMELEN], buf[FNAMELEN], *shell = getenv("SHELL");
	int color;

	sprintf(cmdline, "%s%s %s",
			!is_unix ? "" : (strstr(shell, "/csh") || strstr(shell, "/tcsh"))
			? "setenv LANG C ; " : "LANG=C ",
			man_use_program(type), topic);
	sprintf(buf, "man-%s", topic);
	quiet_set_bookmark();
	to_buffer(buf);
	man_mode();
	if (!size() || type != _man_program) {
		set_read_only(0);
		zap(buf);
		translation_type = FILETYPE_AUTO;
		pipe_text(NULL, buf, cmdline, NULL,
				  PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0);
		point = 0;
		if (parse_string(1, "Reformatting.*please wait.*\n"))
			delete(0, matchend);
		simple_re_replace(1, "\n\n\n+", "\n\n");
		set_character_color(0, size(), -1);
		point = 0;
		while (re_search(1, ".\b.")) {
			if (character(point - 3) == '_')
				color = color_class man_underlined;
			else
				color = color_class man_bold;
			delete(point - 3, point - 1);
			set_character_color(point - 1, point, color);
			point--;
		}
		filename = "fake-man-file.man";	// So find-linked-files works.
		force_save_as = FSA_NEWFILE;
		discardable_buffer = buffer_not_saveable = 1;
	}
	modified = point = 0;
	_man_program = type;
	set_read_only(3);
}

info_guess_node_position(char *node)
{
	char pat[FNAMELEN];
	int pos, endtags, endhead, atend = 0;

	save_var point = 0, narrow_start, narrow_end;
	if (!re_search(1, "\n" INFO_CSTR))
		return 0;
	endhead = point;
	point = size();
	if (narrowed_search(REGEX | REVERSE, "\n" INFO_CSTR "\f?\nEnd Tag Table", 500)) {
		re_search(-1, "\n" INFO_CSTR "\f?\nTag Table:");
		atend = 1;
	} else
		point = 0;
	if (!narrowed_search(REGEX, "\n" INFO_CSTR "\f?\nTag Table:",
						 MAX(10000, size() / 20)))
		return 0;
	narrow_start = point;
	if (!re_search(1, "\n" INFO_CSTR "\f?\nEnd Tag Table"))
		return 0;
	endtags = point;
	sprintf(pat, "\nNode: %s\x7F", node);
	if (!search(-1, pat))
		return 0;
	point = matchstart;
	if (grab_numbers(bufnum, &pos) != 1)
		return 0;
	return pos + (atend ? 0 : (endtags - endhead)) - 10000;
}

scan_binding_ranges(int b1, int b2)
{
	char init[FNAMELEN], def1[FNAMELEN], def2[FNAMELEN];
	int s1, e1, s2, e2;
	save_var bufnum = b1;
	point = 0;
	while (re_search(1, "~[^ \t\n:]+ !.*\" \".*\": ")) {
		grab(matchstart, matchend, init);
		int *kcode = get_keycode();
		if (!kcode || kcode[0] != 2)
			continue;
		s1 = kcode[1];
		if (!parse_string(1, " \""))
			continue;
		point++;
		kcode = get_keycode();
		if (!kcode || kcode[0] != 2)
			continue;
		e1 = kcode[1];
		if (!parse_string(1, ": .*", def1))
			continue;
		if (!find_lowest_binding_range(b2, init, s1, &s2, &e2, def2))
			continue;
		if (s1 >= s2 && e1 <= e2)	// This range is contained within other.
			continue;
		if ((s1 != s2 || e1 != e2) && s2 <= e1) {
			delete(give_begin_line(), give_end_line() + 1);
			int pos = point;
			insert_new_binding_range(b1, s1, s2 - 1, init, def1);
			insert_new_binding_range(b1, MAX(s1, s2), MIN(e1, e2), init, def1);
			insert_new_binding_range(b1, e2 + 1, e1, init, def1);
			point = pos;
		}
	}
}

int make_difference_list(int orig, int cur)
{
	noteput("Comparing...");
	save_var show_status = 0;
	scan_binding_ranges(orig, cur);
	scan_binding_ranges(cur, orig);
	save_var bufnum = orig;
	save_var case_fold = 1;
	do_sort_region(0, size(), 0, 0);
	save_var bufnum = cur;
	save_var case_fold = 1;
	do_sort_region(0, size(), 0, 0);
	int out = tmp_buf();
	do_compare_sorted(orig, cur, "original-settings",
					  bufnum_to_name(out), NULL);
	bufnum = out;
	point = 0;
	save_var case_fold = 1;
	custom_apply_immediately();
	// Drop vars starting with -, but not color scheme ones.
	do_drop_matching_lines(REGEX, "^%+-[^o][^u][^r][^-]", 1);
	drop_boring_change_lines();
	custom_fix_mode_vars();
	insert_key_unbindings(name_to_bufnum("original-settings"));
	do_drop_matching_lines(REGEX, "^%+(push-cmd|echo-line|permanent-menu"
						   "|avoid-(top|bottom)-lines|bufed-width"
						   "|mode-start|mode-end) ", 1);
	do_drop_matching_lines(REGEX, "^&.* color class for after-exiting: .*$", 1);
	do_drop_matching_lines(REGEX, "^&.* color class for .*: "
						   "%[0x4000000 on 0x0%]$", 1);
	do_drop_matching_lines(REGEX, "^~.*: case-indirect$", 1);
	return out;		// Don't count vars saved in session file, others.
}

int compile_handler(int activity, int buf, int from, int to);

do_compile(cmd)
char *cmd;
{
	int err = 0, nerr, sep_buf_ok = 0;
	int obuf = bufnum;
	char *procbuf = PROCBUF, curdir[FNAMELEN];

	_last_was_grep = 0;
	if (maybe_save_all())
		return 3;
	if (start_process_in_buffer_directory)
		get_buffer_directory(curdir);
	else
		getcd(curdir);
	drop_final_slash(curdir);
	if (!(has_feature & FEAT_MULT_CONCUR) && !another_process())
		save_var restart_concurrent = 0;

	// If we're allowed to run non-concur, then use separate buffer.
	if (concurrent_make < 2 && (has_feature & FEAT_MULT_CONCUR)
		  && compile_in_separate_buffer)
		sep_buf_ok = 1;
	else if (concurrent_make == 2)
		create_concur("", PROCBUF);
	else if (!sep_buf_ok && concurrent_make == 0)
		try_exit_concurrent();
	if (!sep_buf_ok && clear_process_buffer)
		zap(PROCBUF);
	if (sep_buf_ok) {
		int i;
		char newname[FNAMELEN];
		bufnum = create(procbuf = "compilation-output");
		for (i = 2; is_process_buffer(bufnum); i++) {
			sprintf(procbuf = newname, "compilation-output-%d", i);
			bufnum = create(procbuf);
		}
		if (clear_process_buffer)
			zap(procbuf);
        process_mode();
		point = size();
		set_error_spot();
		_had_errors = 0;
		bprintf("Running `%s' in directory `%s'\n", cmd, curdir);
		type_point = size();
		if (compile_in_separate_buffer == 2)
			locate_window(procbuf, "");
		if (!concur_shell("", cmd, curdir, procbuf)) {
			update_mode_more_extra(0);
			when_activity = compile_handler;
			while (process_exit_status == PROC_STATUS_RUNNING) {
				if (char_avail() && getkey() != MOUSE_MOVE
					  && !IS_WIN_PASSIVE_KEY(key)) {
					ungot_key = full_key;	// Wait until user presses a key
					return 2;			// or process finishes.
				}
				delay(10, COND_KEY, bufnum);	// Delay 0.1 secs
			}
			err = process_exit_status;
			bufnum = obuf;
			check_dates(0);
		} else
			err = -1;
	} else if (concurrent_make == 4 && is_process_buffer(name_to_bufnum(PROCBUF))) {
		save_var bufname = PROCBUF;	// Run the program in the concur
		set_error_spot();			// proc buffer, but don't process
		_had_errors = 0;			// the user's keys until it finishes.
		if (*cmd) {
			point = size();
			bprintf("%s\n", cmd);
		}
		maybe_refresh();
		iter = 1;
		delay(-1, COND_PROC, bufnum);
		if (!is_process_buffer(name_to_bufnum(PROCBUF)) || !process_input())
			return 2;	/* else assume it's compiled */
	} else if (no_running(0)) {	/* goes to end of proc buffer */
		set_error_spot();
		_had_errors = 0;
		to_begin_line();
		bprintf("%pRunning `%s' in directory `%s'\n",
				color_class process_output, cmd, *process_current_directory
				? process_current_directory : curdir);
		type_point = point = size();
		if (*cmd)
			bprintf("%s\n", cmd);
		maybe_refresh();
		iter = 1;
		if (opsys != OS_UNIX)
			delay(-1, COND_PROC | COND_KEY, bufnum);
		if (!is_process_buffer(name_to_bufnum(PROCBUF)) || !process_input())
			return 2;	/* else assume it's compiled */
		if (!(has_feature & FEAT_DETECT_CONCUR_WAIT))
			return 5;
	} else {
		bufnum = create(PROCBUF);
		point = type_point = size();
		buffer_printf("process", "%p\nRunning `%s' in directory `%s'\n",
				color_class process_output, cmd, curdir);
		bufnum = obuf;
		err = do_push(cmd, 1, 0);
	}
	maybe_restart_concurrent();
	nerr = do_next_error(1, cmd);	// scan for error messages
	if (!nerr && err) {	// no message recognized, but error code
		say("Process returned %d", err);
		locate_window(procbuf, "");
		point = size();		/* show end of process buffer */
		return err;
	} else
		bufnum = obuf;
	if (err)
		return 1;
	else if (nerr)
		return 4;
	return 0;
}

tag_lister(char *s, char *(*finder)(), int flags, int mbuf)
{
	char pat[TAGLEN], look_for[TAGLEN];
	int sorted;

	finder = finder;	/* silence eel's warning */
	flags = flags;
	save_var bufnum;
	save_var abort_searching = 0;
	if (tags_from_bsc) {
		buf_printf(mbuf, "%s", s);
		return;
	}
	if (!*s) {
		tags_look_for(look_for);
		if (*look_for) {
			s = look_for;
			save_var tag_list_exact_only = 1;
		}
	}
	if (want_sorted_tags && !*s)
		sort_tags();
	bufnum = _tags_buffer;
	save_var case_fold = !tag_case_sensitive;
	sorted = are_tags_sorted();	/* also goes to top */
	if (*s) {
		sprintf(pat, "^([^\t\n]+(::|'))?%q", s);
		while (search(1, s)) {
			if (parse_string(-1, pat)
				&& (!tag_list_exact_only || curchar() == '\t')) {
				to_end_line();
				buf_xfer(mbuf, matchend, point + 1);
			}
		}
	} else {
		nl_forward();
		buf_xfer(mbuf, point, size());
	}
	bufnum = mbuf;
	if (want_sorted_tags && !sorted && !user_abort) {
		abort_searching = ABORT_ERROR;
		sort_another(bufnum_to_name(mbuf), 0, 0);
	}
	point = 0;
	if (tag_display_width)
		tab_size = tag_display_width;
	else
		tab_size = (screen_cols - _view_left - _view_right) / 2 - 1;
	goal_column = display_column = 0;
}

good_c_tag()	// tag name at point if a valid c tag
{		// return 0 to go back to orig point before searching again
	char func[TAGLEN], tmp[TAGLEN];
	int start = point, i = 1, skip = 0;

	if (!parse_string(1, "[a-zA-Z0-9_]+", func))
		return 0;
	else if (!strcmp(func, "struct") || !strcmp(func, "union"))
		return tag_c_struct(0);		// maybe struct tag
	else if (!strcmp(func, "class"))
		return tag_c_struct(1);		// or similar class def
	else if (!strcmp(func, "interface") && (c_extra_keywords & JAVA_KEYWORDS))
		return tag_c_struct(1);		// for java: like a class
	else if (!strcmp(func, "enum"))
		return tag_c_struct(2);		// or similar enum def
	else if (!strcmp(func, "namespace"))
		return tag_c_struct(3);		// or namespace def
	else if (!strcmp(func, "operator") && parse_string(1, CPP_OPLIST)) {
		grab(matchstart, matchend, func + strlen(func));
		point = matchend;
	} else if (!strcmp(func, "extern")) {
		tag_modifier_kwd |= 1;	// next vars are just declarations
		return 1;
	} else if (!strcmp(func, "typedef")) {
		tag_modifier_kwd |= 2;	// next vars are typedefs
		return 1;
	} else if ((c_extra_keywords & CPP_KEYWORDS) && !strcmp(func, "using")) {
		tag_modifier_kwd |= 4;	// just a declaration 
		return 1;
	} else if (is_java_buffer() && (!strcmp(func, "import")
									|| !strcmp(func, "package"))) {
		tag_modifier_kwd |= 4;
		return 1;
	} else if (!strcmp(func, "__declspec")) {
		return 1;
	} else if (parse_string(-1, "^[ \t]*#[^#\n]*"))
		return tag_c_define(func);	// do preproc line
	c_fwd_space();
	if (curchar() == '_' && parse_string(1, "(_ANSI_ARGS_|__P)</word>")) {
		point = matchend;	// skip over arg-hiding macro
		c_fwd_space();
	}
	if (curchar() == '<')
		tag_check_template();
	if (curchar() == '(') {
		if (matches_at(start, -1, "~[ \t]*")) {
			point = matchend;
			parse_string(1, "~[ \t]*[a-zA-Z0-9_]+", func);
		}
		i = tag_c_func(1);
		if (!(tag_which_items & 1))
			skip = 1;
	} else if (curchar() == ':' && parse_string(1, "::[ \t\n]*"
			"(_?(near|far)[ \t\n]+)?([a-zA-Z0-9_~]+|" CPP_OPSPEC ") *%(")) {
		strcat(func, "::");	// make raw class::func w/o spaces
		grab(find_group(3, 1), find_group(3, 0), func + strlen(func));
		i = tag_c_func(1);
		if (!(tag_which_items & 1))
			skip = 1;
	} else if (tag_c_variable()
			   && ((tag_declarations & 2) || !(tag_modifier_kwd & 1)
				   || (tag_modifier_kwd & 2))) {
		i = 2;
		if (!(tag_which_items & 2))
			skip = 1;
	}
	if ((i == 2 || i == 3) && !isdigit(*func) && !(tag_modifier_kwd & 4)) {
		if (*c_tagging_class && !index(func, ':')) {
			strcpy(tmp, func);	// put class:: before func name
			sprintf(func, "%s::%s", c_tagging_class, tmp);
		}
		if (!skip)
			add_tag(func, start);
	}
	return i;
}

int tex_maybe_break_line(int mode);
int color_tex_range();
int recolor_tex_from_here();

command tex_mode()
{
	mode_default_settings();
	mode_keys = tex_tab;		/* Use these keys. */
	tex_paragraphs = 1;
	buffer_maybe_break_line = tex_maybe_break_line;
	if (tex_force_latex == 2){		// Determine if it's LaTeX.
		save_var point = 0;
		tex_force_latex = search(1, "\\begin{document}");
	}
	major_mode = tex_force_latex ? _latex_mode_name : _tex_mode_name;
	compile_buffer_cmd = tex_force_latex ? compile_latex_cmd : compile_tex_cmd;
	strcpy(comment_start, "([^\\\n]|^)%%[ \t\f]*");
	strcpy(comment_pattern, "[^\\]%%.*$");
	strcpy(comment_begin, "% ");
	strcpy(comment_end, "");
	recolor_range = color_tex_range;	// set up coloring rules
	recolor_from_here = recolor_tex_from_here;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	idle_coloring_size = 3000;	// Do bigger bunches.
	if (auto_show_tex_delimiters)
		auto_show_matching_characters = tex_auto_show_delim_chars;
	try_calling("tex-mode-hook");
	drop_all_colored_regions();
	make_mode();
}

when_loading()
{
	cc_tab[CTRL('[')] = alt_prefix;
	cc_tab[CTRL('^')] = cc_tab[ALT(CTRL('^'))] = ctrl_prefix;
	ftp_compatible_dirs |= 2;	// New flag is on by default.
}
