/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2005 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 12.00 or later, or beta version.  We call other
// updater files for each individual version change past 12.00.

// For 12.00 and earlier, preserve fwd-, rev-search-key, which we're
// resizing.  Do this before #include files resize and discard old vals.
#define NT_VAR		5	/* normal variable */
#define TYPE_SHORT	2		/* a 16-bit integer */
#define TYPE_INT	3		/* a 32-bit integer */

when_loading()
{
	int i, old;

	i = find_index("fwd-search-key");
	if (vartype(i) == TYPE_SHORT) {
		old = get_num_var(i);
		drop_name("fwd-search-key");
		new_variable("fwd-search-key", NT_VAR, TYPE_INT);
		set_num_var(i, old);
	}
	i = find_index("rev-search-key");
	if (vartype(i) == TYPE_SHORT) {
		old = get_num_var(i);
		drop_name("rev-search-key");
		new_variable("rev-search-key", NT_VAR, TYPE_INT);
		set_num_var(i, old);
	}
}


#include "eel.h"

when_loading()
{
	int dorest = 0;
	// Old format 12.00b4, new format 12b5
	if (!strncmp(eel_version, "12.00b", 6) || !strncmp(eel_version, "12b", 3)) {
		save_var bufnum = zap(HELPBUF);
		bprintf("%s is from a beta version of Epsilon.\n"
				"The above state file is for Epsilon version %s.\n"
				"This Epsilon executable is from Epsilon version %s.\n\n"
				"I must first load a separate updater to bring this\n"
				"state file to 12.00.  Press Enter to continue.\n",
				state_file, eel_version, EEL_MINOR_VERSION);
		view_help(" Beta Version Detected ", 0);
		update_autoload("find-file", "oquickup");	// Always load.
		dorest = 1;
	}
	if (!strcmp(eel_version, "12.00") || dorest) {
		update_autoload("find-file", "quick01");
		dorest = 1;
	}
	if (!strcmp(eel_version, "12.01") || dorest) {
		update_autoload("find-file", "quick02");
		dorest = 1;
	}
	if (!strcmp(eel_version, "12.02") || dorest) {
		update_autoload("find-file", "quick03");
		dorest = 1;
	}
	if (!strcmp(eel_version, "12.03") || dorest) {
		update_autoload("find-file", "quick04");
		dorest = 1;
	}
	if (!strcmp(eel_version, "12.04") || dorest) {
		update_autoload("find-file", "quick05");
		dorest = 1;
	}
	if (!strcmp(eel_version, "12.05") || dorest) {
		update_autoload("find-file", "quick06");
		dorest = 1;
	}
	if (!strcmp(eel_version, "12.06") || dorest) {
		update_autoload("find-file", "quick07");
		dorest = 1;
	}
	if (!strcmp(eel_version, "12.07") || dorest) {
		update_autoload("find-file", "quick08");
		dorest = 1;
	}
	if (!strcmp(eel_version, "12.08") || dorest) {
		update_autoload("find-file", "quick09");
		dorest = 1;
	}
	if (!strcmp(eel_version, "12.09") || dorest) {
		update_autoload("find-file", "quick10");
		dorest = 1;
	}

	if (state_file && get_file_read_only(state_file) == 1) {
		save_var bufnum = zap(HELPBUF);
		if (strcmp(eel_version, EEL_MINOR_VERSION))
			bprintf("%s is read-only.\n"
					"Therefore it cannot be updated.\n"
					"The above state file is for Epsilon version %s.\n"
					"This Epsilon executable is from Epsilon version %s.\n"
					"Perhaps EPSPATH is set incorrectly?\n",
					state_file, eel_version, EEL_MINOR_VERSION);
		else
			bprintf("%s is read-only.\n"
					"Therefore it cannot be updated.\n"
					"But this Epsilon executable is from version %s,\n"
					"and so is the above state file.\n"
					"Therefore it appears no update is needed.\n",
					state_file, eel_version);
		view_help(" Update Not Performed ", 0);
		return;
	}

	set_chars(eel_version, 0, sizeof(eel_version) / sizeof(char));
	strcpy(eel_version, EEL_MINOR_VERSION);
	delay(1, COND_KEY);	// Make sure the X window is up.
	refresh();
	sayput("Update complete, preparing to write state file...");
	delay(300, COND_KEY);	// Give the user a chance to abort.
	check_abort();
	if (!*state_file)
		write_state();		// Prompt for name.
	else
		do_save_state(state_file);	// Rewrite current state file.
	exit();
}
