/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2003 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 12 beta 4 or later to 12.00.

#include "eel.h"
#include "proc.h"
#include "html.h"
#include "c.h"
#include "colcode.h"
#include "vbasic.h"
#include "perl.h"
#include "filter.h"
#include "tags.h"
#include "hex.h"
#include "epsnet.h"
#include "proc.h"
#include "bufed.h"
#include "kill.h"
#include "session.h"
#include "dired.h"
#include "grep.h"
#include "vhdl.h"
#include "misclang.h"
#include "tex.h"
#include "browse.h"

command quoted_insert()
{
	mention("Quote: ");
	do {
		getkey();
		if (IS_WIN_KEY(key)) {
			iter = 0;
			say("");
			return;
		}
	} while (!IS_ASCII_KEY(key & ~KEY_ALT));
	if (key == CTRL('Q') && ungot_key != -1) {	// auto-quoted
		key = ungot_key;	// Strip extra auto-quoting.
		ungot_key = -1;
	}
	say("");
	if (user_abort == 2)
		check_abort();
	user_abort = 0;
	if (!IS_ASCII_KEY(full_key))
		full_key = generic_key(full_key);
	if (IS_ALT_KEY(full_key) && KEY_PLAIN(full_key) < 128)
		full_key |= 128;			// turn on meta bit
	delete_if_highlighted();
	while (iter-- > 0)
		insert(full_key);
}

/* Delete trailing whitespace, then break line here and indent it. */
break_and_indent_line()
{
	delete(point, point - parse_string(-1, "[ \t]*"));
	insert('\n');
	do_reindent_previous_line();
	(*indenter)();
}

/* insert or overwrite according to mode */
command normal_character() on reg_tab['\n']		/* more bindings, see below */
{
	int i;

	check_abort();
	delete_if_highlighted();
	if (!call_maybe_break_line(3))
		return;
	if (IS_ALT_KEY(key) && KEY_PLAIN(key) < 128)
		key |= 128;			// turn on meta bit
	if (IS_ASCII_KEY(key) && !IS_ASCII_KEY(full_key))
		full_key = key;
	if (full_key == '\n') {
		if (over_mode) {
			if (!nl_forward())
				insert('\n');
		} else if (auto_indent && indenter)
			break_and_indent_line();
		else
			insert('\n');
	} else if (over_mode && point < size() && curchar() != '\n') {
		if (curchar() == '\t')
			i = display_width('\t', current_column());
		else
			i = 0;
		replace(point++, full_key);
		while (--i > 0)		/* insert i-1 spaces after pt */
			insert(' '), point--;
	} else
		insert(full_key);
}

do_shift_selects()
{
	if (!shift_selects || IS_ASCII_KEY(key)
		  || (NUMPLAIN(key) < KEYEND || NUMPLAIN(key) > KEYPGUP)
		  && (NUMPLAIN(key) < NUMDIGIT(0) || NUMPLAIN(key) > NUMDIGIT(9))
	   ) {	// only affects cursor keys or numpad keys
		if (shift_selecting == 2)		// except at prompts
			highlight_off();
		return;
	}
	if (was_key_shifted()) {
		if (!is_highlight_on())		// if shift key is down, begin/extend
			do_set_mark(point);		// selection
		set_region_type(REGNORM);
		highlight_on();
		shift_selecting = 1;
	} else if (shift_selecting)	// if shift was just removed, remove highlight
		highlight_off();
}

full_getkey(char *prompt, int code)
{			/* get a key using ^Q, ^[, and ^^ prefixes */
	char msg[80];
	int allcode = 0, refreshed = 0;

	strcpy(msg, prompt);
	for (;;) {
		switch (code) {
			case ALTIFY_KEY:  strcat(msg, "<Alt> "); break;
			case CTRLIFY_KEY: strcat(msg, "<Ctrl> "); break;
		}
		allcode |= code, code = 0;
		for (;;) {
			mention("%s", msg);
			getkey();
			if (key != MOUSE_MOVE && !IS_WIN_KEY(key))
				break;
			if (!refreshed)
				refresh();
			refreshed = 1;
		}
		if (allcode & ALTIFY_KEY)
			key = generic_key(full_key = make_alt(full_key));
		if (allcode & CTRLIFY_KEY)
			key = generic_key(full_key = make_ctrl(full_key));
		if (root_keys[key] == (short) alt_prefix)
			code = ALTIFY_KEY;
		else if (root_keys[key] == (short) ctrl_prefix)
			code = CTRLIFY_KEY;
		else
			return;
	}
}

command brief_drop_bookmark()
{
	int b;

	b = KEY_PLAIN(key);
	b = (b < 256) ? (b & 127) : 0;	/* Alt-3 => '3' */
	if (!isdigit(b)) {
		b = get_number("Drop bookmark [1-10]: ");
		if (b < 1 || b > 10)
			return;
	} else			/* set b to 1 ... 10 */
		b = (b == '0') ? (10) : (b - '0');
	do_set_bookmark(b - 1 + 10);	/* use bookmarks A ... M */
	say("Bookmark %d dropped.", b);
}

#define BROWSER_BUF "Symbols"

#define BSC_DLL "bsc.dll" /* Compiler-specific DLL file from Microsoft. */

// This takes the address of a pointer, and makes the pointer point
// to a BSC handle.
#define BSCOpen(browsefile, ppbsc) \
	((int) call_dll(BSC_DLL, \
					"BSCOpen", "c", "sr", browsefile, ppbsc))

// This takes a pointer to a BSC handle.
#define BSCClose(pbsc) ((int) call_dll(BSC_DLL, "BSCClose", "c", "p", pbsc))

#define BSCGetOverloadArray(pbsc, name, mbf, ppinst, pnum) \
	((int) call_dll(BSC_DLL, \
					"BSCGetOverloadArray", "c", "psirr", \
					pbsc, name, mbf, ppinst, pnum))


#define BSCIinstInfo(pbsc, inst, name, typ, atr) \
	((int) call_dll(BSC_DLL, \
					"BSCIinstInfo", "c", "pirpp", \
					pbsc, inst, name, typ, atr))

#define BSCFormatDname(pbsc, name) \
	((int) call_dll(BSC_DLL, \
					"BSCFormatDname", "c", "pp", \
					pbsc, name))

#define BSCGetDefArray(pbsc, inst, ppidef, pnum) \
	((int) call_dll(BSC_DLL, \
					"BSCGetDefArray", "c", "pirr", \
					pbsc, inst, ppidef, pnum))

#define BSCGetRefArray(pbsc, inst, ppidef, pnum) \
	((int) call_dll(BSC_DLL, \
					"BSCGetRefArray", "c", "pirr", \
					pbsc, inst, ppidef, pnum))


#define BSCIdefInfo(pbsc, idef, pmodname, plineno) \
	((int) call_dll(BSC_DLL, \
					"BSCIdefInfo", "c", "pirp", \
					pbsc, idef, pmodname, plineno))													  

#define BSCSzFrTyp(pbsc, typ) \
	((int) call_dll(BSC_DLL, \
					"BSCSzFrTyp", "c", "pi", \
					pbsc, typ))

#define BSCSzFrAtr(pbsc, atr) \
	((int) call_dll(BSC_DLL, \
					"BSCSzFrAtr", "c", "pi", \
					pbsc, atr))

#define BSCDisposeArray(pbsc, any_array) \
	((int) call_dll(BSC_DLL, \
					"BSCDisposeArray", "c", "pp", \
					pbsc, any_array))


void get_inst_data(int *pbsc, int inst, char *sym_name, char *sym_descr)
{
	char sym_type[512];
	char sym_attr[512];
	byte *undecor_name;
	char typ[1];
	short atr[1];

	BSCIinstInfo(pbsc, inst, &undecor_name, typ, atr);

	bytes_to_chars(sym_name, (byte *) BSCFormatDname(pbsc, undecor_name));
	bytes_to_chars(sym_type, (byte *) BSCSzFrTyp(pbsc, *typ));
	bytes_to_chars(sym_attr, (byte *) BSCSzFrAtr(pbsc, *atr));

	sprintf(sym_descr, "%s%s%s",
			sym_attr[0] ? sym_attr : "",
			sym_attr[0] ? " " : "",
			sym_type
		   );
}

int dump_defs(int *pbsc, char *bsc_filename, char *symbol, int *insts, int num_insts)
{
	char sym_name[512];
	char sym_descr[512];
	int r = 0;
	int i, j;

	bufnum = zap(BROWSER_BUF);

	use_black(); bprintf("Browser File: ");
	use_blue();  bprintf("%s\n", bsc_filename);

	use_black(); bprintf("Symbol: ");
	use_blue();  bprintf("%s\n", symbol);

	use_black();bprintf("\n%s:\n", "Definitions");

	for (j = 0; j < num_insts; j++){
		int *defs, num_defs;
		int inst = insts[j];

		get_inst_data(pbsc, inst, sym_name, sym_descr);
		BSCGetDefArray(pbsc, inst, &defs, &num_defs);
		for (i = 0; i < num_defs; i++){
			char modname[1024];
			byte *temp;
			short line_num[1];
			BSCIdefInfo(pbsc, defs[i], &temp, line_num);
			r++;
			bytes_to_chars(modname, temp);

			// The file name.
			use_black();
			bprintf("    %s(%d): ", modname, *line_num);

			// The symbol name.
			use_blue();
			bprintf("%s", sym_name);

			// The symbol description.
			use_green();
			bprintf(" (%s)", sym_descr);
			bprintf("\n");
			use_black(); // flush
		}
	}
	point = 0;
	return r;
}

/*
	This gathers def or ref data for the indicated symbol, using
	the indicated browser database.  All the data goes into a
	buffer, and the browser database is closed, and the dll is
	unloaded.  Subsequently, you can use browser_find_match()
	to retrieve information on the matches.
*/
int browser_query(char *bsc_filename, char *symbol, int flags)
{
	int *bsc;
	int *Inst, num_insts;
	int r = 0;
	byte temp[FNAMELEN];

	save_var bufnum;
	UNREFERENCED_PARAMETER(flags);
	chars_to_bytes(temp, bsc_filename);
	if (BSCOpen(temp, &bsc) == 0)
		return -1;
	chars_to_bytes(temp, symbol);
	BSCGetOverloadArray(bsc, temp, MBFALL, &Inst, &num_insts);	

	// flags will indicate defs or refs.
	r = dump_defs(bsc, bsc_filename, symbol, Inst, num_insts);

	if (num_insts)
		BSCDisposeArray(bsc, Inst);
	BSCClose(bsc);
		// We unload the DLL to be sure that all the resources are released.
	call_dll(BSC_DLL);
	return r;
}

bufed_set_sort()
{
	_bufed_sort = full_key;
	do_bufed(has_arg ? BUFED_ALL : 0);
}

// Return nonzero if the regex pattern appears in the range.
matches_in(int start, int end, char *pat)
{
	save_var narrow_start = start, point = end;
	return re_search(-1, pat);
}

delete_user_buffer(bname)	/* delete a buffer that might have coloring */
char *bname;
{
	if (!exist(bname))
		return;
	drop_coloring(name_to_bufnum(bname));
	highlight_off();
	save_var bufname = bname;
	if (telnet_id)
		telnet_send(telnet_id, NULL);
	restore_vars();
	delete_buffer(bname);
}

char *preserve_buffer_type(char *buf)
{
	struct scp_job *p;
	save_var bufname = buf;

	for (p = scp_job; p; p = p->next)
		if (p->buf == bufnum)
			return p->busy ? "Scp" : 0;
	if (is_process_buffer(bufnum) || can_send_invisible())
		return major_mode;
	return 0;
}

check_drop_buffer(bname) /* check if buffer can be deleted safely */
char *bname;		/* ask user if questionable */
{			/* return 1 if done, 0 if user said don't */
	char msg[FNAMELEN], msg2[FNAMELEN], *origbuf = bufname, *type;
	int mod, noerr = 0;

	if (!exist(bname))
		error("No buffer %s.", bname);
	if (type = preserve_buffer_type(bname)) {
		sprintf(msg2, "Confirm deleting %s buffer", type);
		switch (process_warn_on_killing) {
			case 1:			// Ask if we should kill it.
				sprintf(msg, "%s is running in the buffer, exit it? ", type);
				if (!ask_yn(msg2, msg, "Yes", "No", 2))
					break;		// Else fall through.
			case 0:			// Always try to kill it.
				if (!strcmp(type, "Telnet")) {
					noerr = 1;
					break;
				}
				exit_this_process(bname);
				if (type = preserve_buffer_type(bname)) {
					sprintf(msg, "Couldn't exit cleanly from %s, kill it? ", type);
					if (ask_yn(msg2, msg, "Yes", "No", 2))
						if (telnet_id)
							telnet_send(telnet_id, NULL);
						else if (!process_kill(name_to_bufnum(bname)))
							error("Couldn't kill process in \"%s\".", bname);
				}

		}		// Other values: just give error.
		if ((type = preserve_buffer_type(bname)) && !noerr)
			error("A %s job is running in buffer \"%s\".", type, bname);
	}
	bufname = bname;
	mod = modified && size() > 0 && !discardable_buffer;
	if (ftp_job && ftp_job->buf == bufnum
		  && ftp_job->operation == FTP_SEND)
		mod = 1;
	if (!strcmp(bname, MSG_BUFFER))
		mod = 0;
	bufname = origbuf;
	if (mod) {
		sprintf(msg, "Buffer %.40s not saved.  Delete it? ", bname);
		if (!ask_yn("Delete buffer", msg, "&Delete", "&Cancel", 2))
			return 0;
	}
	to_another_buffer(bname);
	return 1;
}

c_indenter()		/* replace surrounding indentation with new */
{
	if (c_extra_keywords & IDL_KEYWORDS)
		return;
	if (in_shell_buffer)
		prev_indenter();
	else if (!(in_c_comment(c_look_back) & IN_COMMENT))
		indent_to_column(c_compute_indent());
	else if (reindent_c_comments && !in_perl_buffer)
		indent_c_comment();
}

int find_template_end()
{
	int opos = point, level = 1, parens = 0, bad = 0;

	if (curchar() != '<')
		return 0;
	point++;
	for (;; point++) {
		if (parse_string(1, "[ \n\t,a-zA-Z0-9_]+|//.*"))
			point = matchend - 1;
		else
			switch (curchar()) {
				case '/':
					if (character(point + 1) != '*')
						bad = 1;
					point += 2;
					if (!search(1, "*/"))
						bad = 1;
					break;
				case '<':
					if (parens == 0)
						level++;
					break;
				case '>':
					if (--level <= 0) {
						point++;
						return 1;
					}
					break;
				case '(':
					parens++;
					break;
				case ')':
					if (--parens <= 0)
						bad = 1;
					break;
				default:
					bad = 1;
					break;
			}
		if (bad) {
			point = opos;
			return 0;
		}
	}
}

int find_c_class_info(int start, char *type, char *class, char *func)
{
	char this[FNAMELEN], last[FNAMELEN];
	int class_start = point, i;

	point--;	// See if "class foo" is before this.
	strcpy(last, "");
	for (i = 20; i-- > 0; ) {	// Look through some words before the {.
		if (cind_try_space() || give_begin_line() > start)
			return 0;
		if (strchr("<>:,", character(point - 1)))
			point--;
		else {
			if (!parse_string(-1, (c_extra_keywords
								   & (CSHARP_KEYWORDS | JAVA_KEYWORDS))
							  ? "[a-zA-Z0-9_.]+" : "[a-zA-Z0-9_]+", this))
				return 0;
			if (!strcmp(this, "class") || !strcmp(this, "struct")
				  || !strcmp(this, "union") || !strcmp(this, "namespace")
				  || !strcmp(this, "enum")
				  || (c_extra_keywords == CSHARP_KEYWORDS) && !strcmp(this, "interface")) {
				if (class) {
					if (*class && ((display_c_options & 1)
								   || strcmp(type, "namespace"))) {
						strcat(class, (c_extra_keywords == CSHARP_KEYWORDS) ? "." : "::");
						class += strlen(class);
					}
					strcpy(class, last);	// Is place where we started within func?
				}
				if (type)
					strcpy(type, this);		// Word to the right of type is class.
				if (find_inner_func_name(start, class_start, func))
					return CF_INFO_TYPE | CF_INFO_CLASS | CF_INFO_FUNC;
				if (func)
					*func = 0;
				return CF_INFO_TYPE | CF_INFO_CLASS;
			}
			strcpy(last, this);		// Go on to the next word.
		}
	}
	return 0;		// Couldn't recognize anything here.
}

suffix_acf()
{
	suffix_idl();
}

command do_c_indent()
{
	int orig = point, ind = get_c_indent();
	int orig_column = current_column();

	if (maybe_indent_rigidly(0))
		return;
	to_indentation();
	if (!c_tab_always_indents && orig_column > current_column() || in_shell_buffer) {
					/* if not in indentation */
		point = orig;
		if (this_cmd != CMD_INDENT_REG)
			to_column(orig_column + tab_size - orig_column % tab_size);
	} else if (prev_cmd == C_INDENT)	/* repeated, make bigger */
		to_column(orig_column + ind);
	else if ((c_extra_keywords & IDL_KEYWORDS))
		prev_indenter();
	else {
		c_indenter();
		if (in_perl_buffer && !get_indentation(point) && !orig_column
			  && this_cmd != CMD_INDENT_REG)
			to_column(ind);	// 0 indent=>0 indent, but indent anyway
	}
	if (this_cmd != CMD_INDENT_REG)
		this_cmd = C_INDENT;
}


after_many_comments()
{
	char pat[FNAMELEN];
	int cnt = 0;

	if (comment_repeat_indentation_lines < 0)
		return 0;
	save_var point;
	sprintf(pat, "[ \t]*%s", get_comment_start());
	while (nl_reverse()) {
		to_begin_line();
		if (!parse_string(1, pat))
			break;
		if (++cnt >= comment_repeat_indentation_lines)
			return 1;
	}
	return 0;
}

indent_after_statement(orig) // orig is place needing indentation, for toplevel
{
	int lastcont, i, c, atparen = -1;

	for (;;) { /* was only the end of a larger statement, find start */
		do {	/* look back for line that ends with ;{}: */
			lastcont = point;
			to_begin_line();
			if (cind_try_space())
				break;
			if ((c = character(point - 1)) == ')') {
				atparen = lastcont;
				if (!c_rev_skip_level('(', ')'))
					break;
				to_begin_line();
			}
		} while (!index(";{}:", c));
		if (c == ':' && parse_string(-1, "%)[ \t]*:"))
			continue;	// Not a label but C++ inheritance or similar.
		point = lastcont;	/* use indentation of line */
		to_c_indentation();	/* just after it */
		if (parse_string(RE_FORWARD, "(}[ \t\f]*)?e!lse[^a-zA-Z0-9_]")) {
			point = matchend;
			if (!to_matching_if())
				break;
		} else if (parse_string(RE_FORWARD,
								"(}[ \t\f]*)?!(catch|finally)[^a-zA-Z0-9_]")) {
				point = matchend;
				if (!to_matching_try())
					break;
			} else
			break;
	}
	i = current_column();
	if (!i && atparen != -1 && (Topindent || !c_statement_start()))
		i = get_indentation(atparen);	// 2nd func param decl
	if ((atparen = parse_string(1, "for[ \t\f]*%([ \t\f]*"))
	    && for_unclosed(point, orig))
		i = get_column(point + atparen); // part of "for" statement
	if (i < c_top_braces)
		return 0;	// 1st line of new func
	return i;
}

c_matches_in(int start, int end, char *pat)
{
	save_var narrow_start = start, point = end;
	while (re_search(-1, pat))
		if (!in_c_comment(c_look_back))
			return 1;
	return 0;
}

indent_c_contin(ind, orig, lastl)	/* indent continuation line */
{
	int c, plevel = 0, qlevel = 0, findc, alevel = 0;

	if (!c_align_contin_lines)	// never line up
		return c_default_indent(ind, orig, 1);
 	if (matches_at(orig, -1, ":\n[ \t]*")
 		  && !c_matches_in(give_begin_line(), lastl, "%?"))
 		return get_indentation(point) + c_align_inherit;
	to_begin_line();
	save_var narrow_start = point;
	point = orig;
	findc = character(orig);
	if (c_align_open_paren && parse_string(1, "%(([ \t]|//|/%*)*$"))
		return ind;
	while (re_search(-1, "[()?:\"'\n<>]|[=!<>]?=")) {
		c = curchar();
		if (in_perl_buffer && (in_perl_comment(point) || strchr("\"'\n", c)))
			continue;	// Skip over comments differently.
		switch (c) {
			case '(':	/* indent after unmatched lparen */
				if (plevel-- > 0)
					break;
				if (c_align_open_paren && findc != ')')
					return ind + get_c_indent();
indhere:		if (findc != ')')
					point++;
				if (!parse_string(1, "[ \t\f]*/[*/]"))
					point += parse_string(1, "[ \t\f]*![^ \t\f\n]");
				return limit_c_contin_indent(ind);
			case ')':
				plevel++;
				break;
			case '?':	/* indent : to matching ? */
				if (qlevel-- > 0 || plevel > 0 || findc != ':')
					break;
				point--;
				goto indhere;
			case ':':
				qlevel++;
				break;
			case '=': case '!':
				if (character(point + 1) == '=' || plevel > 0
				    || character(lastl - 1) == ',' || findc == ')')
					break;
				goto indhere;
			case '\n':		 // skip over comments
				cind_try_space(); // at ends of lines
				break;
			case '"': case '\'':
				point++;
				c_skip_item(c, RE_REVERSE);
				break;
			case '<':
				if (character(point + 1) == '=')
					break;
				if (alevel-- > 0 || !(c_extra_keywords & (CPP_KEYWORDS
					| JAVA_KEYWORDS | CSHARP_KEYWORDS)))
					break;
				// Someday, test if there's an identifier before the <
				// that's the name of a template, not just if it's a
				// template keyword.
				if (parse_string(-1, "</word>template[ \t\n]*")) {
					point++;
					point += parse_string(1, "[ \t]*!<word>");
					return limit_c_contin_indent(ind);
				}
				break;
			case '>':
				if (character(point + 1) != '=')
					alevel++;
				break;
		}
	}
	if (c_matches_in(point, orig, ":")
		  && !c_matches_in(point, orig, "%?"))
		return get_indentation(point) + c_align_inherit;

	if (character(lastl - 1) == ',') {
		if (character(lastl - 2) != ')') {
			point = lastl;
			to_begin_line();
		}
		point += parse_string(1, "(.*[ \t\f]on)?[ \t\f]+");
		if (in_c_comment(c_look_back))
			to_indentation();
		return limit_c_contin_indent(ind);
	}
	return c_default_indent(ind, orig, 2);
}

c_default_indent(ind, orig, defmult)
{
	if (matches_at(orig, 1, "[ \t]*//"))
		return ind;
	if ((c_extra_keywords & (CPP_KEYWORDS | JAVA_KEYWORDS | CSHARP_KEYWORDS))
		  && matches_at(orig, 1, "[ \t]*class</word>"))
		return ind;		// Put class keyword in continuation back on left.
	if (ind)
		return ind + defmult * get_c_indent() + c_contin_offset;
	if (parse_string(1, "[ \t\f]*(struct|union|class|namespace)[ \t\f\n]+"
					 "[a-zA-Z0-9_]*[ \t\f\n]*{[ \t\f\n]*[\n/]"))
		return ind + c_top_struct;
	if (just_open(orig))
		return c_top_braces;
	point = orig;
	cind_space();
	return is_contin_statement(point, 0, orig)
			? c_top_contin : c_param_decl;
}

restore_check_version()
{
	char exe_version[30], *s;

	// To deal with change in version string format, process
	// versioned_file_string into a standard format: strip colon and
	// make part after b two-digit.
	strcpy(exe_version, versioned_file_string);
	if (s = strchr(exe_version, ':'))
		strcpy(s, s + 1);
	s = exe_version + strlen(exe_version) - 1;	// Last char.
	if (s[-1] == 'b') {	// One digit, make two.
		s[2] = 0;
		s[1] = s[0];
		s[0] = '0';
	}

	if (!load_from_state || strchr(exe_version, 'a'))
		return;
	if (strcmp(eel_version, exe_version) >= 0)
		if (!strchr(eel_version, 'b') || strchr(exe_version, 'b'))
			return;
	// Each user has a private state file, so we can't update them
	// all automatically.
	save_var bufnum = zap(HELPBUF);
	bprintf("Your version %s state file requires updating.\n"
			"You need to start Epsilon with its -quickup flag.\n",
			eel_version);
	bprintf("Until you do this, Epsilon will not work correctly.\n");
	view_help(" Update Needed! ", 0);
}

color_go_back_lines()
{
	save_var narrow_start = point - 10000;
	nl_reverse();
	nl_reverse();
	nl_reverse();
}

color_for_narrowed_area(from, to)
{
	int i, j, give_up = 0;

	save_var point, abort_searching = 0;
	init_color_code();
	to = MIN(to, size() - narrow_end);
	for (;;) {
		if (get_tagged_region("needs-color", from, &i, &j)) {
			if (i < from) {
				point = from;		// From isn't colored, so start before it.
				color_go_back_lines();	// Color a few extra lines before
										// to speed up scrolling.
				if (point > i) // Next, find starting place known to be on 
					i = recolor_from_here(i); // the edge of a colored region.
			}				// (Or use known safe place if it's closer.)
		} else {
			if (j >= to) {		// From is colored, see if good region extends
				return;			// far enough.  If not, get following bad reg.
			} else if (!get_tagged_region("needs-color", j, &i, &j)) {
				int a, b, res;

				say("Coloring internal error 1");
				buffer_printf("debug-color", "Coloring internal error 1:\n");
				buffer_printf("debug-color", "bad color 0 from %d to %d in %s, size %d\n",
							  i, j, bufname, size());
				res = get_tagged_region("needs-color", from, &a, &b);
				buffer_printf("debug-color", "color %d at %d (from %d to %d) in %s, size %d\n",
							  res, from, a, b, bufname, size());
				res = get_tagged_region("needs-color", j - 1, &a, &b);
				buffer_printf("debug-color", "color %d at %d (from %d to %d) in %s, size %d\n",
							  res, j - 1, a, b, bufname, size());
				return;
			}
		}
		j = MIN(j, to);		// Color only some of it.
		if (i >= j)
			return;
		j = recolor_range(i, j);
		set_tagged_region("needs-color", i, j, 0);
		if (give_up++ > 100) {
			say("Coloring internal error 2");
			when_displaying = 0;
			return;
		}
	}
}

paren_follows() // is there whitespace followed by a "(" at point in buffer?
{		// if so, we assume it's a function name
	for (;; point++) {
		switch (curchar()) {
			case ' ': case '\t': case '\n':
				continue;
			case '(':		return 1;
			case '<':		if (find_template_end()) {
								point--;
								break;
							}
			default:		return 0;
		}
	}
}

preserve_color_settings(int *orig_fg, int *orig_bg, int *orig_schemes,
						int orig_scheme)
{
	int i, j, k, n, classes = number_of_color_classes();
	int schemes = count_color_schemes(), fg_color, bg_color;

	*get_color_scheme_variable() = selected_color_scheme;
	if (want_record_customizations()) {
		k = prep_customizations();
		if (selected_color_scheme != orig_scheme)
			list_one(k, color_scheme_variable_index(), 1);
		for (i = 0; i < schemes; i++) {
			selected_color_scheme = orig_schemes[i];
			for (j = 0; j < classes; j++) {
				fg_color = get_foreground_color(j, 1);
				bg_color = get_background_color(j, 1);
				n = i * classes+ j;
				if (fg_color != orig_fg[n] || bg_color != orig_bg[n])
					list_one_color_setting(k, j, 1);
			}
		}
		selected_color_scheme = *get_color_scheme_variable();
		say("");
	} else
		show_binding("To remember these colors for "
					 "future sessions, use %s.", "list-customizations");
}

#define FG(c)		((c) & 15)	/* extract foreground & background */
#define BG(c)		(((c) >> 4) & 7)
#define COLOR_COL	56
#define COLOR_LINE	2
#define SCRPLACE(c)	(COLOR_COL + FG(c)), (COLOR_LINE + BG(c))

command set_color()		/* query user to set colors */
{
	int class, color, i, j, *orig, classes, docbuf, wantwin;
	int swin, ccwin, cwin, infowin, newcolor, newscheme, orig_scheme;
	int *orig_fg, *orig_bg, *orig_schemes;
	int fg_color, bg_color;
	char buf[100];
	int sbuf, ccbuf, cbuf, infobuf, orig_win = window_handle, win, page;
	char *new, *old;
	jmp_buf this_level;

	if (is_unix == IS_UNIX_XWIN) {
		set_color_xwin();
		return;
	}
	save_var top_level = &this_level;
	classes = number_of_color_classes();

	sbuf = tmp_buf();	/* list of color schemes */
	bufnum = sbuf;	/* now highlight current line in buffers */
	add_region(point_spot, point_spot, color_class menu_highlight,
			   REGLINE);
	make_color_scheme_list(sbuf);

	orig_scheme = selected_color_scheme;
	record_color_settings(&orig_fg, &orig_bg, &orig_schemes);
	selected_color_scheme = orig_scheme;
	orig = (int *) malloc(classes * sizeof(int));
	for (i = 0; i < classes; i++)
		orig[i] = alter_color(i, -1);

	class = iter = page = 0;

	bufnum = docbuf = tmp_buf();	// read help text for classes into buf
	on_exit buf_delete(docbuf);
	if (new = lookpath("colclass.txt"))
		file_read(new, FILETYPE_AUTO);
	ccbuf = tmp_buf();	/* list of color classes */
	cbuf = tmp_buf();	/* window with colors */ 
	infobuf = tmp_buf();	/* help buffer with options */ 

	bufnum = ccbuf;
	add_region(point_spot, point_spot, color_class menu_highlight,
			   REGLINE);
	for (i = 0; i < classes; i++)
		bprintf("%s\n", name_color_class(i));
	delete(point - 1, point);

winagain:
	swin = add_popup(4, 1, 20, 10, ALL_BORD(BSINGLE), sbuf);
	window_title(swin, BTOP, TITLECENTER, " Color Schemes ");
	set_menu_look(swin);

	ccwin = add_popup(29, 1, 21, 10, ALL_BORD(BSINGLE), ccbuf);
	display_column = 0;		/* prevent continuation lines */
	window_title(ccwin, BTOP, TITLECENTER, " Color Classes ");
	set_menu_look(ccwin);

	if (!(has_feature & FEAT_ANYCOLOR)) {		// Matrix of color choices.
		cwin = add_popup(55, 1, 18, 10, ALL_BORD(BSINGLE), cbuf);
		window_title(cwin, BTOP, TITLECENTER, " Colors ");
		invisible_window = 1;
		insert('\n');
		if (selectable_colors > 128)
			window_title(cwin, BBOTTOM, TITLERIGHT(1), " More ");
	}

	infowin = add_popup(2, 14, 75, 8, _view_border, infobuf);
	display_column = 0;		/* prevent continuation lines */
	mouse_cursor = &left_pointer;
	text_color = is_gui ? color_class text : color_class menu;
	display_scroll_bar = 0;
	newscheme = 2;
	if (has_feature & FEAT_COLOR_DIALOG) {
		display_dialog_box("SetColorBox", "Set colors",
						   swin, ccwin, infowin, "OK", "Cancel", "Help");
		save_var permit_window_keys = PERMIT_SCROLL_KEY;
		save_var return_raw_buttons = 1;
		wantwin = ccwin;
	}

	if (setjmp(top_level))	/* come here if subcommand's aborted */
		;
	for (;;) {
		color = alter_color(class, -1);
		fg_color = get_foreground_color(class, 0);
		bg_color = get_background_color(class, 0);

		if ((color >= 128) != page && !newscheme)
			newscheme = 1;
		page = (color >= 128);
		if (newscheme) {
			if (!(has_feature & FEAT_ANYCOLOR)) {	// Sample window has 2 pages.
				window_handle = cwin;
				point = page;
			}
			window_handle = swin;
			if (newscheme == 2) {
				make_color_scheme_list(bufnum);
				delete(size() - 1, size());	/* no final \n */
			}
			point = 0;
			col_search(name_name(selected_color_scheme), 0);
			if (!(has_feature & FEAT_ANYCOLOR)) {
				j = page ? 128 : 0;
				for (i = 0; i < 128; i++, j++)
					term_write_attr(SCRPLACE(i), '+', j);
			}
		}
		window_handle = ccwin;
		point = 0;
		sprintf(buf, "^%s$", name_color_class(class));
		re_search(1, buf);
		to_begin_line();
		buf_zap(infobuf);
		window_handle = infowin;
		if (!(has_feature & FEAT_COLOR_DIALOG)) {
			bprintf("Type F or B to change the color scheme, D to "
					"define a new color scheme.\n");
			bprintf("Or use the arrow keys to change the "
					"%s color class.\n", name_color_class(class));
		} else
			bprintf("Sample text using the %s color class:\n",
					name_color_class(class));
		i = point;
		bufnum = docbuf;
		point = 0;
		sprintf(buf, "^%s\t+", name_color_class(class));
		if (re_search(1, buf))
			buf_xfer(infobuf, point, give_end_line());
		else
			buf_printf(infobuf, "This is sample text.");
		bufnum = infobuf;
		j = point + 1;
		if (!(has_feature & FEAT_COLOR_DIALOG)) {
			bprintf("\nType N or P to set other color classes, "
					"T to try out\n");
			bprintf("these colors temporarily, S to set "
					"these colors permanently,\n");
			bprintf("or C to cancel.  Color was initially "
					"0x%02x, this color is 0x%02x.",
					orig[class], color);
		}
		set_character_color(i, j, class);
		point = 0;

		dialog_checkboxes = ((fg_color & EFONT_BOLD) ? 1 : 0)
							| ((fg_color & EFONT_ITALIC) ? 2 : 0)
							| ((fg_color & EFONT_UNDERLINED) ? 4 : 0);

		if (has_feature & FEAT_ANYCOLOR) {
			window_handle = wantwin;
			maybe_refresh();
			getkey();
			wantwin = window_handle;
		} else {
			window_handle = cwin;
			do {
				maybe_refresh();
				term_write_attr(SCRPLACE(color), 'C', color);
				full_getkey("Your choice: ", 0);
				term_write_attr(SCRPLACE(color), '+', color);
			} while (!IS_TRUE_KEY(key) && !IS_MOUSE_DOWN(key));
		}
		key = toupper(key);
		newcolor = -100;	// Assume it's an invalid key.
		newscheme = 0;
		if (binding_of(key, reg_tab) == (short) next_page
			  || binding_of(key, reg_tab) == (short) previous_page) {
			do_command(binding_of(key, reg_tab));
			newcolor = -99;		// Mark as a valid key.
		}
		if (has_feature & FEAT_COLOR_DIALOG) {
			window_handle = ccwin;	// Scrolling might have changed these.
			class = lines_between(0, point);

			window_handle = swin;
			grab_full_line(bufnum, buf);
			i = find_index(buf);
			if (i > 0 && name_type(i) == NT_COLSCHEME)
				selected_color_scheme = i;
		}
		if (key_is_button == 4) {
			i = comm_dlg_color(fg_color, "Foreground color");
			if (i != -1)
				set_color_pair(class, i | (fg_color & EFONT_STYLE), bg_color);
		} else if (key_is_button == 5) {
			i = comm_dlg_color(bg_color, "Background color");
			if (i != -1)
				set_color_pair(class, fg_color, i);
		} else if (key_is_button == 8) {
			set_color_pair(class, fg_color ^= EFONT_BOLD, bg_color);
		} else if (key_is_button == 9) {
			set_color_pair(class, fg_color ^= EFONT_ITALIC, bg_color);
		} else if (key_is_button == 10) {
			set_color_pair(class, fg_color ^= EFONT_UNDERLINED, bg_color);
		} else if (key == 'C' || key == 'Q' || key == ESC || user_abort
				   || key_is_button == 2) {
			restore_color_settings(orig_fg, orig_bg, orig_schemes);
			selected_color_scheme = orig_scheme;
			say("");
			break;
		} else if (key == 'S' || key == ' ' || IS_ENTER_KEY(key)
				   || key_is_button == 1) {
			preserve_color_settings(orig_fg, orig_bg, orig_schemes,
									orig_scheme);
			break;
		} else if (IS_MOUSE_DOWN(key)) {
			win = window_at_coords(mouse_y, mouse_x, mouse_screen);
			if (!is_window(win))
				continue;
			window_handle = win;
			get_text_area(win);
			if (win == cwin) {
				if (bar_on_border()) {
					i = scroll_bar_code();
					if ((i ? (i > 0) : !page) != point)
						newcolor = color ^ 128;
				} else {
					newcolor = (mouse_y - COLOR_LINE) * 16;
					newcolor += (mouse_x - COLOR_COL);
					if (page)
						newcolor += 128;
				}
			} else if (in_text_area(mouse_x, mouse_y))
				point_to_pos(mouse_x, mouse_y, mouse_screen);
			else if (display_scroll_bar && bar_on_border())
				mouse_scroll_bar();
			else
				continue;
			if (win == ccwin)
				class = lines_between(0, point);
			else if (win == swin)
				newscheme = 1;
			else if (win == infowin && isupper(curchar())
					 && character(point - 1) == ' '
					 && character(point + 1) == ' ')
				ungot_key = curchar();
			if (win != infowin)
				wantwin = win;
		} else if ((has_feature & FEAT_COLOR_DIALOG)
				   && full_key == NUMSHIFT(GREYTAB)) {
			prev_dialog_item();
			wantwin = window_handle;
			continue;
		} else if ((has_feature & FEAT_COLOR_DIALOG) && generic_key(key) == '\t') {
			next_dialog_item();
			wantwin = window_handle;
			continue;
		} else if ((has_feature & FEAT_COLOR_DIALOG) && key == GETFOCUS) {
			win = window_at_coords(mouse_y, mouse_x, mouse_screen);
			if (!is_window(win) || win == infowin)
				continue;
			wantwin = win;
		}
		if (!(has_feature & FEAT_ANYCOLOR))
			switch (get_direction()) {
				case BTOP:		newcolor = color + 256 - 16; break;
				case BBOTTOM:	newcolor = color + 16; break;
				case BLEFT:		newcolor = color + 256 - 1; break;
				case BRIGHT:	newcolor = color + 1; break;
			}
		else
			switch (get_direction()) {
				case BTOP: case BLEFT:
					key = (win == ccwin) ? 'P' : 'B'; break;
				case BBOTTOM: case BRIGHT:
					key = (win == ccwin) ? 'N' : 'F'; break;
			}
		if ((has_feature & FEAT_COLOR_DIALOG) && key == 'N' && win == swin)
			key = 'F';
		if ((has_feature & FEAT_COLOR_DIALOG) && key == 'P' && win == swin)
			key = 'B';
		switch (key) {
			case 'N': class++; break;
			case 'P': class--; break;
			case 'T':
				show_binding("Continue changing colors "
					"by typing %s", "exit-level");
				remove_window(swin);
				remove_window(ccwin);
				remove_window(cwin);
				remove_window(infowin);
				recursive_edit();
				goto winagain;
				break;
			case 'F':
				wantwin = window_handle = swin;
				if (!nl_forward())
					point = 0;
				newscheme = 1;
				break;
			case 'B':
				wantwin = window_handle = swin;
				if (!nl_reverse())
					point = size();
				newscheme = 1;
				break;
			case 'D':
				wantwin = swin;
				maybe_refresh();
				get_strpopup(buf, "New scheme name: ", "", "set-color");
				if (*buf) {
					new_variable(buf, NT_COLSCHEME, 0, 0);
					new = varptr(find_index(buf));
					old = varptr(selected_color_scheme);
					for (i = 0; i < ptrlen(old, 1); i++)
						*new++ = *old++;
					newscheme = 2;
				}
				break;
			default:
				if (binding_of(key, root_keys) == (short) help
					|| key == 'h' || key == '?'
					|| key_is_button == 3) {
					if (!(has_feature & FEAT_COLOR_DIALOG))
						say("Press <Space> to continue.");
					help_on_command("set-color");
					if (!(has_feature & FEAT_COLOR_DIALOG))
						newscheme = 1;
				} else if (IS_TRUE_KEY(key) && newcolor == -100) {
					maybe_ding(bell_on_bad_key);
					newscheme = 1;
				}
				break;
		}
		if (newcolor > -99)	/* did it change? */
			alter_color(class, newcolor % selectable_colors);
		else if (newscheme) {
			window_handle = swin;
			grab_full_line(bufnum, buf);
			i = find_index(buf);
			if (i > 0 && name_type(i) == NT_COLSCHEME)
				selected_color_scheme = i;
		} else
			class = (class + classes) % classes;
	}
	remove_window(swin);
	remove_window(ccwin);
	if (!(has_feature & FEAT_ANYCOLOR))
		remove_window(cwin);
	remove_window(infowin);
	buf_delete(ccbuf);
	buf_delete(cbuf);
	buf_delete(infobuf);
	if (is_window(orig_win))
		window_handle = orig_win;
	buf_delete(sbuf);
	full_redraw = 1;
	*get_color_scheme_variable() = selected_color_scheme;
	free(orig);
	free(orig_fg);
	free(orig_bg);
}

full_comp_read(trystr, prmpt, finder, flags, edit, def)
char *trystr, *prmpt, *def;
char *(*finder)();
{
	int i, j, num_match = 0, maxlen = ptrlen(trystr) - 1, pbuf, res;
	int pwin, editwin, owin = window_handle, recall_cnt = 0;
	int olen = 0, in_menu = 0, menu = -1, mbuf = 0;
	jmp_buf this_level;
	int *(*lister)();
	char password_display_class[256];

	save_var last_index;
	save_var _help_on_cmd = last_index;
	save_var _doing_input = DI_LINEINPUT;
	save_var _doing_completion = (finder != 0);
	save_var top_level = &this_level;

	while (!in_macro() && !char_avail() && unseen_msgs())
		check_abort();	// Wait for unseen msgs to display so we don't cover them.
	pwin = add_popup(0, echo_line, screen_cols - 1, 1, 0,
					 pbuf = tmp_buf());	/* make system window with prompt */
	stuff(prmpt);
	i = current_column();
	system_window = 1;	/* don't let user switch to it */
	point = display_column = 0;
	text_color = color_class echo_area;
	_display_class = _echo_disp_class;

	editwin = add_popup(i, echo_line, screen_cols - i, 1,
						0, edit);
	display_column = 0;	/* horizontally scroll this window */
	text_color = color_class echo_area;
	_display_class = _echo_disp_class;
	if (flags & PASSWORD_PROMPT) {
		set_chars(password_display_class, '*', 256);
		_display_class = password_display_class;
		build_first = 1;
		save_var disable_recording_last_keys = 1;
	}
	mouse_handler = allow_mouse_switching;
	if (insert_default_response == 1) {
		highlight_on();
		shift_selecting = 2;
	}
	if ((flags & POP_UP_PROMPT) && (has_feature & FEAT_GUI_DIALOGS)) {
		remove_window(pwin);
		do_display_dialog_box("OneLineBox", prmpt, editwin, 0, 0,
							  "OK", "Cancel", "Help", NULL, NULL);
	}
	maybe_refresh();
	if (setjmp(top_level)) {
		if (in_menu == 2) {	/* menu is half-created */
			remove_window(menu);
			buf_delete(mbuf);
			in_menu = mbuf = 0;
			if (is_gui && want_gui_prompts) {
				res = INP_ABORT;
				goto done;
			}
		}
	}
	for (;;) {
		if (mbuf) {
			window_handle = menu;
			if (in_menu && line_gets_highlight())
				make_line_highlight();
			else
				remove_line_highlight();
		}
		window_handle = in_menu ? menu : editwin;
		mode_keys = in_menu ? menu_tab : comp_tab;
		_doing_input = (response_tracks_list && mbuf)
					   ? DI_LISTBOX : DI_LINEINPUT;
		if (_doing_input == DI_LISTBOX) {
			maybe_refresh();
			getkey();
			cmd_result = 0;
			run_topkey();
			res = cmd_result;
		} else
			res = do_recursion();	/* let user edit input */
		if (!is_window(editwin) || in_menu && !is_window(menu)) {
			res = INP_ABORT;
			break;
		} else if (window_handle == (in_menu ? editwin : menu))
			in_menu = !in_menu;
		window_handle = editwin;
		if (res == INP_ADD) {
			window_handle = editwin;
			run_topkey();
			if (search_in_menu && mbuf) {
				window_handle = editwin;
				grab(0, maxlen, trystr);
				if (strlen(trystr) >= olen) {
					window_handle = menu;
					point = 0;
					col_search(trystr, 0);
				}
			} else
				in_menu = 0;
		} else if (res == INP_SEEN)
			in_menu = 0;
		else if (res == INP_TO_LIST && mbuf) {
			window_handle = menu;
			in_menu = 1;
		} else if ((res == INP_GET || res == INP_GET_EXIT) && mbuf) {
			window_handle = menu;
			get_menu_line(edit);
			window_handle = editwin;
			in_menu = 0;
		} else if (!res) {
			if (_doing_input == DI_LISTBOX)
				if (in_menu) {
					window_handle = menu;
					get_menu_line(edit);
				} else {
					window_handle = editwin;
					grab(0, maxlen, trystr);
					window_handle = menu;
					point = 0;
					col_search(trystr, 0);
				}
			continue;
		}
		restart_file_paths();
		if (res == INP_ABORT || res == INP_GET_EXIT || res == INP_EXIT)
			break;
		else if (res == INP_GRAB_WORD) {
			grab_buffer_word(owin, 0);
			continue;
		} else if (res == INP_GRAB_LINE) {
			grab_buffer_word(owin, 1);
			continue;
		} else if (res == INP_DEFAULT) {
			if (!size())	/* type in default text */
				stuff(def);
			in_menu = 0;
			continue;
		} else if (res == INP_LAST_RECALL || res == INP_NEXT_RECALL
				   || res == INP_PREV_RECALL) {
			if (res == INP_LAST_RECALL)
				get_last_recall(finder, trystr);
			else
				recall_cnt = get_last_recall_from_end(finder, trystr,
					recall_cnt + (res == INP_NEXT_RECALL ? -1 : 1));
			if ((*trystr || res != INP_LAST_RECALL) && !mbuf) {
				delete(0, size());
				stuff(trystr);
				if (recall_response_selected && *trystr) {
					highlight_on();
					shift_selecting = 2;
				}
			}
			in_menu = 0;
		} else if (res == INP_GET_OTHER) {
			if (inp_get_other_name && *inp_get_other_name) {
				delete(0, size());
				stuff(inp_get_other_name);
			}
			in_menu = 0;
		} else if (res == INP_SHOW_MATCHES || res == INP_RECALL) {
show_matches:		grab(0, maxlen, trystr);	/* typed '?' */
					if (is_highlight_on() || res == INP_RECALL) {
						if (flags & COMP_FILE)
							*get_tail(trystr, 0) = 0;
						else
							*trystr = 0;
						highlight_off();
					}
					olen = strlen(trystr);
					if (!mbuf) {
						menu = new_menu(mbuf = tmp_buf());
						if (is_gui && want_gui_prompts) {
							remove_window(pwin);
							say("");
							make_prompt_box(res == INP_RECALL, prmpt, editwin, menu, finder);
						}
					}
					in_menu = 2;		/* initializing */
					if (res == INP_RECALL)
						lister = list_recalls;
					else if (completion_lister)
						lister = completion_lister;
					else
						lister = list_matches;
					make_menu(menu, trystr, finder, flags, editwin, mbuf, lister);
					in_menu = 1;
					mode_keys = menu_tab;
					cursor_to_column = -1;
		} else if (res == INP_TRY || res == INP_COMPLETE) {
			if (res == INP_TRY && cursor_to_column
				  == current_column() + 1)
				break;	/* exactly 1 match last time */
			grab(0, maxlen, trystr);
			i = strlen(trystr);
			num_match = complete(trystr, finder, flags);
			j = strlen(trystr);
			if ((flags & SPACE_VALID) && key == ' ' && trystr[i] == ' '
				  && could_add_space(trystr, finder, flags) == 1)
				i++, num_match++;
			else if (num_match && i != j)
				i = j;
			else if (num_match != 1) {
				if ((flags & SPACE_VALID) && key == ' ' && is_window(menu)
//					  && trystr[i] == ' '
// Given dirs c:\program files and c:\programs, this commented-out code
// made selecting the former fail.
					  && could_add_space(trystr, finder, flags)) {
					trystr[i++] = ' ';
				} else {
					maybe_ding(bell_on_completion);
					if (completion_pops_up)
						ungot_key = '?';
				}
			}
			if ((flags & COMP_FILE) && num_match == 1 && i > 0
				  && is_path_separator(trystr[i - 1]))
				num_match++;	/* might be directory */
			trystr[i] = 0;
			buf_zap(edit);
			stuff(trystr);
			if (num_match == 1) {
				if (res == INP_COMPLETE)
					break;	/* <Esc> worked */
				maybe_refresh();
				cursor_to_column = current_column() + 1;
			}
		}
	}
done:
	if (res != INP_ABORT)
		grab(0, maxlen, trystr);
	else if (!user_abort)		/* user deleted edit window */
		user_abort = 1;
	if (res == INP_EXIT && must_match(trystr, finder, flags))
		goto show_matches;
	if (is_window(owin))
		window_handle = owin;
	remove_window(editwin);
	remove_window(pwin);
	buf_delete(pbuf);
	if (mbuf) {
		remove_window(menu);
		buf_delete(mbuf);
	}
}

set_case_indirect(short *table)	// Does nothing now.
{
	UNREFERENCED_PARAMETER(table);
}

set_list_keys(table)	// set keytable for list of items, with n & p moving
short *table;
{
	table['n'] = (short) down_line;
	table['p'] = (short) up_line;
}

int binding_of(int k, short *tab)
{
	return (k & CMD_INDEX_KEY) ? (k & ~CMD_INDEX_KEY) :
			(tab[k] > 0) ? tab[k] : tab[generic_key(k)];
}

check_running_processes()
{
	char *buf, *open = 0, *s, msg[FNAMELEN];

	if (!process_warn_on_exit)
		return;
	buf = buffer_list(1);
	save_var bufnum;
	do {
		if (preserve_buffer_type(buf))
			if (process_warn_on_exit == 2 || !exit_this_process(buf)) {
				open = buf;
				break;
			}
	} while (buf = buffer_list(0));
	restore_vars();
	if (open) {
		tiled_only();
		locate_window(open, "");
		refresh();
		s = preserve_buffer_type(open);
		if (process_warn_on_exit != 2) {
			if (s && !strcmp(s, "Process"))
				error("Couldn't exit from concurrent process.");
			else
				error("Couldn't exit %s job.", s);
		} else {
			if (s && !strcmp(s, "Process"))
				sprintf(msg, "A process is running, confirm exit? ");
			else
				sprintf(msg, "%s job is running, confirm exit? ", s);
			if (!ask_yn("Confirm exit", msg, "Yes", "No", 2))
				error("Exiting canceled.");
		}
	}
}

chars_to_bytes(byte *b, char *s)
{
	while (*b++ = *s++)
		;
}

bytes_to_chars(char *s, byte *b)
{
	while (*s++ = *b++)
		;
}

load_eel_from_path(char *f, int gripe)
{
	char *p = lookpath(f), orig_dir[FNAMELEN], new_dir[FNAMELEN];
	int res, tmp = tmp_buf();

	on_exit buf_delete(tmp);
	if (!p) {
		if (gripe == 2)
			say("Can't find EEL source file %f.", f);
		return 1;
	}
	getcd(orig_dir);
	strcpy(new_dir, p);
	*get_tail(new_dir, 0) = 0;
	on_exit chdir(orig_dir);
	chdir(new_dir);
	res = eel_compile(p, 1, "-n -q", bufnum_to_name(tmp));
	if (buf_size(tmp)) {
		save_var bufnum = create("eel-errors");
		point = size();
		bprintf("\nLoading %s:\n", p);
		grab_buffer(tmp);
	}
	if (res == 1) {
		if (gripe)
			show_text(0, 300, "Error loading %f, see %s buffer.",
					  p, "eel-errors");
		return 2;
	}
	return 0;
}

dired_create_directory()
{
	save_var _show_help_on = "dired-mode";
	do_create_directory(1);
}

dired_shell_command()
{
	char dir[FNAMELEN], file[FNAMELEN];

	save_var _show_help_on = "dired-mode";
	dired_getname(file, 1);
	dired_getdir(dir);
	do_shell_command(file, dir);
}

int dired_do_delete();
int dired_do_copy();
int dired_do_move();
int dired_do_uppercase();
int dired_do_lowercase();
int dired_do_regex_rename();
char *dired_prompt_regex_replace();

dired_execute()
{
	char url[FNAMELEN];
	int del, copy, move, uc, lc, rr;

	save_var dired_confirmation, _show_help_on = "dired-mode";
	if (is_remote_file(get_dired_buffer_pattern()) && exist(REMOTE_FTP_OP_BUFFER))
		zap(REMOTE_FTP_OP_BUFFER);
	del = dired_do_no_param("Delete", "\nD ", "deleted", dired_do_delete, 0);
	copy = dired_do_param("Copy", "\nC ", "copied", dired_do_copy);
	move = dired_do_param("Move", "\nM ", "moved", dired_do_move);
	uc = dired_do_no_param("Uppercase", "\nU ", "uppercased",
						   dired_do_uppercase, 0);
	lc = dired_do_no_param("Lowercase", "\nL ", "lowercased",
						   dired_do_lowercase, 0);
	rr = dired_do_no_param("Regex-rename", "\nR ", "regex-renamed",
						   dired_do_regex_rename, dired_prompt_regex_replace);
	if (del + copy + move + uc + lc + rr == 0)
		error("No files marked");
	modified = 0;
	if (!is_remote_file(get_dired_buffer_pattern()) && (move + copy + uc + lc + rr))
		dired_refresh();	// could be rename or similar, make new listing
	if (is_remote_file(get_dired_buffer_pattern())) {
		dired_getdir(url);
		send_remote_ftp_op(url);
	}
	fix_window_start();
	iter = 0;
}

dired_strip_parent(char *tag)
{
	char name[FNAMELEN];

	save_var point = 0;
	while (search(1, tag) == 1) {
		dired_getname(name, 3);
		if (!strcmp(get_tail(name, 0), "..")) {
			save_var modified;	// leave modify bit alone
			dired_unmark();
			say("Skipping .. for safety; delete this hierarchy separately.");
		}
	}
}

dired_do_no_param(op, tag, past, func, extra)
char *op, *tag, *past, *(*extra)();
int (*func)();
{
	int count = dired_ask(op, tag, "? ", NULL);

	if (count > 1 && tag[1] == 'D')
		dired_strip_parent(tag);
	if (count > 0)
		dired_do(past, tag, count, func, extra ? (*extra)() : NULL);
	return count;
}

int dired_do_set_attributes(char *fname, char *dir)
{
	char flags[FNAMELEN], cmd[FNAMELEN], buf[FNAMELEN], msg[FNAMELEN];
	char abbrev[FNAMELEN];

	save_var _show_help_on = "dired-mode";
	strcpy(abbrev, fname);
	abbreviate_file_name(abbrev, screen_cols - 50);
	if (is_unix) {
		sprintf(msg, "Flags for chmod command on %s", abbrev);
		get_str_auto_def(flags, msg);
		if (!*flags)
			return 0;
		sprintf(cmd, "chmod %s '%s'", flags, fname);
		strcpy(buf, "-dired-set-attributes-result");
		pipe_text(NULL, buf, cmd, dir,
				  PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0);
		if (buffer_size(buf)) {
			to_buffer(buf);
			discardable_buffer = 1;
			point = 0;
			return 0;
		}
	} else {
		struct file_info fi;
		int attrs;

		if (!check_file(fname, &fi))
			error("Can't get attributes for %f", fname);
		attrs = fi.opsysattr;
		sprintf(cmd, "%s%s%s%s",
				(attrs & DOS_ATTR_RDONLY) ? "R" : "",
				(attrs & DOS_ATTR_SYSTEM) ? "S" : "",
				(attrs & DOS_ATTR_HIDDEN) ? "H" : "",
				(attrs & DOS_ATTR_ARCHIVE) ? "A" : "");
		sprintf(msg, "Set attributes (RO, Sys, Hid, Arch) for %s to: ", abbrev);
		get_strnone(flags, msg, cmd);
		save_var case_fold = 1;
		attrs = (strstr(flags, "R") ?  DOS_ATTR_RDONLY : 0)
				| (strstr(flags, "S") ? DOS_ATTR_SYSTEM : 0)
				| (strstr(flags, "H") ? DOS_ATTR_HIDDEN : 0)
				| (strstr(flags, "A") ? DOS_ATTR_ARCHIVE : 0);
		set_file_opsys_attribute(fname, attrs);
	}
	return 1;
}

dired_getindex(s,i)	/* put current line in s, starting at column i */
char *s;
{
	save_var point;
	move_to_column(i);
	re_search(1, " *");		/* skip initial spaces */
	if (parse_string(1, FILECHARS) >= ptrlen(s))
		error("File name too long.");
	parse_string(1, FILECHARS, s);
}

when_setting_tab_size()
{
	char varname[FNAMELEN], msg[FNAMELEN];
	int i;

	sprintf(varname, "%s-tab-override", major_mode);
	i = find_index(varname);
	sprintf(msg, "Set the mode-specific tab size variable %s to %d also?",
			varname, tab_size.default);
	if (i && tab_size == tab_size.default && get_num_var(i) != tab_size.default
		  && get_num_var(i) > 0 && !now_parsing_cmds)
		if (ask_yn("Set override too?", msg, "Yes", "No", 1))
			set_num_var(i, tab_size.default);
}

show_window_caption()
{
	if ((has_feature & FEAT_SET_WIN_CAPTION) && is_window(window_handle) == ISTILED) {
		char *s = get_tail(filename, 1);
		char title[FNAMELEN], hostname[FNAMELEN];

		if (s && *s && force_save_as != FSA_NEWFILE)
			sprintf(title, window_caption_file, s);
		else
			sprintf(title, window_caption);
		if (want_display_host_name && opsys == OS_UNIX
			  && !gethostname(hostname, want_display_host_name))
			sprintf(title + strlen(title), "@%s", hostname);
		if (display_func_name_method() == 2 && *display_func_name)
			sprintf(title + strlen(title), " - %s", display_func_name);
		set_window_caption(window_handle, title);
	}
}

assemble_mode_line(char *line)
{
	int want_func = 0, want_mode = 0, want_name = 0, size = 0, maxsize;
	char bname[FNAMELEN * 2], funcname[FNAMELEN], mode[FNAMELEN], *s, *t, *q;
	char extra[FNAMELEN];

	strcpy(extra, mode_more_extra);
	if (mode_extra)
		strcat(extra, mode_extra);
	for (s = mode_format; *s; s++)
		if (*s == '%' && s[1])
			switch (*++s) {
				case 'b':	want_name = 1; break;
				case 'M':	want_mode = 1; break;
				case 'f':	want_func = 1; break;
				case 'E':	size += strlen(extra);
							break;
				default:	// Approximate the size of various options.
					if (strchr("dDmsSanehH", *s))
						size += 2;
					else if (strchr("cClPp", *s))
						size += 4;
					else
						size++;
					break;
			}
		else
			size++;
	*bname = *funcname = *mode = 0;
	if ((want_func) && display_func_name_method() == 1) {
		strncpy(funcname, display_func_name, ptrlen(funcname) - 2);
		if (*funcname)
			strcat(funcname, " ");
	}
	size += strlen(funcname);
	if (want_mode)
		make_mode_for_mode_line(mode);
	size += strlen(mode);
	maxsize = MIN(window_width, ptrlen(line));
	if (want_name)
		make_name_for_mode_line(bname, maxsize - size);
	for (s = mode_format, t = line; (*t++ = *s) && ptrlen(t) > 2; s++)
		if (*s == '%' && s[1]) {
			q = 0;
			switch (*++s) {
				case 'b':	q = bname; break;
				case 'M':	q = mode; break;
				case 'f':	q = funcname; break;
				case 'E':	q = extra; break;
				default:	*t++ = *s; break;
			}
			if (q) {
				strncpy_double_percents(--t, q, ptrlen(t) - 1);
				t += strlen(t);
			}
		}
	*t = 0;
}

get_standard_color_scheme()
{
	char *s;

	if (has_feature & FEAT_ANYCOLOR)
		s = "standard-gui";
	else if (monochrome)
		s = "standard-mono";
	else if (opsys == OS_UNIX && getenv("USE_DEFAULT_COLORS"))
		s = "xterm-color";
	else
		s = "standard-color";
	selected_color_scheme = find_index(s);
}

color_scheme_variable_index()
{
	if (has_feature & FEAT_ANYCOLOR)
		return find_index("_our_gui_scheme");
	else if (monochrome)
		return find_index("_our_mono_scheme");
	else if (opsys == OS_UNIX && getenv("USE_DEFAULT_COLORS"))
		return find_index("_our_unixconsole_scheme");
	return find_index("_our_color_scheme");
}


ftp_activity(int activity, int buf, int from, int to)
{
	char msg[FNAMELEN];
	int show_log = 0, orig_buf = bufnum;

	if (!buf_exist(buf))
		return;
	save_var iter, has_arg;
	save_var bufnum = buf;
	if (!ftp_job)
		return;
	switch (activity) {
		case NET_DONE:
			if (ftp_job->operation == FTP_SEND)
				show_text(0, 1, "%s written.", filename);
			if (translation_type == FILETYPE_AUTO
				  && ftp_job->operation == FTP_RECV)
				translation_type = new_buffer_translation_type;
			if (!modified && !point) {
				save_var undo_size = 0;
				undo_mainloop();
				go_line(ftp_go_to_line);
			}
			free(ftp_job);		// Fall through.
		case NET_LOG_DONE:
			ftp_job = 0;
			set_mode_message(NULL);
			check_filevars();
			break;
		case NET_RECV:
			if (!from) {		// First block.
				position_in_all_windows(buf, 0);
				if (mode_keys == buf_tab)		// Switch out of Fundamental mode.
					call_mode(filename);
			}
			show_ftp_percent("got", from);
			break;
		case NET_SEND:
//			ftp_job->so_far = from;
			ftp_job->file_size = to;
			show_ftp_percent("sent", from);
			break;
		case NET_LOG_WRITE:
			save_var point = from;	// Get file size from msg.
			if (re_search(1, "^150 ") && re_search(1, "[0-9]+ bytes")) {
				char filesize[100];
				grab(matchstart, matchend, filesize);
				if (numtoi(filesize) && !got_bad_number)
					ftp_job->file_size = numtoi(filesize);
			}
			point = from;
			if (found_ftp_status("^220 "))	// Welcome msg.
				set_both_mode_message("[FTP: Sending user name] ");
			if (found_ftp_status("^331 "))	// Host wants password.
				set_both_mode_message("[FTP: Sending password] ");
			if (found_ftp_status("^230 "))	// User logged in.
				set_both_mode_message("[FTP: Working] ");
			if (found_ftp_status("^530 "))	// Bad password.
				discard_password_data();
			if (re_search(1, "^[45][0-9][0-9] ") && !parse_string(-1, "PASV\n... ")) {
				parse_string(1, ".*", msg);		// Some error code.
				say("FTP error: %s", msg);
				// Don't go to log if we're just creating a new file.
				if (ftp_job->operation != FTP_RECV
					  || !parse_string(-1, "^550 .*"))
					show_log = 1;
			}
			break;
	}
	if (buffer_ftp_activity)	// Do some buffer-specific FTP work.
		buffer_ftp_activity(activity, buf, from, to);
	switch (activity) {		// These don't automatically refresh later.
		case NET_DONE:
		case NET_LOG_DONE:
		case NET_SEND:
			refresh();
			break;
	}
	if (show_log) {
		if (num_of_popups() > 0 || orig_buf != ftp_job->buf) {
			say("FTP errors occurred, see buffer %s.",
				  buf_exist(ftp_job->log_buf) ?
				  bufnum_to_name(ftp_job->log_buf) : "<deleted>");
			if (buf_exist(ftp_job->buf) && ftp_job->operation == FTP_SEND) {
				save_var bufnum = ftp_job->buf;
				modified = 1;
			}
		} else {
			restore_vars();
			save_var iter, has_arg;
			save_var bufnum = buf;
			if (buf_exist(ftp_job->buf)) {
				to_buffer_num(ftp_job->buf);
				if (ftp_job->operation == FTP_SEND)
					modified = 1;
			}
			if (buf_exist(ftp_job->log_buf))
				to_buffer_num(ftp_job->log_buf);
			point = size();
			refresh();
			restore_vars();
			bufnum = window_bufnum;
		}
	}
	restore_vars();
	if (num_of_popups() > 0 && !strcmp(bufname, CONNECT_LIST_BUF)) {
		int line = lines_between(0, point, 0) + 1, col = current_column();
		make_connect_list();
		go_line(line);
		move_to_column(col);
		refresh();
	}
}

int scp_sftp_handle_list(struct scp_job *p)
{
	int i, isdata;
	char msg[FNAMELEN];

	save_var case_fold = 1;
	while (nl_forward()) {
		i = point;
		point--;
		to_begin_line();
		isdata = 0;
		if (parse_string(1, "[-dcblsp][-r][-w][-sxt][-r][-w][-sxt]"
						 "[-r][-w][-sxt] .*" UNIX_DATE))
			isdata = 1;
		if ((p->flags & SCP_CHECK_IS_DIR) && parse_string(1, ".* not a directory"))
			isdata = 1;
		if (parse_string(1, ".* No such file or directory")
			  && p->cmd && !strncmp(p->cmd, "ls ", 3)) {
			if (buf_exist(p->destbuf))	// Error part way through, so don't
				set_buf_mode_message(p->destbuf, NULL);	// copy any more.
			p->flags |= SCP_CMD_ERROR;
		}
		if (isdata) { // Got some data, copy where it belongs.
			if (buf_exist(p->destbuf) && !(p->flags & SCP_CMD_ERROR)) {
				if (*p->file && !(p->flags & SCP_CHECK_IS_DIR))
					buf_printf(p->destbuf, p->file);
				buf_xfer(p->destbuf, point, i);
			}
		} else if (parse_string(1, "Remote (working )?directory(:| is) ")) {
			grab(matchend, give_end_line(), msg);
			if (p->lastdir)
				free(p->lastdir);
			p->lastdir = strsave(msg);
			if (!p->homedir)
				p->homedir = strsave(msg);
		} else if ((p->flags & SCP_IS_XFER)	// Ignore this msg.
				   && parse_string(1, "((Fetching|Uploading) .* to |"
								   "(remote|local):.* => (remote|local):|"
								   ".*--:-- ETA)")) {
		} else if (!parse_string(1, "(Exited|Connecting to "
								 "|Connection closed|Listing directory "
								 "|Using username \")")) {
			grab(point, MIN(give_end_line(),
							point + sizeof(msg) / sizeof(*msg) - 1), msg);
			if (*msg)		// Any other line is an error msg.
				say("%s", msg);
			if ((p->flags & SCP_IS_XFER)
				  && parse_string(1, ".*: (Permission denied|No such file"
								  "|Device out of space|Read-only file system)"))
				p->flags |= SCP_CMD_ERROR;
		}
		point = i;
	}
	if (parse_string(-1, "p?sftp%> *")) {
		if (!p->cmd || !*p->cmd) {
			if (buf_exist(p->destbuf))
				set_buf_mode_message(p->destbuf, NULL);
			if (p->flags & SCP_IS_XFER)
				scp_handle_done(p);
		}
		p->busy = 0;			// Got a prompt for another cmd.
		return 1;			// Send a pending command if any.
	}
	return 0;
}

scp_handle_list(struct scp_job *p, int from, int to)
{
	int start, old_modified, old, old_busy, nextcmd;

	UNREFERENCED_PARAMETER(to);
	save_var point = from;
	if (buf_exist(p->destbuf)) {
		old = get_buf_point(p->destbuf);
		old_modified = get_buf_modified(p->destbuf);
		start = buf_size(p->destbuf);
		old_busy = p->busy;
		set_buf_point(p->destbuf, start);
	}
	if (p->flags & SCP_CLIENT_MODE)
		nextcmd = scp_client_handle_list(p);
	else
		nextcmd = scp_sftp_handle_list(p);
	if (buf_exist(p->destbuf)) {
		set_buf_point(p->destbuf, old);
		set_buf_modified(p->destbuf, old_modified);
		if (p->op == FTP_LIST) {	// Say we've got more dired results.
			if (!(p->flags & SCP_CMD_ERROR))
				dired_ftp_activity(NET_RECV, p->destbuf, start,
								   buf_size(p->destbuf));
			if (old_busy == 1 && !p->busy && !p->cmd)
				dired_ftp_activity(NET_DONE, p->destbuf, 0, 0);
		} else if ((p->flags & SCP_IS_XFER) && (p->flags & SCP_CMD_ERROR)
				   && p->op == FTP_SEND)
			set_buf_modified(p->destbuf, 1);
	}
	if (nextcmd)	// If there's a pending command, send it now.
		scp_next_cmd(p);
	if (!p->busy && !p->cmd)
		p->destbuf = 0;	// Finished the operation, so don't copy any more.
}

telnet_send_line()
{
	char telnet_cbuf[200], hist_buf[FNAMELEN];

	if (point < type_point || !telnet_id) {
		enter_key();
		return;
	}
	sprintf(hist_buf, "-process-history-%s", bufname);
	create(hist_buf);
	process_cmds_to_end(hist_buf);
	while (point > type_point) {
		int i = MIN(point - type_point,
					sizeof(telnet_cbuf) / sizeof(*telnet_cbuf) - 1);
		grab(type_point, type_point + i, telnet_cbuf);
		if (telnet_server_echoes(telnet_id))
			delete(type_point, type_point + i);
		else
			type_point += i;
		telnet_send(telnet_id, telnet_cbuf);
		if (i > 0)
			buffer_printf(hist_buf, "%s", telnet_cbuf);
	}
	if (!telnet_server_echoes(telnet_id))
		insert('\n'), type_point = point++;
	telnet_send(telnet_id, "\r\n");
	buffer_printf(hist_buf, "\n");
}

telnet_mode()
{
	mode_default_settings();
	mode_keys = telnet_tab;
	telnet_tab[CTRL('C')] = find_index("telnet-ctrlc-tab");
	telnet_tab[ALT('p')] = find_index("process-previous-cmd");
	telnet_tab[ALT('n')] = find_index("process-next-cmd");
	major_mode = _telnet_mode_name;
	tab_size = 8;
	make_mode();
}

int ssh_activity();
char *scp_variable();

void do_ssh(char *host, char *bname)
{
	char cmd[FNAMELEN], *args[2], justhost[FNAMELEN], username[FNAMELEN], *s;
	struct url_parts parts;

	zap(bname);
	to_buffer(bname);
	type_point = 0;
	ssh_mode();
	retain_forced_mode = 1;
	discardable_buffer = 1;
	buffer_not_saveable = 1;
	set_read_only(0);
	when_activity = ssh_activity;
	process_echo = 4;
	memset(&parts, 0, sizeof(parts));
	strcpy(justhost, host);
	*username = 0;
	if (s = strchr(host, '@')) {
		strcpy(justhost, s + 1);
		strncpy(username, host, s - host);
		parts.have_usr = 1;
	}
	strcpy(parts.service, "ssh");
	strcpy(parts.host, justhost);
	strcpy(parts.usr, username);
	args[0] = username;
	args[1] = justhost;
	if (expand_string_template(cmd, scp_variable(&parts, ""), "uh", args))
		error("Invalid ssh template.");
	if (concur_shell("", cmd, NULL, bufname))
		error("Couldn't start ssh process");
}

// The file has been modified on disk; maybe re-read it without prompting.
maybe_auto_read(struct file_info *now)
{
	int pos;

	// Only auto-read if size on disk is > 90% of current size, or
	// file is very small.
	if (!modified && (now->fsize / 9 >= size() / 10 || size() < 300)) {
		pos = point;
		if (!read_file(filename, reading_translation_type()))
			say("%f: Modified file read from disk.", filename);
		else
			auto_read_changed_file = _force_auto_read_changed_file = 0;
		point = pos;
		to_begin_line();
		maybe_add_readonly_once(now);
		return 1;
	}
	return 0;
}

show_date_warning(temp, orig, then, now, save)
struct file_info *now, *then;
{
	char msg[30], *file;
	int mod, warn, auto;

	save_var bufnum = orig;
	file = filename;
	mod = modified;
	warn = want_warn;
	auto = auto_read_changed_file;
	bufnum = temp;
	buf_zap(temp);
	bprintf("%s\nhas changed since it was read from disk", file);
	bprintf(mod ? "!\n" : ".\nFortunately, you haven't edited it.\n");

	stuff_file_date("\nThe version being edited was dated %s on disk.\n",
		then);
	stuff_file_date("         The copy on disk is dated %s.\n\n", now);
	current_time_and_date(msg);
	bprintf("                         It is now %s.\n\n", msg);

	bprintf("Options:\n");
	bprintf("C) Compare the two versions by reading the disk version ");
	bprintf("in its own buffer.\n");
	bprintf("   You may then use compare-windows or diff to ");
	bprintf("see the changes.\n");
	bprintf("R) Read in the new version now, replacing the ");
	bprintf("version already read in.\n");
	if (save)
		bprintf("S) Save your changes, replacing ");
	else
		bprintf("I) Ignore ");
	bprintf("the version on disk.\n");
	bprintf("!) %s this warning for this buffer.  ",
		warn ? "Disable" : "Re-enable");
	bprintf("(You may then choose\n");
	bprintf("   another option.)\n");

	bprintf("A) %s auto-read.  Read changed files without prompting, "
			"if it's safe.\n", auto ? "Disable" : "Enable");
	bprintf("   (You may then choose another option.)\n");

	bprintf("\nYour choice? ");
	point = 0;
}

command insert_file()
{
	char file[FNAMELEN];
	int flags = OFN_HIDEREADONLY;

	save_var show_mouse_choices = 1;
	if (use_common_file_dialog()) {
		strcpy(file, _region_file);
		if (!use_common_open_file_dlg(file, "Insert file", &flags, 0))
			return;
		save_var file_pattern_wildcards = 0; // []'s etc in file name are literal.
	} else {
		if (!insert_file_remembers_file || !*_region_file)
			get_buffer_directory(_region_file);
		get_file(file, "Insert file", _region_file);
	}
	if (!*file)
		return;
	strcpy(_region_file, file);
	mark = point;			/* put mark before it */
	set_region_type(REGNORM);
	build_first = 1;
	do_insert_file(file, ask_line_translate());
	iter = 0;
}

prepare_save_file()
{
	int i = get_mode_variable("%s-delete-trailing-spaces");

	if (got_bad_number)
		i = default_delete_trailing_spaces;
	if (i && translation_type != FILETYPE_BINARY)
		simple_re_replace(1, "[ \t]+$", "");
}

do_save_file(backup, checkdate, getdate) /* 0=ok, 1=err, 2=user canceled */
{				/* always prints status in echo area */
	struct file_info *finfo;
	char othername[FNAMELEN], msg[FNAMELEN];
	int err;

	iter = 0;
	if (!*filename || force_save_as)
		return write_file();
	if (checkdate)
		if (check_dates(1))
			return 2;
	if ((buffer_flags(0) & BF_UNICODE) && !GET_ENCODING(translation_type)
		  && (has_feature & FEAT_UNICODE)) {
		if (!buffer_has_pattern("<#256-#65535>"))
			save_var translation_type = SET_ENCODING(translation_type, 1);
		else		// Force to raw temporarily.
			pick_write_encoding();	// Have 16-bit chars, get an encoding for them.
		if (!GET_ENCODING(translation_type))
			error("Invalid encoding.");
	}
	if (_read_aborted) {
		relative(filename, othername);
		sprintf(msg, "Only part of %s was read from disk.  "
				"Save anyway? ", othername);
		if (!ask_yn("Confirm save", msg, "Save", "Cancel", 0))
			return 2;
	}
	if (backup)
		do_write_backup();
	prepare_save_file();
	finfo = getdate ? &file_info : 0;
	if (err = file_convert_write(filename, translation_type, finfo)) {
		auto_read_changed_file = 0;
		file_info.check_type = 0;	// Prevent checking date.
		file_error(err, filename, "write error");
		maybe_ding(bell_on_write_error);
	} else {
		if (!save_without_prompt && backup != 2)
			note_written(filename);
		undo_set_flag(UNDO_WRITE_FILE);
		if (finfo)
			update_readonly_warning(finfo);
	}
	return (err != 0);
}

command find_file()
{
	char fname[FNAMELEN];
	int want_readonly = 0, flags = 0;

	tiled_only();
	was_quoted = 0;
	save_var show_mouse_choices = 1;
	if (use_common_file_dialog()) {
		*fname = 0;
		if (!use_common_open_file_dlg(fname, "Open file", &flags, 0))
			return;
		if (flags & OFN_READONLY)
			want_readonly = 1;
		save_var file_pattern_wildcards = 0; // []'s etc in file name are literal.
	} else
		get_file_dir(fname, "Find file: ");
	if (!*fname)
		dired_one(fname);
	else if (was_quoted && !is_directory(fname) && !is_remote_file(fname))
		find_it(fname, ask_line_translate());
	else
		do_find(fname, ask_line_translate());
	if (want_readonly)
		force_save_as = FSA_READONLY;
	iter = 0;
}

#define URL_PATTERN	"|((ftp|http):[/\\][/\\][^\n\t >]+)"

default_find_linked_file(char *file)
{
	char *pat;

	save_var case_fold = 1;
	if (opsys == OS_UNIX)
		pat = "[+-9A-z~]+[/.][+-9A-z]*[--9+A-z]" URL_PATTERN;
	else
		pat = "([a-zA-Z]:)?[+-9A-z~]+<.>[+-9A-z]*[--9+A-z]" URL_PATTERN;
	point -= parse_string(-1, "[^ \t\n!-*]*");
	if (!re_search(1, pat)) {
		to_begin_line();
		if (!re_search(1, pat))
			error("Can't find any file name on this line.");
	}
	if (strchr("\"'`<", character(matchstart)))
		matchstart++;
	if (strchr("\"'`.,>", character(matchend - 1)))
		matchend--;
	grab(matchstart, matchend, file);
}

int look_for_filevar_pattern(char *var, char *value, char *oldvalue)
{
	char fullpat[FNAMELEN];

	save_var point = 0;
	save_var narrow_end = size() - 6000;	// Might have very long lines.
	if (parse_string(1, "#!"))	// Look in first line or two only.
		nl_forward();
	narrow_end = size() - give_end_line();
	point = 0;
	if (oldvalue) {
		sprintf(fullpat, ".*-%%*-[ \t]*%s[ \t]*-%%*-", oldvalue);
		if (re_search(1, fullpat))
			return 1;
	}
	point = 0;
	sprintf(fullpat, ".*-%%*-.*</word>%s:[ \t]*%s[ \t]*(;.*)?[ \t]*-%%*-",
			var, value);
	if (re_search(1, fullpat))
		return 1;
	restore_vars();
	save_var point = size() - 3000;
	sprintf(fullpat, "</word>%s:[ \t]*%s", var, value);
	if (search(1, "Local Variables:") && re_search(1, fullpat))
		return 1;
	return 0;
}

#define C_COMMENT_FILLER		"[^0-9a-zA-Z\x80-\xFF_(\n]*"
#define C_COMMENT_PREFIX		"^" C_COMMENT_FILLER

get_comment_marker(char *cm, int generic)
{
	char *p;

	strcpy(cm, generic ? get_comment_begin() : "//");
	for (p = cm + strlen(cm); strlen(cm) > 1 && p[-1] == ' '
			 && !isalnum(p[-2]); )
		*--p = 0;		// Remove trailing spaces.
}

do_fill_comment(int generic, int margin)
{
	int pbuf, width = 1000, i, old_lines, new_lines, orig_lines;
	int col, col2, orig = bufnum, type;
	char last_prefix[FNAMELEN], orig_prefix[FNAMELEN], pat[FNAMELEN], comm[30];

	save_var margin_right = margin;
	save_spot point;
	type = locate_c_comment();	// 1 if block comment, 2 if one-line
	if (!type)					// comment, or 0 if not a comment.
		error("Can't fill comment here, not in a comment.");
	point += parse_string(1, C_COMMENT_FILLER);	// Past 1st line's prefix.
	sprintf(pat, "[ \t]*%s[^0-9a-z_\n]*", generic ? get_comment_start() : "/[*/]");
	grab(point, point - parse_string(-1, pat), orig_prefix);
	save_var narrow_start = point;		// Narrow to the paragraph.
	point = mark;
	if (!generic)
		point -= parse_string(-1, "<*>/");
	save_var narrow_end = size() - point + 1;

	if (type == 1 && !generic)		// Combine adjacent block comments.
		simple_re_replace(1, "[ \t]*<*>/[ \t]*\n?[ \t]*/<*>[ \t]*", " ");
	point = 0;
	while (nl_forward()) {	// Find width of narrowest prefix.
		i = parse_string(1, C_COMMENT_PREFIX);
		col = get_column(point + i);
		width = MIN(width, col);
	}
	pbuf = tmp_buf();
	point = 0;
	get_comment_marker(comm, generic);
	sprintf(pat, "[ \t]*%s", comm);
	while (nl_forward()) {	// Move each prefix to our temp buffer.
		i = parse_string(1, C_COMMENT_PREFIX);
		col = get_column(point + i);
		if (col > width)	// Make sure it's not too wide.
			col = width;
		if (type == 2) {	// But if it's made of one-line comments,
			i = parse_string(1, pat);
			col2 = get_column(point + i);
			if (col < col2)		// each prefix must include a "//".
				col = col2;
		}
		i = column_to_pos(col);
		buf_xfer(pbuf, point - 1, i);
		delete(point, i);
	}
	if (width == 1000)	// No prefix, so use default.
		width = 3;
	orig_lines = old_lines = lines_between(0, size(), 0) + 1;
	region_fill_first_line(narrow_start, size() - narrow_end,
						   MAX(margin_right, 20),
						   MAX(margin_right - width, 20));
	new_lines = lines_between(0, size(), 0) + 1;
	point = 0;
	bufnum = pbuf;
	point = 0;
	*last_prefix = 0;	// Now copy the prefixes back.
	for (; new_lines > 0 && old_lines > 0; old_lines--, new_lines--) {
		grab(point, give_end_line(), last_prefix);
		buf_xfer(orig, point, give_end_line());
		nl_forward();
		bufnum = orig;
		nl_forward();
		bufnum = pbuf;
	}
	if (orig_lines < 2) {		// Default for new lines if 1 line originally.
		strcpy(last_prefix, orig_prefix);
		if (type == 2 && !strstr(last_prefix, comm))
			sprintf(last_prefix, "%s ", comm);
		else if (type == 1 && !generic)
			if (!strstr(last_prefix, "/*"))
				strcpy(last_prefix, (auto_fill_comment_rules & 4) ? " " : " * ");
			else {
				replace_in_string(last_prefix, '/', ' ');
				if (fill_c_comment_plain)
					replace_in_string(last_prefix, '*', ' ');
			}
	}
	bufnum = orig;
	narrow_start = 0;
	for (; new_lines > 0; new_lines--) {	// More lines than before.
		stuff(last_prefix);			// Add some prefixes.
		if (generic) {
			to_begin_line();
			if (indenter)
				(*indenter)();
		}
		nl_forward();
	}
	buf_delete(pbuf);
	if (display_column > 0)
		display_column = 0;
	fix_window_start();	/* win shouldn't start in middle of line */
}

int narrow_to_new_comment()
{
	char pat[50];

	if (strcmp(major_mode, "C"))
		sprintf(pat, "[ \t]*%s", get_comment_start());
	else
		strcpy(pat, "[ \t]*//");
	save_var point;
	to_begin_line();
	if (!parse_string(1, pat)) {
		narrow_start = point;	// This line doesn't start with one.
		restore_vars();
		to_end_line();
		if (!re_search(-1, pat))
			return 0;		// Can't find comment start.
		narrow_start = point;
	}
	while (nl_forward())		// Find the last such line.
		if (!parse_string(1, pat))
			break;
	narrow_end = size() - point;
	do {						// Find the first such line.
		point--;
		to_begin_line();
		if (!parse_string(1, pat)) {
			nl_forward();
			break;
		}
	} while (point > narrow_start);
	narrow_start = point;
	return 1;
}

comm_break_line_here(int c_like, int comm, int starts_with_comm, int starts_in_comm,
					 int comm_start, spot next_ends, spot after, char *fill)
{
	int offset, generic = strcmp(major_mode, "C");
	char rest[FNAMELEN], *new_comm;

	delete_horizontal_space();
	insert('\n');
	if (display_column > 0)	// Horizontally scroll back to left margin.
		display_column = 0;
	if (key == '\n')	// User might be trying to break line elsewhere as
		key = ' ';		// we're breaking it here.  Prevent a mess.

	if (comm == IN_NEW_COMMENT || !c_like) {		// A //-style C++ comment.
		new_comm = generic ? get_comment_begin() : "// ";
		if (starts_with_comm)	// This line began with a comment, so just
			stuff(fill);		// repeat its prefix (indentation and such).
		else if (next_ends && get_column(comm_start) > get_column(*next_ends)) {
			grab(point, give_end_line(), rest);	// Continue it on the next line
			offset = give_end_line() - *after;	// since there's no comment
			delete(point - 1, give_end_line());	// there.
			point = *next_ends;
			to_column(get_column(comm_start) - 2);
			bprintf("%s%s%s", isspace(character(point - 1)) ? "" : " ",
					new_comm, rest);
			*after = point - offset;	// Point might be in the middle of moved text.
		} else {			// Make a new line after the commented code line we
			stuff(new_comm);	// broke.
			point -= 3;
			if (c_like)
				indent_to_column(c_compute_indent());
			else if (indenter)
				(*indenter)();
		}
	} else {			/* Old-style K&R comment. */
		if (strstr(fill, "*/"))
			*fill = 0;
		if (starts_in_comm)		// Just repeat the *'s or whatever at start of
			stuff(fill);		// old line.
		else if (starts_with_comm) {
			if (!strcmp(fill, "/**"))	// A Javadoc comment.
				fill = " * ";
			replace_in_string(fill, '/', ' ');
			if (fill_c_comment_plain)	// Like above, but convert that first /*.
				replace_in_string(fill, '*', ' ');
			stuff(fill);
		} else {
			delete(point - 1, point);	// Delete the \n we just inserted.
			stuff(" */\n/* ");			// And put this instead.
			to_begin_line();
			indent_to_column(c_compute_indent());
		}
	}
}

int in_comment()
{
	int start = point;
	save_var point;
	to_begin_line();
	save_var narrow_end = size() - give_end_line();
	if (re_search(1, get_comment_start()) && start >= matchstart)
		return IN_NEW_COMMENT;
	return 0;
}

#define TEX_PARAG_SEP	"^\\(begin|end)|^[ \t]*(%%|%$%$|\\[][])"

command forward_paragraph()
{
	char pat[100];

	do_shift_selects();
	re_search(1, ".");
	sprintf(pat, "\n%s|\f",
			indents_separate_paragraphs ? "[ \t\n]" : "[ \t]*\n");
	if (tex_paragraphs) {	/* skip formatter lines */
		to_begin_line();
		re_search(1, "(([@.]|" TEX_PARAG_SEP ").*\n)*");
		re_search(1, "[^ \t\n@.\\]");
		strcat(pat, "|" TEX_PARAG_SEP);
	}
	if (re_search(1, pat)) {
		point--;
		if (curchar() != '\f') {
			if (!indents_separate_paragraphs)
				re_search(-1, "[ \t]*");
			if (tex_paragraphs)
				re_search(-1, "^(([@.]|" TEX_PARAG_SEP ").*\n)*");
		}
	} else
		point = size();
}

command backward_paragraph()
{
	do_shift_selects();
	if (tex_paragraphs) {
		if (character(point - 1) == '\n')
			re_search(-1, "^(([@.]|" TEX_PARAG_SEP ").*\n)*");
		re_search(-1, "[^ \t\n]");
		if (matches_at(give_begin_line(), 1, TEX_PARAG_SEP))
			to_begin_line();
		if (indents_separate_paragraphs) {
			if (re_search(-1, "\n[ \t\n]|\f|" TEX_PARAG_SEP))
				re_search(1, "\n*");
		} else
			re_search(-1, "(\n[ \t]*\n|\f|" TEX_PARAG_SEP ")!");
		if (curchar() != '\f')
			re_search(1, "^(([@.\f]|" TEX_PARAG_SEP ").*\n)*");
	} else {
		re_search(-1, "[^ \t\n]");
		if (indents_separate_paragraphs) {
			if (re_search(-1, "\n[ \t\n]|\f|\n(" TEX_PARAG_SEP ")"))
				re_search(1, "\n*");
		} else
			re_search(-1, "(\n[ \t]*\n|\f)!");
	}
}

when_setting_sentence_end_double_space()
{
	if (now_parsing_cmds)
		return;
	if (ask_yn("Set sentence-end too?",
			   "Set the sentence-end variable to match?", "Yes", "No", 1))
		sprintf(sentence_end, "([.?!]<rsquare|rparen|squote|dquote>*"
				"[ \t\n]*(%s[ \t]|\n))", sentence_end_double_space ? "[ \t]" : "");
}

command enter_key()
{
	full_key = key = '\n';
	normal_character();
}

prepare_grep_buffer(char *buf, int flags, char *pat, char *filepat)
{
	char search_type[FNAMELEN], dir[FNAMELEN];
	int before_head;

	create(buf);
	if (grep_empties_buffer)
		zap(buf);
	locate_window(buf, "");
	point = size();
	if (grep_empties_buffer)
		modified = 0;
	save_var modified;
	if (size())
		bprintf("\n");	/* separate from previous query */
	make_search_prompt(search_type, flags & SMODES);
	before_head = point;
	getcd(dir);
	bprintf("Found 0 lines with %s\"%s\" in %s (cur dir \"%s\")\n",
			search_type, pat, filepat, dir);
	set_character_color(before_head, point, color_class grep_header);
	if (display_column.default != -1)
		display_column = 0;	/* horizontally scroll this window */
	grep_mode();
	return before_head;
}

do_grep(buf, pat, flags, filepat, inbufs)
char *buf, *pat, *filepat;
{
	int res = 0, aborted = 0, cnt = 0, tempbuf = 0, i, skipped = 0;
	int orig, before_head, this_err;
	char *s, rel[FNAMELEN], abs[FNAMELEN], *(*finder)();
	char *months = "Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec ";
	char extra[FNAMELEN];
	spot start;

	before_head = prepare_grep_buffer(buf, flags, pat, filepat);
	start = alloc_spot(0);
	save_var modified;
	save_var abort_file_matching = ABORT_ERROR;
	finder = give_grep_finder(inbufs, filepat, &s);	// get 1st match
	save_var readonly_warning, no_popup_errors = 1;
	if (grep_on_changed_file & 1)
		readonly_warning |= ROWARN_GREP;
	readonly_warning &= ~(ROWARN_MSG | ROWARN_BELL);
	orig = bufnum;
	for (; s; s = (*finder)(filepat, 0)) {
		if (user_abort) {
			aborted = 1;
			break;
		}
		grep_make_names(rel, abs, s, inbufs);
		if (inbufs)
			bufname = s;
		else if (!ok_grep_file_match(s, abs)) {
			skipped++;
			continue;
		}
		noteput("%d matches, searching %f...", cnt, rel);
		if (inbufs)	// skip over grep results buffer
			this_err = !strcmp(s, buf);
		else if (!look_file(abs)) {
			this_err = find_it(abs, default_translation_type);
			if (!grep_keeps_files)
				tempbuf = bufnum;
		} else {
			this_err = 0;
			if (grep_on_changed_file & 2)
				save_var auto_read_changed_file = 1;
			if (check_dates(0) == 2)
				grep_date_warning(buf, filename);
		}
		if (this_err == EREADABORT || user_abort) {
			aborted = 1;
			break;
		} else if (this_err) {
			window_bufnum = orig;
			if (tempbuf)
				buf_delete(tempbuf);
			tempbuf = 0;
			continue;
		}
		*extra = 0;
		if (grep_include_timestamp && !inbufs && file_info.check_type == CHECK_FILE)
			if (grep_include_timestamp == 2)
				sprintf(extra, "[%04d-%02d-%02d %02d:%02d:%02d] ",
						file_info.year, file_info.month, file_info.day,
						file_info.hour, file_info.minute, file_info.second);
			else
				sprintf(extra, "[%02d-%.3s-%04d %02d:%02d:%02d] ",
						file_info.day,
						(file_info.month >= 1 && file_info.month <= 12)
						? months + (file_info.month - 1) * 4 : "???",
						file_info.year,
						file_info.hour, file_info.minute, file_info.second);
		if (grep_show_absolute_path == 2)
			convert_to_8_3_filename(abs);
		res = copy_matches(name_to_bufnum(buf), flags, pat,
						   grep_show_absolute_path ? abs : rel, extra);
		window_bufnum = orig;
		if (res < 0)	/* error occurred during search */
			break;
		else if (res > 0) {
			cnt += res;
			build_first = 1;
			update_grep_title(before_head, cnt);
			refresh();
		}
		if (tempbuf)
			buf_delete(tempbuf);
		tempbuf = 0;
	}
	window_bufnum = orig;
	point = i = *start;
	free_spot(start);
	build_first = 1;
	if (tempbuf)
		buf_delete(tempbuf);
	say("");
	if (aborted)
		quick_abort();
	if (res == -DSBAD)
		error("Illegal search pattern: %s", pat);
	else if (res != -DSABORT) {
		locate_window(buf, "");
		point = i;
		if (skipped)
			say("Skipped %d file%s due to grep-ignore-file- variables",
				  skipped, (skipped == 1) ? "" : "s");
	}
}

do_file_replace(str, with, flags, filepat, inbufs)
char *str, *with, *filepat;	// (query-) replace across files/buffers
{
	int res, oldbuf = 1, prevbuf = 0, total = 0, cnt = 0;
	char rel[FNAMELEN], abs[FNAMELEN], *s, *(*finder)(), msg[FNAMELEN];
	int total_found = 0, total_changed = 0, total_readonly = 0, old_readonly;

	old_readonly = readonly_warning;
	save_var readonly_warning;			// Don't show msg or ring bell
	readonly_warning &= ~(ROWARN_MSG | ROWARN_BELL); // unless matches.
	save_var abort_file_matching = ABORT_ERROR;
	save_var bufnum, window_bufnum;
	finder = give_grep_finder(inbufs, filepat, &s);	// Get 1st match
	for (; s; s = (*finder)(filepat, 0)) {
		grep_make_names(rel, abs, s, inbufs);
		if (inbufs)
			bufname = s;
		else if (!ok_grep_file_match(s, abs))
			continue;
		noteput("Searching %s...", rel);
		if (!inbufs) {
			oldbuf = look_file(abs);
			if (find_it(abs, default_translation_type) == EREADABORT)
				user_abort = 1;
		} else
			to_buffer(s);
		undo_mainloop();
		if (!user_abort) {
			res = one_file_replace(str, with, flags, oldbuf, old_readonly);
			total_found += replace_num_found;
			total_changed += replace_num_changed;
			total++;
			if (res == -2)
				total_readonly++;
			else if (res >= 0)
				cnt++;
		} else {
			res = DSABORT;
			user_abort = 0;
		}
		if (prevbuf && buf_in_window(prevbuf) == -1) {
			buf_delete(prevbuf);	// delete untouched new bufs
			prev_forget(_prev_num - 1); // fix default
		}
		prevbuf = (oldbuf || modified) ? 0 : bufnum;
		if (res == 0 && (flags & QUERY) && key == '!'
			  && ask_continue_replace("Replace without asking"))
			flags &= ~QUERY;	// no more querying
		else if (res == DSABORT || res == 1 && ungot_key != -1) {
			if (!ask_continue_replace("Continue replacing"))
				break;
		} else if (res > 0) {	// User said to stop, or bad pattern.
			if (res == DSBAD)
				show_replace(str, with, flags | BAD);
			user_abort = 0;
			return;
		}
		ungot_key = -1;
		if (user_abort)
			break;
	}
	ungot_key = -1;
	user_abort = 0;
	sprintf(msg, "Replaced %d of %d matches in %d of %d %s",
			total_changed, total_found, cnt, total,
			inbufs ? "buffers" : "files");
	if (total_readonly)
		sprintf(msg + strlen(msg), ", %d read-only, see %s buffer",
				total_readonly, MSG_BUFFER);
	say("%s.", msg);
	restore_vars();
	if (prevbuf && buf_in_window(prevbuf) == -1) {
		buf_delete(prevbuf);	// delete untouched new bufs
		prev_forget(_prev_num - 1); // fix default
	}
}

one_file_replace(str, with, flags, oldbuf, old_readonly)
char *str, *with;	// return 0=ok, 1=user said to stop,
{			// DSABORT=aborted, DSBAD=bad pat, -1=no matches, -2=readonly
	int res;

	replace_num_found = replace_num_changed = 0;
	save_var point = 0;
	res = do_color_searching(flags, str);
	if (res != 1)
		return res ? res : -1;
	if (!oldbuf && (flags & FILEGREP))	// had to read file to search
		replace_in_readonly_hook(old_readonly);
	else
		replace_in_existing_hook(old_readonly);
	if (_buf_readonly) {
		if ((flags & FILEGREP) && file_info.check_type == CHECK_FILE
			  && (file_info.attr & ATTR_READONLY))
			show_text(0, 1, "File %r is read-only.", filename);
		else
			show_text(0, 1, "Buffer %s is read-only.", bufname);
		return -2;
	}
	point = matchstart - 1;
	show_replace(str, with, flags);
	return string_replace(str, with, flags);
}

#define MANCONF_PATH	"^(MANDATORY_|OPTIONAL_)?MANPATH" \
						"(_MAP[ \t]+[^ \t\n]+)?[ \t]+([^ \t\n]+)$"

build_man_path(char *dest)
{
	char *path = getenv("MANPATH");
	int tmp;

	if (path) {
		strcpy(dest, path);
		return;
	}
	save_var bufnum = tmp = tmp_buf();
	if (!file_read("/etc/manpath.config", FILETYPE_AUTO)
		  || !file_read("/etc/man.config", FILETYPE_AUTO)) {
		do_drop_matching_lines(REGEX, MANCONF_PATH, 0);
		string_replace(MANCONF_PATH, "#3", REGEX);
	}
	stuff("/usr/man\n/usr/share/man\n");
	do_sort_region(0, size(), 0, 0);
	do_uniq(1, 1, 0);
	simple_re_replace(0, "\n", ":");
	simple_re_replace(0, ":<eof>", "");
	grab(0, size(), dest);
	restore_vars();
	buf_delete(tmp);
}

int literal_if_args();
char *man_match(), *man_use_program();

show_man_page(int type)
{
	char topic[FNAMELEN], buf[FNAMELEN], pr[FNAMELEN], *prog = man_use_program(type);

	save_var show_status = 0, abort_searching = 0;
	grab_man_word_from_buffer(bufnum, point, buf,
							  "[-+_0-9a-z.:]+(%(..?%))?");
	if (type == 1 && !strcmp(buf, "use")) {
		save_var point;
		point += parse_string(1, "<word>*<space>*");
		grab_man_word_from_buffer(bufnum, point, buf,
								  "[-+_0-9a-z.:]+(%(..?%))?");
	}
	man_rewrite(buf);
	save_var completion_literal_char = literal_if_args;
	sprintf(pr, "Show %s entry for", type ? prog : "manual");
	prompt_comp_read(topic, pr, prog ? man_match : 0,
					 COMP_FOLD | SPACE_VALID | CAUTIOUS, buf);
	if (!*topic)
		return;
	make_man_page(topic, type);
}

hex_cc_prefix()
{
	int cmd;

	table_keys[table_count++] = key;
	table_prompt();
	getkey();
	cmd = binding_of(key, hex_tab);
	if (cmd <= 0 || cmd == hex_cc_prefix)
		cmd = binding_of(key, cc_tab);
	save_var hex_overtype_mode = 2;
	say("");
	do_command(cmd);
}

char *hex_prefix()
{
	if (major_mode != _hex_overtype_mode_name)
		return "";
	if (!hex_tab[CTRL('C')] && hex_tab[CTRL('K')] == hex_cc_prefix)
		return "Ctrl-K ";
	return "Ctrl-C ";
}

color_hex_range(from, to)
{
	int cnt, pos, oldbuf = 0, dump, ascii, i, j;
	char oldbytes[20], newbytes[20];

	if (from >= to)
		return to;
	if (strlen(bufname) > sizeof(HEX_BUFNAME_PREFIX) / sizeof(char) - 1)
		oldbuf = name_to_bufnum(ORIG_BUF(bufname));
	save_var point, matchstart, matchend;
	point = from;
	to_begin_line();
	save_var narrow_start = give_begin_line();
	point = to;
	nl_forward();
	save_var narrow_end = size() - point;
	point = 0;
	do {
		dump = column_to_pos(HEX_DUMP_COL);
		ascii = column_to_pos(HEX_ASCII_COL);
		set_character_color(point, dump, color_class hex_address);
		set_character_color(dump, ascii, -1);
		set_character_color(ascii, column_to_pos(HEX_ASCII_COL_END),
							color_class hex_ascii);
		if (!oldbuf)
			continue;
		pos = hex_get_offset();
		cnt = hex_extract(dump, newbytes, 0);
		if (cnt < 0)
			continue;
		buf_grab_bytes(oldbuf, pos, pos + cnt, oldbytes);
		for (i = 0; i < cnt; i++)
			if (oldbytes[i] != newbytes[i]) {
				j = dump + HEX_TO_CHAR(i);
				set_character_color(j, j + 2 * hex_size(),
									color_class hex_change);
				set_character_color(ascii + i, ascii + i + 1,
									color_class hex_change);
			}
	} while (nl_forward());
	return point;
}

int hex_maybe_overtype()
{
	int pos, col = current_column();

	if (!is_hex_buffer(bufnum) || hex_overtype_mode == 2)
		return 0;
	if (col < HEX_ASCII_COL || col >= HEX_ASCII_COL_END)
		return 0;
	if (major_mode != _hex_overtype_mode_name
		  || !IS_ASCII_KEY(full_key) || full_key < ' ')
		return 0;
	save_var call_on_modify = 0;
	pos = hex_get_offset();
	if (pos >= hex_get_original_buffer_size())
		return 0;
	hex_goto_place(pos, 1);
	hex_modify_byte(pos & ((hex_size() > 1) ? 0x7 : 0xf), full_key & 0xffff);
	hex_goto_place(pos + 1, 1);
	return 1;
}

hex_help()
{
	if (hex_maybe_overtype())
		return;
	help_on_command("hex-mode");
}

hex_search_literal()
{
	jmp_buf this_level;
	char orig[FNAMELEN];

	if (is_hex_buffer(bufnum)) {
		strcpy(orig, ORIG_BUF(bufname));
		int pos = hex_get_offset();
		to_buffer(orig);
		point = pos;
	} else
		error("Not in hex buffer.");
	check_abort();
	save_var top_level = &this_level;
	if (key != CTRL('S') && key != CTRL('R'))
		key = CTRL('S');
	if (!setjmp(top_level))
		do_topkey();
	restore_vars();
	if (!is_hex_buffer(bufnum))
		hex_back_to_hex_buffer();
}

hex_overtype()
{
	int col = current_column();

	if (hex_maybe_overtype())
		return;
	if (major_mode == _hex_overtype_mode_name)
		major_mode = _hex_mode_name;
	else if (major_mode == _hex_mode_name) {
		if (col < HEX_DUMP_COL || col >= HEX_ASCII_COL_END)
			move_to_column(HEX_ASCII_COL);
		else if (col < HEX_ASCII_COL)
			hex_tab_key();
		major_mode = _hex_overtype_mode_name;
		say("Press %so to leave overtype submode.", hex_prefix());
	}
	make_mode();
}

html_recolor_from_here(safe)
{
	int start = point;

	safe = safe;			// avoid compiler warning
	if (re_search(-1, "%<%!%[CDATA%[|%]%]%>!"))
		return point;
	point = start;
	if (re_search(-1, sublanguage_code == SUBLANG_XML
				  ? "%<%!--|--[ \t\n]*>|%<%?php"
				  : "%<%!--|-[ \t\n]*>|%<[%?]|[%?]%>") && curchar() != '-')
		return point;
	return start;
}

color_html_comment()
{
	int s;

	search(-1, "<");	// Find start of comment.
	s = point;
	if (parse_string(1, "%<%!%[CDATA%[")) {	// xml cdata
		re_search(1, "%]%]%>");
		set_character_color(s, point, color_class html_value);
		return;
	} else if (parse_string(1, "%<[ \t\n]*%![ \t\n]*--")) {
		// A legal comment?
		point = matchend;
		re_search(1, "--[ \t\n]*>");	// Search for end of legal comment.
	} else if (parse_string(1, "%<%!DOCTYPE")) {
		search(1, ">");
		set_character_color(s, point, color_class html_value);
		return;
	} else	// Some odd flavor of comment; don't be too picky about its end.
		search(1, ">");
	set_character_color(s, point, color_class html_comment);
}

color_html_include()
{
	int s, t, c, lang, orig = point;
	char *end;

	search(-1, "<");	// Find start of include.
	c = character(point + 1);
	s = point;
	switch (c) {
		case '%':	if (parse_string(1, "%<%%--")) {
						search(1, "--%>");	// Found a JSP "hidden comment".
						set_character_color(s, point, color_class html_comment);
						return 1;
					}
					point += parse_string(1, "%<%%[a-z0-9]*");
					end = "%>";
					lang = get_html_asp_language(sublanguage_code
						== SUBLANG_XML ? xml_asp_coloring : html_asp_coloring);
					break;
		case '?':	if (sublanguage_code != SUBLANG_XML
						  || parse_string(1, "%<%?php"))
						lang = html_php_coloring;
					else
						lang = xml_php_coloring;
					point += parse_string(1, "%<%?[a-z0-9]*");
					end = "?>";
					break;
		default:	point++;
					end = ">";
					lang = get_html_asp_language(html_asp_coloring);
					break;
	}
	if (lang == HT_IGNORE) {
		point = orig;
		return 0;
	}
	t = point;
	set_character_color(s, point, color_class html_embedded);
	c = search(1, end) ? matchstart : point;
	set_character_color(c, point, color_class html_embedded);
	set_tagged_region("color-as-unit", s, point, 99);
	color_embedded_code(t, c, lang);
	return 1;
}

html_in_quoted_attribute()
{
	save_var point;
	if (!re_search(-1, "['\"]"))
		return 0;
	if (!parse_string(-1, "%<" HTML_NAME "([ \t\n]+" HTML_ATTRIBUTE ")*"
					  "[ \t\n]+" HTML_NAME HTML_EQUALS))
		return 0;
	restore_vars();
	point = matchend + 1;
	return 1;
}

color_html_range(from, to) // recolor just this section
{			// last colored region may go past to
	int s, script, t, i;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	if (from < to) {
		set_character_color(from, to, color_class html_text);
		set_tagged_region("needs-color", from, to, -1);	// Say it needs coloring.
	}
	point = from;
	save_var case_fold = 1;
	color_html_move_back();
	while (point > narrow_start
		   && get_tagged_region("color-as-unit", point - 1, &from) != -1)
		point = from;		// Move before script areas.
	if (to > size() - narrow_end)
		to = size() - narrow_end;
	set_tagged_region("color-as-unit", from, to, -1);
	from = point;
	for (s = point; point < to; ) {
		if (curchar() == '&') {
			set_character_color(s, point, color_class html_text);
			s = point++;	 // Color an entity like &amp; or &#251;
			if (parse_string(1, "([a-z]+|#[0-9]+|#x[0-9a-f]+);?"))
				point = matchend;
			set_character_color(s, point, color_class html_entity);
			s = point;
			continue;
		}
		if (curchar() != '<') {
			point++;
			continue;
		}
		set_character_color(s, point, color_class html_text);
		i = t = point++;
		while (strchr(" \t\n", curchar()))
			point++;	// Parse < followed by whitespace.
		set_character_color(t, point, color_class html_punctuation);
		if (curchar() == '!')
			color_html_comment();
		else if (strchr("%?", curchar()) && color_html_include())
			;
		else if (s = parse_string(1, "/?[-_.a-z0-9]*")) { // Parse the tag.
			set_character_color(point, point + s, color_class html_tag);
			script = (s == 6 && parse_string(1, "script"));
			if (s == 3 && parse_string(1, "jsp:(declaration|expression"
									   "|scriptlet)</word>"))
				script = 2;
			point += s;			// Look for javascript tag.
			if (curchar() == ':' && parse_string(1, ":[-_.a-z0-9]*")) {
				set_character_color(point - s, point + 1,
									color_class html_namespace);
				set_character_color(point + 1, matchend,
									color_class html_tag);
				point = matchend;
			}
			if (parse_string(1, "[ \t\n]*/[ \t\n]*>")) {
				set_character_color(point, matchend, color_class html_tag);
				point = matchend;
			} else {
				for (i = point; strchr(" \t\n", curchar()); )
					point++;	// Parse whitespace.
				set_character_color(i, point, color_class html_punctuation);
				color_html_attributes(script, t);
			}
		}
		s = point;
	}
	re_search(1, "![<&]");
	if (s < point)
		set_character_color(s, point, color_class html_text);
	set_tagged_region("needs-color", from, point, 0);	// Say it's colored.
	return point;
}

do_html_indent(int manual)
{
	int region, start;

	recolor_buffer_range(point, point + 1);
	region = get_tagged_region("color-as-unit", point, &start);
	if (IS_ENTER_KEY(key) && !(html_auto_indent &
							   (region == 1 ? 2 : region == 2 ? 4
								: region == 3 ? 8 : 1)))
		return;
	save_var narrow_start = start;
	switch (region) {
		case HT_JSCRIPT:
			html_exclude_script_head();
			save_var c_extra_keywords |= JAVA_KEYWORDS;
			do_c_indent();
			return;
		case HT_VBSCRIPT:
			html_exclude_script_head();
			do_vbasic_indent();
			return;
		case HT_PHP:
			html_exclude_script_head();
			save_var in_perl_buffer = 1;
			save_var sublanguage_code = SUBLANG_PHP;
			do_c_indent();
			return;
	}
	if (manual)
		indent_previous();
	else
		prev_indenter();
}

html_fill_paragraph()
{
	int region, start, end;

	recolor_buffer_range(point, point + 1);
	region = get_tagged_region("color-as-unit", point, &start, &end);
	switch (region) {
		case HT_JSCRIPT:
			save_var major_mode = _c_mode_name;
			fill_comment();
			return;
		case HT_VBSCRIPT:
		case HT_PHP:
			fill_comment();
			return;
	}
	fill_paragraph();
}

int html_auto_show_delimiter()
{
	int region, start, end;

	save_var point;
	recolor_buffer_range(point, point + 1);
	region = get_tagged_region("color-as-unit", point, &start, &end);
	save_var narrow_start = start, narrow_end = size() - end;
	switch (region) {
		case HT_JSCRIPT:
			html_exclude_script_head();
			save_var auto_show_matching_characters = c_auto_show_delim_chars;
			save_var mode_move_level = c_move_level;
			break;
		case HT_VBSCRIPT:
			html_exclude_script_head();
			save_var auto_show_matching_characters = vbasic_auto_show_delim_chars;
			save_var mode_move_level = 0;
			break;
		case HT_PHP:
			html_exclude_script_head();
			save_var auto_show_matching_characters = php_auto_show_delim_chars;
			save_var mode_move_level = perl_move_level;
			break;
	}
	return default_auto_show_delimiter();
}

command html_find_matching_tag()
{
	char tag[FNAMELEN], pat[FNAMELEN];
	int level, dir, start = point;

	if (curchar() != '<' && !search(-1, "<")) {
		point = start;
		error("No tag found.");		// Get to opening <.
	}
	point += parse_string(1, ".[ \t\n]*");
	if (curchar() == '/') {
		dir = -1;
		level = 0;
		point++;
	} else
		level = dir = 1;
	point += parse_string(1, "[ \t\n]");
	if (!parse_string(1, "[-_./a-z0-9:]+", tag)) {
		point = start;
		error("No tag found.");
	}
	sprintf(pat, "%%<(/?%q)</word>", tag);
	while (re_search(dir, pat)) {
		if (!(level += (character(find_group(1, 1)) == '/') ? -1 : 1))
			break;
	}
	if (level) {
		point = start;
		say("No match for %s tag.", tag);
		expire_message = 1;
	}
}

javascript_func_name_finder()
{
	int from, to, orig = point;

	save_var point, case_fold = 1;
	to_end_line();
	while (do_color_searching(REVERSE | REGEX, "^[ \t]*<^c:*comment><^c:*string>"
							  "(function)</word>")) {
		if (!parse_string(1, "^[ \t]*(function)[ \t]+([a-zA-Z_]<word>*)"))
			continue;
		from = find_group(2, 1);
		to = find_group(2, 0);
		to = MIN(to, from + ptrlen(display_func_name) - 1);
		grab(from, to, display_func_name);
		point = matchend;
		point += parse_string(1, "[ \t]*");
		if (curchar() == '(')
			move_level(1, "(", ")", 0, 0);
		move_level(1, "{", "}", 0, 0);
		if (give_end_line() < orig)
			*display_func_name = 0;
		return 1;
	}
	*display_func_name = 0;
	return 1;
}

inner_php_func_name_finder()
{
	int from, to, orig = point;

	save_var point, case_fold = 1;
	to_end_line();
	while (do_color_searching(REVERSE | REGEX, "^[ \t]*<^c:*comment><^c:*string>"
							  "(function)</word>")) {
		if (!parse_string(1, "^[ \t]*(function)[ \t]+([a-zA-Z_]<word>*)"))
			continue;
		from = find_group(2, 1);
		to = find_group(2, 0);
		to = MIN(to, from + ptrlen(display_func_name) - 1);
		grab(from, to, display_func_name);
		point = matchend;
		point += parse_string(1, "[ \t]*");
		if (curchar() == '(')
			move_level(1, "(", ")", 0, 0);
		move_level(1, "{", "}", 0, 0);
		if (give_end_line() < orig)
			*display_func_name = 0;
		return 1;
	}
	*display_func_name = 0;
	return 1;
}

html_func_name_finder()
{
	int start, end, region, res;
	recolor_buffer_range(point, point + 1);
	region = get_tagged_region("color-as-unit", point, &start, &end);
	save_var narrow_start = start, narrow_end = size() - end;
	*display_func_name = 0;
	if (region > 0 && html_display_definition) {
		html_exclude_script_head();
		switch (region) {
			case HT_JSCRIPT:
				save_var mode_move_level = c_move_level;
				res = javascript_func_name_finder();
				if (html_display_definition == 2 && res && !*display_func_name)
					strcpy(display_func_name, "JavaScript");
				return res;
			case HT_VBSCRIPT:
				save_var mode_move_level = 0;
				res = vbasic_func_name_finder();
				if (html_display_definition == 2 && res && !*display_func_name)
					strcpy(display_func_name, "VBScript");
				return res;
			case HT_PHP:
				save_var mode_move_level = perl_move_level;
				res = inner_php_func_name_finder();
				if (html_display_definition == 2 && res
					  && !*display_func_name && strcmp(major_mode, "PHP"))
					strcpy(display_func_name, "PHP");
				return res;
		}
	}
	return 1;
}

php_func_name_finder()		// Uses same rules as HTML.
{
	return html_func_name_finder();
}

char *html_comment_start_override()
{
	recolor_buffer_range(point, point + 1);
	switch (get_tagged_region("color-as-unit", point)) {
		case HT_PHP:
			return "(/</|*>|#)[ \t\f]*";
		case HT_JSCRIPT:
			return "(/</|*>)[ \t\f]*";
		case HT_VBSCRIPT:
			return "(rem |')[ \t]*";
	}
	return 0;
}

char *html_comment_pattern_override()
{
	recolor_buffer_range(point, point + 1);
	switch (get_tagged_region("color-as-unit", point)) {
		case HT_PHP:
			return "(//|#).*$|/<*>(.|<newline>)*<*>/<FirstEnd>";
		case HT_JSCRIPT:
			return "//.*$|/<*>(.|<newline>)*<*>/<FirstEnd>";
		case HT_VBSCRIPT:
			return "(rem |').*$";
	}
	return 0;
}

char *html_comment_begin_override()
{
	recolor_buffer_range(point, point + 1);
	switch (get_tagged_region("color-as-unit", point)) {
		case HT_PHP:
			switch (php_comment_style) {
				case 1: return "# ";
				case 2: return "// ";
			}
			return "/* ";
		case HT_JSCRIPT:
			return new_c_comments ? "// " : "/* ";
		case HT_VBSCRIPT:
			return "' ";
	}
	return 0;
}

char *html_comment_end_override()
{
	recolor_buffer_range(point, point + 1);
	switch (get_tagged_region("color-as-unit", point)) {
		case HT_PHP:
			switch (php_comment_style) {
				case 1:
				case 2: return "";
			}
			return " */";
		case HT_JSCRIPT:
			return new_c_comments ? "" : " */";
		case HT_VBSCRIPT:
			return "";
	}
	return 0;
}

char *php_comment_start_override()
{
	return html_comment_start_override();
}

char *php_comment_pattern_override()
{
	return html_comment_pattern_override();
}

char *php_comment_begin_override()
{
	return html_comment_begin_override();
}

char *php_comment_end_override()
{
	return html_comment_end_override();
}

int html_maybe_break_line(int mode)
{
	switch (mode) {
		case 0:
		case 1:
			html_auto_fill_mode = (mode != 0);
			return 0;
		case 2:
			return html_auto_fill_mode;
	}
	if (!html_auto_fill_mode)		// We're not in auto-fill mode.
		return 1;
	recolor_buffer_range(point, point + 1);
	if (((get_tagged_region("color-as-unit", point) == -1) ? 2 : 3)
		  == html_auto_fill_mode)
		return 1;	// Only do one sort of filling?
	switch (get_tagged_region("color-as-unit", point)) {
		case HT_PHP:			 // Break only if in a comment.
		case HT_VBSCRIPT:
			return do_maybe_break_line(0);
		case HT_JSCRIPT:
			return do_maybe_break_line(1);
	}
	return maybe_break_this_line();	// Break always as text.
}

int html_indenter(), html_move_level(), html_recolor_from_here();

html_like_mode(int force_xml)
{
	mode_default_settings();
	sublanguage_code = 0;
	mode_keys = html_tab;		/* Use these keys. */
	if (force_xml || matches_at(0, 1, "<wspace>*%<%?xml"))
		sublanguage_code = SUBLANG_XML;
	if (sublanguage_code == SUBLANG_XML)
		major_mode = _xml_mode_name;
	else
		major_mode = _html_mode_name;

	strcpy(comment_start, "%<%!-- *");
	strcpy(comment_pattern, "%<%!--(.|<newline>)*--%><FirstEnd>");
	strcpy(comment_begin, "<!-- ");
	strcpy(comment_end, " -->");
	buffer_maybe_break_line = html_maybe_break_line;
	indenter = html_indenter;
	auto_indent = 1;
	mode_move_level = html_move_level;
	if (auto_show_html_delimiters)
		auto_show_matching_characters = html_auto_show_delim_chars;
	mode_auto_show_delimiter = html_auto_show_delimiter;
	recolor_range = color_html_range;	// set up coloring rules
	recolor_from_here = html_recolor_from_here;
	idle_coloring_size = 5000;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	try_calling("html-mode-hook");
	drop_all_colored_regions();
	make_mode();
}

command php_mode()
{
	html_like_mode(0);
	major_mode = _php_mode_name;
	sublanguage_code = SUBLANG_PHP;
	if (auto_show_php_delimiters)
		auto_show_matching_characters = php_auto_show_delim_chars;
	if (php_tab_override > 0)
		tab_size = php_tab_override;
	try_calling("php-mode-hook");
}

suffix_php()		// PHP server-side page
{
	php_mode();
}

suffix_php3()		// PHP server-side page
{
	php_mode();
}

suffix_php4()		// PHP server-side page
{
	php_mode();
}

suffix_sphp()		// PHP server-side page
{
	php_mode();
}

command backward_delete_character()
{
	if (delete_if_highlighted())
		return;
	if (virtual_space && cursor_to_column > 0)
		virtual_move_back();
	else {
		int many = (delete_hacking_tabs & 2) && character(point - 1) == ' '
				   && !has_arg;
		if (delete_hacking_tabs & 1)
			hack_tabs(-iter);
		do_delete_characters(-1);
		if (many)
			while (character(point - 1) == ' '
				   && current_column() % tab_size)
				delete(point - 1, point);
	}
}


do_reindent_after_yank()
{
	int tmp, orig = bufnum, from, to, stuff_before, max;
	jmp_buf this_level;

	if (_doing_input)	// Don't indent at prompts.
		return;
	max = get_mode_variable("reindent_after_%s_yank");
	if (got_bad_number)
		max = reindent_after_yank;
	if (!max || max > 0 && point - mark > max)
		return;	// Text is too large.
	stuff_before = matches_at(mark, -1, "^.*[^ \n\t].*") > 0;
	if (lines_between(mark, point, 0) < 1) {
		if (parse_string(1, "[ \t]*[^ \t\n]"))
			return;
		if (stuff_before)	// Other stuff on the line, don't reindent.
			return;
	}
	save_var narrow_start = 0, narrow_end = 0;
	if (reindent_c_comments == 1)
		save_var reindent_c_comments = 0;
	if (reindent_one_line_c_comments == 1)
		save_var reindent_one_line_c_comments = 0;
	prev_cmd = 0;
	from = mark;	// Make a copy of the buffer and do indenting
	to = point;		// there, so undo records just one operation.
	tmp = tmp_buf();	// (Yank-pop requires undo.)
	buf_xfer(tmp, 0, size());
	bufnum = tmp;
	mark = from;
	point = to;
	safe_copy_buffer_variables(tmp, orig);	// Duplicate bufvars into tmp.
	save_var top_level = &this_level;
	if (!setjmp(top_level)) {
		spot end = alloc_spot(1);
		save_var iter = 1, has_arg = 0;
		if (stuff_before) {		// Don't indent the first line;
			exchange_point_and_mark();	// there was already an indent.
			nl_forward();
			exchange_point_and_mark();
		}
		indent_region();
		if (!buffers_identical(orig, tmp)) {
			fix_region();
			if (stuff_before)	// Indent-region can expand the region.
				point = from;
			else
				from = point;
			to = size() - *end;
			delete(0, point);	// Discard all but the region.
			delete(*end, size());
			bufnum = orig;
			delete(from, size() - to);
			grab_buffer(tmp);
			_oldfrom = mark;
			_oldto = point;		/* end of yank */
		}
	}					// User aborted, just clean up.
	bufnum = orig;
	buf_delete(tmp);
}

fix_extension_list(int buf)
{
	save_var bufnum = buf;
	point = 0;
	do_drop_matching_lines(REGEX, "b[0-9]+patch%.e$", 1);
	string_replace("[\\\"]", "\\#0", REGEX);
	string_replace("^.+$", "(load-eel-from-path \"#0\" 2)", REGEX);
}

// Scan the rest of the current buffer for EEL-created things that
// couldn't be defined.  If there are any, remove them from the buffer,
// show them to the user, and return nonzero.  If none, return zero.
list_undefined_customizations()
{
	int list = 0;

	while (re_search(1, "^[^ \n\t:]+: (command|subroutine)\n")) {
		if (!list) {
			list = zap("-unknown-customizations");
			save_var bufnum = list;
			bprintf("Epsilon couldn't find definitions for the functions listed "
					"below.\nIf your previous setup included any EEL "
					"extension files, you\nmay want to rerun this command "
					"and include their names.\n\n");
			restore_vars();
		}
		buf_xfer(list, matchstart, matchend);
		delete(matchstart, matchend);
	}
	if (list) {
		set_buf_point(list, 0);
		view_buf(list, 1);
	}
	return list != 0;
}

// Comment out old customizations before this position.
comment_old_customizations(int pos)
{
	save_var narrow_end = size() - pos;
	point = 0;		// First delete previous batch of cleanings.
	simple_re_replace(0, "^![-(+&~]", "### ");
}


custom_apply_immediately()
{
	int tmp = tmp_buf();
	buf_xfer(tmp, 0, size());
	save_var bufnum = tmp, show_status = 0;
	do_drop_matching_lines(REGEX, "^%+(push-cmd|_region_file|_grep_files) ", 0);
	restore_vars();
	save_var _abort_load_on_error = 2;
	if (!parse_cmds())
		buf_delete(tmp);
}

custom_fix_mode_vars()
{
	char start[FNAMELEN], end[FNAMELEN];
	save_var point = 0;
	*start = *end = 0;
	if (re_search(1, "^%+mode-start .*: [0-9]+ \"(.*)\"$"))
		grab(find_group(1, 1), find_group(1, 0), start);
	point = 0;
	if (re_search(1, "^%+mode-end .*: [0-9]+ \"(.*)\"$"))
		grab(find_group(1, 1), find_group(1, 0), end);
	point = size();
	if (*start || *end)
		bprintf("+mode-format char array variable: %d \"%s%%b%%E%%M%s\"\n",
				ptrlen(mode_format), *start ? start : " ",
				*end ? end : " %d%p %s%f");
}

int make_difference_list(int orig, int cur)
{
	save_var show_status = 0;
	save_var bufnum = orig;
	save_var case_fold = 1;
	do_sort_region(0, size(), 0, 0);
	save_var bufnum = cur;
	save_var case_fold = 1;
	do_sort_region(0, size(), 0, 0);
	int out = tmp_buf();
	do_compare_sorted(orig, cur, "original-settings",
					  bufnum_to_name(out), NULL);
	bufnum = out;
	point = 0;
	save_var case_fold = 1;
	custom_apply_immediately();
	// Drop vars starting with -, but not color scheme ones.
	do_drop_matching_lines(REGEX, "^%+-[^o][^u][^r][^-]", 1);
	drop_boring_change_lines();
	custom_fix_mode_vars();
	do_drop_matching_lines(REGEX, "^%+(push-cmd|echo-line|permanent-menu"
						   "|avoid-(top|bottom)-lines|bufed-width"
						   "|mode-start|mode-end) ", 1);
	do_drop_matching_lines(REGEX, "^&.* color class for after-exiting: .*$", 1);
	return out;		// Don't count vars saved in session file, others.
}

// Check each line in the range to see if it names an existing file.
// Return 0 if everything's OK, otherwise ask the user if he wants to
// try again and return 1 if yes, 2 to give up.
test_valid_files(int start, int end)
{
	char file[FNAMELEN], msg[FNAMELEN], *s;
	int ret = 0;

	save_var point = start, narrow_end = size() - end;
	while (!ret && grab_line(bufnum, file)) {
		if (s = strstr(file, " (not found)"))
			*s = 0;
		if (check_file(file) != CHECK_FILE)
			ret = 1;
		else if (is_relative(file))
			ret = 3;
	}
	if (!ret)
		return ret;
	if (ret == 3)
		sprintf(msg, "\"%s\" not an absolute path, try again? ", file);
	else
		sprintf(msg, "File \"%s\" not found, try again? ", file);
	return ask_yn("Try again?", msg, "Yes", "No", 1) ? 1 : 2;
}

command list_customizations()
{
	char cmd[FNAMELEN], *args, time[FNAMELEN];
	char ext_file[FNAMELEN], settings_file[FNAMELEN], *s;
	int tmp, cur, ext, i, max, err;
	int cust = prep_customizations(), clist;

	iter = 1;
	tiled_only();
	get_customization_directory(ext_file);
	strcat(ext_file, "extensions-list");
	get_customization_directory(settings_file);
	strcat(settings_file, "settings-list");
	delete_file(settings_file);
	i = name_match("-loaded-eel-file-", STARTMATCH);
	save_var bufnum = ext = zap("-list-customizations");
	give_binding(cmd, "exit-level");
	bprintf("Epsilon will include the following EEL extension files as part\n"
			"of your customizations.  Edit this list if needed, then\n"
			"press %s to begin.\n\n%p\n", cmd, color_class viewed_text);
	point--;
	for (; i; i = name_match("", 0))	// List new extensions by searching
		if (name_type(i) == NT_VAR) {	// for id variables.
			s = get_str_var(i);
			bprintf("%s%s\n", s, (check_file(s) == CHECK_FILE)
					? "" : " (not found)");
		}
	to_buffer_num(ext);
	point = 0;
	do {
		recursive_edit();
		if (!buf_exist(ext))
			return;
		point = 0;
		point += parse_string(1, "(.|\n.)*\n*");
		i = test_valid_files(point, size());
		if (i == 2) {
			buf_zap(ext);
			return;
		}
	} while (i);
	point = 0;		// Drop the explanatory text.
	delete(0, parse_string(1, "(.|\n.)*\n*"));
	if (err = file_write(ext_file, FILETYPE_BINARY)) {
		file_error(err, ext_file, "write error");
		return;
	}
	strcpy(cmd, argv[0]);
	convert_to_8_3_filename(cmd);
	sayput("Running %s to create original settings list...", cmd);
	args = " -s original -p0 -noinit -r build-settings-file";
	err = shell_with_args(cmd, args);
	if (err)
		error("Error running %s %s", cmd, args);
	sayput("Reading original settings list...");
	bufnum = tmp = tmp_buf();
	if (err_file_read(settings_file, FILETYPE_BINARY))
		return;
	bufnum = cur = tmp_buf();
	max = final_index();
	for (i = 1; i <= max; i++)
		list_one(bufnum, i, 3);
	use_default = 0;
	bufnum = cur;
	clist = make_difference_list(tmp, cur);
	restore_vars();
	buf_delete(tmp);
	buf_delete(cur);
	to_buffer_num(cust);
	current_time_and_date(time);
	save_spot point;
	i = point;
	bprintf("\n# List-customizations created these settings %s.\n", time);
	fix_extension_list(ext);
	grab_buffer(ext);
	grab_buffer(clist);
	buf_delete(clist);
	delete_file(settings_file);
	delete_file(ext_file);
	bprintf("\n");
	point = i;
	if (!has_arg)
		comment_old_customizations(point);
	if (!list_undefined_customizations())
		say("");
}

build_settings_file()
{
	char ext_file[FNAMELEN], settings_file[FNAMELEN], eel_file[FNAMELEN];
	int tmp = tmp_buf(), i, err, max;

	get_customization_directory(ext_file);
	strcat(ext_file, "extensions-list");
	get_customization_directory(settings_file);
	strcat(settings_file, "settings-list");
	save_var bufnum = tmp;
	if (err_file_read(ext_file, FILETYPE_BINARY)) {
		show_text(0, 300, "Can't read extensions list file, aborting.");
		leave(1);
	}
	point = 0;
	while (grab_line(bufnum, eel_file)) {
		sayput("Loading %f...", eel_file);
		if (load_eel_from_path(eel_file, 2))
			return;
	}
	buf_zap(bufnum = tmp);
	sayput("Listing original settings...");
	max = final_index();
	for (i = 1; i <= max; i++)
		list_one(bufnum, i, 3);
	use_default = 0;
	bufnum = tmp;
	if (err = file_write(settings_file, FILETYPE_BINARY)) {
		file_error(err, settings_file, "write error");
		delete_file(settings_file);	// Ensure caller notices failure.
	}
	say("");
	leave(0);
}

command clean_customizations()
{
	int dup;

	tiled_only();
	to_buffer_num(prep_customizations());
	point = 0;		// First delete previous batch of cleanings.
	simple_re_replace(0, "###.*\n", "");
	point = 0;
	while (point < size() - narrow_end) {
		dup = 0;
		if (parse_string(1, "^(%+<word|dash>+ )"))
			dup = check_dup_customization(1, "^<>");
		else if (parse_string(1, "^(%&<word|dash>+ color class "
							  "for <word|dash>+: )"))
			dup = check_dup_customization(1, "^<>");
		else if (parse_string(1, "^%(bind-to-key \"[^\"\n]+\" (\"[^\"\n]+\")%)"))
			dup = check_dup_customization(1, "^%(bind-to-key \"[^\"\n]+\" <>%)");
		else if (parse_string(1, "^(<word|dash>+: macro )"))
			dup = check_dup_customization(1, "^<>");
		else if (parse_string(1, "^(%(load-eel-from-path \"[^\"\n]+\" 2%)$)"))
			dup = check_dup_customization(1, "^<>");
		else if (parse_string(1, "^(~<^wspace>+ \".*\": )"))
			dup = check_dup_customization(1, "^<>");
		if (dup) {
			stuff("### ");
			nl_forward();
			continue;
		}
		switch (curchar()) {
			case '(':	skip_paren_cmd(); break;
			case ';':
			case '#':	nl_forward(); break;
			default:
				nl_forward();
				break;
		}
	}
}

command import_customizations()
{
	int tmp = zap("import-customizations"), i, ver = 0, err, clean = !has_arg;
	int debug = (iter == 99);
	char buf[2000], maindir[FNAMELEN], bindir[FNAMELEN];
	char origsta[FNAMELEN], cursta[FNAMELEN], *s, *exe_ext = "", pat[FNAMELEN];
	char answer[6][2000], ranswer[6][2000], oldname[FNAMELEN];
	char origdir[FNAMELEN], *epsout = lookpath("epsout.e");

	iter = 1;
	tiled_only();
	getcd(origdir);
	on_exit chdir(origdir);
	putenv("EPSRUNS");		// Prevent error msg in old Epsilons.
	save_var bufnum;
	if (!epsout)
		error("The file epsout.e is missing.");
	for (i = EELVERSION / 10 - 1; i >= 8 ; i--) {
		if (is_win32) {
			sprintf(buf, "EPSPATH%02d", (i < 10) ? (i * 10) : i);
			if (!verenv(buf))
				continue;
		} else {
			sprintf(buf, "~/.epsilon/epsilon-v%d.sta", i);
			if (!check_file(buf))
				continue;
		}
		buf_printf(tmp, "%d\n", ver = i);
	}
	set_buf_point(tmp, 0);
	switch (count_lines_in_buf(tmp, 0)) {
		case 0:		ver = 0; break;
		case 1:		break;
		default:
			*buf = 0;
			buf_delete_final_newlines(tmp);
			if (!get_choice(tmp, buf, " Update from version ",
							"Update from which version? ",
							"OK", "Cancel", ""))
				return;
			ver = numtoi(buf);
			say("");
	}
	*maindir = *bindir = *origsta = *cursta = 0;
	if (is_win32 && ver) {
		exe_ext = ".exe";
		sprintf(buf, "EPSPATH%02d", (ver < 10) ? (ver * 10) : ver);
		if (s = verenv(buf)) {
			strcpy(maindir, s);
			absolute(maindir);
			sprintf(origsta, "%s%coriginal.sta", maindir, path_sep);
		}
		sprintf(buf, "BINPATH%02d", (ver < 10) ? (ver * 10) : ver);
		if (*maindir)
			sprintf(bindir, "%s\\bin", maindir);
		s = verenv(buf);
		if (s)
			strcpy(bindir, s);
		absolute(bindir);
		if (*maindir)
			sprintf(cursta, "%s\\epsilon.sta", maindir);
	} else if (ver) {		// Set vars for Unix.
		sprintf(buf, "/usr/local/epsilon%d*/bin/epsilon", ver);
		int max = 0;		// Find newest match of pattern.
		struct file_info fi;
		s = file_match(buf, STARTMATCH);
		*buf = 0;
		while (s && check_file(s, &fi)) {
			if (!max || fi.raw_file_date_low > max) {
				strcpy(buf, s);
				max = fi.raw_file_date_low;
			}
			s = file_match("", 0);
		}
		if (*buf) {
			get_tail(buf, 0)[-1] = 0;
			strcpy(bindir, buf);
			strcpy(origsta, buf);
			strcpy(get_tail(origsta, 0), "original.sta");
		}
		sprintf(cursta, "~/.epsilon/epsilon-v%d.sta", ver);
	}
	save_var window_bufnum;
	to_buffer_num(tmp);
	buf_zap(bufnum);
	sprintf(oldname, "Epsilon %d", ver);
	if (!ver)
		strcpy(oldname, "the old version");
	bprintf("Epsilon will try to import any customizations you saved in your\n");
	bprintf("state file in %s.  Please correct any of\n", oldname);
	give_binding(buf, "exit-level");
	bprintf("these settings as necessary, then press %s\n", buf);
	bprintf("to begin the import process.\n\n");

	if (!ver) {
		bprintf("Epsilon couldn't find any %s,\n", is_win32
				? "Epsilon registry entries"
				: "customized state file in ~/.epsilon");
		bprintf("so you'll have to fill in this information.\n\n");
	}
	bprintf("1. Your customized state file for %s is:\n%p%s\n%p\n",
			oldname, color_class viewed_text,
			*cursta ? cursta : "full path for your .sta file", -1);
	if (*cursta && check_file(cursta) != CHECK_FILE)
		bprintf("(This file doesn't exist.)\n\n");
	bprintf("2. The original pre-customization state file for %s\n", oldname);
	bprintf("is:\n%p%s\n%p\n",
			color_class viewed_text,
			*origsta ? origsta : "full path for original.sta", -1);
	if (*origsta)
		if (check_file(origsta) != CHECK_FILE)
			bprintf("(This file doesn't exist.)\n\n");
	bprintf("3. If you customized %s by loading EEL files,\n", oldname);
	bprintf("please list each .e file's pathname here, one per line:\n%p\n%p\n",
			color_class viewed_text, -1);
	sprintf(buf, "full path for epsilon%s", exe_ext);
	if (*bindir)
		sprintf(buf, "%s%c%s%s", bindir, path_sep,
				(ver < 10) ? "e32" : "epsilon", exe_ext);
	bprintf("4. The Epsilon%s executable file for %s is:\n%p%s\n%p\n",
			is_win32 ? " (GUI, not console version)" : "", oldname,
			color_class viewed_text, buf, -1);
	if (*bindir && check_file(buf) != CHECK_FILE)
		bprintf("(This file doesn't exist.)\n\n");
	sprintf(buf, "full path for eel%s", exe_ext);
	if (*bindir)
		sprintf(buf, "%s%ceel%s", bindir, path_sep, exe_ext);
	bprintf("5. The EEL compiler executable for %s is:\n%p%s\n%p\n",
			oldname, color_class viewed_text, buf, -1);
	if (*bindir && check_file(buf) != CHECK_FILE)
		bprintf("(This file doesn't exist.)\n\n");
	get_customization_directory(buf);
	strcat(buf, "upd");
	mkdir(buf);
	bprintf("6. Use this working directory for intermediate files:\n%p%s\n%p\n",
			color_class viewed_text, buf, -1);
	if (buf && check_file(buf) != CHECK_DIR)
		bprintf("(Warning: Couldn't create this directory.)\n\n");
	for (;;) {
		err = point = 0;
		recursive_edit();
		if (!buf_exist(tmp))
			return;
		bufnum = tmp;
		point = 0;		// Now retrieve the answers.
		for (i = 0; i < 6; i++) {
			sprintf(pat, "^%d%%. [^:]+:\n((.|\n.)*)\n\n", i + 1);
			if (!re_search(1, pat)) {
				say("Can't interpret response %d", i + 1);
				err = 1;
				break;
			}
			grab(find_group(1, 1), find_group(1, 0), answer[i]);	// 8.3 version.
			grab(find_group(1, 1), find_group(1, 0), ranswer[i]);	// For display.
			if (i == 2) {
				err = test_valid_files(find_group(1, 1), find_group(1, 0));
				if (err == 2)
					quick_abort();
				err = err ? 3 : 0;	// Don't prompt again.
			} else if (check_file(answer[i]) != ((i == 5)
				? CHECK_DIR : CHECK_FILE)) {
				say("Response %d: Can't read %s \"%f\".",
					  i + 1, (i == 5) ? "directory" : "file", answer[i]);
				err = 1;
				break;
			} else
				convert_to_8_3_filename(answer[i]);
		}
		if (!err)
			break;
		if (err != 3
			  && !ask_yn("Try again", "Would you like to correct your response? ",
						 "Yes", "No", 0))
			quick_abort();
	}
	if (chdir(answer[5]))
		error("Couldn't set directory to %f.", answer[5]);
	buf_zap(tmp);	// Compile epsout & extensions, then save list of them.
	compile_for_import_customizations(epsout, answer[4], answer[5], 0);
	compile_for_import_customizations(answer[2], answer[4], answer[5], tmp);
	if (err = file_write("extensions-list", FILETYPE_BINARY)) {
		file_error(err, "extensions-list", "write error");
		return;
	}

	// Run the old version with orig state file.
	sprintf(buf, " -s%s -l%s%cepsout.b -r%sbuild_settings_file",
			answer[1], answer[5], path_sep, debug ? "debug_" : "");
	show_text(-2, 1, "Run %s %s", answer[3], buf);
	shell_with_args(answer[3], buf);
	buf_zap(tmp);
	if (err = file_write("extensions-list", FILETYPE_BINARY)) {
		file_error(err, "extensions-list", "write error");
		return;
	}
	if (err_file_read("settings-list", FILETYPE_BINARY))
		return;
	// Run the old version with new state file.
	sprintf(buf, " -s%s -l%s%cepsout.b -r%sbuild_settings_file",
			answer[0], answer[5], path_sep, debug ? "debug_" : "");
	show_text(-2, 1, "Run %s %s", answer[3], buf);
	shell_with_args(answer[3], buf);
	int cur = tmp_buf();
	on_exit buf_delete(cur);
	bufnum = cur;
	if (err_file_read("settings-list", FILETYPE_BINARY))
		return;
	int clist = make_difference_list(tmp, cur);
	on_exit buf_delete(clist);	// Now format the list of differences.
	int cust = prep_customizations();
	bufnum = cust;
	current_time_and_date(buf);
	save_var point;
	i = point;
	bprintf("\n# Import-customizations created these settings %s\n", buf);
	bprintf("# from %s.\n", ranswer[0]);
	int ext = tmp_buf();
	on_exit buf_delete(ext);
	buf_printf(ext, "%s\n", ranswer[2]);
	fix_extension_list(ext);
	grab_buffer(ext);
	grab_buffer(clist);
	bprintf("\n");
	restore_vars();
	to_buffer_num(cust);
	point = i;
	if (clean)
		comment_old_customizations(point);
	if (!list_undefined_customizations())
		say("Done.");
}

color_batch_range(from, to)
{
	int t = -1;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend, case_fold = 1;
	point = to;				// Color entire lines.
	nl_forward();
	to = point;
	save_var narrow_end = size() - to;
	set_character_color(from, to, -1);
	point = from;

	while (point < to) {
		if (!re_search(1, "[-a-z0-9.%/\\]+|[<>|]")) {
			t = size();
			break;
		}
		t = matchstart;
		if (strchr("<>|", character(point - 1))) {
			set_character_color(t, point, color_class batch_punctuation);
			continue;
		}
		switch (is_batch_keyword(t)) {
			case 1:
				set_character_color(t, point, color_class batch_keyword);
				break;
			case 2:
				nl_forward();
				set_character_color(t, point, color_class batch_comment);
				break;
			default:
				set_character_color(t, point, -1);
				break;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	return point;
}

drag_drop_handler()
{
	char buf[FNAMELEN], *s, dir[FNAMELEN];
	int line = -1, column = -1, j, dir_flags = directory_flags, first = 1;
	int had_action = 0;

	tiled_only();
	if (mouse_screen >= 0)
		to_mouse_window(mouse_x, mouse_y, mouse_screen);
	while (j = drag_drop_result(buf)){
		if (*buf == '+') {			// a +linenum parameter
			line = numtoi(buf + 1);
			s = strchr(buf, ':');
			if (s)
				column = numtoi(s + 1);
		} else if (*buf == '-') {
			switch (buf[1]) {
				case 'd':
					s = index(buf + 2, '=');
					if (!s)		// accept either VAR=VAL or VAR!VAL
						s = index(buf + 2, '!');
					if (!s)
						break;
					*s = 0;
					if (j = find_index(buf + 2)) {
						set_var_val(j, s + 1, 1);
						continue;
					} else
						say("No such variable %s.", buf + 2);
					break;
				case 'l':
					load_commands(buf + 2);
					break;
				case 'r':
					had_action = 1;
					s = index(buf + 2, '=');
					if (!s)		// accept either CMD=ARG or CMD!ARG
						s = index(buf + 2, '!');
					if (s)
						*s = 0;
					if (j = find_index(buf + 2)) {
						if (s)
							(*(int (*)())j)(s + 1);
						else
							(*(int (*)())j)();
					} else
						say("No such function %s.", buf + 2);
					break;
				case 'w':
					dir_flags = numtoi(buf + 2);
					break;
			}
			continue;
		} else if (*buf) {
			windows_foreground();
			had_action = 1;
			absolute(buf);
			if (is_directory(buf) || is_remote_file(buf) || (is_pattern(buf) & 2))
				do_find(buf, default_translation_type);
			else
				find_it(buf, default_translation_type);
			if (line >= 0) {
				go_line(line);
				line = -1;
			}
			if (column >= 0) {
				move_to_column(column);
				column = -1;
			}
			if ((dir_flags & DIR_CWD_FROM_FILE) && first) {
				strcpy(dir, buf);
				*get_tail(dir, 1) = 0;
				chdir(dir);
			}
			first = 0;
			if (j == 2)
				show_binding("Use %s to release the Epsilon client when you finish.",
							 "resume-client");
		}
	}
	if (!had_action)			// Ensure we're not minimized, unless we got
		windows_foreground();	// a -r command but no file.
}

color_this_perl_range(from, to)
{
	int t = -1, talk, c, talk_at = 0, after_here_doc = 0;
	char *langname;

	save_var point, matchstart, matchend, case_fold = 0;
	talk = (to - from > 30000);	// Show status during long delays.
	point = from;
	langname = sublanguage_code == SUBLANG_PHP ? "PHP"
											   : in_shell_buffer ? "Shell" : "Perl";

	while (point < to) {
		if (!re_search(1, "[0-9A-Za-z_]+|[\"'`/#$&<=\n]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (c = character(point - 1)) {		// check last char
			case '#':			// found comment
				to_end_line();
				set_character_color(t, point, color_class perl_comment);
				break;
			case '&':
				if (in_shell_buffer)
					break;
				if (parse_string(1, "((<word>+('|::))*|'|::)<word>+"))
					set_character_color(t, point = matchend,
										color_class perl_function);
				break;
			case '$':
				if (in_shell_buffer && curchar() == '"') {	// Translated str.
					set_character_color(t, point, color_class perl_string);
					break;
				}
				if (in_shell_buffer || sublanguage_code == SUBLANG_PHP) {
					if (parse_string(1, "<word>+|%{([^{}\n]|%{[^{}\n]*%})+%}|[^\n`]"))
						set_character_color(t, point = matchend,
											color_class perl_variable);
					break;
				}
				if (!isident(curchar()) && !strchr("#{}", curchar())) {
					if (character(point++) == '^')
						point++;	// $', $?, $^Z or similar special var
					set_character_color(t, point, color_class perl_variable);
				} else if (parse_string(1, "#?((<word>+('|::))*|'|::)<word>+"))
					set_character_color(t, point = matchend,
										color_class perl_variable);
				else if (parse_string(1, "#?{"))
					set_character_color(t, point = matchend - 1,
										color_class perl_variable);
				break;
			case '\'':
				if (in_shell_buffer ? character(t - 1) != '\\'
									: (!isident(character(t - 1))
									   && !strchr("&$@%*", character(t - 1))))
					color_perl_fwd_quote(c, t, 0);	// old syntax pkg'func
				break;
			case '"':		// found a string literal
			case '`':
				if (in_shell_buffer && character(t - 1) == '\\')
					break;
				color_perl_fwd_quote(c, t, 1);
				break;
			case '/':
				if (sublanguage_code == SUBLANG_PHP) {
					if (curchar() == '/') {
						nl_forward();
						set_character_color(t, point, color_class perl_comment);
					} else if (curchar() == '*') {
						search(1, "*/");
						set_character_color(t, point, color_class perl_comment);
					}
					break;
				}

				// if previous thing looks like an expression, skip it
				if (!in_shell_buffer && !perl_after_expression(t, 1)) {
					point--;
					color_perl_delimiters(0, 1, 2);
				}
				break;
			case '<':
				if (sublanguage_code == SUBLANG_PHP) {	// Look for <<<word
					if (parse_string(1, "%<%<(<alpha|_>+)"))
						color_perl_here_doc(t, 1);
					break;
				}
				if (perl_after_expression(t, 0) || character(t - 1) == '<')
					break; // It's a shift or lessthan op, not "here" doc or input op.
				if (curchar() == '<') {		// found <<, possible "here" document.
					point++;	// Next must be an identifier or text in quotes.
					if (parse_string(1, in_shell_buffer
									 ? "[ \t]*[-\\]?(<alpha|_>+)"
									 : "(<alpha|_>+)"))
						color_perl_here_doc(t, 1);
					else if (parse_string(1, "[ \t]*\"([^\"\n]+)\""))
						color_perl_here_doc(t, 1);
					else if (parse_string(1, "[ \t]*'([^'\n]+)'"))
						color_perl_here_doc(t, 0);
					else if (parse_string(1, "[ \t]*`([^`\n]+)`"))
						color_perl_here_doc(t, 1);
					else
						break;
					after_here_doc = 1;
					break;
				}
				if (in_shell_buffer || sublanguage_code == SUBLANG_PHP)
					break;
				point--;	// File glob operator.
				color_perl_delimiters(0, 1, 0);
				break;
			case '\n':
				if (after_here_doc)
					if (get_tagged_region("color-perl-as-unit", point) != -1)
						to_end_line();	// Already colored this line.
					else
						after_here_doc = 0;
				break;
			case '=':			// Found POD (documentation).
				if (in_shell_buffer || sublanguage_code == SUBLANG_PHP)
					break;
				if (parse_string(-1, "^=")) {
					search(1, "\n=cut");
					set_character_color(t, point, color_class perl_comment);
				}
				break;
			default:		// found identifier, kywd, or number
				color_perl_word(t);
				break;
		}
		if (talk && point > talk_at + 2000) {
			note("Coloring %s program: %d%% complete...",
				 langname, muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	if (talk)
		note("");
	return point;
}

command push()
{
	char cmd[FNAMELEN];

	_last_was_grep = 0;
	save_var clear_process_buffer = (clear_process_buffer & 3) != 0;
	do_push(get_cmdline("Push With Command", cmd), capture_output, 1);
}

command start_process()
{
	char cmd[FNAMELEN], *cmdline, *sh, fixup[FNAMELEN], *p;
	char cust1[FNAMELEN], cust2[FNAMELEN], buf[FNAMELEN];
	int i = 1;

	save_var clear_process_buffer = (clear_process_buffer & 5) != 0;
	_last_was_grep = 0;
	tiled_only();
	error_if_input(1);
	strcpy(buf, PROCBUF);
	if (has_arg && (has_feature & FEAT_MULT_CONCUR)) {
		if (is_process_buffer(name_to_bufnum(buf)))
			do {
				sprintf(buf, "process-%d", i++);
			} while (is_process_buffer(name_to_bufnum(buf)));
	} else if (no_running(1))
		return;
	iter = has_arg = 0;
	cmdline = get_cmdline("Start Process With Command", cmd);
	create_concur(cmdline, buf);
	locate_window(buf, "");
	if (opsys == OS_UNIX && (sh = getenv("SHELL"))) {
		*cust1 = *cust2 = *fixup = 0;
		if (p = lookpath("epsilon.sh", 0))
			strcpy(cust1, p);
		if (p = lookpath("epsilon.csh", 0))
			strcpy(cust2, p);
		if (!*cust1 || !*cust2)
			sprintf(fixup, "echo \"Customization file not found, "
					"no customizations performed.\">/dev/null");
		else if (strstr(sh, "/sh") || strstr(sh, "/bash") || strstr(sh, "/ksh"))
			sprintf(fixup, ". %s; set +o emacs", cust1);
		else if (strstr(sh, "/csh") || strstr(sh, "/tcsh"))
			sprintf(fixup, "source %s", cust2);
		else
			sprintf(fixup, "echo \"Unrecognized shell, no customizations "
					"performed.\">/dev/null");
		delay(40, 0, bufnum);	// Let's first get a prompt if we can.
		point = type_point;
		bprintf("%s\n", fixup);
	}
	point = size();
	process_mode();
}

command compile_buffer()
{		// compile the current buffer (mode-specific)
	int res, buf = bufnum;
	char cmd[FNAMELEN * 4], oscmd[FNAMELEN * 4], full[FNAMELEN * 4], *ext;

	save_var clear_process_buffer = (clear_process_buffer & 9) != 0;
	tiled_only();
	if (use_compile_command_file_variable && look_for_filevar_pattern(
		"compile-command", "([^-\" \n\t;]([^-;\n]|-[^*])*|\"[^\n\"]+\")", NULL)) {
		grab(find_group(1, 1), find_group(1, 0), oscmd);
		if (*oscmd == '"')
			drop_quotes(oscmd);
		if (use_compile_command_file_variable != 1)
			get_strdef(cmd, "Command to compile this buffer", oscmd);
		else
			strcpy(cmd, oscmd);
		if (*cmd)
			save_var compile_buffer_cmd = cmd;
	}
	if (!compile_buffer_cmd) {
		ext = get_extension(filename);
		sprintf(cmd, "compile-%s-cmd", *ext ? (ext + 1) : "none");
		sprintf(oscmd, "%s-%s", cmd, is_unix ? "unix" :
				is_win32 ? "win32" : "other");
		if ((ext = get_str_var(find_index(oscmd)))
			  || (ext = get_str_var(find_index(cmd))))
			save_var compile_buffer_cmd = ext;
	}
	if (!compile_buffer_cmd) {
		get_string(cmd, "Command to compile this buffer: ");
		compile_buffer_cmd = strsave(cmd);
	}
	save_var concurrent_make;
	if (pre_compile_hook)	// hook can change above vars
		if ((*pre_compile_hook)())
			return;		// hook says don't do compile
	if (concurrent_compile < 3)
		concurrent_make = concurrent_compile;
	if (concurrent_compile == 0 && (is_gui || is_win32))	// Run in the
		concurrent_make = 4;		// concur proc if any, but keep control.
	if (start_process_in_buffer_directory) {
		char dir[FNAMELEN], old[FNAMELEN];

		getcd(old);
		get_buffer_directory(dir);
		chdir(dir);
		build_filename(full, compile_buffer_cmd, filename);
		chdir(old);
		if (start_process_in_buffer_directory == 2)	// For error msgs.
			save_var start_process_in_buffer_directory = 1;
	} else
		build_filename(full, compile_buffer_cmd, filename);
	res = do_compile(full);
	if (post_compile_hook)
		(*post_compile_hook)(res, buf);
}

load_this_bytecode_file(res, buf)	// load .b file made from current file
{
	char file[FNAMELEN];

	if (res && res != 4)
		return;
	save_var bufnum = buf;
	strcpy(file, filename);
	strcpy(get_extension(file), byte_extension);
	if (check_file(file, NULL) == CHECK_FILE) {
		load_commands(filename);
		say("%f compiled and loaded.%s", *filename ? filename : "Buffer",
			  res == 4 ? "  (There were warnings.)" : "");
	}
}

compile_via_dll()	// Try to compile .e file using a DLL, return 1 if ok.
{
	if (!(has_feature & FEAT_EEL_COMPILE))
		return 0;			// Not supported under this platform.
	_had_errors = 0;
	save_var point;
	switch (eel_compile(bufname, 0, compile_eel_dll_flags, "eel-errors")) {
		case 2:		// Couldn't access DLL.
			say("Unable to compile internally, running process.");
			return 0;
		case 1:
			break;
		case 0:
			say("%f compiled and loaded.%s",
				  filename ? filename : "Buffer",
				  buffer_size("eel-errors") ? "  (There were warnings.)" : "");
			break;
		default:
			say("Unable to compile internally, running process.");
			return 0;
	}
	restore_vars();
	save_var bufname = "eel-errors";
	if (size()) {
		point = 0;			// Starting at the beginning,
		set_error_spot();	//    scan in this buffer.
		do_next_error(1, NULL);	// Scan for error messages.
	} else {
		restore_vars();
		delete_buffer("eel-errors");
	}
	return 1;
}

command make()
{
	char cmd[FNAMELEN];

	save_var clear_process_buffer = (clear_process_buffer & 17) != 0;
	tiled_only();
	get_cmdline("Make using command", cmd);
	save_var start_process_in_buffer_directory
			= start_make_in_buffer_directory;
	do_compile(cmd);
}


update_mode_more_extra(int offset)
{
	char *s;
	int cnt = 0;

	for (s = buf_match("compilation-output", STARTMATCH); s
			 ; s = buf_match("compilation-output", 0)) {
		if (is_process_buffer(name_to_bufnum(s)))
			cnt++;
	}
	cnt += offset;
	switch (cnt) {
		case 0:		*mode_more_extra = 0; break;
		case 1:		strcpy(mode_more_extra, "[Compiling] "); break;
		default:	sprintf(mode_more_extra, "[Compiling %d] ", cnt); break;
	}
	all_must_build_mode = 1;
	maybe_refresh();
}

compile_handler(int activity, int buf, int from, int to)
{
	UNREFERENCED_PARAMETER(to);
	if (activity == NET_DONE) {
		save_var bufnum = buf;
		point = size();
		if (size() && character(point - 1) != '\n')
			bprintf("\n");
		bprintf("Process exited with status %d\n", from);
		update_mode_more_extra(-1);
	}
}

do_compile(cmd)
char *cmd;
{
	int err = 0, nerr, sep_buf_ok = 0;
	int obuf = bufnum;
	char *procbuf = PROCBUF, curdir[FNAMELEN];

	_last_was_grep = 0;
	if (maybe_save_all())
		return 3;
	if (start_process_in_buffer_directory)
		get_buffer_directory(curdir);
	else
		getcd(curdir);
	drop_final_slash(curdir);
	if (!(has_feature & FEAT_MULT_CONCUR) && !another_process())
		save_var restart_concurrent = 0;

	// If we're allowed to run non-concur, then use separate buffer.
	if (concurrent_make < 2 && (has_feature & FEAT_MULT_CONCUR)
		  && compile_in_separate_buffer)
		sep_buf_ok = 1;
	else if (concurrent_make == 2)
		create_concur("", PROCBUF);
	else if (!sep_buf_ok && concurrent_make == 0)
		try_exit_concurrent();
	if (!sep_buf_ok && clear_process_buffer)
		zap(PROCBUF);
	if (sep_buf_ok) {
		int i;
		char newname[FNAMELEN];
		bufnum = create(procbuf = "compilation-output");
		for (i = 2; is_process_buffer(bufnum); i++) {
			sprintf(procbuf = newname, "compilation-output-%d", i);
			bufnum = create(procbuf);
		}
		if (clear_process_buffer)
			zap(procbuf);
		process_mode();
		point = size();
		set_error_spot();
		_had_errors = 0;
		bprintf("Running `%s' in directory `%s'\n", cmd, curdir);
		type_point = size();
		if (compile_in_separate_buffer == 2)
			locate_window(procbuf, "");
		if (!concur_shell("", cmd, curdir, procbuf)) {
			update_mode_more_extra(0);
			when_activity = compile_handler;
			while (process_exit_status == PROC_STATUS_RUNNING) {
				if (char_avail() && getkey() != MOUSE_MOVE && !IS_WIN_KEY(key)) {
					ungot_key = key;	// Wait until user presses a key
					return 2;			// or process finishes.
				}
				delay(10, COND_KEY, bufnum);	// Delay 0.1 secs
			}
			err = process_exit_status;
		} else
			err = -1;
	} else if (concurrent_make == 4 && is_process_buffer(name_to_bufnum(PROCBUF))) {
		save_var bufname = PROCBUF;	// Run the program in the concur
		set_error_spot();			// proc buffer, but don't process
		_had_errors = 0;			// the user's keys until it finishes.
		if (*cmd) {
			point = size();
			bprintf("%s\n", cmd);
		}
		maybe_refresh();
		iter = 1;
		delay(-1, COND_PROC, bufnum);
		if (!is_process_buffer(name_to_bufnum(PROCBUF)) || !process_input())
			return 2;	/* else assume it's compiled */
	} else if (no_running(0)) {	/* goes to end of proc buffer */
		set_error_spot();
		_had_errors = 0;
		to_begin_line();
		bprintf("%pRunning `%s' in directory `%s'\n",
				color_class process_output, cmd, *process_current_directory
				? process_current_directory : curdir);
		type_point = point = size();
		if (*cmd)
			bprintf("%s\n", cmd);
		maybe_refresh();
		iter = 1;
		if (opsys != OS_UNIX)
			delay(-1, COND_PROC | COND_KEY, bufnum);
		if (!is_process_buffer(name_to_bufnum(PROCBUF)) || !process_input())
			return 2;	/* else assume it's compiled */
		if (!(has_feature & FEAT_DETECT_CONCUR_WAIT))
			return 5;
	} else {
		bufnum = create(PROCBUF);
		point = type_point = size();
		buffer_printf("process", "%p\nRunning `%s' in directory `%s'\n",
					  color_class process_output, cmd, curdir);
		bufnum = obuf;
		err = do_push(cmd, 1, 0);
	}
	maybe_restart_concurrent();
	nerr = do_next_error(1, cmd);	// scan for error messages
	if (!nerr && err) {	// no message recognized, but error code
		say("Process returned %d", err);
		locate_window(procbuf, "");
		point = size();		/* show end of process buffer */
		return err;
	} else
		bufnum = obuf;
	if (err)
		return 1;
	else if (nerr)
		return 4;
	return 0;
}

color_process_range(from, to)
{
	if (from >= to || spot_to_buffer(process_input_starts_at) <= 0)
		return to;
	save_var point, matchstart, matchend;
	point = *process_input_starts_at;
	if (get_character_color(point - 1) != color_class process_prompt)
		return to;
	to_end_line();
	set_character_color(*process_input_starts_at, point,
						color_class process_input);
	if (curchar() == '\n')
		set_character_color(point, point + 1, -1);
	return to;
}

try_grab_directory_name(int from, int to)
{
	char dir[FNAMELEN];

	if (!is_win32 || !use_process_current_directory
		  || to - from > 200 || can_get_process_directory == 1)
		return;					// Too long for a likely dir name.
	if (*process_prompt_pattern) {
		save_var narrow_end = size() - to;
		if (!matches_at(from, 1, process_prompt_pattern))
			return;
		grab(find_group(1, 1), find_group(1, 0), dir);
	} else {
		if (character(--to) != '>')
			return;
		grab(from, to, dir);
	}
	convert_cygwin_filename(dir, 1);
	if (is_directory(dir)) {
		process_current_directory = dir;
		if (use_process_current_directory == 2)
			chdir(dir);
		can_get_process_directory = 2;
	}
}

command process_complete()
{
	char pat[FNAMELEN], pref[FNAMELEN];
	int cnt, start, end, buf = create(_MATCH_BUF);
	int offset = IS_SHIFT_KEY(full_key) ? -1 : 1;

	start = end = point;
	if (start < type_point) {
		if (!maybe_indent_rigidly(IS_SHIFT_KEY(full_key)))
			indent_like_tab();
		return;
	}
	*pat = 0;
	if (process_line_inside_quoting() || parse_string(-1, "[^\n\t <>\"]+", pat)) {
		start = point;
		point = end;
		if (process_line_inside_quoting())
			grab(start = matchend + 1, end, pat);
		if (start < type_point && type_point < end) {
			start = type_point;		// exclude prompt
			grab(start, end, pat);
		}
	}
	if (!*pat)
		strcpy(pat, "*");
	if (!(process_completion_style & 1) || prev_cmd != CMD_PROC_COMPL) {
		cnt = proc_complete_do_list_matches(pat, all_blanks(type_point, start),
											process_completion_style & 2);
		offset = 0;
		if (!cnt)
			error("No matches for %s.", pat);
		set_buf_point(buf, 0);
	}
	if (!(process_completion_style & 1))
		find_buffer_prefix_folding(name_to_bufnum(_MATCH_BUF), pref, !is_unix);
	else {
		buf_delete_final_newlines(buf);
		grab_line_offset(buf, pref, offset, 1);
	}
	if ((process_completion_style & 1) || strlen(pref) > end - start) {
		delete(start, end);
		if (strchr(pref, ' ') && !process_line_inside_quoting())
			stuff("\"");
		stuff(pref);
		if ((process_completion_style & 1)) {
			if (is_path_separator(character(point - 1)))
				delete(point - 1, point);
			this_cmd = CMD_PROC_COMPL;
			return;
		}
		if (cnt == 1 && !is_path_separator(character(point - 1))) {
			if (strchr(pref, ' ') && process_line_inside_quoting())
				stuff("\"");
			stuff(" ");
		}
	} else if (prev_cmd == CMD_PROC_COMPL) {
		proc_complete_show_matches();
		this_cmd = CMD_PROC_COMPL;
	} else if (cnt == 1 && !is_path_separator(character(point - 1))) {
		if (strchr(pref, ' ') && process_line_inside_quoting())
			stuff("\"");
		stuff(" ");
	} else if (!(process_completion_style & 4)
			   && buffer_size(_MATCH_BUF) < screen_cols - 20 - strlen(pat)) {
		save_var bufname = _MATCH_BUF;
		simple_re_replace(1, "\n", " ");
		grab(0, size(), pref);
		say("%d matches for %s: %s", cnt, pat, pref);
		expire_message = 1;
	} else {
		say("%d matches for %s: press again to see matches.", cnt, pat);
		expire_message = 1;
		this_cmd = CMD_PROC_COMPL;
	}
}

show_line()		/* display point for a moment, then restore */
{
	int oldstart = window_start;
	int time;

	if (in_macro())
		return;
	refresh();
	time = (window_start == oldstart) ? near_pause : far_pause;
	delay(time, COND_TRUE_KEY);
	window_start = oldstart;
	build_first = 1;
}

to_line_comment()	/* go to start of comment on this line, or eol */
{			/* return 1 if there was comment, else 0 */
	int end;

	save_var narrow_start = give_begin_line();
	to_end_line();
	end = point;
	if (re_search(RE_REVERSE, get_comment_start()))
		return 1;
	point = end;
	return 0;
}

command indent_for_comment()
{
	int pt = point, scom;

	if (has_arg) {
		if (iter < 0) {
			to_begin_line();
			if (!re_search(RE_REVERSE, get_comment_start())) {
				point = pt;
				error("No previous comment found.");
			}
		}
		if (!re_search(RE_FORWARD, get_comment_start())) {
			point = pt;
			error("No comment found.");
		}
		iter = 0;
		return;
	}
	if (to_line_comment()) {	/* line has comment */
		if (pt > point) {	/* if we were in the comment */
			scom = point;	/* remember where comment was */
			point = pt;	/* stay there after */
			save_spot point = scom;
			to_comment_column();	/* now reindent it */
		} else {
			to_comment_column();	/* reindent it */
			re_search(RE_FORWARD, get_comment_start());
		}			/* go to start of comment */
	} else {
		to_comment_column();	/* make a new comment on this line */
		stuff(get_comment_begin());
		pt = point;
		stuff(get_comment_end());
		point = pt;		/* insert comment text here */
	}
}

command set_comment_column()
{				/* put future comments in this column */
	int newcol, orig = point;

	if (iter < 0) {
		kill_comment();
		return;
	}
	if (has_arg) {	/* copy comment column from last comment */
		to_begin_line();
		re_search(RE_REVERSE, get_comment_start());
		newcol = current_column();
		point = orig;
		iter = 0;
	} else
		newcol = current_column(); /* start comments here */
	comment_column = newcol;
	if (!to_line_comment()) {
		point = orig;
		return;
	}
	to_comment_column();
}

command kill_comment()		/* kill any comment on this line */
{
	int start;

	iter = 0;
	if (!to_line_comment())
		error("This line has no comment.");
	re_search(RE_REVERSE, "[ \t]*");
	start = point;
	re_search(RE_FORWARD, get_comment_pattern());
	do_save_kill(start, point);
}

command comment_region()
{
	char pat1[FNAMELEN], pat2[FNAMELEN];
	char pat1plain[FNAMELEN], pat2plain[FNAMELEN], *s;

	iter = 0;
	save_spot point, mark;
	save_var narrow_start, narrow_end;
	narrow_to_region();
	sprintf(pat1, "%q", get_comment_begin());
	sprintf(pat2, "%q", get_comment_end());
	sprintf(pat1plain, "%q", get_comment_begin());
	while (strlen(pat1plain) > 0 && isspace(pat1plain[strlen(pat1plain) - 1]))
		pat1plain[strlen(pat1plain) - 1] = 0;
	s = get_comment_end();	// We also want to match "#" if comment_begin is "# "
	while (isspace(*s))	// and similar.  So make versions of comment delims
		s++;			// with stripped whitespace.
	sprintf(pat2plain, "%q", s);
	point = 0;
	do {
		if (point >= size() - narrow_end)
			return;		// Skip last line.
		if (has_arg) {
			if (*pat1 && parse_string(1, pat1)
				  || *pat1plain && parse_string(1, pat1plain))
				delete(point, matchend);
			to_end_line();
			if (*pat2 && parse_string(-1, pat2)
				  || *pat2plain && parse_string(-1, pat2plain))
				delete(matchend, point);
		} else {
			stuff(get_comment_begin());
			to_end_line();
			stuff(get_comment_end());
		}
	} while (nl_forward());
}

// See if there's an override function for the current mode.  If there
// is and it returns non-null, use that instead.
char *get_str_by_mode(char *varname)
{
	char override[FNAMELEN], *s;

	sprintf(override, "%s-%s-override", major_mode, varname);
	int i = find_index(override);
	if (i && (s = (*(char * (*)()) i)()))
		return s;
	return varptr(find_index(varname));
}

struct {
	char prefix[100];		// Find words that start with this.
	char found[200];		// The current candidate replacement.
	char last[200];			// Previous replacement we found.
	int len;				// Length of the current candidate replacement.

	// These apply to the destination buffer.
	int savepos;		// Buffer position of the start of the prefix.

	// The rest apply to the source buffer we're searching for matches.
	int window_src;		// Get matches from this window.
	int pullsrc;		// Region we're highlighting.
	spot start;			// Its start and end.
	spot end;

	int tag_position;	// Position in tag buffer we're searching.

	spot ignore_start;	// Ignore matches in this range.
	spot ignore_end;

	int match_current_word;	// At start, match word we're inside.

	int orig;			// Position in search buffer where we started.
} pull;

int pull_find_from_tags(int dir)
{
	char filter_pattern[200];
	int res;

	if (pull.tag_position == -1)
		return 0;
	save_var bufname = "-tags";
	point = pull.tag_position;
	sprintf(filter_pattern, "^([^\t\n]+(::|'))?(%s)(%s)",
			pull.prefix, word_pattern);
	for (;;) {
		res = re_search(dir, filter_pattern);
		pull.tag_position = point;
		*pull.found = 0;
		if (!res)
			return 0;
		grab(find_group(3, 1), find_group(4, 0), pull.found);
		if (!strcmp(pull.found, pull.last))
			continue;
		dismiss_echo_context();
		note("Tag found: %s", pull.found);
		*pull.end = *pull.start;
		expire_message = 1;
		return 1;
	}
}

command pull_word()
{
	int pl_key, old_mod = modified, orig_key = full_key;

	check_modify(bufnum);
	iter = 1;
	alloc_pl();

	for (pl_key = pl_trans(full_key); 1; pl_key = pl_getkey()){
		if (full_key == orig_key && pl_key != 1 && pl_key != -1)
			pl_key = 1;
		switch(pl_key){
			case 1:
			case -1:
				if (pl_search(pl_key) == 0){
					expire_message = 1;
					pl_restore();
					dismiss_echo_context();
					maybe_refresh();
					continue;
				}
				break;
			case 7:
				pl_restore();
				free_pl();
				if (*pull.found)
					user_abort = 0;
				else
					ungot_key = full_key;
				modified = old_mod;
				return;
			case '?':
				help_on_command("pull-word");
				break;
			default:
				free_pl();
				/* Run the key that caused us
				 * to exit.
				 */
				ungot_key = full_key;
				return;
		}
	}
}

int pl_trans(int k)
{
	int r = reg_tab[k];
	if (r <= 0)
		r = reg_tab[generic_key(k)];
	if (r == (short) pull_word)
		return -1;
	else if (r == (short) pull_word_fwd)
		return 1;
	else if (r == (short) help)
		return '?';
	else if (r == (short) abort)
		return 7;
	else
		return 0;
}

int pl_getkey()
{
	for (;;) {
		switch (getkey()) {
			case MOUSE_MOVE:
			case GETFOCUS:
			case LOSEFOCUS:
			case LOST_SELECTION:
			case WIN_MOUSE_LEAVES:
			case WIN_MOUSE_ENTERS:
				break;
			default:
				return pl_trans(full_key);
		}
	}
}

search_read(trystr, prmpt, flags)
char *trystr, *prmpt;
{
	int lastok, goodchars, oldflags, origwin = window_handle;
	int editbuf, pbuf, pwin, editwin, s_code, ostart, cnt;
	int oreg_type, old_mark = -1, search_contin = search_continuation != 0;
	int restrict_region = 0, recall_cnt = 0;
	char tmpbuf[FNAMELEN];
	old_place orig, mstart, okpoint;

	if (_doing_input == DI_SEARCH)
		error_if_input(0);
	save_var _highlight_control;
	if (search_in_region && is_highlight_on()) {
		save_var restrict_search_type;
		restrict_region = make_restriction(flags);
	}
	if (is_highlight_on() && point != mark && ABS(point - mark) < 200
		  && !*trystr && typing_deletes_highlight && typing_deletes_highlight != 3
		  && !in_macro(1) && _len_def_mac < 1 && search_defaults_from == 1) {
		old_mark = mark;	// Use this as the initial search string.
		highlight_off();
	}
	save_var iter, has_arg, this_cmd, prev_cmd;
	save_var _doing_input = DI_SEARCH;
	save_var prompt_orig_buf = bufnum;
	save_var prompt_orig_win = window_handle;
	save_var last_index;
	save_var mark_to_column = -1, mark = point;
	save_var cursor_to_column = -1;
	save_var postpone_asynchronous_notify = 1;
	save_var macro_runs_immediately = 0;	// Make special keys work in macros.
	oreg_type = region_type();
	set_region_type(REGNORM);
	flags &= ~SEARCH_EMPTY | SEARCH_CONTIN;
	remember_old_place(&orig, point);
	remember_old_place(&mstart, point);
	remember_old_place(&okpoint, point);
	ostart = window_start;
	goodchars = lastok = 0;	/* 2 versions of len of last good string */
				/* (only lastok is set by getting default) */

	editbuf = tmp_buf();		/* make buffer for cmd line editing */
	pwin = add_popup(0, echo_line, screen_cols - 1, 1, 0,
					 pbuf = tmp_buf());	/* prompt in window */
	stuff(trystr);
	trystr = malloc(FNAMELEN);
	display_column = 0;
	system_window = 1;	/* don't let user switch to it */
	text_color = color_class echo_area;

	editwin = add_popup(1, echo_line, screen_cols - 1, 1,
						0, editbuf);	/* window on top of prompt window */
	text_color = color_class echo_area;
	_display_class = _echo_disp_class;
	display_column = 0;

	if (old_mark != -1) { // Start out with highlighted region.
		window_handle = origwin;
		buf_xfer(editbuf, old_mark, point);
		window_handle = editwin;
		point = 0;
		if (flags & REGEX)	// Quote special chars.
			string_replace("[[%.!#()*+|?^$<]", "%#0", REGEX);
		point = size();
		highlight_on();		// Mark it so user can delete by typing.
		shift_selecting = 2;
	} else if (search_defaults_from == 2 && !size()) {
		if (flags & UNKNOWN)
			flags = known_flags(flags);
		recall_cnt++;
		stuff(default_search_string(flags));
		point = size();
		highlight_on();		// Mark it so user can delete by typing.
		shift_selecting = 2;
	}

	for (;;) {
		do {
			show_search_prompt(prmpt, flags, origwin,
							   pwin, editwin);
			getkey();
		} while (key == MOUSE_MOVE);
		if (IS_MOUSE_KEY(key))
			if (mouse_y == echo_line)
				flags &= ~INCR;
			else {
				flags |= SONCE;
				ungot_key = key;
				break;
			}
		cnt = 1;
		if (key == CTRL('^'))
			full_getkey("", CTRLIFY_KEY);
		window_handle = editwin;
		s_code = 0;
		oldflags = flags;
		flags &= ~(BAD | FAILING | SONCE);	/* assume ok */
		if (is_search_dir()) {
			window_handle = origwin;
			to_old_place(&okpoint);
			window_handle = editwin;
			if ((key == CTRL('R')) != ((flags & REVERSE) != 0)) {
				flags ^= REVERSE;	/* switch dir */
				if (goodchars)
					flags |= SONCE;
				lastok = size();
			} else if (!size()) {
				if (flags & UNKNOWN)
					flags = known_flags(flags);
get_defaults:
				stuff(default_search_string(flags));
				lastok = size();
			} else if (oldflags & FAILING) {
				if (search_contin) {
					flags |= SEARCH_CONTIN;
					s_code = FHERE;			/* Try other records. */
				} else if (search_wraps) {
					flags |= (flags & SEARCH_WRAPPED) ? SEARCH_OVERWRAPPED
						: SEARCH_WRAPPED;
					window_handle = origwin; /* Wrap to other end of buffer. */
					point = (flags & REVERSE) ? size() : 0;
					window_handle = editwin;
					s_code = FHERE;			/* And search again. */
				} else
					flags |= FAILING; /* no more matches, stop */
			} else
				flags |= SONCE;
			if ((flags & (INCR | SONCE)) && !(flags & FAILING))
				flags |= SONCE, s_code = FHERE;
		} else if (set_flag_modes(&flags)) {
			if ((flags & INCR) && size())
				s_code = FMATCH; /* if modes changed, search again */
		} else if (key == CTRL('O')) {
			flags ^= INCR;
			point = size();
			if (flags & INCR)
				s_code = FMATCH;
		} else if (binding_of(full_key, reg_tab) == (short) help)
			help_on_command(help_on_search(flags));
		else if (IS_ENTER_KEY(key)) {
			if (oldflags & SONCE)
				flags |= SONCE;
			break;
		} else if (binding_of(full_key, reg_tab)
				   == (short) backward_delete_character && (flags & INCR)) {
			run_topkey();
			window_handle = editwin;
			if (!(oldflags & FAILING) || size() <= lastok)
				s_code = FORIG;
			else		/* don't bother searching */
				flags |= FAILING;
		} else if (key == abort_key && (oldflags &
										((flags & INCR) ? FAILING : SONCE))) {
			if (flags & INCR)
				delete(goodchars, size());
		} else if (is_recall_key(flags)) {
			if ((recall_prior_response_options & 2) && !IS_ALT_KEY(key))
				recall_cnt = pick_prev_search_from_end(&flags,
					recall_cnt + ((is_recall_key(flags))));
			else
				flags = pick_prev_search(flags) & ~INCR;
			if (user_abort)
				break;
		} else if (binding_of(full_key, reg_tab) == (short) pull_word
				   || binding_of(full_key, reg_tab) == (short) pull_word_fwd) {
			flags &= ~UNKNOWN;
			window_handle = origwin;
			_highlight_control = 0;
			window_handle = editwin;
			pull_word();
		} else if (user_abort || (flags & INCR) && exit_incr())
			break;
		else {
			flags &= ~UNKNOWN;
			if ((binding_of(full_key, comp_tab) == (short) inp_grab_word
				 || binding_of(full_key, comp_tab) == (short) inp_grab_line)
				  && (IS_ALT_KEY(key) /*|| !(flags & INCR) */ )) {
				if (!(oldflags & FAILING)) {
					if (grab_search_word(flags, origwin,
										 (binding_of(full_key, comp_tab) == (short) inp_grab_line))) {
						window_handle = origwin;	// Move origin
						remember_old_place(&orig, point);
						remember_old_place(&mstart, point);	// to start of word.
						window_handle = editwin;
						s_code = FORIG;
					} else
						s_code = (flags & REVERSE) ? FHERE : FMREGEX;
				}
			} else if (key == '\t')
				normal_character();
			else
				run_topkey();
			window_handle = editwin;
			if (flags & INCR)
				if (flags & (REGEX | WORD))
					s_code = FMREGEX;
				else if (oldflags & FAILING)
					flags |= FAILING;
				else
					s_code = FMADD;
		}
		_doing_input = DI_SEARCH;
		if (!is_window(origwin) || !is_window(editwin)
			  || !is_window(pwin)) {	// user deleted our window
			user_abort = 1;
			goto end;
		}
		user_abort = 0;
		grab_expanding(0, size(), &trystr, 256);
		if (!s_code)
			continue;
		window_handle = origwin;
		if (s_code == FORIG || s_code == FMREGEX)
			to_old_place(&orig);
		else if (s_code >= FMATCH) {
			to_old_place(&mstart);
			if (flags & REVERSE)
				point++;
		}
		_highlight_control = 0;
search_again:
		switch (do_color_searching(flags, trystr)) {
			case DSBAD:
				flags = BAD | (flags & ~INCR);
				flags |= FAILING;
				to_old_place(&okpoint);
				if (in_macro(1))
					goto end;
				else if (!(oldflags & FAILING))
					maybe_ding(bell_on_search);
				break;
			case 0:
				if ((flags & SEARCH_CONTIN) && search_contin
					  && (*search_continuation)(cnt++, flags, trystr))
					goto search_again;
				flags |= FAILING;
				to_old_place(&okpoint);
				if (in_macro(1))
					goto end;
				else if (!(oldflags & FAILING))
					maybe_ding(bell_on_search);
				break;
			case 1:
				remember_old_place(&okpoint, point);
				remember_old_place(&mstart, matchstart);
				mark = matchstart;
				highlight_on();
				if ((flags & SEARCH_WRAPPED)
					  && past_old_place(&orig) == !(flags & REVERSE))
					flags |= SEARCH_OVERWRAPPED;
				window_handle = editwin;
				goodchars = lastok = size();
				break;
			case DSABORT:
				to_old_place(&okpoint);
				window_handle = editwin;
				if ((flags & INCR) && s_code == FMADD)
					delete(size() - 1, size());
				while (ungot_key == -1 && char_avail()
					   && getkey() == abort_key || !IS_TRUE_KEY(key))
					;
				if (key != abort_key && IS_TRUE_KEY(key))
					ungot_key = full_key;
				drop_pending_says();
				break;
		}
	}
	window_handle = origwin;
	if (key == abort_key || user_abort) {
		to_old_place(&orig);
		window_start = ostart;
	} else {
		window_handle = editwin;
		if (size()) {
			_last_was_regex = ((flags & REGEX) != 0);
			*(_last_was_regex ? &_default_regex_flags
							  : &_default_flags) = flags;
			grab_expanding(0, size(), &trystr, 256);
			if (_last_was_regex)
				grab_expanding(0, size(), &_default_regex_search, 256);
			else
				grab_expanding(0, size(), &_default_search, 256);
			save_search_recall(flags, trystr);
		} else
			flags |= SEARCH_EMPTY;
		if ((flags & INCR) && !IS_ENTER_KEY(key) && key != ESC)
			ungot_key = full_key;
	}
end:
	strncpy(tmpbuf, trystr, FNAMELEN - 1);
	free(trystr);
	if (is_window(origwin))
		window_handle = origwin;
	set_region_type(oreg_type);
	remove_window(editwin);
	remove_window(pwin);
	buf_delete(editbuf);
	buf_delete(pbuf);
	if (restrict_region)
		remove_region(restrict_region);
	if ((flags & FAILING) && in_macro(1))
		error("%s: not found", tmpbuf);
	check_abort();
	say("");
	restore_vars();
	maybe_refresh();	/* put cursor back in prev window */
	return flags;
}

exit_incr()	/* does this key exit an incremental search? */
{
	return key != '\t' && key != CTRL('Q') && (key < ' ' || !IS_ASCII_KEY(key))
			&& ((binding_of(full_key, comp_tab) != (short) inp_grab_word
				 && binding_of(full_key, comp_tab) != (short) inp_grab_line)
				|| !IS_ALT_KEY(key)) && !norm_binding();
}

is_recall_key(int flags)
{
	if (IS_ALT_KEY(key) || !(flags & INCR))
		if (binding_of(full_key, comp_tab) == (short) inp_recall)
			return 1;
		else if (binding_of(full_key, comp_tab) == (short) inp_grab_word
				 && (recall_prior_response_options & 2) && !IS_ALT_KEY(key))
			return -1;
	return 0;
}

int grab_search_word(flags, origwin, line)
{
	char text[FNAMELEN];
	int orig, len = size(), ret = 0;

	save_var window_handle = origwin;
			// First time, if we're inside a word, go to its start.
	if (!len && !line && !(flags & FAILING) && parse_string(1, word_pattern)
		  && parse_string(-1, word_pattern)) {
		point -= parse_string(-1, word_pattern);
		ret = 1;
	}
	if (!(flags & REVERSE))
		save_var point;
	else if (len && !(flags & FAILING))
		point = matchstart;
	orig = point;
	if (!(flags & FAILING))
		if (line) {
			if (curchar() == '\n')
				point++;
			to_end_line();
		} else
			re_search(1, word_pattern);		// Get the next word.
	grab(orig, point, text);
	restore_vars();
	if (search_defaults_from && shift_selecting == 2 && is_highlight_on()) {
		delete(0, size());
		shift_selecting = 0;
	}
	point = size();
	bprintf((flags & REGEX) ? "%q" : "%s", text);
	return ret;
}

gen_replace(flags)	/* get input, then do replace */
{
	char *str, with[FNAMELEN];
	int restrict_region = 0, want_restrict = 0;

	save_var mark, mark_to_column, point, cursor_to_column;
	if (replace_in_region && is_highlight_on()) {
		want_restrict = 1;
		save_var search_in_region = 1, typing_deletes_highlight = 3;
	}
	if (_buf_readonly)
		check_modify(bufnum);
	if (has_arg)
		flags |= WORD;
	restore_vars();
	save_var mark, mark_to_column, point, cursor_to_column;
	if (want_restrict) {
		save_var search_in_region = 1, typing_deletes_highlight = 3;
		save_var restrict_search_type;
		restrict_region = make_restriction(flags);
		on_exit remove_region(restrict_region), restrict_region = 0;
	}
	flags = get_replace_text(flags, with);
	restore_vars();
	if (want_restrict) {
		if (restrict_region)
			remove_region(restrict_region);
		save_var restrict_search_type;
		restrict_region = make_restriction(flags);
	}
	str = default_search_string(flags);
	if (flags != -1) {
		if (!want_restrict && point > mark && is_highlight_on() && !(flags & REVERSE))
			exchange_point_and_mark();
		show_replace(str, with, flags);
		string_replace(str, with, flags | INCR);
	}
	if (restrict_region)
		remove_region(restrict_region);
}

string_replace(str, with, flags)
char *str, *with;
{
	spot orig, prev, oldmark;
	int skip = 0, show = 0, ret, evershow = 0, oldstart, i;
	int null_detect = -1, new_null_detect;
	int perform_case_replace = 0, this_case_replace;

	if (_buf_readonly)
		error("Buffer is read-only.");
	if (!*str)
		return 0;
	if (replace_by_case)
		perform_case_replace = want_case_replace(str, with, flags);
	orig = alloc_spot();
	prev = alloc_spot();
	save_var postpone_asynchronous_notify = 1;
	replace_num_found = replace_num_changed = 0;
	for (;;) {
		skip = 0;
		*prev = point;
		oldstart = matchstart;
		ret = do_color_searching(flags, str);
		if (!ret || ret == DSBAD) {
			point = *orig;
			break;
		} else if (ret == DSABORT)
			break;
		new_null_detect = (flags & REVERSE) ? point : (size() - point);
		if (null_detect == new_null_detect) {
			if ((flags & REVERSE) ? (point-- <= 0) : (point++ >= size())) {
				point = *orig;
				break;	// Make sure replacing a nullable pattern like $
			}			// moves forward.
			continue;
		} else
			null_detect = new_null_detect;
		replace_num_found++;
ask:
		if (show) {
			show_replace(str, with, (flags & ~INCR));
			evershow = 1;
			show = 0;
		}
		if (!(flags & QUERY)) goto do_replace;
		if (!get_replace_key(flags, str, skip)) { // show match, then get key
			point = *orig;		// match went away (maybe flags changed)
			break;
		}
		if (key == '?' || key == 'h'
			  || binding_of(full_key, root_keys) == (short) help) {
			help_on_command("query-replace");
			show = 1;
			point = *prev;
			null_detect = -1;
			continue;
		} else if (user_abort || key == abort_key) {
			point = *orig;
			ret = DSABORT;
			goto done;
		}
		switch(tolower(key)) {
			case '^': point = *prev; matchstart = oldstart;
					  null_detect = -1;
			case '\b':
			case GREYBACK:
			case 'n': skip = 1;	/* go to the next match */
			case ' ':
			case 'y':		/* do the replace */
			case '!':
			case ',':
			case '.': break;
			case CTRL('R'):
				say("");
			null_detect = -1;
			oldmark = alloc_spot();
			*oldmark = mark;	/* save mark */
			mark = matchstart;
					/* set temporary mark before text */
			recursive_edit();
			if (spot_to_buffer(oldmark) < 0) {
				ret = DSABORT;	// user deleted buf
				goto done;
			}
			mark = *oldmark;	/* restore mark */
			free_spot(oldmark);
			show = 1;
			continue;
			default:	if (set_flag_modes(&flags)) {
							show = 1;
							goto ask;
						}
						if (!IS_ENTER_KEY(key))
							ungot_key = key;
			case ESC:	ret = 1; goto done;
		}
do_replace:
		if (!skip) {
			save_spot matchstart;
			if (perform_case_replace)
				this_case_replace = (perform_case_replace == 1)
									? check_for_case_replace(matchstart, point)
									: check_alt_case_replace(matchstart, point);
			if (flags & REGEX)
				fromorig(with);
			else {
				delete(matchstart, point);
				stuff(with);
			}
			if (perform_case_replace == 1)
				switch (this_case_replace) {
					case 1: do_uppercase_region(matchstart, point); break;
					case 2: do_capitalize_region(matchstart, point); break;
				}
			else if (perform_case_replace == 2) {
				i = MIN(matchstart, point);
				switch (this_case_replace) {
					case 1: do_uppercase_region(matchstart, point); break;
					case 2: do_lowercase_region(matchstart, point); break;
					case 3: do_uppercase_region(i, i + 1); break;
					case 4: do_lowercase_region(i, i + 1); break;
				}
			}
			replace_num_changed++;
			restore_vars();
		}
		if (flags & QUERY)
			switch (key) {
				case ',': case '^':	skip = 1; goto ask;
				case '.':		ret = 1; goto done;
				case '!':		flags &= ~QUERY; show = 1;
			}
	}
done:
	if (ret == DSABORT)
		say("Replacement canceled");
	else if (ret == DSBAD)
		show_replace(str, with, (flags & ~INCR) | BAD);
	else if (!in_macro(1))
		if (evershow || (flags & INCR))
			say("Replaced %d of %d.", replace_num_changed,
				  replace_num_found);
		else if (evershow)
			say("");
	free_spot(orig);
	free_spot(prev);
	return ret;
}

session_save_misc(int b)
{
	buf_printf(b, "%d %d %d %d\n",
			   _default_flags, _default_regex_flags, _last_was_regex,
			   dired_sorts_files);
	buf_printf(b, "%s" SES_DELIM "%s" SES_DELIM,
			   _default_search ? _default_search : "",
			   _default_regex_search ? _default_regex_search : "");
	buf_printf(b, "%s" SES_DELIM "%s" SES_DELIM,
			   _default_replace, _default_regex_replace);
	buf_printf(b, "%s" SES_DELIM "%s" SES_DELIM,
			   push_cmd, _region_file);
	buf_printf(b, "%s" SES_DELIM "\n", _grep_files);
}

session_save_buffer(int b, char *sdir)
{
	char file[FNAMELEN];
	int transl = GET_LINE_TRANSLATE(translation_type);
	int uni_compat = 0;

	if (translation_type & FORCED_TRANS)
		transl = GET_LINE_TRANSLATE(translation_type);
	if (has_feature & FEAT_UNICODE) {
		char *s = encoding_to_name(GET_ENCODING(translation_type));
		if (s)
			uni_compat = strstr(s, "UTF-16") != 0;
		if (uni_compat)
			transl = FILETYPE_BINARY;
	}
	buf_printf(b, "%s\n", bufname);
	if (is_dired_buf()) {
		insert_char_in_buf(b, 0);	// Hide file pattern from old versions.
		buf_printf(b, "%s\n", get_dired_buffer_pattern());
	} else {
		relative(filename, file, sdir);
		buf_printf(b, "%s\n", file);
	}
	buf_printf(b, "%d %d %d %d %d %d",
			   point, mark, transl,
			   is_dired_buf(), dired_buffer_pattern_wildcards,
			   dired_sorts_files);
	if ((has_feature & FEAT_UNICODE) && (translation_type & FORCED_TRANS))
		buf_printf(b, " %d @%s", GET_LINE_TRANSLATE(translation_type),
				   encoding_to_name(GET_ENCODING(translation_type)));
	buf_printf(b, "\n");
	buf_printf(b, "%d %d %d %d\n", over_mode, fill_mode,
			   indent_with_tabs, margin_right);
	transl = FILETYPE_AUTO;
	if (uni_compat || (translation_type & FORCED_TRANS))
		transl = GET_LINE_TRANSLATE(translation_type);
	// This flavor of the translation type is used by non-Unicode/
	// pre-v12 versions only.
	buf_printf(b, "%d %d %d %d %d %d %d %d %d\n", tab_size,
			   case_fold, 0, want_warn,
			   _buf_readonly, buf_accessed,
			   (file_io_converter == oem_file_converter),
			   uni_compat,
			   transl);
}

int session_restore_one_file(int b, char *file, char *sdir, int *readerrs)
{
	int i, trans = FILETYPE_AUTO;
	int num1[10], num1cnt, num2[10], num2cnt, num3[10], num3cnt;
	char firstline[150];

	grab_line(b, firstline);
	if ((num1cnt = break_into_numbers(firstline, num1)) < 3)
		return 2;
	if ((num2cnt = grab_numbers(b, num2)) < 4)
		return 3;
	if ((num3cnt = grab_numbers(b, num3)) < 5)
		return 4;
	if (session_restore_files)
		filename = file;
	if (num1cnt >= 4 && num1[3] && session_restore_files) {	// restore dired buf
		filename = "";
		if (num1cnt >= 5)
			save_var file_pattern_wildcards = num1[4];
		save_var show_status = 0;
		if (num1cnt >= 6)
			dired_sorts_files = num1[5];
		create_dired_listing(file, 0);
		if (is_remote_file(get_dired_buffer_pattern()))
			dired_orig_position = -1;
	} else if (*file && session_restore_files) {
		if (!is_remote_file(file))
			absolute(file, sdir);
		if (has_feature & FEAT_UNICODE) {
			if (num1cnt >= 7) {		 // If explicit trans, force.
				char *p;
				if (p = strchr(firstline, '@')) {
					i = encoding_from_name(p + 1);
					if (i)
						trans = MAKE_TRANSLATE(i, num1[6], 1);
				}
			}
		} else {
			if (num3cnt > 8 && num3[7])
				trans = num1[2];	// If explicit trans, or common unicode, force.
		}
		i = do_file_read(file, trans);
		force_save_as = 0;
		if (user_abort || i == EREADABORT)
			return 0;
		if (!i || i == 2) {
			if (!i)
				set_readonly_from_file();
			window_bufnum = bufnum;		// make suffix- kbd macros work
			call_mode(filename);
			make_mode();
		} else if (!*readerrs++) {
			file_error(i, file, "read error");
			maybe_ding(bell_on_read_error);
		} else
			show_text(0, -1, "%d files could not be read", *readerrs);
	}
	if (session_restore_files) {
		if (bufnum != b)	/* same tmp name as session buffer */
			point = num1[0];
		mark = num1[1];
		if (is_dired_buf())
			move_to_column(goal_column);
	}
	if (session_restore_files) {
		session_restore_one_file_1(num2, num2cnt);
		session_restore_one_file_2(num3, num3cnt);
	}
	return -1;
}

int session_restore_misc(int b)
{
	int nums[10], i;
	char file[FNAMELEN];

	i = grab_numbers(b, nums);
	if (i < 3)
		return 1;
	_default_flags = nums[0];
	_default_regex_flags = nums[1];
	_last_was_regex = nums[2];
	if (i >= 4)
		dired_sorts_files = nums[3];
	grab_string_expanding(b, &_default_search, SES_DELIM, FNAMELEN);
	grab_string_expanding(b, &_default_regex_search, SES_DELIM, FNAMELEN);
	grab_string(b, _default_replace, SES_DELIM);
	grab_string(b, _default_regex_replace, SES_DELIM);
	grab_string(b, push_cmd, SES_DELIM);
	grab_string(b, _region_file, SES_DELIM);
	grab_string(b, _grep_files, SES_DELIM);
	restore_session_hook(b);
	restore_recall(b);
	grab_line(b, file);		// skip \n after recall buffer
	grab_line(b, file);
	if (!strncmp(file, "CWD: ", 5)) {
		switch (session_restore_directory) {
			default:	if (directory_flags & DIR_CWD_RESTORE)
			case 1:			chdir(file + 5);
			case 0:		break;
		}
		grab_line(b, file);
		if (!strncmp(file, "OPENCWD: ", 9))
			common_open_curdir = file + 9;
	}
	return 0;
}

do_setting(int i, char *tmp, int def)
{
	save_var use_default = def;
	switch (vartype(i)) {
		case TYPE_CHAR:
		case TYPE_SHORT:
		case TYPE_INT:
			if (*tmp)
				set_num_var(i, numtoi(tmp));
			break;
		case TYPE_CARRAY:
			strncpy(varptr(i), tmp, ptrlen(varptr(i)) - 1);
			int j = ptrlen(varptr(i)) - strlen(tmp);
			if (j > 0)
				set_chars(varptr(i) + strlen(tmp), 0, j);
			else
				say("%s truncated: %d characters, room for %d.",
					  name_name(i), strlen(tmp), ptrlen(varptr(i)) - 1);
			break;
		case TYPE_CPTR:
			set_str_var(i, strsave(tmp)); break;
	}
}

set_var_val(i, tmp, def)
char *tmp;
{
	char when_setting[FNAMELEN];
	int old_record_cust = record_customizations;

	if (strstr(" selected-color-scheme -our-mono-scheme -our-color-scheme "
			   "-our-gui-scheme -our-unixconsole-scheme ", name_name(i))
		  && isalpha(*tmp)) {
		if (name_type(find_index(tmp)) != NT_COLSCHEME)
			error("%s is not a color scheme", tmp);
		sprintf(tmp, "%d", find_index(tmp));
	}
	do_setting(i, tmp, def);
	sprintf(when_setting, "when-setting-%s", name_name(i));
	try_calling(when_setting);
	if (def) {
		save_customization(i);
		// Also record turning it off.
		if (old_record_cust && !record_customizations && !noinit_flag
			  && !now_parsing_cmds && i == find_index("record-customizations"))
			list_one(prep_customizations(), i, 0);
	}
	if (def & 2)
		switch (name_type(i)) {
			case NT_WINVAR:
				save_var window_handle;
				int orig = window_number;
				do {
					if (!system_window || (def & 1))
						do_setting(i, tmp, 0);
				} while (++window_number != orig);
				break;
			case NT_BUFVAR:
				save_var bufnum;
				int buf = buf_list(0, 2);
				do {
					bufnum = buf;
					if (*bufname != '-' || (def & 1))
						do_setting(i, tmp, 0);
				} while (buf = buf_list(-1, 1));
				break;
		}
}

do_set_variable(i)
{
	char msg[80], res[2048], *tname, *p, tmp[FNAMELEN];
	int def;

	tname = variable_type_name(vartype(i));
	if (!(p = get_var_val(i, tmp, 0)))
		error("Can't interpret type of %s", name_name(i));
	sprintf(msg, "Set %s variable %s to", tname, name_name(i));
	if (run_by_mouse && (has_feature & FEAT_GUI_DIALOGS)) {
		get_strpopup(res, msg, p, name_name(i));
		def = 1;			// set both
	} else {
		strcat(msg, ": ");
		get_strnone(res, msg, p);
		def = ask_default("Set", name_name(i), 1);
	}
	if (def)
		set_var_val(i, res, 1);		/* set default */
	if (def >= 0)
		set_var_val(i, res, 0);		/* set current */
	do_show_variable(name_name(i), def);
	full_redraw = 1;	/* might have set a display-type variable */
}

void eval_template(char *expr, char *src, int in_buf)
{
	char quoted[300], *p, *q;

	save_var bufnum = zap(src);
	for (p = expr, q = quoted; *q = *p; p++, q++)
		if (*p == '"') {
			*q++ = '\\';
			*q = '"';
		}
	ev_header();
	bprintf("ev_template(){\n");
	bprintf("char *strval;\n");
	bprintf("  if (sizeof(%s) <= sizeof(int))\n", expr);
	if (in_buf)
		bprintf("    bprintf(\"%%d\", (%s));\n", expr);
	else
		bprintf("    say(\"%%s == %%d\", \"%s\", (%s));\n", quoted, expr);
	bprintf("  else {\n");
	bprintf("    strval = (char *)(%s);\n", expr);
	bprintf("    if (!strval) strval = \"<null>\";\n");
	if (in_buf)
		bprintf("    bprintf(\"%%s\", strval);\n");
	else
		bprintf("    say(\"%%s == %%s\", \"%s\", strval);\n", quoted);
	bprintf("}\n");
	bprintf("}\n");
}

sort_region_cmd(rev)	// sort region of any type in place,
{			// use num arg as column
	if (region_type() == REGRECT) {
		int width, orig = bufnum, b = tmp_buf();
		width = extract_rectangle(b, 0);
		save_var bufnum = b;
		mark = 0;
		sort_region_with_settings(0, size(), has_arg ? iter : 0,
								  rev, orig);
		xfer_rectangle(orig, width, 1);
		buf_delete(b);
	} else {
		save_spot point, mark;
		fix_region();		// handle other types of regions
		do_sort_region(point, mark, has_arg ? iter : 0, rev);
		if (window_start == point && mark < point)	// If region is now just
			window_start = mark;	// above window, reposition.
	}
	iter = 0;
}

sort_region_with_settings(from, to, col, rev, buf)
{	/* sort a region in place, using case-folding from buf */
	char *temp1 = temp_buf(), *temp2 = temp_buf();
	int this_tab_size, this_case_fold, this_sort_case_fold;
	int res = ABORT_JUMP, orig = bufnum, sz;
	jmp_buf this_level;

	bufnum = buf;
	this_case_fold = case_fold;
	this_sort_case_fold = sort_case_fold;
	this_tab_size = tab_size;

	bufnum = orig;
	save_var top_level = &this_level;
	if (!setjmp(top_level)) {
		point = from;
		if (curchar() == '\n')
			point++;
		else
			to_begin_line();
		from = point;
		point = to;
		if (curchar() == '\n')
			point++;
		else
			to_begin_line();
		to = point;
		xfer(temp1, from, to);
		bufname = temp1;
		case_fold = this_case_fold;
		sort_case_fold = this_sort_case_fold;
		tab_size = this_tab_size;
		res = do_buffer_sort(temp2, col, rev);
		if (!res) {
			bufnum = orig;
			point = MIN(from, to);
			sz = buffer_size(temp2);
			grab_buffer(name_to_bufnum(temp2));
			delete(fix_narrowing(from) + sz, fix_narrowing(to) + sz);
		}
	}
	bufnum = orig;
	delete_buffer(temp1);
	delete_buffer(temp2);
	return res;
}

jmp_buf jbuf;

command import_colors()
{
	char name[FNAMELEN];
	int tmp, i;

	get_file(name, "Load changes from file", "changes");
	tmp = tmp_buf();
	to_buffer_num(tmp);
	if (read_file(name, FILETYPE_AUTO))
		quick_abort();

	strcpy(name, "mycolors.e");
	absolute(name);
	if (warn_existing_file(name) > 0)
		return;
	do_find(name, FILETYPE_AUTO);
	buf_zap(bufnum);
	grab_buffer(tmp);
	point = 0;
	do_drop_matching_lines(REGEX, "^&.* color class for ", 0);
	do_drop_matching_lines(0, "[0x4000000 on 0x0]", 1);	// Unknown color.
	while (re_search(1, ": *([0-9a-fx]+)$")) {
		grab(find_group(1, 1), find_group(1, 0), name);
		point = find_group(1, 1);		// Change old-style color numbers
		delete(point, find_group(1, 0));	// to names.
		i = numtoi(name);
		give_color_combo(name, i);
		stuff(name);
	}
	point = 0;
	for (i = 0; i < 10; i++) {	// Assume class names have no more than 10 -'s.
		noteput("Working... %d%%", 20 + i * 5);
		string_replace("(color class for .*)-", "#1_", REGEX);	// Conv to _'s.
		if (!replace_num_changed)
			break;
	}
	for (i = 0; i < 10; i++) {
		string_replace("(=<word>*)-", "#1_", REGEX);	// Conv to _'s.
		if (!replace_num_changed)
			break;
	}

	noteput("Working... 70%%");
	// Make into valid EEL definitions.
	string_replace("^&(.*) color class for (.*): (.*)",
				   "color_scheme \"#1\" color_class #2 #3;", REGEX);
	noteput("Working... 80%%");
	// Convert pre-v8 color names to EEL names.
	string_replace(" (blue|green|cyan|red|magenta)", " dark_#1", REGEX);
	string_replace(" white", " grey", REGEX);
	string_replace(" brightblack", " dark_grey", REGEX);
	noteput("Working... 90%%");
	string_replace(" bright(blue|green|cyan|red|magenta|white)", " #1", REGEX);

	string_replace(" *= *", " = color_class ", REGEX);
	string_replace("%[(.* on .*)%]", "#1", REGEX);
	string_replace(" on 0x1000000;", ";", REGEX);

	stuff("// Automatically generated by import-colors.\n");
	stuff("// Save this file, then press Alt-F3 to import colors.\n\n");
	stuff("#include \"eel.h\"\n\n");
	point = 0;
	buf_delete(tmp);
	add_color_class_declarations();
	say("Done.");
}

add_color_class_declarations()
{
	int tmp;

	tmp = tmp_buf();
	buf_xfer(tmp, 0, size());
	save_var bufnum = tmp, show_status = 0;
	point = 0;
	do_drop_matching_lines(0, "= color_class", 0);
	point = 0;
	string_replace(".* = ", "", REGEX);
	point = 0;
	do_sort_region(0, size(), 0, 0);
	point = 0;
	do_uniq(1, 1, 0);
	restore_vars();
	point = 0;
	re_search(1, "(#include .*\n)+\n*");
	to_begin_line();
	grab_buffer(tmp);
	stuff("\n");
	buf_delete(tmp);
}

parse_cmds()	/* load cmds in current buf, return 1 if error occurred */
{
	point = 0;
	save_var _errors_in_cmd_file = 0;
	save_var now_parsing_cmds = 1;
	if (setjmp(&jbuf))			/* errors come here */
		if (_abort_load_on_error)
			return 1;
	for (;;) {
		re_search(1, "[ \n\t]*");	/* skip whitespace */
		if (point >= size() - narrow_end)
			break;
		if (_abort_load_on_error != 2)
			noteput("Loading: %d%%...", muldiv(point, 100, size()));
		switch (curchar()) {
			case ';':		/* found comment */
			case '#': nl_forward(); break;
			case '(': paren_cmds(); break;
			case '+': load_var(); break;
			case '~': load_binding(); break;
			case '&': load_color(); break;
			default:  load_other(); break;
		}
	}
	copy_bindings();
	int i = *get_color_scheme_variable();
	if (name_type(i) == NT_COLSCHEME)
		selected_color_scheme = i;
	if (!_abort_load_on_error
		  || !_errors_in_cmd_file && _abort_load_on_error != 2)
		say("%d error%s detected.", _errors_in_cmd_file,
			  _errors_in_cmd_file == 1 ? "" : "s");
	else if (_abort_load_on_error != 2)
			say("%d error%s detected, see buffer %s.", _errors_in_cmd_file,
				  _errors_in_cmd_file == 1 ? "" : "s", MSG_BUFFER);
	return _errors_in_cmd_file != 0;
}

load_var()	/* variable definition */
{
	int i, new_type, vtype, hasnum;
	char name[80], junk[80], num[20], *p, *tmp;

	point++;			/* skip initial + */
	parse_string(1, "[^ \t\n:]+", name);
	i = find_index(name);
	new_type = parse_string(1, " buffer", junk)
			   ? NT_BUFVAR : parse_string(1, " window", junk)
			   ? NT_WINVAR : NT_VAR;
	if (parse_string(1, " (char ptr )?variable: %<complex>", junk)) {
		if (i <= 0 || name_type(i) != new_type)
			init_warning("Variable %s too complex to define.", name);
		nl_forward();
		return;
	} else if (parse_string(1, " char array ", junk))
		vtype = TYPE_CARRAY;
	else if (parse_string(1, " int ", junk))
		vtype = TYPE_INT;
	else if (parse_string(1, " short ", junk))
		vtype = TYPE_SHORT;
	else if (parse_string(1, " char ", junk))
		vtype = TYPE_CHAR;
	else
bad:		init_error("Bad definition for variable %s.", name);
	if (!parse_string(1, "variable: ", junk))
		goto bad;
	hasnum = parse_string(1, "[-0-9][a-zA-Z0-9]*", num);
	if (vtype == TYPE_CARRAY) {
		if (i <= 0 && hasnum) {
			i = new_variable(name, new_type, TYPE_CARRAY, numtoi(num));
			set_name_user(i, 1);		// Make it a user variable.
		}
		if (!parse_string(1, " *<dquote>", junk)
			  || name_type(i) != new_type
			  || vartype(i) != TYPE_CARRAY)
			goto bad;
		use_default = 1;
		tmp = p = varptr(i);
		use_default = 0;
		for (; curchar() != '"'; point++) {
			if (curchar() != '\\') {
				*p++ = curchar();
				continue;
			}
			point++;
			switch (curchar()) {
				case 't': *p++ = '\t'; break;
				case 'r': *p++ = '\r'; break;
				case 'n': *p++ = '\n'; break;
				case 'z': *p++ = 0; break;
				default: *p++ = curchar();
			}
		}
		while (p - tmp < ptrlen(tmp))
			*p++ = 0;
		if (!parse_string(1, "<dquote>$", junk))
			goto bad;
	} else if (vtype) {
		if (i <= 0) {
			i = new_variable(name, new_type, vtype, 1);
			set_name_user(i, 1);		// Make it a user variable.
		}
		if (!hasnum && parse_string(1, ".*", junk))
			sprintf(num, "%d", find_index(junk));
		set_var_val(i, num, 2);
	}
}

load_color()
{
	int i, j, fore, back;
	char scheme[80], class[80], num[50];

	point++;			/* skip initial & */
	parse_string(1, "[^ \t\n:]+", scheme);
	i = find_index(scheme);
	if (!parse_string(1, " color class for ", class)
		  || !parse_string(1, "[^ \t\n:]+", class)
		  || !parse_string(1, ": ", num)
		  || !parse_string(1, "[[=]?[-a-zA-Z0-9]+", num)) {
		init_warning("Bad color definition for %s.", scheme);
		nl_forward();
		return;
	}
	if (i <= 0)
		i = new_variable(scheme, NT_COLSCHEME, 0);
	for (j = 0; strcmp(class, name_color_class(j)); )
		if (++j >= number_of_color_classes()) {
			init_warning("Undefined color class %s.", class);
			nl_forward();
			return;
		}
	if (*num == '=') {		// copy definition
		for (fore = 0; strcmp(num + 1, name_color_class(fore)); )
			if (++fore >= number_of_color_classes()) {
				init_warning("Undefined color class %s.", num + 1);
				nl_forward();
				return;
			}
		fore |= ECOLOR_COPY;
		back = 0;
	} else if (*num == '[') {		// rgb-style definition
		fore = numtoi(num + 1);
		if (!parse_string(1, " on ", num)
			  || !parse_string(1, "[0-9a-fx]+]", num)) {
			init_warning("Bad rgb color definition for %s.", scheme);
			nl_forward();
			return;
		}
		back = numtoi(num);
	} else {
		fore = find_color_code(num);
		if (fore < 0) {
			init_warning("Bad color name %s.", num);
			nl_forward();
			return;
		}
		if (parse_string(1, " on ", num)
			  && parse_string(1, "[-a-zA-Z0-9]+", num)) {
			back = find_color_code(num);
			if (back < 0) {
				init_warning("Bad color name %s.", num);
				nl_forward();
				return;
			}
		}
		fore = attr_to_rgb(fore);
		back = attr_to_rgb(back);
	}
	set_color_pair(j, fore, back, i);
}

list_one_color_setting(int buf, int class, int raw)
{
	char colorpair[100];
	int fg, bg;

	fg = get_foreground_color(class, raw);
	if ((fg & ETRANSPARENT) && !raw)
		fg = get_foreground_color(0);
	bg = get_background_color(class, raw);
	if ((bg & ETRANSPARENT) && !raw)
		bg = get_background_color(0);
	if (fg & ECOLOR_COPY)
		sprintf(colorpair, "=%s", name_color_class(COLOR_STRIP_ATTR(fg)));
	else if (!(fg & ~0xffffff) && fg == attr_to_rgb(rgb_to_attr(fg))
			 && !(bg & ~0xffffff) && bg == attr_to_rgb(rgb_to_attr(bg)))
		give_color_combo(colorpair, alter_color(class, -1));
	else
		sprintf(colorpair, "[0x%x on 0x%x]", fg, bg);
	buf_printf(buf, "&%s color class for %s: %s\n",
			   name_name(selected_color_scheme),
			   name_color_class(class), colorpair);
}

list_table(int i)
{
	char tabname[FNAMELEN];
	int j, onekey[2], cnt;
	short *k;

	strcpy(tabname, name_name(i));
	bprintf("%s: table\n", tabname);
	k = index_table(i);
	onekey[0] = 2;
	for (j = 0; j < MAXKEYS; j += cnt) {
		cnt = get_range(k, j, k[j]);
		if (k[j] > 0) {
			bprintf("~%s ", tabname);
			onekey[1] = j;
			stuff_macro(onekey, 1);
			if (cnt > 1) {
				bprintf(" ");
				onekey[1] = j + cnt - 1;
				stuff_macro(onekey, 1);
			}
			if (name_name(k[j]))
				bprintf(": %s\n", name_name(k[j]));
			else
				bprintf(": %d\n", k[j]);
		}
	}
}

// Append to buffer b a description of name table entry i.  Flags:
// 1=ignore non-user variables, 2=include color schemes, 4=use final
// colors in schemes, ignoring flags to build colors from others.
list_one(int b, int i, int flags)
{
	int j, last, val, type;
	char *p, *nm, *tname;

	save_var bufnum = b;
	save_var point = size();
	save_var use_default;
	nm = name_name(i);
	switch (name_type(i)) {
		case NT_COMMAND:
		case NT_AUTOLOAD:
			bprintf("%s: command\n", nm); break;
		case NT_SUBR:
		case NT_AUTOSUBR:
			bprintf("%s: subroutine\n", nm); break;
		case NT_MACRO:
			bprintf("%s: macro ", nm);
			stuff_macro(get_macro(i), 1);
			insert('\n');
			break;
		case NT_TABLE:
			list_table(i);
			break;
		case NT_WINVAR:
		case NT_BUFVAR: case NT_VAR:
			use_default = 1;
			p = varptr(i);
			type = 0;
			if ((flags & 1) && !name_user(i) && (strncmp(nm, "-our-", 5)
				|| !strstr(nm, "-scheme")))
				break;
			switch (vartype(i)) {
				case TYPE_CHAR:	val = *(char *) p;
								tname = "char ";
								break;
				case TYPE_SHORT: val = *(short *) p;
								 tname = "short ";
								 break;
				case TYPE_INT:	val = *(int *) p;
								tname = "int ";
								break;
				case TYPE_CPTR:	tname = "char ptr ";
								type = 2; break;
				case TYPE_CARRAY: tname = "char array ";
								  type = 1; break;
				default:	tname = ""; type = 2; break;
			}
			bprintf("+%s %s%svariable", nm,
					(name_type(i) == NT_WINVAR)
					? "window " :
					(name_type(i) == NT_BUFVAR)
					? "buffer " : "", tname);
			if (type == 0) {
				if (val && !strncmp(nm, "-our-", 5)
					  && parse_string(-1, "^%+-our-(color|mono|gui"
									  "|unixconsole)-scheme .*"))
					bprintf(": %s\n", name_name(val));
				else
					bprintf(": %d\n", val);
			} else if (type == 2)
				stuff(": <complex>\n");
			else {
				last = ptrlen(p);
				bprintf(": %d \"", last);
				while (last > 0 && !p[last - 1])
					last--;
				for (j = 0; j < last; j++)
					switch (p[j]) {
						case '\t':	stuff("\\t"); break;
						case '\r':	stuff("\\r"); break;
						case '\n':	stuff("\\n"); break;
						case 0:		stuff("\\z"); break;
						case '\\':
						case '"':	insert('\\');
						default:	insert(p[j]); break;
					}
				stuff("\"\n");
			}
			use_default = 0;
			break;
		case NT_COLSCHEME:
			if (flags & 2)
				list_color_scheme(i, !(flags & 4));
			break;
	}
}

int tag_lister();

tag_searcher_c(char *tag)
{
	char classname[FNAMELEN], fulltag[FNAMELEN], pr[FNAMELEN];
	int err, cnt = 0, selected = 0, try = 1;

	if (parse_string(-1, "-%>|%.[ \t\n]"))	// Ignore func's class.
		*classname = 0;
	else
		find_c_func_info(NULL, classname, NULL, 0);
again:
	strcpy(fulltag, tag);
	if (!strchr(tag, ':')) {		// Class not known yet.
withclass:
		if (*classname && strcmp(classname, tag)) {		// See if current class
			sprintf(fulltag, "%s::%s", classname, tag);	// has such a member.
			if (!(err = go_tag(fulltag, 1)))
				return err;
			bufnum = window_bufnum;
		}			// Is there more than one function by that name in any class?
		*fulltag = 0;
		bufnum = window_bufnum;
		if (tag_match(tag, STARTMATCH | EXACTONLY)) {
			do {
				cnt++;
			} while (tag_match(tag, EXACTONLY));
			if (cnt != 1) {
				sprintf(pr, "%d matches for %s: ", cnt, tag);
				save_var menu_width = screen_cols - _view_left - _view_right;
				save_var completion_lister = tag_lister;
				save_var insert_default_response = 2, tag_list_exact_only = 1;
				ungot_key = '?';
				comp_read(fulltag, pr, tag_match, (tag_case_sensitive ? 0 : COMP_FOLD)
						  | MUST_MATCH | NONE_OK | CAUTIOUS, tag);
				selected = 1;
			} else
				strcpy(fulltag, tag_match(tag, STARTMATCH | EXACTONLY));
		} else
			strcpy(fulltag, tag);
		err = go_tag(fulltag, 1);
	} else {
		err = go_tag(fulltag, 1);
		if (err && *classname)
			goto withclass;
	}
	strcpy(_last_tag, tag);
	bufnum = window_bufnum;
	if (err && !selected && find_tagger(0)) {
		// Try retagging current file before giving up.
		sprintf(pr, "Tag %s has moved, retag file %s? ",
				fulltag, filename);
		if (!tag_ask_before_retagging || try > 1
			  || ask_yn("Retag file", pr, "Yes", "No", 1)) {
			if (try < 2)
				do_retag_file(filename);
			err = go_tag(fulltag, 1);
			if (try < 2)
				do_save_tags();
			if (err && try++ < 2)
				goto again;
			if (!err)
				say("Found %s in this file.", *tag ? tag : "tag");
		}
	}
	if (!err && !selected && strcmp(tag, fulltag))
		say("Searched for %s, found %s", tag, fulltag);
	report_tag_error(err);
}

tag_make_loose_pattern(char *tag, char *pat)
{
	*pat++ = '(';
	for (; *tag; tag++) {
		sprintf(pat, "%q", tag);
		if (*pat == '%')
			pat++;
		pat++;
		if (tag[1] &&  !(isalnum(tag[0]) || tag[0] == '_')
			  && (isalnum(tag[1]) || tag[1] == '_')) {
			strcpy(pat, ")?(");
			pat += 3;
		}
	}
	strcpy(pat, ")[^a-zA-Z0-9_]");
}

do_go_tag(s, n_fwd)
char *s;
{
	char pat[2048], tdir[FNAMELEN], *file, *p;
	int pos, oldbuf = bufnum, i;

	i = get_mode_based_index("tag_to_next_%s");
	bufnum = _tags_buffer;
	save_var case_fold = !tag_case_sensitive;
	if (*s) {
		if (p = index(s, '\t')) {
			strcpy(_last_tag_file, p + 1);
			*p = 0;		/* remove file name from s */
			if (p = index(_last_tag_file, '\t'))
				*p = 0;
		} else
			*_last_tag_file = 0;
		strcpy(_last_tag, s);
		if (!*_last_tag)	/* no last tag */
			return 0;
		if (*_last_tag_file && p && !strchr(p + 1, '\t'))
			sprintf(pat, "^%q\t%q\t%q", _last_tag, _last_tag_file, p + 1);
		else if (*_last_tag_file)
			sprintf(pat, "^%q\t%q\t",
					_last_tag, _last_tag_file);
		else
			sprintf(pat, "^%q\t", _last_tag);
		point = 0;
		if (!re_search(1, pat))		/* look for 1st */
			return 1;
	} else {		/* find next matching tag */
		pos = point;
		sprintf(pat, "^%q\t", _last_tag);
		switch (name_type(i)) {		// Call mode-specific version.
			case NT_SUBR: case NT_COMMAND:
			case NT_AUTOSUBR: case NT_AUTOLOAD:
				if ((*(int (*)())i)(pat, n_fwd))
					return 1;
				break;
			default:
				if (!tag_search_by_offset(pat, n_fwd))
					return 1;
		}
	}
	grab(give_begin_line(), give_end_line(), pat);	/* tag, file, pos */
	file = index(pat, '\t');
	*file++ = 0;

	p = index(file, '\t');		/* find start of pos */
	pos = strtoi(p + 1, 10);
	*p = 0;
	p = index(p + 1, '\t');		/* if p, p+1 has line's text */
	strcpy(_last_tag_file, file);
	put_directory(tdir);
	bufnum = oldbuf;
	absolute(_last_tag_file, tdir);
	locate_window("", _last_tag_file);
	find_it(_last_tag_file, default_translation_type);
	if (p)
		sprintf(tdir, "%q[^a-zA-Z0-9_]", _last_tag);
	else			// Might not include class info.
		tag_make_loose_pattern(_last_tag, tdir);
	point = pos;
	if (!parse_string(1, tdir, NULL))
		if (!p || !prox_line_search(p + 1))	/* search by text */
			return 2;
	to_begin_line();
	window_start = prev_screen_line(show_tag_line);
	return 0;
}

good_c_tag()	// tag name at point if a valid c tag
{		// return 0 to go back to orig point before searching again
	char func[TAGLEN], tmp[TAGLEN];
	int start = point, i = 1, skip = 0;

	if (!parse_string(1, "[a-zA-Z0-9_]+", func))
		return 0;
	else if (!strcmp(func, "struct") || !strcmp(func, "union"))
		return tag_c_struct(0);		// maybe struct tag
	else if (!strcmp(func, "class"))
		return tag_c_struct(1);		// or similar class def
	else if (!strcmp(func, "interface") && (c_extra_keywords & JAVA_KEYWORDS))
		return tag_c_struct(1);		// for java: like a class
	else if (!strcmp(func, "enum"))
		return tag_c_struct(2);		// or similar enum def
	else if (!strcmp(func, "namespace"))
		return tag_c_struct(3);		// or namespace def
	else if (!strcmp(func, "operator") && parse_string(1, CPP_OPLIST)) {
		grab(matchstart, matchend, func + strlen(func));
		point = matchend;
	} else if (!strcmp(func, "extern")) {
		tag_modifier_kwd |= 1;	// next vars are just declarations
		return 1;
	} else if (!strcmp(func, "typedef")) {
		tag_modifier_kwd |= 2;	// next vars are typedefs
		return 1;
	} else if (is_java_buffer() && (!strcmp(func, "import")
									|| !strcmp(func, "package"))) {
		tag_modifier_kwd |= 4;
		return 1;
	} else if (!strcmp(func, "__declspec")) {
		return 1;
	} else if (parse_string(-1, "^[ \t]*#[^#\n]*"))
		return tag_c_define(func);	// do preproc line
	c_fwd_space();
	if (curchar() == '_' && parse_string(1, "(_ANSI_ARGS_|__P)</word>")) {
		point = matchend;	// skip over arg-hiding macro
		c_fwd_space();
	}
	if (curchar() == '<')
		tag_check_template();
	if (curchar() == '(') {
		if (matches_at(start, -1, "~[ \t]*")) {
			point = matchend;
			parse_string(1, "~[ \t]*[a-zA-Z0-9_]+", func);
		}
		i = tag_c_func(1);
		if (!(tag_which_items & 1))
			skip = 1;
	} else if (curchar() == ':' && parse_string(1, "::[ \t\n]*"
												"(_?(near|far)[ \t\n]+)?([a-zA-Z0-9_~]+|" CPP_OPSPEC ") *%(")) {
		strcat(func, "::");	// make raw class::func w/o spaces
		grab(find_group(3, 1), find_group(3, 0), func + strlen(func));
		i = tag_c_func(1);
		if (!(tag_which_items & 1))
			skip = 1;
	} else if (tag_c_variable()
			   && ((tag_declarations & 2) || !(tag_modifier_kwd & 1)
				   || (tag_modifier_kwd & 2))) {
		i = 2;
		if (!(tag_which_items & 2))
			skip = 1;
	}
	if ((i == 2 || i == 3) && !isdigit(*func) && !(tag_modifier_kwd & 4)) {
		if (*c_tagging_class && !index(func, ':')) {
			strcpy(tmp, func);	// put class:: before func name
			sprintf(func, "%s::%s", c_tagging_class, tmp);
		}
		if (!skip)
			add_tag(func, start);
	}
	return i;
}

tag_check_template()
{
	int pos = point;
	if (!find_template_end() || !parse_string(1, "[ \t\n]*%("))
		point = pos;
}

tag_c_skip_before_body(int show)
{
	if (parse_string(1, "#.*(\\\n.*)*")) {
		point = matchend;
		return 1;
	}
	if (parse_string(1, "(try|const|volatile|NOTHROW)![^_A-Za-z0-9]")) {
		point = matchend;
		return 1;
	}
	if (parse_string(1, "throws[ \t]")) {	// for java
		point = matchend;
		skip_identifier_list();
		return 1;
	}
	if (parse_string(1, "throw[ \t]*!%(")) {	// for C++
		point = matchend;
		if (skip_c_delims('(', ')', 1)) {
			if (show)
				buffer_printf("-tag-errors", "Couldn't parse %s at line %d\n",
							  filename, lines_between(0, point, 0) + 1);
			return 0;
		}
		return 1;
	}
	return 0;
}

tag_c_struct(int flavor)
{
	char func[TAGLEN], fullname[TAGLEN];
	int start;

	c_fwd_space();
	start = point;
	if (!parse_string(1, "[a-zA-Z0-9_]+", func)) {
		if (flavor != 1)
			*func = 0;	// all but class don't require a name
		else
			return 1;
	}
	c_fwd_space();	// skip compiler-specific junk betw "class" & name
	if (flavor == 1 && !is_java_buffer() && parse_string(1, "[a-zA-Z0-9_]+")) {
		parse_string(1, "[a-zA-Z0-9_]+", func);
		c_fwd_space();
	}
	if (curchar() != '{' && (flavor != 1 || !parse_other_class()))
		return 1;	// classes can have other stuff before {
	if (flavor == 1 || flavor == 3 || flavor == 0) {
		point++;	// skip over { so we do stuff within
		if (*c_tagging_class)
			strcat(c_tagging_class, "::");
		strcat(c_tagging_class, func);	// remember, now in class
	} else if (flavor == 2) {
		tag_c_enum();
		strcpy(fullname, func);
		if (*c_tagging_class)
			sprintf(fullname, "%s::%s", c_tagging_class, func);
		if (*func && (tag_which_items & 32))
			add_tag(fullname, start);
		return 1;
	}
	if (*func && (tag_which_items & 16))
		add_tag(c_tagging_class, start);
	return 1;
}

tag_c_enum()	// just before { of enum list; tag them all
{
	char name[TAGLEN];

	do {
		point++;
		c_fwd_space();
		if (!parse_string(1, "[a-zA-Z0-9_]+", name))
			return;
		if (tag_which_items & 8)
			add_tag(name, matchstart);
		c_fwd_space();
		if (curchar() == '=')
			skip_initializer();
	}  while (curchar() == ',');
	point += re_search(1, "<wspace>*}");
}

int list_definitions_when_displaying();

command list_definitions()
{
	char tag[FNAMELEN], pat[FNAMELEN], *title;
	int tags, tmp, pos, decls = 0;
	int orig = bufnum, oldpoint = point, starthere = -1, old_ws = window_start;
	int orig_tab_size;

	if (when_displaying == list_definitions_when_displaying)
		return;		// Ignore double-presses.
	error_if_input(1);
	tags = zap("-list-definitions");
	tmp = tmp_buf();
	delete_buffer("-tag-errors");
	iter = 0;
	save_var point = 0, _tags_buffer = tags, tag_show_percent = 1;
	save_var tag_by_text = 1, tag_which_items = list_which_definitions;
	save_var tags_from_bsc = 0;
	if (has_arg)
		save_var tag_declarations = decls = 3;
	if (!*filename || !find_tagger(1))
		error("Don't know how to list definitions in this buffer.");
	orig_tab_size = tab_size;
	if (!buffer_size("-list-definitions")) {
		save_var tag_declarations = decls = 3;
		point = 0;
		find_tagger(1);
	}
	if (!buffer_size("-list-definitions"))
		error("No definitions or declarations found in this buffer.");
	say("");
	restore_vars();
	if (want_code_coloring && recolor_range)
		recolor_buffer_range(0, size());	// Update colors for whole buffer.
	save_var point, bufnum = tags;
	point = 0;		// Copy colored lines of original file into tmp, based on tags.
	while (point < size() && parse_string(1, "^[^\t\n]+\t[^\t\n]+\t([^\t\n]+)\t")) {
		grab(find_group(1, 1), find_group(1, 0), pat);
		pos = strtoi(pat, 10);		// Position in original buffer.
		bufnum = orig;
		point = pos;
		if (give_begin_line() > oldpoint && starthere < 0)	// Look for starting pos.
			starthere = buffer_size(bufnum_to_name(tmp)) - 1;
		buf_xfer_colors(tmp, give_begin_line(), give_end_line() + 1);
		bufnum = tags;
		nl_forward();
	}
	bufnum = tmp;
	delete_final_newlines();
	point = (starthere == -1) ? size() : starthere;
	to_begin_line();	// Begin at corresponding point to original location.
	goal_column = 0;
	tab_size = orig_tab_size;
	when_displaying = list_definitions_when_displaying;
	restore_vars();
	*tag = 0;
	save_var menu_width = screen_cols - _view_left - _view_right;
	save_var _bufed_orig_window = window_handle;	// Save it here.
	save_var tag_declarations = decls;
	title = tag_declarations ? "Select a declaration or definition"
							 : "Select a definition";
	if (get_choice(tmp, tag, title, title, "OK", "Cancel", "")) {
		save_var bufnum = tmp;
		pos = tag_line_to_position(lines_between(0, point, 0) + 1, NULL);
		restore_vars();
		point = oldpoint;
		quiet_set_bookmark();
		point = pos;
		position_to_tag_line();
	} else
		point = oldpoint, window_start = old_ws;
	say("");
	buf_delete(tmp);
	buf_delete(tags);
}

color_tex_cmd(int s)
{
	char cmd[200];
	int color = color_class tex_literal;

	cmd[0] = '|';
	if (parse_string(1, "[a-z@]*", cmd + 1)) {
		strcat(cmd, "|");
		if (strstr("|begin|end|", cmd))
			color = color_class tex_begin_end;
		else if (strstr("|chapter|section|subsection|subsubsection"
						"|appendix|bibliography|title|tableofcontents"
						"|input|include|listoffigures|listoftables|part"
						"|usepackage|requirepackage|paragraph|subparagraph|",
						cmd))
			color = color_class tex_section;
		else
			color = color_class tex_command;
	} else if (curchar() == '[') {
		int orig = point;
		if (re_search(1, "\n[ \t]*\n|\\par![^a-z]|\\]")
			  && character(point - 1) == ']') {
			color_tex_math(orig - 1, point, color_class tex_display_math);
			return;
		}
		point = orig + 1;
	} else if (curchar() == '(') {
		int orig = point;
		if (re_search(1, "\n[ \t]*\n|\\par![^a-z]|\\%)")
			  && character(point - 1) == ')') {
			color_tex_math(orig - 1, point, color_class tex_math);
			return;
		}
		point = orig + 1;
	} else
		point++;
	set_character_color(s, point, color);
}

color_tex_dollar(int s)
{
	char pat[30];
	int orig, color = color_class tex_math;

	strcpy(pat, "\n[ \t]*\n|\\par![^a-z]|\\|%$");
	if (curchar() == '$') {	// A $$ that signals displayed math?
		point++;			// Get to after $$.
		strcat(pat, "%$");
		color = color_class tex_display_math;
	}
	orig = point;
	while (re_search(1, pat))
		if (character(point - 1) == '\\') {	// skip quoted ones
			if (curchar() == '$')
				point++;
		} else if (character(point - 1) == '$') {
			color_tex_math(s, point, color);
			break;
		} else {
			point = orig;
			break;
		}
}

jump_to_dvi_by_command()
{
	char dvi[FNAMELEN], *args[4], lnum[20], cmd[FNAMELEN * 3], buf[FNAMELEN];
	char basename[FNAMELEN];
	int i = 1;

	*dvi = 0;
	if (!has_arg) {
		if (*jump_to_dvi_main_file)
			strcpy(dvi, jump_to_dvi_main_file);
		if ((!*dvi || check_file(dvi) != CHECK_FILE) && *filename) {
			strcpy(dvi, filename);
			strcpy(get_extension(dvi), ".dvi");
		}
		if (*dvi && check_file(dvi) != CHECK_FILE)
			*dvi = 0;
	}
	save_var only_file_extensions = "|.dvi|";
	if (!*dvi)
		get_file_dir(dvi, "DVI file to view: ");
	strcpy(jump_to_dvi_main_file, dvi);
	if (!*dvi) {
		strcpy(dvi, filename);
		strcpy(get_extension(dvi), ".dvi");
		if (check_file(dvi) != CHECK_FILE)
			error("Can't locate .dvi file.");
	}
	sprintf(lnum, "%d", lines_between(0, point, 1) + 1);
	strcpy(basename, get_tail(filename, 1));
	*get_extension(basename) = 0;
	args[0] = lnum;
	args[1] = filename;
	args[2] = dvi;
	args[3] = basename;
	if (expand_string_template(cmd, jump_to_dvi_command, "lfdb", args))
		error("Invalid jump-to-dvi-command setting.");
	do {
		sprintf(buf, "-jump-to-dvi-output-%d", i++);
	} while (exist(buf) && is_process_buffer(name_to_bufnum(buf)));
	pipe_text(NULL, buf, cmd, NULL, 0, 0);
	iter = 0;
}

int vbasic_indent_more()
{
	char kwd[FNAMELEN], pat[FNAMELEN];

	if (vbasic_indent_subroutines
		  && parse_string(1, "[ \t]*(%<[^\n>]+%>)?[ \t]*"
						  "(</word>(Private|Public|Default|Friend"
						  "|Static|Protected|Shadows|MustInherit"
						  "|NotInheritable|Overloads|Overrides|Overridable"
						  "|NotOverridable|MustOverride|Shared|ReadOnly"
						  "|WriteOnly)</word>[ \t]*)*"
						  "</word>(Sub|Function|Property|Class|Enum"
						  "|Module|Interface|Namespace|Structure)</word>"))
		return 1;
	if (parse_string(1, "[ \t]*</word>(If)</word>.*"
					 "</word>Then</word>[ \t]*[^ \t\n':]"))
		return 0;		// Old one-line form of If statement.
	if (parse_string(1, "[ \t]*</word>(Do|While|Type|Else|With|For|Select"
					 "|Case|Try|Catch|Finally|SyncLock|If|Elseif)</word>")) {
		grab(find_group(1, 1), find_group(1, 0), kwd);
		sprintf(pat, ".*:[ \t]*</word>End[ \t]+%s</word>", kwd);
		if (!parse_string(1, pat))	// Multiple statements: ignore block.
			return 1;
	}
	if (parse_string(1, "[ \t]*(Get|Set)[ \t]*(%([ \t]*%))?[ \t]*[\n':]"))
		return 1;	// Get/set keyword with nothing after it starts block.
	return 0;
}

int vbasic_indent_less()
{
	return parse_string(1, "[ \t]*</word>(End|Wend|Else|Elseif"
						"|Loop|Next|Case|Catch|Finally)</word>") != 0;
}

vbasic_indenter()
{
	int orig = point, ind;

	if (matches_at(give_begin_line(), 1, "[ \t]*#")) {
		indent_to_column(0);
		return;
	}
	if (vbasic_is_continuation()) {
		vbasic_indent_continuation(orig);
		return;
	}
	do {
		to_begin_line();
		if (!re_search(-1, "[^ \t\n]"))	/* Find previous non-blank line */
			break;
		to_indentation();
	} while (parse_string(1, "['#]"));	// that's not a comment or preproc.
	vbasic_before_continuation();
	ind = get_indentation(point);
	if (vbasic_indent_more())
		ind += vbasic_step();
	point = orig;
	if (vbasic_indent_less())
		ind -= vbasic_step();
	to_indentation();		/* go to current line's indent */
	indent_to_column(ind);
}

color_vhdl_range(from, to) // recolor just this section
{			// last colored region may go past to
	int t = -1, talk, s, talk_at = 0;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	if (from < to)
		set_character_color(from, to, -1);
	point = from;
	save_var case_fold = 1;
	talk = (to - from > 20000);	// show status during long delays
	while (point < to) {
		if (!re_search(1, "[a-z_][a-z0-9_]*('[a-z0-9_]+)*"
					   "|[0-9.]([a-z0-9._]|e[-+][0-9])*l?|--|[\"'\\]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(point - 1)) {		// check last char
			case '-':			// found comment
				nl_forward();
				set_character_color(t, point, color_class vhdl_comment);
				break;
			case '"':		// found a string literal
				t = --point;
				if (strchr("BOX", toupper(character(point - 1)))
					  && parse_string(-1, "</word>[box]"))
					t = --point;
				if (re_search(1, "[box]?\"([^\"\\\n]|\\.)*[\"\n]"))
					set_character_color(t, point, color_class vhdl_string);
				if (get_character_color(point, (int *) 0, &s) ==
					  color_class vhdl_string && s > to)  // fix up after
					if (point < (to = s))	 // quoted "'s
						set_character_color(point, to, -1);
				break;
			case '\'':
//				if (re_search(1, "([^'\\\n]|\\.)*['\n]"))
				if (re_search(1, "([^'\n])*['\n]"))
					set_character_color(t, point, color_class vhdl_string);
				if (get_character_color(point, (int *) 0, &s) ==
					  color_class vhdl_string && s > to)  // fix up after
					if (point < (to = s))	 // quoted "'s
						set_character_color(point, to, -1);
				break;
			case '\\':		// \arbitrary name\ is an identifier.
				if (re_search(1, "[\\\n]"))
					set_character_color(t, point, color_class vhdl_identifier);
				break;
			default:		// found identifier, kywd, or number
				set_character_color(t, point, vhdl_keyword_color(t));
				break;
		}
		if (talk && point > talk_at + 2000) {
			note("Coloring VHDL program: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	if (talk)
		note("");
	return point;
}

vhdl_func_name_finder()
{
	int from, to, pos, color, orig = point;
	char pat[FNAMELEN * 2];

	save_var point, case_fold = 1;
	to_end_line();
	recolor_buffer_range(0, point);
	while (re_search(-1, "^[ \t]*(procedure|function)</word>")) {
		pos = point;
		color = get_character_color(matchend);
		if (color == color_class vhdl_comment
			  || color == color_class vhdl_string)
			continue;
		if (!parse_string(1, "^[ \t]*(procedure|function)[ \t]+([a-zA-Z_]<word>*)"))
			continue;
		from = find_group(2, 1);
		to = find_group(2, 0);
		to = MIN(to, from + ptrlen(display_func_name) - 1);
		grab(from, to, display_func_name);
		point = matchend;
		point += parse_string(1, "[ \t]*");
		if (curchar() == '(')
			move_level(1, "(", ")", 0, 0);
		if (!parse_string(1, "[ \t]*(return[ \t\n]+[a-z0-9_]+[ \t]*)?is</word>")) {
			point = pos;
			continue;
		}
		if (matchend - matchstart < FNAMELEN) {
			sprintf(pat, "^[ \t]*end[ \t]+%s</word>",
					display_func_name);
			if (re_search(1, pat) && point < orig)
				*display_func_name = 0;
		}
		return 1;
	}
	*display_func_name = 0;
	return 1;
}

tag_mode_vhdl()
{
	char func[TAGLEN];
	int start, color;

	vhdl_mode();
	recolor_buffer_range(0, size());	// Just to skip over comments
	save_var point = 0, case_fold = 1;				// & strings.
	while (re_search(1, "^[ \t]*(procedure|function|entity)"
					 "[ \t]+([a-zA-Z_]<word>*)")) {
		color = get_character_color(matchstart);
		if (color == color_class vhdl_comment
			  || color == color_class vhdl_string)
			continue;
		grab(start = find_group(2, 1), find_group(2, 0), func);
		add_tag(func, start);
	}
}

color_class html_embedded = color_class html_attribute;

when_loading()
{
	int i, j, k;

	if (!strcmp("12.00b4", eel_version)) { // Going from 12 B4 (old ver format).
		reg_tab[NUMSHIFT(NUMDIGIT(0))] = find_index("yank");
		ssh_tab[NUMSHIFT(NUMDIGIT(0))] = find_index("process_yank");
		process_tab[NUMSHIFT(NUMDIGIT(0))] = find_index("process_yank");
	}
	if (!strcmp("12.00b4", eel_version) || !strcmp("12b5", eel_version)) {
		reg_tab[NUMSHIFT(NUMCTRL(GREYTAB))] = find_index("switch_buffers");
	}
	if (html_auto_indent == 6)	// New default value.
		html_auto_indent = 14;

	k = find_index("case-indirect");
	if (name_type(k) == NT_COMMAND)
		for (i = 1; i <= final_index() ; i++)
			if (name_type(i) == NT_TABLE) {
				short *tab = index_table(i);
				j = -1;	// Replace case_indirect with -1 in all keytables.
				while ((j = find_next_entry(tab, k, j)) != -1)
					set_range(tab, j, -1, get_range(tab, j, k));
			}
	drop_name("case-indirect");

	// Fix up incorrect bindings made by beta 6's quickup.
#define REPAIR_KEYTABLE(s)	\
	{		\
		short *tab = index_table(find_index(s));				\
		for (i = 'A'; i <= 'Z'; i++)							\
			if (tab[i] == 0)									\
				set_range(tab, i, -1, get_range(tab, i, 0));	\
	}
	REPAIR_KEYTABLE("dired-tab");
	REPAIR_KEYTABLE("bufed-tab");
	REPAIR_KEYTABLE("blist-tab");
	REPAIR_KEYTABLE("grep-tab");
	REPAIR_KEYTABLE("info-tab");
}
